
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const HeartbeatParms_cxx_Version = "$Id$";


#include "cpLog.h"
#include "HeartbeatParms.hxx"
#include "ProvisionInterface.hxx"
#include "VNetworkException.hxx"
#include "VIoException.hxx"

static const string GlobConfigfilename("GlobalConfigData");

HeartbeatParms* HeartbeatParms::theHeartbeatParms = 0;

/**
 * Returns the only heartbeatParms or create it
 * if it doesn't already exist. 
 */
HeartbeatParms&
HeartbeatParms::instance()
{
    if ( theHeartbeatParms == 0 )
    {
        theHeartbeatParms = new HeartbeatParms();
    }
    return *theHeartbeatParms;
}

/**
 * Constructor
 */
HeartbeatParms::HeartbeatParms()
        : itsMulticastPort(0),
        itsMaxMissedHeartbeats(0),
        itsHeartbeatInterval(0)
{
    VGlobalConfigData globalConfig;
    try
    {
        /// get pointer to global config object
        ProvisionInterface::instance().getConfigData(globalConfig);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed ProvisionInterface::instance().getConfigData, reason %s", 
	      e.getDescription().c_str());
    }

    try
    {
        /// register interest in the following Global Config file
        ProvisionInterface::instance().registerForUpdate( updateHeartbeatParms, 
							  GlobConfigfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed ProvisionInterface::instance().registerForUpdate, reason %s",
	      e.getDescription().c_str());
    }
    
    /// assign host
    itsMulticastHost = globalConfig.getMultiCast();
    /// assign port
    itsMulticastPort = globalConfig.getMultiCastPort();
    /// assign host:port
    itsMulticastAddr = itsMulticastHost;
    itsMulticastAddr += ":";
    itsMulticastAddr += itsMulticastPort;
    /// assign max number of heartbeats missed before server is considered down
    itsMaxMissedHeartbeats = globalConfig.getMaxMissedHB();
    /// assign interval time between heartbeats
    itsHeartbeatInterval = globalConfig.getHBInterval();
    
    cpLog( LOG_DEBUG, "multicast addr=%s", itsMulticastAddr.logData());
    
    /// display configured params
    displayHeartbeatParms();
    
}

/**
 * Copy Constructor
 */
HeartbeatParms::HeartbeatParms(const HeartbeatParms& rhs)
{}

/**
 * Overload operator assignment =
 */
HeartbeatParms&
HeartbeatParms::operator=(const HeartbeatParms& rhs)
{
    return *this;
}


/**
 * Get Multicast Address   
 * @param  address  the global multicast ip/port used by the system.
*/
void
HeartbeatParms::getMulticastAddr( Data & address )
{
    address = itsMulticastAddr;
}


/**
 * Get Multicast Host   
 * @param  host  the global multicast ip used by the system.
*/
void
HeartbeatParms::getMulticastHost( Data & host )
{
    host = itsMulticastHost;
}


/**
 * Get Multicast Port   
 * @return  the global multicast port used by the system.
*/
int
HeartbeatParms::getMulticastPort( ) const
{
    return itsMulticastPort;
}


/**
 * Get Maximum Missed Heartbeats   
 * @return  the global max missed heartbearts used by the system.
*/
int
HeartbeatParms::getMaxMissedHeartbeats( ) const
{
    return itsMaxMissedHeartbeats;
}


/**
 * Get Heartbeat Interval    
 * @return  the global heartbeat interval used by the system.
*/
int
HeartbeatParms::getHeartbeatInterval( ) const
{
    return itsHeartbeatInterval;
}


/**
 * Set Multicast Port    
 * @param  multicastPort  the multicast port to be used by the system.
*/
void
HeartbeatParms::setMulticastPort( const int multicastPort )
{
    itsMulticastPort = multicastPort;
}


/**
 * Set Maximum Missed Heartbeats    
 * @param  maxMissedHeartbeats  the maximum missed heartbeats tolerated before a
 *                              server is considered dead.
*/
void
HeartbeatParms::setMaxMissedHeartbeats( const int maxMissedHeartbeats )
{
    itsMaxMissedHeartbeats = maxMissedHeartbeats;
}


/**
 * Set Heartbeat Interval     
 * @param  heartbeatInterval  the heartbeat interval used to schedule heartbeat
 *                            transmission and to schedule housekeeping.  
*/
void
HeartbeatParms::setHeartbeatInterval( const int heartbeatInterval )
{
    itsHeartbeatInterval = heartbeatInterval;
}


/**
 * Destructor
 */
HeartbeatParms::~HeartbeatParms( )
{}


/**
 * Display Heartbeat Parameters stored in this object.  
 */
void
HeartbeatParms::displayHeartbeatParms( )
{
    cpLog( LOG_DEBUG, "-----------------------------");
    cpLog( LOG_DEBUG, "Heartbeat Parameters:");
    cpLog( LOG_DEBUG, "multicast Host: %s", itsMulticastHost.logData());
    cpLog( LOG_DEBUG, "          Port: %d", itsMulticastPort);
    cpLog( LOG_DEBUG, "          Addr: %s", itsMulticastAddr.logData());
    cpLog( LOG_DEBUG, "Max Missed HB: %d", itsMaxMissedHeartbeats);
    cpLog( LOG_DEBUG, "HB Interval: %d", itsHeartbeatInterval);
    cpLog( LOG_DEBUG, "-----------------------------");
}


/**
 * The Provisioning Server is notifying us of a change in the Global Config Data.
 */
void
updateHeartbeatParms(const string& data, const string& fileName, bool)
{
    cpLog( LOG_DEBUG, "updating Heartbeat Parms");
    VGlobalConfigData globalConfigData;
    try
    {
        /// get pointer to global config object
        ProvisionInterface::instance().getConfigData(globalConfigData);
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed ProvisionInterface::instance().getConfigData, reason %s", 
	      e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed  ProvisionInterface::instance().getConfigData, reason %s", 
	      e.getDescription().c_str());
    }

    HeartbeatParms& hbParms = HeartbeatParms::instance();
    
    /// assign new max number of heartbeats missed if changed
    int newValue = globalConfigData.getMaxMissedHB();
    int oldValue = hbParms.getMaxMissedHeartbeats();
    cpLog( LOG_DEBUG, "old maxMissed: %d new: %d", oldValue, newValue );
    if ( oldValue != newValue)
    {
        hbParms.setMaxMissedHeartbeats( newValue );
    }
    /// display configured params
    hbParms.displayHeartbeatParms();

}


