
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const HeartbeatRxThread_cxx_Version =
    "$Id: HeartbeatRxThread.cxx,v 1.28.2.1 2002/04/25 01:07:03 sprajpat Exp $";

#include <unistd.h>
#include <netinet/in.h>
#include "cpLog.h"
#include "HeartbeatRxThread.hxx"
#include "HeartbeatParms.hxx"
#include "HeartbeatCommon.h"

/**
 * Ctor
 * @param serverContainer the container of all servers.
 */
HeartbeatRxThread::HeartbeatRxThread( BaseServerContainer& serverContainer ) 
  : ThreadIf(),
    itsServerContainer( serverContainer )
{
    HeartbeatParms& parms = HeartbeatParms::instance();

    cpLog( LOG_DEBUG, "Assign HeartbeatRx MulticastAddr" );
    Data multicastAddr;
    parms.getMulticastAddr( multicastAddr );
    LocalScopeAllocator lo;
    itsMulticastGroup = NetworkAddress( multicastAddr.getData(lo));

    int heartbeatPort = parms.getMulticastPort();
    itsRxUdpStack = new UdpStack(&itsMulticastGroup, heartbeatPort, heartbeatPort, sendrecv, false, true);
    assert( itsRxUdpStack != 0);

    char hostName[256];
    memset (hostName, 0, sizeof(hostName));
    if (gethostname(hostName, 256) == 0)
    {
        itsIface = NetworkAddress(hostName);
    }
    else
    {
        cpLog (LOG_ERR, "Failed gethostname().");
    }
    /// join Multicast group
    int ifaceInexe = 0;
    itsRxUdpStack->joinMulticastGroup(itsMulticastGroup, &itsIface, ifaceInexe);

}


/**
 * Ctor - for applications that do not talk to provisioning. 
 * @param  serverContainer the container of all servers.
 * @param  multicastIP    IP on which heartbeat messages are received.
 * @param  multicastPort  Port on which heartbeat messages are received.
 */
HeartbeatRxThread::HeartbeatRxThread(
    BaseServerContainer& serverContainer, 
    const char * multicastIP,
    u_int16_t multicastPort
    )
  : ThreadIf(),
    itsMulticastGroup(multicastIP, multicastPort),
    itsServerContainer( serverContainer )
{
    itsRxUdpStack = new UdpStack(&itsMulticastGroup, multicastPort, multicastPort, sendrecv, false, true);
    assert( itsRxUdpStack != 0);

    char hostName[256];
    memset (hostName, 0, sizeof(hostName));
    if (gethostname(hostName, 256) == 0)
    {
        itsIface = NetworkAddress(hostName);
    }
    else
    {
        cpLog (LOG_ERR, "Failed gethostname().");
    }

    /// join Multicast group
    int ifaceInexe = 0;
    itsRxUdpStack->joinMulticastGroup(itsMulticastGroup, &itsIface, ifaceInexe);

}


/**
 * Destructor 
 */
HeartbeatRxThread::~HeartbeatRxThread()
{
    assert( itsRxUdpStack != 0);

    /// leave Multicast group
    int ifaceInexe=0;
    itsRxUdpStack->leaveMulticastGroup(itsMulticastGroup, &itsIface, ifaceInexe);
}


/**
 * Process events from the udp stack.
 */
void
HeartbeatRxThread::thread()
{
    int bytesRead = 0;
    char buf[256];
    NetworkAddress sender;

    while (true)
    {
        try
        {

            bytesRead = itsRxUdpStack->receiveFrom(buf, 256, &sender);

	    cpLog(LOG_DEBUG_HB, "the udp stack receive from: %s", sender.getIpName().c_str());

            heartbeatPortData* hbPortData = (heartbeatPortData*) & buf;

            Data serverAddress( sender.getIpName() );
            serverAddress += ":";
            serverAddress += ntohs(hbPortData->port);

            if ( (u_int16_t)HeartbeatData == ntohs(hbPortData->dataType) )
            {     
                cpLog( LOG_DEBUG_HB, "HB: %s", serverAddress.logData() );

                itsServerContainer.processHeartbeat(serverAddress);
            }
            else
            {
                cpLog( LOG_ERR, "Unrecognized Mesg received on Heartbeat Port from %s", 
		       serverAddress.logData() );
            }


        }
        catch ( ... )
        {
            cpLog( LOG_DEBUG, "Exception from udpstack for HeartbeatRx" );
        }
    }

}
// End of File
