#ifndef ROUNDROBINSERVERS_HXX_
#define ROUNDROBINSERVERS_HXX_

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const RoundRobinServers_hxx_Version =
    "$Id: RoundRobinServers.hxx,v 1.16 2001/04/27 11:02:42 bko Exp $";


#include "Sptr.hxx"
#include "Data.hxx"
#include "ServerContainer.hxx"

/**
 * RoundRobinServer class
 */

class RoundRobinServers
{
public:
        /**
	 * Constructor
	 */
        RoundRobinServers();
        /**
	 * Obtain the singleton of this object
	 * @param serverType the type of the server such as RS, MS, etc.
	 * @return Reference to the singleton
	 */
        static RoundRobinServers& instance( const ServerType serverType );
        /// Destructor
        virtual ~RoundRobinServers();
        /// Display the list of servers stored in this object
        void displayServerList() const;
        /**
	 * Obtain a server using a hash function and strip off the port before returning.
	 * @param  hash_value  Value with which to hash the list of servers
	 * @return  Host address of chosen server
	 */
        const Data getHost( const Data& hash_value );
        /**
	 * Obtain a server port by hashing into the server list and stripping
	 * off the host before returning.
	 * @param  hash_value  Value with which to hash the list of servers
	 * @return  Port of chosen server
	 */
        const int getPort( const Data& hash_value );

        /**
	 * Obtain a server host and port by hashing into the server list and stripping
	 * off the host before returning.
	 * @param  hash_value  Value with which to hash the list of servers
	 */
        void getHostNPort( const Data& hash_value , Data& host, int& port);

        /**
	 * Determine if the given host and port match a valid server
	 * @param  host  Host portion of address to be verified
	 * @param  port  Port portion of address to be verified
	 * @return  True if server exists, false if not
	 */
        bool isExistingServer( const Data& host, const int port ) const;
        /**
	 * Determine if given server matches an existing server in the list
	 * @param  searchString  Server address
	 * @return  True if server exists, false if not
	 */
        bool isExistingServer( Sptr < Data > searchServer ) const;
        /**
	 * Test code to exercise obtaining the server lists
	 * @param  server
	 * @param  port
	 */
        void updateServers( const char* server, const int port );
        /**
	 * Obtain a number using a hashing algorithm
	 * @param  hash_key  Key to be used in the hash
	 * @return  Integer result of the hash
	 */
        unsigned int hashingFunction( const Data& hash_key ) const;

protected:
        /// This is the single instance of RoundRobinServers
        static RoundRobinServers* _theRoundRobinServers;
        /**
	 * Copy Constructor
	 * @param  rhs  Object from which to copy
	 */
        RoundRobinServers( const RoundRobinServers& rhs );
        /**
	 * Copy member data from given object into this one
	 * @param  rhs  Object from which to copy
	 * @return  Reference to this object with new data
	 */
        RoundRobinServers& operator=( const RoundRobinServers& rhs );

private:

        Data _pServer;
        int _pServerPort;

        /// list of servers
        Sptr < vector < Data > > _listOfServers;
        /// types of servers
        ServerType _serverType;
        /**
	 * Store the server type and obtain the list of servers for that type
	 * @param  serverType  Type of server
	 */
        void updateServerContainer( const ServerType );
        /**
	 * Obtain a server from the list using a hash function
	 * @param  hash_key  Key to be used in the hash
	 * @return  Server resulting from the hash
	 */
        string hashIntoServer( const Data& hash_key );
        /**
	 * Copy member data from given object into this one
	 * @param  rhs  Object from which to copy
	 */
        void copyRhsToThis( const RoundRobinServers& rhs);
        
};

#endif  // ROUNDROBINSERVERS_HXX_
