#ifndef SERVER_HXX
#define SERVER_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ServerVersion = "$Id: Server.hxx,v 1.14 2001/07/31 22:56:29 sprajpat Exp $";

#include "Data.hxx"
#include "Sptr.hxx"
#include "HeartbeatCommon.h"


/// This class describes a Server in the system

class Server
{
public:
        /**
	 * Constructor
	 * @param  serverType  Type of server to be created
	 */
        Server( const ServerType serverType);
        /**
	 * Constructor provided for applications that don't talk to provisioning.
	 * @param
	 * @return
	 */
        Server( const ServerType serverType, const int maxMissedHB);
        /**
	 * Copy Constructor
	 * @param  rhs  Object from which to copy
	 * @return
	 */
        Server::Server( const Server& rhs );
        /**
	 * Operator overload for "="
	 * @param  rhs  Object from which to copy
	 * @return  Reference to this object with new data
	 */
        Server& Server::operator=( const Server& rhs );
        /// Destructor
        virtual ~Server();
        /**
	 * Obtain the server's type
	 * @return  The server's type
	 */
        ServerType getType() const;
        /**
	 * Obtain the status of this server
	 * @return  This server's status
	 */
         ServerStatus getStatus() const;
         /**
	  * Obtain the current heartbeat count
	  * @return  The number of heartbeats counted for this server
	  */
        int getHeartbeatCount() const;
        /**
	 * Obtain the number of heartbeats missed for this server
	 * @return  Number of heartbeats missed
	 */
        int getNumOfHeartbeatMissed() const;
        /**
	 * Obtain the maximum number of heartbeats allowed before the server
	 * is considered to be down.
	 * @return  Max number of missed heartbeats
	 */
        int getMaxMissedHB() const;
        /**
	 * Store the maximum number of missed heartbeats
	 * @param  Maximum number of missed heartbeats
	 */
        void setMaxMissedHB( const int maxMissedHB);
        /**
	 * Handle an incoming heartbeat by updating the server's counts
	 * @param  key  Server's address
	 */
        void processHeartbeat( const Data& key);
        /**
	 * Check if given server is no longer heartbeating and
	 * change its state accordingly. Clear the counts if necessary
	 * @param  serverId  Address of server
	 */
        void doHouseKeeping( const Data& serverId);
        /**
	 * Check to see if a server is no longer heartbeating.
	 * @return  True if the max number of heartbeats missed has exceeded the
	 *          allowed limit, false if not.
	 */
        bool serverIsDead() const;

        /**
	 * Store the server's current status
	 * @param  status  Status of server
	 */
        void setStatus (const ServerStatus status);

        ///Displays server status in log
        void displayStatus(const Data& key);

protected:
        /**
	 * Store the server type
	 * @param  type  Type of server
	 */
        void setType (const ServerType type);
        /// Increment the count of heartbeats received
        void incrHeartbeatCountByOne();
        /// Increment the count of heartbeats missed
        void incrNumOfHeartbeatMissed();
        /// Reset the heartbeat count to zero.
        void clearHeartbeatCount();
        /// Reset the missed heartbeat count to zero
        void clearNumOfHeartbeatMissed();
        /**
	 * Handle the first heartbeat for the given server by changing its state to active.
	 * @param  key  Server's address
	 */
        void gotFirstHeartbeat(const Data& key);
        /**
	 * Handle receiving a heartbeat after one or more was missed by making
	 * the server active.
	 */
        void gotHiccup();

        /// Server type
        ServerType itsType;
        /// Server's current status
        ServerStatus itsStatus;
        /// Number of heartbeats received
        int heartbeatCount;
        /// Number of heartbeats missed
        int numOfHeartbeatMissed;
        /// Maximum number of heartbeats a server may miss before it is considered inactive.
        int itsMaxMissedHB;

private:
        /**
	 * Copy member data from given object into this one
	 * @param  rhs  Object from which to copy
	 */
        void Server::copyRhsToThis( const Server& rhs );
};

#endif // SERVER_HXX
