
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const HBServerContainer_cxx_Version =
    "$Id: HBServerContainer.cxx,v 1.22 2002/02/25 05:46:19 sprajpat Exp $";


#include "ProvisionInterface.hxx"
#include "HBServerContainer.hxx"
#include "ReadLockHelper.hxx"
#include "WriteLockHelper.hxx"
#include "ServerWithSnmp.hxx"
#include "HeartbeatUtil.h"
#include "VIoException.hxx"

static const string RSfilename("ListOfRedirectServers");
static const string MSfilename("ListOfMarshalServers");
static const string FSfilename("ListOfFeatureServers");
static const string JSfilename("ListOfJtapiServers");
static const string POSfilename("ListOfPdpServers");
static const string CDRfilename("ListOfCdrServers");
static const string GlobConfigfilename("GlobalConfigData");

Sptr < HBServerContainer > HBServerContainer::theRSSrvContainer = 0;

/**
 * Return the one Server container or create it if it doesn't already exist. 
 * @return  Reference to the singleton
 */
HBServerContainer&
HBServerContainer::instance()
{
    if ( theRSSrvContainer == 0 )
    {
        theRSSrvContainer = new HBServerContainer;
    }
    return *theRSSrvContainer;
}

/**
 * Constructor
 */
HBServerContainer::HBServerContainer()
        : BaseServerContainer()
{
    /// get config data from PS
    configRedirectServers();
    configMarshalServers();
    configFeatureServers();
    configJtapiServers();
    configPolicyServers();
    configCdrServers();

    /// register interest in the following files
    try
    {
	ProvisionInterface::instance().registerForUpdate( updateHBRedirectServers, RSfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBRedirectServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBMarshalServers, MSfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBMarshalServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBFeatureServers, FSfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBFeatureServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBJtapiServers, JSfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBJtapiServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBPolicyServers, POSfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBPolicyServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBCdrServers, CDRfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBCdrServers, reason %s", 
	      e.getDescription().c_str());
    }
    try
    {
        ProvisionInterface::instance().registerForUpdate( updateHBHeartbeatParms, GlobConfigfilename,
							  ProvisionInterface::VCONFIG);
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed updateHBHeartbeatParms, reason %s", e.getDescription().c_str());
    }

    displayLists();
}


/**
 * Get Servers from Provisioning   
 */
void
HBServerContainer::getServersFromPS( const ServerType type, vector < Data > & retList)
{
    /// get instance of Provisioning Server
    ProvisionInterface& pif = ProvisionInterface::instance();

    try
    {
        Sptr < list < string > > tmpList = new list < string > ;

        switch (type)
        {
            case SERVER_RS:
            {
                pif.getRedirectServers( *tmpList );
            }
            break;
            case SERVER_JS:
            {
                pif.getJtapiServers( *tmpList );
            }
            break;
            case SERVER_FS:
            {
                pif.getFeatureServers( *tmpList );
            }
            break;
            case SERVER_MS:
            {
                pif.getMarshalServers( *tmpList );
            }
            break;
            case SERVER_POS:
            {
                pif.getPolicyServers( *tmpList );
            }
            break;
            case SERVER_CDR:
            {
                pif.getCdrServers( *tmpList );
            }
            break;
            default:
            break;
        }

        /// add server to vector
        for (list < string > ::iterator i = tmpList->begin(); i != tmpList->end(); ++i)
        {
            /// check for duplicate entries in list
            bool duplicateEntryFound = false;
            for (vector < Data > ::iterator j = retList.begin(); j != retList.end(); ++j)
            {
                char buf[156];
                j->getData(buf, 156);
                string serverInList = buf;
                if ( *i == serverInList )
                {
                    duplicateEntryFound = true;
                    break;
                }
            }

            if (duplicateEntryFound)
            {
                cpLog( LOG_ALERT, "*** Duplicate Entry Found In LIST: %s ***", i->c_str() );
                cpLog( LOG_ALERT, "*** DO NOT add to Container ***");
            }
            else
            {
                Data addressPort(*i);
                cpLog( LOG_DEBUG, "ip:port=%s", addressPort.logData());
                retList.push_back( addressPort );
            }
        }

    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }
}


/**
 * Remove Duplicates entries in the List of Servers passed in.   
 * @param  list  list of servers to be added to server container.  
 */
void
HBServerContainer::removeDuplicatesFromList( vector < Data > & list )
{
    ReadLockHelper tableHelper(tableLock);

    bool done = false;
    while (!done)
    {
        TableIter iter;

        bool duplicateEntryFound = false;
        for (vector < Data > ::iterator i = list.begin();
                i != list.end(); ++i)
        {
            iter = serverList.find(*i);
            if (iter != serverList.end())
            {
                cpLog( LOG_ALERT, "*** Duplicate Entry Found in CONTAINER: %s ***", i->logData() );
                cpLog( LOG_ALERT, "*** DO NOT add to Container ***");
                duplicateEntryFound = true;
                list.erase(i);
                break;
            }
        }
        if (!duplicateEntryFound) done = true;
    }
}


/**
 * Configure Redirect Servers    
 */
void
HBServerContainer::configRedirectServers()
{
    WriteLockHelper rsHelper(rsLock);

    /// get RSlist
    getServersFromPS( SERVER_RS, itsRSList );

    removeDuplicatesFromList(itsRSList);

    /// create Redirect Servers
    for (vector < Data > ::iterator i = itsRSList.begin();
            i != itsRSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new RS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        addServerWithSnmp( smartServer, SERVER_RS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Marshal Servers    
 */
void
HBServerContainer::configMarshalServers()
{
    WriteLockHelper msHelper(msLock);

    /// get MSlist
    getServersFromPS( SERVER_MS, itsMSList );

    removeDuplicatesFromList(itsMSList);

    /// create Marshal Servers
    for (vector < Data > ::iterator i = itsMSList.begin(); i != itsMSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new MS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// create new Marshal
        addServerWithSnmp( smartServer, SERVER_MS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Feature Servers    
 */
void
HBServerContainer::configFeatureServers()
{
    WriteLockHelper fsHelper(fsLock);

    /// get FSlist
    getServersFromPS( SERVER_FS, itsFSList );

    removeDuplicatesFromList(itsFSList);

    /// create Feature Servers
    for (vector < Data > ::iterator i = itsFSList.begin(); i != itsFSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new FS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// create new Feature
        addServerWithSnmp( smartServer, SERVER_FS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Jtapi Servers    
 */
void
HBServerContainer::configJtapiServers()
{
    WriteLockHelper jsHelper(jsLock);

    /// get JSlist
    getServersFromPS( SERVER_JS, itsJSList );

    removeDuplicatesFromList(itsJSList);

    /// create Jtapi Servers
    for (vector < Data > ::iterator i = itsJSList.begin(); i != itsJSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new JS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        /// create new Marshal
        addServerWithSnmp( smartServer, SERVER_JS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure Policy Servers    
 */
void
HBServerContainer::configPolicyServers()
{
    WriteLockHelper posHelper(posLock);

    /// get POSlist
    getServersFromPS( SERVER_POS, itsPOSList );

    removeDuplicatesFromList(itsPOSList);

    /// create Policy Servers
    for (vector < Data > ::iterator i = itsPOSList.begin();
            i != itsPOSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new POS: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        addServerWithSnmp( smartServer, SERVER_POS );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Configure CDR Servers    
 */
void
HBServerContainer::configCdrServers()
{
    WriteLockHelper cdrHelper(cdrLock);

    /// get CDRlist
    getServersFromPS( SERVER_CDR, itsCDRList );

    removeDuplicatesFromList(itsCDRList);

    /// create Cdr Servers
    for (vector < Data > ::iterator i = itsCDRList.begin();
            i != itsCDRList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "from PS new CDR: %s", i->logData() );
        /// get server id
        Data smartServer(*i);
        addServerWithSnmp( smartServer, SERVER_CDR );

        /// display updated data
        display( smartServer );
    }
}


/**
 * Destructor    
 */
HBServerContainer::~HBServerContainer( )
{}


/**
 * Get List of All Servers in Server Container
 * @param  listofServers  list of all servers in server container  
 */
void
HBServerContainer::getListofAllServers( Sptr < vector < Data > > & listofServers )
{
    ReadLockHelper rsHelper(rsLock);
    ReadLockHelper msHelper(msLock);
    ReadLockHelper fsHelper(fsLock);
    ReadLockHelper jsHelper(jsLock);
    ReadLockHelper posHelper(posLock);
    ReadLockHelper cdrHelper(cdrLock);

    cpLog( LOG_DEBUG_HB, "HBServerContainer::getListofAllServers");

    for (vector < Data > ::iterator i = itsRSList.begin();
            i != itsRSList.end(); ++i)
    {
        listofServers->push_back(*i);
    }

    for (vector < Data > ::iterator i = itsMSList.begin();
            i != itsMSList.end(); ++i)
    {
        listofServers->push_back(*i);
    }

    for (vector < Data > ::iterator i = itsFSList.begin();
            i != itsFSList.end(); ++i)
    {
        listofServers->push_back(*i);
    }
    for (vector < Data > ::iterator i = itsJSList.begin();
            i != itsJSList.end(); ++i)
    {
        listofServers->push_back(*i);
    }
    for (vector < Data > ::iterator i = itsPOSList.begin();
            i != itsPOSList.end(); ++i)
    {
        listofServers->push_back(*i);
    }
    for (vector < Data > ::iterator i = itsCDRList.begin();
            i != itsCDRList.end(); ++i)
    {
        listofServers->push_back(*i);
    }
}


/**
 * Get List of Active Servers of a certain type  
 * @param  type  the type of server to search for   
 * @param  listofServers  list of all servers in container that matches the type.  
 */
void
HBServerContainer::getListofActive( const ServerType type, Sptr < vector < Data > > & listofServers )
{
    /// function not used in heartbeat server at this time
}


/**
 * Add a server object that send messages to snmp.   
 * @param  key  Server's address   
 * @param  serverType  type of server.  
 */
void
HBServerContainer::addServerWithSnmp( const Data& key,
                                      const ServerType serverType )
{
    cpLog( LOG_DEBUG, "HBServerContainer::addServerWithSnmp %s",
           convertToString(serverType) );

    /// create a Server with 'new'
    Sptr < ServerWithSnmp > server = new ServerWithSnmp(serverType);

    WriteLockHelper tableHelper(tableLock);

    /// insert the Server in the map
    //serverList.insert(pair< Sptr<Data>, Sptr<Server> >(key, server));
    serverList[key] = server;

}


/**
 * Deal with a heartbeat event from the given server
 * @param  key  Server who sent the heartbeat
 * @return  True if server is recognized, false if not
 */
bool
HBServerContainer::processHeartbeat( const Data& key)
{
    Sptr < Server > serverPtr;

    if (search( key, serverPtr) )
    {
        ReadLockHelper tableHelper(tableLock);

        //cpLog(LOG_DEBUG, "Server found");
        Sptr < ServerWithSnmp > serverWithSnmpPtr;

        /// print data contained in the server object
        switch (serverPtr->getType() )
        {
            case SERVER_JS:
            case SERVER_FS:
            case SERVER_MS:
            case SERVER_RS:
            case SERVER_POS:
            case SERVER_CDR:
            {
                serverWithSnmpPtr.dynamicCast(serverPtr);
                serverWithSnmpPtr->processHeartbeat(key);
            }
            break;
            default:
            break;
        };

        displayServer(*serverPtr);
        return true;
    }
    else
    {
        return false;
    }
}


/**
 * Check all servers to see if they are still actively heartbeating
 */
void
HBServerContainer::doHouseKeeping()
{
    Data serverId;
    Sptr < Server > serverPtr;

    cpLog( LOG_DEBUG_HB, "");
    cpLog( LOG_DEBUG_HB, "****** do HouseKeeping *****");

    ReadLockHelper tableHelper(tableLock);

    if ( !serverList.empty() )
    {
        Sptr < ServerWithSnmp > serverWithSnmpPtr;

        for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
        {
            serverId = i->first;
            serverPtr = i->second;
            //cpLog( LOG_DEBUG, "%s", serverId.logData() );
            if ( serverPtr->getStatus() != Inactive )
                cpLog( LOG_DEBUG_HB, "%s", serverId.logData() );
            switch (serverPtr->getType() )
            {
                case SERVER_JS:
                case SERVER_FS:
                case SERVER_MS:
                case SERVER_RS:
                case SERVER_POS:
                case SERVER_CDR:
                {
                    serverWithSnmpPtr.dynamicCast(serverPtr);
                    serverWithSnmpPtr->doHouseKeeping(serverId);
                }
                break;
                default:
                break;
            }
        }
    }
    else
    {
        cpLog( LOG_DEBUG_HB, "nothing in server container");
    }

    cpLog( LOG_DEBUG_HB, "****** finished HouseKeeping *****");
    cpLog( LOG_DEBUG_HB, "");
}


/**
 * A change was made to the RS's in the provisioning server. 
 * Delete the stored RS server list and recreate it with the new data.
 */
void
HBServerContainer::updateRSsFromPS()
{

    {
        WriteLockHelper rsHelper(rsLock);

        /// delete all RSs in Server Container
        for (vector < Data > ::iterator i = itsRSList.begin();
                i < itsRSList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete RS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear RSList
        itsRSList.clear();
    }

    /// reconfigure RSs from scratch
    configRedirectServers();

    displayLists();
}


/**
 * A change was made to the MS's in the provisioning server. 
 * Delete the stored MS server list and recreate it with the new data.
 */
void
HBServerContainer::updateMSsFromPS()
{

    {
        WriteLockHelper msHelper(msLock);

        /// delete all MSs in Server Container
        for (vector < Data > ::iterator i = itsMSList.begin();
                i < itsMSList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete MS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear MSList
        itsMSList.clear();
    }

    /// reconfigure MSs from scratch
    configMarshalServers();

    displayLists();
}


/**
 * A change was made to the FS's in the provisioning server. 
 * Delete the stored FS server list and recreate it with the new data.
 */
void
HBServerContainer::updateFSsFromPS()
{

    {
        WriteLockHelper fsHelper(fsLock);

        /// delete all FSs in Server Container
        for (vector < Data > ::iterator i = itsFSList.begin();
                i < itsFSList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete FS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear FSList
        itsFSList.clear();
    }

    /// reconfigure FSs from scratch
    configFeatureServers();

    displayLists();
}


/**
 * A change was made to the JS's in the provisioning server. 
 * Delete the stored JS server list and recreate it with the new data.
 */
void
HBServerContainer::updateJSsFromPS()
{

    {
        WriteLockHelper jsHelper(jsLock);

        /// delete all JSs in Server Container
        for (vector < Data > ::iterator i = itsJSList.begin();
                i < itsJSList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete JS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear JSList
        itsJSList.clear();
    }

    /// reconfigure JSs from scratch
    configJtapiServers();

    displayLists();
}


/**
 * A change was made to the POS's in the provisioning server. 
 * Delete the stored POS server list and recreate it with the new data.
 */
void
HBServerContainer::updatePOSsFromPS()
{

    {
        WriteLockHelper posHelper(posLock);

        for (vector < Data > ::iterator i = itsPOSList.begin();
                i < itsPOSList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete POS: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear POSList
        itsPOSList.clear();
    }

    /// reconfigure POSs from scratch
    configPolicyServers();

    displayLists();
}


/**
 * A change was made to the CDR's in the provisioning server. 
 * Delete the stored CDR server list and recreate it with the new data.
 */
void
HBServerContainer::updateCDRsFromPS()
{

    {
        WriteLockHelper cdrHelper(cdrLock);

        for (vector < Data > ::iterator i = itsCDRList.begin();
                i < itsCDRList.end(); ++i)
        {
            cpLog( LOG_DEBUG, "delete CDR: %s", i->logData() );
            Data server(*i);
            /// delete from Server Container
            remove(server);
        }

        /// clear CDRList
        itsPOSList.clear();
    }

    /// reconfigure CDRs from scratch
    configCdrServers();

    displayLists();
}


/**
 * A change was made to the Global Config Parms Data on the provisioning server. 
 * Delete the stored parameter(s) and replace it with new values.
 */
void
HBServerContainer::updateHBParmsFromPS()
{
    cpLog( LOG_DEBUG, "updating Heartbeat Parms in RSServerContainer");

    ProvisionInterface& pif = ProvisionInterface::instance();
    try
    {
        WriteLockHelper tableHelper(tableLock);

        /// get pointer to global config object
        VGlobalConfigData globalConfigData;
        pif.getConfigData(globalConfigData);

        int newValue = globalConfigData.getMaxMissedHB();
        int oldValue = serverList.begin()->second->getMaxMissedHB();
        cpLog( LOG_DEBUG, "maxMissedHB old:%d new: %d", oldValue, newValue );

        if (( !serverList.empty() ) && ( oldValue != newValue ))
        {
            Data serverId;
            Sptr < Server > serverPtr;

            for ( TableIter i = serverList.begin(); i != serverList.end(); ++i)
            {
                serverId = i->first;
                serverPtr = i->second;
                serverPtr->setMaxMissedHB( newValue );
            }
        }
    }
    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to get data from Provisioning Server, reason %s", e.getDescription().c_str());
    }
    catch (VException& e)
    {
        cpLog(LOG_ALERT, "Failed to connect to Provisioning Server, reason %s", e.getDescription().c_str());
    }
}


/**
 * Display lists of server by type.
 */
void
HBServerContainer::displayLists()
{
    ReadLockHelper jsHelper(jsLock);
    ReadLockHelper fsHelper(fsLock);
    ReadLockHelper msHelper(msLock);
    ReadLockHelper rsHelper(rsLock);
    ReadLockHelper posHelper(posLock);
    ReadLockHelper cdrHelper(cdrLock);

    cpLog( LOG_DEBUG, "----------- Jtapi Servers ---------");
    for (vector < Data > ::iterator i = itsJSList.begin();
            i < itsJSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "JS: %s", i->logData() );
    }

    cpLog( LOG_DEBUG, "----------- Feature Servers ---------");
    for (vector < Data > ::iterator i = itsFSList.begin();
            i < itsFSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "FS: %s", i->logData() );
    }

    cpLog( LOG_DEBUG, "----------- Marshal Servers ---------");
    for (vector < Data > ::iterator i = itsMSList.begin();
            i < itsMSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "MS: %s", i->logData() );
    }

    cpLog( LOG_DEBUG, "----------- Redirect Servers --------");
    for (vector < Data > ::iterator i = itsRSList.begin();
            i < itsRSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "RS: %s", i->logData() );
    }

    cpLog( LOG_DEBUG, "----------- Policy Servers --------");
    for (vector < Data > ::iterator i = itsPOSList.begin();
            i < itsPOSList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "POS: %s", i->logData() );
    }

    cpLog( LOG_DEBUG, "----------- Cdr Servers --------");
    for (vector < Data > ::iterator i = itsCDRList.begin();
            i < itsCDRList.end(); ++i)
    {
        cpLog( LOG_DEBUG, "CDR: %s", i->logData() );
    }
}


/**
 * The Provisioning Server is notifying us of a change in RS's.
 */
void
updateHBRedirectServers(const string& data, const string& fileName, bool)
{
    /// update RS list
    HBServerContainer::instance().updateRSsFromPS();
}



/**
 * The Provisioning Server is notifying us of a change in MS's.
 */
void
updateHBMarshalServers(const string& data, const string& fileName, bool)
{
    /// update MS list
    HBServerContainer::instance().updateMSsFromPS();
}



/**
 * The Provisioning Server is notifying us of a change in FS's.
 */
void
updateHBFeatureServers(const string& data, const string& fileName, bool)
{
    /// update FS list
    HBServerContainer::instance().updateFSsFromPS();
}


/**
 * The Provisioning Server is notifying us of a change in JS's.
 */
void
updateHBJtapiServers(const string& data, const string& fileName, bool)
{
    /// update JS list
    HBServerContainer::instance().updateJSsFromPS();
}

/**
 * The Provisioning Server is notifying us of a change in POS's.
 */
void
updateHBPolicyServers(const string& data, const string& fileName, bool)
{
    /// update POS list
    HBServerContainer::instance().updatePOSsFromPS();
}

/**
 * The Provisioning Server is notifying us of a change in CDR's.
 */
void
updateHBCdrServers(const string& data, const string& fileName, bool)
{
    /// update CDR list
    HBServerContainer::instance().updateCDRsFromPS();
}

/**
 * The Provisioning Server is notifying us of a change in Heartbeat Parms.
 */
void
updateHBHeartbeatParms(const string& data, const string& fileName, bool)
{
    /// update Heartbeat Parms
    HBServerContainer::instance().updateHBParmsFromPS();
}

// End of File
