/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const PSProtocolService_cxx_Version =
    "$Id: PSProtocolService.cxx,v 1.4 2001/10/03 00:03:10 maivu Exp $";

#include <stdio.h>
#include <util/PlatformUtils.hpp>
#include <util/TransService.hpp>
#include <util/XMLException.hpp>
#include <sax2/SAX2XMLReader.hpp>
#include <sax2/XMLReaderFactory.hpp>
#include <framework/MemBufInputSource.hpp>

#include "VNetworkException.hxx"
#include "XmlContent.hxx"
#include "HttpHeader.hxx"
#include "PSProtocolService.hxx"
#include "PSReqXmlContentHandlers.hxx"
#include "PSRspXmlContentHandlers.hxx"
#include "cpLog.h"

#include "XmlBlobContentHandler.hxx"
#include "PSBlobResult.hxx"



static const char* saxMemBufId = "psResponse"; 

PSProtocolService* PSProtocolService::myInstance = 0;
///
PSProtocolService&
PSProtocolService::getInstance() throw (VException&)
{
   if (!myInstance)
   {
        myInstance  =  new PSProtocolService();
        try 
        {
            // init the XML4C2 system
            XMLPlatformUtils::Initialize();
        }
        catch (const XMLException& toCatch)
        {
	    string err = "Error during Initialization: ";
	    err += StrX(toCatch.getMessage()).localForm();
            cpLog(LOG_ERR, err.c_str());
	    VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	    throw exception;
        }
    }
    return *myInstance;
}


/// Destructor
PSProtocolService::~PSProtocolService()
{
    cpLog(LOG_DEBUG, "~PSProtocolService: terminate XML4C2 system");
    XMLPlatformUtils::Terminate();
}

///
Sptr<PSDBRequest>
PSProtocolService::parseRequest( PSMessage::msgType type, 
                                  const string& xmlBody )
                                  throw( VException& )  
{
     cpLog(LOG_DEBUG, "parseRequest: msgType= %d, xml=%s",
           (int) type, xmlBody.c_str());

     Sptr<PSDBRequest> request = new PSDBRequest( type );

     if (xmlBody.size() == 0)
     {
         cpLog( LOG_DEBUG, "0 length Xml content");
         return request;    /* empty result set */
     } 
     // now parse the xml content and fill in PSDBResult
     // create a SAX parser object.
     SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
     try
     {
         PSReqXmlContentHandlers handler( request );
         parser->setContentHandler(&handler);
         parser->setErrorHandler(&handler);
         MemBufInputSource* memBufIS = new MemBufInputSource(
            (const XMLByte*)xmlBody.c_str(), xmlBody.size(),
            saxMemBufId, false);
         parser->parse(*memBufIS);
         delete memBufIS;
     }
     catch (const XMLException& toCatch)
     {
         delete parser;
	 string err = "parseRequest Error: ";
	 err += StrX(toCatch.getMessage()).localForm();
         cpLog(LOG_ERR, err.c_str());
	 VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	 throw exception;
     }
     delete parser; 
     
     // debug:
     request->print( LOG_DEBUG );
     return request;
}



///
Sptr<PSDBResult>
PSProtocolService::parseResponse( PSMessage::msgType type, 
                                  const string& rawHttpMsg )
                                  throw( VException& )  
{
      cpLog(LOG_DEBUG, "parseResponse: msgType= %d, %s",
            (int)type, rawHttpMsg.c_str());

     Http::Message decodeMsg;

     try
     {
         decodeMsg.decode( rawHttpMsg );
     }
     catch ( Http::Exception & e )
     {
         string err = "parseResponse Error: " + e.getDescription();
	 cpLog (LOG_ERR, err.c_str());
	 VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	 throw exception;
     }

     Sptr<PSDBResult> result = new PSDBResult( type );

     // get status
     const Http::Status& stsObj = decodeMsg.status();
     result->setStatus( stsObj.code());    
     if (stsObj.isSuccess() == true)  // status code in the 200s
     {
         cpLog( LOG_DEBUG, "successful status: %d", stsObj.code());
     }
     else  // this is an error response
     {
     }
     /**  don't care for headers for now
     const list<Http::Header>& hdrs = decodeMsg.headers();
     if (hdrs.size() == 0)
     {
         cpLog( LOG_WARNING, "no Http headers decoded!");
         return result;
     }
     */
     const string& xmlBody = decodeMsg.body();
     if (xmlBody.size() == 0)
     {
         cpLog( LOG_DEBUG, "0 length Xml content");
         return result;    /* empty result set */
     } 
     // now parse the xml content and fill in PSDBResult
     // create a SAX parser object.
     SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
     try
     {
         PSRspXmlContentHandlers handler( result );
         parser->setContentHandler(&handler);
         parser->setErrorHandler(&handler);
         MemBufInputSource* memBufIS = new MemBufInputSource(
            (const XMLByte*)xmlBody.c_str(), xmlBody.size(),
            saxMemBufId, false);
         parser->parse(*memBufIS);
         delete memBufIS;
     }
     catch (const XMLException& toCatch)
     {
         delete parser;
	 string err = "parseResponse Error: ";
	 err += StrX(toCatch.getMessage()).localForm();
	 VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	 throw exception;
     }
     delete parser; 
     
     // debug:
     result->print( LOG_DEBUG );
     return result;
}

///
Sptr<PSBlobResult>
PSProtocolService::parseXmlBlob(const string& xmlBlob)
  throw(VException&)  
{
    cpLog (LOG_DEBUG, "parseXmlBlob: %s", xmlBlob.c_str());

    Sptr<PSBlobResult> result = new PSBlobResult();

    if (xmlBlob.size() == 0)
      {
	cpLog (LOG_DEBUG, "No XmlBlob given");
	return result;
     } 

    // create a SAX parser object.
    SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
    try
     {
       XmlBlobContentHandler handler(result);
       parser->setContentHandler(&handler);
       parser->setErrorHandler(&handler);
       MemBufInputSource* memBufIS = 
	 new MemBufInputSource((const XMLByte*)xmlBlob.c_str(), 
			       xmlBlob.size(),
			       saxMemBufId,
			       false);
       parser->parse(*memBufIS);
       delete memBufIS;
     }
    catch (const XMLException& toCatch)
      {
	delete parser;
	string err = "parseXmlBlob Error: ";
	err += StrX(toCatch.getMessage()).localForm();
	cpLog(LOG_ERR, err.c_str());
	VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	throw exception;
      }

    delete parser; 
    return result;
}


///
Sptr<Http::Message> 
PSProtocolService::buildHttpMessage( Http::Type mthod,
                       const string& psReqType,
                       const string& body )
{
    Http::Method method(mthod);   // GET/PUT/POST/DELETE
    Http::Url    url(psReqType);  // e.g. "GetRecordsInDomain"

    /// need other headers???
    list<Http::Header> headers;
    // Http::Header acceptCharset(Http::Header::AcceptCharset, 
    //                            "iso-8859-1,*,utf-8");
    // headers.push_back(acceptCharset);

    char buf[30];
    sprintf( buf, "%d", body.size());
    Http::Header contentLength(Http::Header::ContentLength, buf );
    headers.push_back(contentLength);

    // Assemble a HttpMessage
    Sptr<Http::Message> request = 
         new Http::Message(method, url, headers, body );

    const string& httpMsg = request->encode();
    cpLog(LOG_DEBUG, "Encoded Http Message: %s", httpMsg.c_str());

    return request;

}

// *****************************************************************
// Generic atomic data accessor functions
// *****************************************************************

/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildGetXmlDescriptionReq(
                                const string& xmlCategory, 
                                const string& xmlDescription)
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "buildGetXMLDescription() category %s, description %s",
                     xmlCategory.c_str(), xmlDescription.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildGetXmlDescriptionReq( xmlCategory,
                          xmlDescription);
  
    return( buildHttpMessage( Http::Method::GET,
                PSCommon::GET_XML_DESC_STR,
                *body ) ); 
}

/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildListXmlCatagoriesReq()
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "buildListXMLCatagories()");

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildListXmlCatagoriesReq();
  
    return( buildHttpMessage( Http::Method::GET,
                PSCommon::LIST_XML_CAT_STR,
                *body ) ); 
}


/** Retrieve an XML file decribing the data requested
 */
Sptr<Http::Message> PSProtocolService::buildListXmlDescriptionsReq(
    const string& xmlCategory)
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "buildGetXMLDescription() category %s",
                     xmlCategory.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;
  
    Sptr<string> body = builder.buildListXmlDescriptionsReq( xmlCategory );  
    
    return( buildHttpMessage( Http::Method::GET,
                PSCommon::LIST_XML_DESC_STR,
                *body ) ); 
}


/** retrive record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<Http::Message> PSProtocolService::buildGetRecordsInDomainReq(
                                const string& domain, 
                                const string& filterRules,
                                FieldListtype & fieldList)
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "buildGetRecordsInDomainReq() domain %s, rules %s",
                     domain.c_str(), filterRules.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */ 

    /// now build the body

    PSReqXmlContentHandlers builder;
  
    Sptr<string> body = builder.buildGetRecordsInDomainReq( domain,
                          filterRules, fieldList );
    
    return( buildHttpMessage( Http::Method::GET,
                PSCommon::GET_RECORDS_IN_DOMAIN_STR,
                *body ) ); 
}

///
Sptr<Http::Message> PSProtocolService::buildGetFieldsInDomainReq( 
                                const string& domain)
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "buildGetFieldsInDomainReq() domain %s",
                     domain.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */
    PSReqXmlContentHandlers builder;
  
    Sptr<string> body = builder.buildGetFieldsInDomainReq( domain );
  
    return( buildHttpMessage( Http::Method::GET,
                PSCommon::GET_FIELDS_IN_DOMAIN_STR,
                *body ) ); 

}

/// update existing records with keyword/value pairs provided
Sptr<Http::Message> 
PSProtocolService::buildSetRecordsInDomainReq(
                                const string& domain, 
                                const string& filterRules,
				const map<string, string>& fieldValPairs )
  throw (VException& )
{
    cpLog(LOG_DEBUG, "buildSetRecordsInDomainReq() domain=%s" 
                     " filter=%s",
                     domain.c_str(), filterRules.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSetRecordsInDomainReq( domain,
                            filterRules, fieldValPairs );

    return( buildHttpMessage( Http::Method::PUT,
                PSCommon::SET_RECORDS_IN_DOMAIN_STR,
                *body ) );

}


/// insert a new record
Sptr<Http::Message> 
PSProtocolService::buildAddRecordInDomainReq( const string& domain, 
                               const map< string, string>& fieldValPairs)
                               throw( VException& )
{
    cpLog(LOG_DEBUG, "buildAddRecordInDomainReq() domain %s",
                     domain.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildAddRecordInDomainReq( domain,
                            fieldValPairs );

    return( buildHttpMessage( Http::Method::PUT,
                PSCommon::ADD_RECORD_IN_DOMAIN_STR,
                *body ) );

}


Sptr<Http::Message>
PSProtocolService::buildDeleteRecordsInDomainReq( const string& domain,
						           const string& filterRules )
				                   throw( VException& )
{
    cpLog(LOG_DEBUG, "buildDeleteRecordsInDomainReq() domain %s "
                     "filterRules=%s",
                     domain.c_str(), filterRules.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildDeleteRecordsInDomainReq( domain,
                            filterRules );

    return( buildHttpMessage( Http::Method::DELETE,
                PSCommon::DELETE_RECORDS_IN_DOMAIN_STR,
                *body ) );

}

// *****************************************************************
// Session related functions
// *****************************************************************

Sptr<Http::Message> 
PSProtocolService::buildSessionOpenReq( int timeInSeconds ) 
                              throw(VException&)
{
    cpLog(LOG_DEBUG, "buildSessionOpenReq() expiry time= %d seconds",
                     timeInSeconds);

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionOpenReq( timeInSeconds );

    return( buildHttpMessage( Http::Method::GET,
                PSCommon::SESSION_OPEN_STR,
                *body ) );

}


///
Sptr<Http::Message>
 PSProtocolService::buildSessionCloseReq( const string& sessionID ) throw( VException& )
{
    cpLog(LOG_DEBUG, "buildSessionCloseReq() sessionID=%s ",
                     sessionID.c_str() );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionCloseReq( sessionID );

    return( buildHttpMessage( Http::Method::POST,
                PSCommon::SESSION_CLOSE_STR,
                *body ) );

}


///
Sptr<Http::Message> PSProtocolService::buildSessionSearchDBReq( 
                       const string& sessionID, const string& domain, 
                       const string& filterRules,
                       const FieldListtype& fieldList,  
                       PSCommon::PSSortType sType,
                       const string& sortField ) throw(VException&)
{
    cpLog(LOG_DEBUG, "buildSessionSearchDBReq() sessionID %s "
                     "domain=%s, filterRules=%s",
                     sessionID.c_str(), domain.c_str(), filterRules.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionSearchDBReq( sessionID, domain,
                            filterRules, fieldList, sType, sortField );

    return( buildHttpMessage( Http::Method::POST,
                PSCommon::SESSION_SEARCH_DB_STR,
                *body ) );

}


///
Sptr<Http::Message>
 PSProtocolService::buildSessionGetDBResultReq( const string& sessionID, int startIdx,
                       int endIdx, const FieldListtype& fieldList)
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "buildSessionGetDBResultReq() sessionID=%s "
                     "startIdx=%d endIdx=%d",
                     sessionID.c_str(), startIdx, endIdx );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionGetDBResultReq( sessionID,
                            startIdx, endIdx, fieldList );

    return( buildHttpMessage( Http::Method::GET,
                PSCommon::SESSION_GET_DB_RESULT_STR,   
                *body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionSetDBResultReq( const string& sessionID, 
                       const map< string, string>& fieldValPairs)
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "buildSessionSetDBResultReq() sessionID=%s ",
                     sessionID.c_str() );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionSetDBResultReq( sessionID,
                            fieldValPairs );

    return( buildHttpMessage( Http::Method::PUT,
                PSCommon::SESSION_SET_DB_RESULT_STR,
                *body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionSortDBResultReq( 
                       const string& sessionID, PSCommon::PSSortType sType,
                       const string& sortField)
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "buildSessionSortDBResultReq() sessionID=%s ",
                     sessionID.c_str() );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionSortDBResultReq( sessionID,
                            sType, sortField );

    return( buildHttpMessage( Http::Method::POST,
                PSCommon::SESSION_SORT_DB_RESULT_STR,
                *body ) );

}


///
Sptr<Http::Message>
PSProtocolService::buildSessionDeleteAllReq( const string& sessionID ) 
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "buildSessionDeleteAllReq() sessionID= %s ",
                     sessionID.c_str() );

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildSessionDeleteAllReq( sessionID );

    return( buildHttpMessage( Http::Method::DELETE,
                PSCommon::SESSION_DELETE_ALL_STR,
                *body ) );

}


Sptr<Http::Message> 
PSProtocolService::buildRegisterForUpdateReq( const string& type,
					      const string& hostport,
					      const string& domain,
					      const vector<string> & keyList,
					      callbackFunc func )
  throw (VException&)
{
    cpLog (LOG_DEBUG, "buil%s() hostport = %s, domain = %s",
	   type.c_str(), hostport.c_str(), domain.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildRegisterForUpdateReq(hostport, domain, keyList, func);

    return (buildHttpMessage(Http::Method::POST, type, *body));
}


Sptr<Http::Message> 
PSProtocolService::buildUpdateReq( const string& type,
				   const string& domain,
				   const vector<string>& keyList)
  throw (VException&)
{
    cpLog (LOG_DEBUG, "build-%s-Req() domain = %s",
	   type.c_str(), domain.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = builder.buildUpdateReq(domain, keyList);

    return (buildHttpMessage(Http::Method::PUT, type, *body));
}


Sptr<Http::Message> 
PSProtocolService::buildUnRegisterForUpdateReq( const string& type,
						const string& hostport,
						const string& domain,
						const vector<string> & keyList,
						callbackFunc func )
  throw (VException&)
{
    cpLog (LOG_DEBUG, "buildUnRegister-%s-ForUpdateReq() hostport = %s, domain = %s",
	   type.c_str(), hostport.c_str(), domain.c_str());

    /* build the XML content, then add the HTTP header, then
     * send the message to PServer 
     */

    /// now build the body
    PSReqXmlContentHandlers builder;

    Sptr<string> body = 
      builder.buildUnRegisterForUpdateReq(hostport, domain, keyList, func);

    return (buildHttpMessage(Http::Method::POST, type, *body));
}

