/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const PSHttpXMLInterface_cxx_Version =
    "$Id: PSHttpXMLInterface.cxx,v 1.4 2001/10/03 00:03:30 maivu Exp $";


#include "cpLog.h"
#include "VNetworkException.hxx"
#include "PSCommon.hxx"
#include "PSHttpXMLInterface.hxx"
#include "PSProtocolService.hxx"


ProvisioningInterface&
PSHttpXMLInterface::getInstance(const char* primHost, const int primPort,
                             const char* altHost, const int altPort) 
                             throw (VException&)
{
   if (!myInstance)
   {

        if (primHost == 0 || primPort == 0 || altHost == 0 || altPort == 0)
        {
	    string err = "PServer hostname and port not provided";
	    cpLog (LOG_ERR, err.c_str());
            VNetworkException exception(err.c_str(), __FILE__, __LINE__);
	    throw exception;
        }
        myInstance  =  new PSHttpXMLInterface(primHost, primPort,
                                           altHost, altPort);
        myInstance->setInterfaceType( PSCommon::HTTPXML );
    }
    return *myInstance;
}

///  A concrete implementation of ProvisioningInterface
PSHttpXMLInterface::PSHttpXMLInterface(
           const char* primHostname, const int primPort,
           const char* altHostname, const int altPort):
           ProvisioningInterface(primHostname, primPort,
                altHostname, altPort) 
{
     cpLog(LOG_DEBUG, "PSHttpXMLInterface constructor invoked");
}

/**
  Frees the memory associated the Instance.Gets register with
  atexit(), so that when program exits, it will be called automatically.
*/
void PSHttpXMLInterface::destroy()
{
    ProvisioningInterface::destroy();
}

/// Destructor
PSHttpXMLInterface::~PSHttpXMLInterface(){}


// *****************************************************************
// Generic atomic data accessor functions
// *****************************************************************

/** 
 */
Sptr<PSDBResult> PSHttpXMLInterface::getXmlDescription(
    const string& xmlCategory, 
    const string& xmlDescription)
    throw( VException& ) 
{
    cpLog(LOG_DEBUG, "getXmlDescription category %s, description %s",
	  xmlCategory.c_str(), xmlDescription.c_str());
    
    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildGetXmlDescriptionReq( xmlCategory, 
								 xmlDescription);
    
    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
    
    return(proto.parseResponse( PSMessage::GET_XML_DESC_RSP, resp ));
}


/** 
 */
Sptr<PSDBResult> PSHttpXMLInterface::listXmlCatagories()
    throw( VException& ) 
{
    cpLog(LOG_DEBUG, "listXmlCatagories");

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildListXmlCatagoriesReq();
    
    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
    
    return(proto.parseResponse( PSMessage::LIST_XML_CAT_RSP, resp ));

}


/** 
 */
Sptr<PSDBResult> PSHttpXMLInterface::listXmlDescriptions(
    const string& xmlCategory)
    throw( VException& ) 
{
    cpLog(LOG_DEBUG, "listXmlDescriptions category %s",
                     xmlCategory.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildListXmlDescriptionsReq( xmlCategory);

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::LIST_XML_DESC_RSP, resp ));
}


/** retrive record(s) matching the domain, filter and fields of interest
 *  If fieldList is empty, then get the entire record
 */
Sptr<PSDBResult> PSHttpXMLInterface::getRecordsInDomain( const string& domain, 
							 const string& filterRules,
							 FieldListtype & fieldList )
  throw( VException& ) 
{
    cpLog(LOG_DEBUG, "getRecordsInDomain domain %s, rules %s",
                     domain.c_str(), filterRules.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildGetRecordsInDomainReq( domain, filterRules, fieldList );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::GET_RECORDS_IN_DOMAIN_RSP, resp ));
}

///
Sptr<PSDBResult>
PSHttpXMLInterface::getFieldsInDomain( const string& domain )
                                throw( VException& ) 
{
    cpLog(LOG_DEBUG, "getFieldsInDomain domain %s",
                     domain.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildGetFieldsInDomainReq( domain );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );

    return(proto.parseResponse( PSMessage::GET_FIELDS_IN_DOMAIN_RSP, resp ));
}


/// update existing records with keyword/value pairs provided
Sptr<PSDBResult> PSHttpXMLInterface::setRecordsInDomain( const string& domain, 
                                string filterRules,
                                const map<string, string>& fieldValPairs )
				                throw (VException& )
{
    cpLog(LOG_DEBUG, "setRecordsInDomain domain %s, rules %s",
                     domain.c_str(), filterRules.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildSetRecordsInDomainReq( domain, filterRules, fieldValPairs );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );

    return(proto.parseResponse( PSMessage::SET_RECORDS_IN_DOMAIN_RSP, resp ));
}

/// insert a new record
Sptr<PSDBResult> PSHttpXMLInterface::addRecordInDomain( const string& domain, 
                               const map< string, string>& fieldValPairs)
                               throw( VException& )
{
    cpLog(LOG_DEBUG, "addRecordsInDomain domain %s",
                     domain.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildAddRecordInDomainReq( domain, fieldValPairs );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::ADD_RECORD_IN_DOMAIN_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::deleteRecordsInDomain( 
                                   const string& domain,
						           const string& filterRules )
				                   throw( VException& ) 
{
    cpLog(LOG_DEBUG, "deleteRecordsInDomain domain %s, rules %s",
                     domain.c_str(), filterRules.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildDeleteRecordsInDomainReq( domain, filterRules );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::DEL_RECORDS_IN_DOMAIN_RSP, resp ));
}


// *****************************************************************
// Session related functions
// *****************************************************************

// string sessionId sessionOpen( int timeInSeconds ) throw(VException&)
string PSHttpXMLInterface::sessionOpen( int timeInSeconds ) throw(VException&)
{
    cpLog(LOG_DEBUG, "sessionOpen() expire time= %d seconds",
                     timeInSeconds );

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildSessionOpenReq( timeInSeconds );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse( PSMessage::SESSION_OPEN_RSP, resp );

    string sessionId = "";
    if (result != 0)
      {
	sessionId = result->getSessionID();
      }

    return sessionId;
}

///
bool PSHttpXMLInterface::sessionClose( const string& sessionID ) throw( VException& )
{
    cpLog(LOG_DEBUG, "sessionClose sessionID = %s ", sessionID.c_str() ); 

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildSessionCloseReq( sessionID );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse( PSMessage::SESSION_CLOSE_RSP, resp );
   
    // if status code == 200 OK, return true
    if ((result !=0) && (result->getStatus() == 200))
    {
        return(true);
    }
    else
    {
        return(false);
    }
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionSearchDB( 
                       const string& sessionID, 
                       const string& domain, const string& filterRules,
                       const FieldListtype& fieldList,  
                       PSCommon::PSSortType sType,
                       const string& sortField ) throw(VException&)
{
    cpLog(LOG_DEBUG, "sessionSearchDB session %s, domain %s, rules %s",
                     sessionID.c_str(), domain.c_str(), filterRules.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildSessionSearchDBReq( sessionID, domain,
				     filterRules, fieldList, 
				     sType, sortField );
    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::SESSION_SEARCH_DB_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionGetDBResult( 
                       const string& sessionID, int startIdx,
                       int endIdx, const FieldListtype& fieldList)
                       throw (VException&) 
{
    cpLog(LOG_DEBUG, "sessionGetDBResult sessionID %s, startIdx %d, endIdx %d",
                     sessionID.c_str(), startIdx, endIdx );

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
     Sptr< Http::Message > req = 
       proto.buildSessionGetDBResultReq( sessionID, startIdx, 
					 endIdx, fieldList );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::SESSION_GET_DB_RESULT_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionSetDBResult( const string& sessionID, 
                       const map< string, string>& fieldValPairs)
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "sessionSetDBResult sessionID %s",
                     sessionID.c_str() );

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildSessionSetDBResultReq( sessionID, fieldValPairs );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::SESSION_SET_DB_RESULT_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionSortDBResult( 
                       const string& sessionID, PSCommon::PSSortType sType,
                       const string& sortField)
                       throw (VException&) 
{
    cpLog(LOG_DEBUG, "sessionSortDBResult sessionID %s", sessionID.c_str() );

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = 
      proto.buildSessionSortDBResultReq( sessionID, sType, sortField );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::SESSION_SORT_DB_RESULT_RSP, resp ));
}

///
Sptr<PSDBResult> PSHttpXMLInterface::sessionDeleteAll( const string& sessionID ) 
                       throw (VException&)
{
    cpLog(LOG_DEBUG, "sessionDeleteAll sessionID %s", sessionID.c_str() );

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr< Http::Message > req = proto.buildSessionDeleteAllReq( sessionID );

    // exception will be thrown if bad request
    string resp = myInterface.sendRequest( req->encode() );
   
    return(proto.parseResponse( PSMessage::SESSION_DELETE_ALL_RSP, resp ));
}

// *****************************************************************
// Registration functions
// *****************************************************************

///
void
PSHttpXMLInterface::registerDomainForUpdate( const string& hostport,
					     const string& domain,
					     callbackFunc func )
  throw (VException&)
{
    // Note: if the same hostport, domain, but different callback func ???
    // not support yet.

    cpLog (LOG_DEBUG, "RegisterDomainForUpdate - domain = %s", domain.c_str());

    bool flagDelete = false;
    if (SetCallbackList(flagDelete, myDomainCallbackMap, domain, func) == false)
      {
	cpLog (LOG_DEBUG, "Hostport and Domain are already registered.  Save the callback func.");
	return;
      }

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    vector<string> dummy;

    Sptr<Http::Message> request = 
      proto.buildRegisterForUpdateReq(PSCommon::REGISTER_DOMAIN_FOR_UPDATE_STR,
				      hostport,
				      domain, 
				      dummy,
				      func);

    // if thrown exception, client needs to catch it.
    string response = myInterface.sendRequest(request->encode());

    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse(PSMessage::REGISTER_DOMAIN_FOR_UPDATE_RSP, 
				 response);

    if ((result != 0) && (result->getError().size() != 0))
    {
        cpLog (LOG_ERR, result->getError().c_str());
    }
    return;
}

///
void
PSHttpXMLInterface::registerRecordForUpdate( const string& hostport,
					     const string& domain,
					     const vector<string>& keyList, 
					     callbackFunc func)
  throw (VException&)
{

    cpLog (LOG_DEBUG, "RegisterRecordForUpdate - domain = %s", 
	   domain.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<Http::Message> request = 
      proto.buildRegisterForUpdateReq(PSCommon::REGISTER_RECORD_FOR_UPDATE_STR,
				      hostport,
				      domain, 
				      keyList, 
				      func);

    // if thrown exception, client needs to catch it.
    string response = myInterface.sendRequest(request->encode());

    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse(PSMessage::REGISTER_RECORD_FOR_UPDATE_RSP, 
				 response);

    if ((result != 0) && (result->getError().size() != 0))
    {
        cpLog (LOG_ERR, result->getError().c_str());
	return;
    }

    string temp;
    bool flagDelete = false;
    for (unsigned short i = 0; i < keyList.size(); i++)
      {
	temp = domain + keyList[i];
	SetCallbackList(flagDelete, myRecordCallbackMap, temp, func);
      }

    return;
}
 
///                       
void 
PSHttpXMLInterface::unRegisterDomainForUpdate( const string& hostport,
					       const string& domain,
					       callbackFunc func ) 
  throw (VException&)
{
    cpLog (LOG_DEBUG, "UnRegisterDomainForUpdate - domain: %s", domain.c_str());

    bool flagDelete = true;
    if (SetCallbackList(flagDelete, myDomainCallbackMap, domain, func) == false)
      {
	cpLog (LOG_DEBUG, "More than one callback func registered. No need to send request now.");
	cpLog (LOG_DEBUG, "Just remove the unregistered callback from the callbackMap.");
	return;
      }


    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    vector<string> dummy;

    Sptr<Http::Message> request = 
      proto.buildUnRegisterForUpdateReq(PSCommon::UN_REGISTER_DOMAIN_FOR_UPDATE_STR,
					hostport,
					domain,
					dummy,
					func);

    // if thrown exception, client needs to catch it.
    string response = myInterface.sendRequest(request->encode());

    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse(PSMessage::UN_REGISTER_DOMAIN_FOR_UPDATE_RSP, 
				 response);

    if ((result != 0) && (result->getError().size() != 0))
    {
        cpLog (LOG_ERR, result->getError().c_str());
    }
    return;
}

///
void
PSHttpXMLInterface::unRegisterRecordForUpdate( const string& hostport,
					       const string& domain,
					       const vector<string>& keyList, 
					       callbackFunc func)
  throw (VException&)
{

    cpLog (LOG_DEBUG, "UnRegisterRecordForUpdate - domain: %s", domain.c_str());

    /* let PSProtocolService helper class build the XML content
     * and the Http header(s)       
     */ 
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<Http::Message> request = 
      proto.buildUnRegisterForUpdateReq(PSCommon::UN_REGISTER_RECORD_FOR_UPDATE_STR,
					hostport,
					domain,
					keyList,
					func);

    // if thrown exception, client needs to catch it.
    string response = myInterface.sendRequest(request->encode());

    Sptr<PSDBResult> result = 0;
    result = proto.parseResponse(PSMessage::UN_REGISTER_RECORD_FOR_UPDATE_RSP, 
				 response);

    if ((result != 0) && result->getError().size() != 0)
    {
        cpLog (LOG_ERR, result->getError().c_str());
	return;
    }

    string temp = "";
    bool flagDelete = true;
    // for each domain+key, search and remove from the callbackRecord Map
    for (unsigned short i = 0; i < keyList.size(); i++)
      {
	temp = domain + keyList[i];
	SetCallbackList(flagDelete, myRecordCallbackMap, temp, func);
      }

}


///
void
PSHttpXMLInterface::handleRequest(const string& request)
  throw (VException&)
{
    Http::Message httpMsg;

    // exception will be thrown if bad request
    httpMsg.decode(request.c_str());

    // For now, handle the PUT request only
    if (httpMsg.method().type() == Http::Method::PUT)
    {
        HandlePUTRequest(httpMsg);
    }
    else
      cpLog (LOG_ERR, "Don't know the request: %s", httpMsg.method().type());

}

// does it really need a seperate PUT request ????
// what else it could be ??? GET/PUT/DELETE/POST
void
PSHttpXMLInterface::HandlePUTRequest(const Http::Message& httpMsg)
{
    cpLog (LOG_DEBUG, "PUT request: URL = %s", httpMsg.url().url().c_str());

    PSProtocolService& protocol = PSProtocolService::getInstance();

    if (httpMsg.url().url() == PSCommon::UPDATE_DOMAIN_STR)
    {
        UpdateDomainCallback(protocol.parseRequest(PSMessage::UPDATE_DOMAIN_REQ,
						   httpMsg.body()));
    }
    else if (httpMsg.url().url() == PSCommon::UPDATE_RECORD_STR)
    {
        UpdateRecordCallback(protocol.parseRequest(PSMessage::UPDATE_RECORD_REQ,
						   httpMsg.body()));
    }
    else
    {
	cpLog (LOG_DEBUG, "Unrecognize PUT function");
    }
    return;
}


void
PSHttpXMLInterface::UpdateDomainCallback(Sptr <PSDBRequest> request)
{
   cpLog (LOG_DEBUG, "Update Domain callback: %s", request->getDomain().c_str());

    string domain = request->getDomain();
    callbackMap::iterator iter = myDomainCallbackMap.find(domain);

    if (iter == myDomainCallbackMap.end())
    {
	cpLog(LOG_ERR, "No Domain \"%s\" register for update",
	      domain.c_str());
	return;
    }

    vector<string> dummy;
    callbackList aList = iter->second;
    cpLog (LOG_DEBUG, "The callback list [%s] has %d items.",
	   domain.c_str(), aList.size());
    
    for (unsigned short i = 0; i < aList.size(); i++)
    {
	(*aList[i])(domain, dummy );
    }
    
}


void
PSHttpXMLInterface::UpdateRecordCallback(Sptr <PSDBRequest> request)
{
    cpLog (LOG_DEBUG, "Update Record callback: %s", request->getDomain().c_str());

    string temp;
    string domain = request->getDomain();
    vector<string> keyList = request->getFieldList();
    for (unsigned short i = 0; i < keyList.size(); i++)
    {
	temp = domain + keyList[i];
	callbackMap::iterator iter = myRecordCallbackMap.find(temp);
	if (iter == myRecordCallbackMap.end())
	{
	    cpLog (LOG_ERR, "No Domain \"%s\" and key \"%s\" register for update",
		   domain.c_str(), keyList[i].c_str());
	    return;
	}
	callbackList aList = iter->second;
	
	for (unsigned short j = 0; j < aList.size(); j++)
	{
	    (*aList[j])(domain, keyList);
	}
	
    }

}

// Add new callback entry to the map.  Or remove a callback from the map
// acording to the flagDelete.
bool
PSHttpXMLInterface::SetCallbackList(const bool& flagDelete,
				    callbackMap& aMap,
				    const string& key,
				    const callbackFunc& func)
{
    bool bReturn = true;
    callbackMap::iterator iter = 0;
    iter = aMap.find(key.c_str());

    if (! flagDelete)
    {
	if (iter != aMap.end())
	{
	    // append to the end of the callbackList
	    iter->second.push_back(func);
	    cpLog (LOG_DEBUG, "Append to the end of list");
	    bReturn = false; // b/c key exists, don't send the request again.
	    return bReturn;
	}
	else
	{
	    callbackList aList;
	    aList.push_back(func);
	    aMap[key] = aList;
	    cpLog (LOG_DEBUG, "Add a new callback function to the list");
	    return bReturn;
	}
    }
    else // delete
    {
	if (iter == aMap.end())
	{
	    cpLog (LOG_ERR, "No registration for this key: %s", key.c_str());
	    return bReturn;
	}
	// vector<callbackFunc>
	callbackList& aList = iter->second;
	cpLog (LOG_DEBUG, "Size of callback list for key[%s] is = %d", 
	       key.c_str(), aList.size());
	callbackList::iterator i = aList.begin();
	
	for (unsigned short u = 0; u < aList.size(); u++, i++)
	{
	    if (*func == *i)
	    {
		cpLog (LOG_DEBUG, "Remove this callback b/c it unregistered");
		aList.erase(i);
		if (aList.size() != 0)
		  {
		    cpLog (LOG_DEBUG, "There is more one callback for this key.");
		    bReturn = false;
		    return bReturn;
		  }
	    }
	}
	cpLog (LOG_DEBUG, "After unregistered, the size of callback list = %d", 
	       aList.size());
	return bReturn;
    }

    return bReturn;
}
