#ifndef PROVISIONINGINTERFACE_HXX
#define PROVISIONINGINTERFACE_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisioningInterface_hxx_Version =
    "$Id: ProvisioningInterface.hxx,v 1.2 2001/09/19 20:32:34 maivu Exp $";

#include <memory>
#include <list>
#include <map>
#include <string>
// #include <vector>

#include "PSInterface.hxx"
#include "VMutex.h"

#include "Data.hxx"
#include "PSCommon.hxx"
#include "PSDBResult.hxx"
#include "PSDBCommonType.h"


class NetworkAddress;
class VException;
class PSListener;



/** Singelton Facade class - Only one instance is available per application.
    Initial access of the instance takes in a PServer Client ID to properly
    instantiate the correct  PServer Client Interface.  PRD 2.0 calls
    for a provisioning interface that sends HTTP encapsulated XML messages
    to the server.  In the future, if the requirements change to use
    SNMP or Java Spaces as the communication protocol, the API defined in 
    ProvisioningInterface will remain unchanged, but a different child
    class will be instantiated (see usage in PSClientCreator.hxx)

    Implements the communications to the provision server
    and notify content update as it happens on the Provision server.
    In order to receive notification client has to register
    itself with the class with a callback function along with the item name
    for which update notification is requested. Whenever the item changes on
    the provision server, the callback function is called with the content
    of the item passed as the argument.
*/

class ProvisioningInterface
{
    public:

        /**
           Frees the memory associated the Instance.Gets register with
           atexit(), so that when program exits, it will be called automatically.
        */
        static void destroy();

        /// Destructor
        virtual ~ProvisioningInterface();

        virtual const char* getClassName() const { 
            return "ProvisioningInterface"; 
        }

        ///
        PSCommon::PSIFType getInterfaceType() const { return myIFType; }

        ///
        void setInterfaceType(PSCommon::PSIFType typ) 
        { 
            myIFType = typ;
        }

        /// define the listening port
        void setListenPort(int port)
	  throw (VException&);

        // *****************************************************************
        // Generic atomic data accessor functions
        // *****************************************************************
    
        /// retrive record(s) matching the domain, filter and fields of interest
        virtual Sptr<PSDBResult> getXmlDescription( const string& xmlCategory, 
                                                    const string& xmlDescription)
                                                     throw( VException& ) =0;

        /// retrive record(s) matching the domain, filter and fields of interest
        virtual Sptr<PSDBResult> listXmlCatagories()
                                                     throw( VException& ) =0;

        /// retrive record(s) matching the domain, filter and fields of interest
        virtual Sptr<PSDBResult> listXmlDescriptions( const string& xmlCategory) 
                                                     throw( VException& ) =0;



        /// retrive record(s) matching the domain, filter and fields of interest
        virtual Sptr<PSDBResult> getRecordsInDomain( const string& domain, 
                                                     const string& filterRules,
                                                     FieldListtype & fieldList) 
                                                     throw( VException& ) =0;

        ///
        virtual  Sptr<PSDBResult> getFieldsInDomain( const string& domain )
                                                     throw( VException& ) = 0;

        /// update existing records with keyword/value pairs provided
        virtual Sptr<PSDBResult>  setRecordsInDomain( const string& domain, 
                                                      string filterRules,
                                                      const map<string, string>& fieldValPairs )
                                                      throw (VException& ) = 0;

        /// insert a new record
        virtual Sptr<PSDBResult> addRecordInDomain( const string& domain, 
                                                    const map< string, string>& fieldValPairs)
                                                    throw( VException& ) = 0;

        ///
        virtual Sptr<PSDBResult> deleteRecordsInDomain( const string& domain,
                                                        const string& filterRules )
				                        throw( VException& ) = 0;

        // *****************************************************************
        // Session related functions
        // *****************************************************************

        /** psServer returns -1 if PServer does not return
         *  otherwise, return value is the sessionID
         */
        virtual string sessionOpen( int timeInSeconds ) throw(VException&) = 0; 

        ///
        virtual bool sessionClose( const string& sessionID ) throw( VException& ) = 0;

        ///
        virtual Sptr<PSDBResult>  sessionSearchDB( 
                       const string& sessionID, 
                       const string& domain, const string& filterRules,
                       const FieldListtype& fieldList, 
                       PSCommon::PSSortType sType,
                       const string& sortField ) throw(VException&) = 0;

        ///
        virtual Sptr<PSDBResult> sessionGetDBResult( 
                       const string& sessionID, int startIdx,
                       int endIdx, const FieldListtype & fieldList)
                       throw (VException&) = 0;

        ///
        virtual  Sptr<PSDBResult> sessionSetDBResult( const string& sessionID, 
                       const map< string, string>& fieldValPairs)
                       throw (VException&) = 0;

        ///
        virtual Sptr<PSDBResult> sessionSortDBResult( 
                       const string& sessionID, PSCommon::PSSortType sType,
                       const string& sortField )
                       throw (VException&) = 0;

        ///
        virtual Sptr<PSDBResult> sessionDeleteAll( const string& sessionID ) 
                       throw (VException&) = 0;


        // *****************************************************************
        // Registration functions
        // *****************************************************************

        /**
           Register client for notification for a given domain
           Callback func get called whenever item is updated.
        */

        virtual void registerDomainForUpdate( const string& hostport,
					      const string& domain,
					      callbackFunc func ) 
	  throw (VException&) = 0;

        virtual void registerRecordForUpdate( const string& hostport,
					      const string& domain,
					      const vector<string>& keyList, 
					      callbackFunc func )
	  throw (VException&) = 0;
                       
        virtual void unRegisterDomainForUpdate( const string& hostport,
						const string& domain,
						callbackFunc func ) 
	  throw (VException&) = 0;

        virtual void unRegisterRecordForUpdate( const string& hostport,
						const string& domain,
						const vector<string>& keyList, 
						callbackFunc func)
	  throw (VException&) = 0;
                       

        /// Get the network address of the currently used pServer
        const NetworkAddress& getNetAddress()
        {
            return myInterface.getNetAddress();
        }

    protected:

        friend class PSIFCreator;

        /// pointer to the singleton instance
        static ProvisioningInterface* myInstance;

        /// interface to the PS
        PSInterface myInterface;

        /**
           Constructor initializes with the name of provision server
           for redundancy passed as command line argument. Intentionally
           made private so that it can not be instantiated directly.
           It also creates an instance of listener in a separate thread
           to listen for data updates (through the interface)
        */
        ProvisioningInterface(const char* primHostname, const int primPort,
                           const char* altHostname, const int altPort);

        ///
        virtual void handleRequest(const string& request) = 0;

        /// Map to contain callbacks for a given data item
        callbackMap myDomainCallbackMap;
        callbackMap myRecordCallbackMap;


    private:

        /// suppress equal operator
        bool operator==( const ProvisioningInterface& rhs );

        /// suppress copy contructor
        ProvisioningInterface( const ProvisioningInterface& rhs );

        /// suppress assignment operator
        ProvisioningInterface& operator=( const ProvisioningInterface& rhs );


        /// Provisioning Interface type
        PSCommon::PSIFType myIFType;

        /// Host name the library is running on
        string myHostName;

        /// make mySessionMap thread safe
        VMutex mySessionMutex;

        /// make myCallBackMap thread safe
        VMutex myCallBackMutex;

friend class PSListener;

        /// listener for updates from server
        PSListener* myListener;

};

#endif // PROVISIONINGINTERFACE_HXX

