/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const CacheManager_cxx_Version =
    "$Id: CacheManager.cxx,v 1.1.1.1 2001/08/21 23:30:24 dabryan Exp $";


#include "CacheManager.hxx"
#include "cpLog.h"
#include <stdio.h>
#include <unistd.h>

#ifdef USE_YARROW
#include "YarrowPseudoRandom.hxx"
#endif

#include "CryptoRandom.hxx"
#include "support.hxx"

#define NUM_RANDOMNESS 16  // 128 bits of randomness



#define CACHE_DELETE_TIME  1800     // in seconds (1800 sec = 30 min) 
#define CACHE_POLLING_TIME 1        // in seconds


VRWLock CacheManager::cacheLock;



void*
CacheManager::startCacheManagerThread(void* args)
{
    cpLog (LOG_DEBUG, "Starting Cache Manager Thread");
    CacheManager* self = (CacheManager*) args;
    self->run();
    return NULL;
}


void
CacheManager::run()
{
    for (; ; )
    {
        cleanCache();
        usleep( CACHE_POLLING_TIME * 1000000); // convert to seconds
    }
}

CacheManager::CacheManager()
{
    pthread_create( &myPid, 0, CacheManager::startCacheManagerThread, (void*) this );
}

CacheManager::~CacheManager()
{

}

timeval
CacheManager::getTimeStamp()
{
    timeval now;

    int err = gettimeofday( &now, NULL );
    assert( !err );

    return now;
}


string
CacheManager::genSessionId()
{

#ifdef USE_YARROW
    YarrowPseudoRandom random;
#else
    CryptoRandom random;
#endif

    unsigned char tempId[NUM_RANDOMNESS];
    string sessionid;
    unsigned char temp;

    int len = random.getRandom( tempId, NUM_RANDOMNESS);

    if (len > 0)
    {
        for (int i = 0; i < NUM_RANDOMNESS; i++)
        {
            temp = tempId[i];
            int hi = (temp& 0xf0) /16;
            int low = (temp& 0xf);

            char buf[4];
            buf[0] = '\0';

            sprintf(buf, "%x%x", hi, low);
            sessionid +=buf;   
        }
    }

    return sessionid;
}

void
CacheManager::addCacheRequest( const string sessionId,
                               const string domain,
                               const string filter,
                               const string sortType,
                               const string sortField,
                               DBResultset* resultset)
{
    // Populate the data structure to be cached.

    // Create a new ProvData... need to delete when removing from cache.

    ProvData* pData  = new ProvData;
    pData->sessionId = sessionId;
    pData->domain    = domain;
    pData->filter    = filter;
    pData->sortType  = sortType;
    pData->sortField = sortField;
    pData->resultset = resultset;
    pData->timeStamp = getTimeStamp();

    cpLog( LOG_DEBUG, "TimeStamp():  sec = %ld, usec = %ld", pData->timeStamp.tv_sec,
                                                             pData->timeStamp.tv_usec );

    // insert into map
    cacheLock.WriteLock();

    myCache[ sessionId.c_str() ] = pData;

    cacheLock.Unlock();
}


int
CacheManager::setCache( const string sessionId,
                        const string domain,
                        const string filter,
                        const string sortType,
                        const string sortField )
{
    int retCode = 1;

    ProvData *pData;

    cacheLock.WriteLock();

    ProvRequestMap::iterator itr = myCache.find ( sessionId.c_str() );

    if ( itr!= myCache.end() )
    {
        pData = itr->second;
        pData->domain     = domain;
        pData->filter     = filter;
        pData->sortType   = sortType;
        pData->sortField  = sortField;
        pData->timeStamp  = getTimeStamp();
    }
    else
    {
        cpLog( LOG_DEBUG, "Error setting cache...  Session ID not found");        
        retCode = 0;
    }

    cacheLock.Unlock();

    return retCode; 
}

int
CacheManager::modifyCache( const string sessionId,
                           const string domain,
                           const string filter,
                           const string sortType,
                           const string sortField )
{
    int retCode = 1;
    cacheLock.WriteLock();

    ProvRequestMap::iterator itr = myCache.find ( sessionId.c_str() );
    ProvData* pData = itr->second;

    if ( itr != myCache.end() )
    {
        cpLog( LOG_DEBUG, "*** Modifying Cache Entry <%s> ***", sessionId.c_str() );
        if ( domain != "")
        {
            cpLog( LOG_DEBUG, "domain (old) = %s", pData->domain.c_str() );
            cpLog( LOG_DEBUG, "domain (new) = %s", domain.c_str() );  
            pData->domain = domain;
        }

        if ( filter != "")
        {
            cpLog( LOG_DEBUG, "filter (old) = %s", pData->filter.c_str() );  
            cpLog( LOG_DEBUG, "filter (new) = %s", filter.c_str() );

            pData->filter = filter;
        }

        if ( sortType != "")
        {
            cpLog( LOG_DEBUG, "sortType (old) = %s", pData->sortType.c_str() ); 
            cpLog( LOG_DEBUG, "sortType (new) = %s", sortType.c_str() );

            pData->sortType = sortType;
        }

        if ( sortField != "")
        {
            cpLog( LOG_DEBUG, "sortField (old) = %s", pData->sortField.c_str() );        
            cpLog( LOG_DEBUG, "sortField (new) = %s", sortField.c_str() );

            pData->sortField = sortField;
        }

        pData->timeStamp = getTimeStamp();

        cpLog (LOG_DEBUG, "***  End Modify ***");
    }
    else
    {
        cpLog( LOG_DEBUG, "Error modifing cache... Session ID not found");
        retCode = 0;
    }

    cacheLock.Unlock();
    return retCode;

}

void
CacheManager::getCacheRequest( const string sessionId, ProvData& pData )
                               throw (VMissingDataException&)
{
    // NOTE:  This only locks and unlocks the cache.  The struct still
    //        needs to be lock when accessing.

    cacheLock.ReadLock();
    ProvRequestMap::iterator itr = myCache.find( sessionId.c_str() );
    cacheLock.Unlock();
    
    if ( itr != myCache.end() )
    {
        // returning an actual copy
        pData.domain    = (itr->second)->domain.c_str();
        pData.filter    = (itr->second)->filter.c_str();
        pData.sortType  = (itr->second)->sortType.c_str();
        pData.sortField = (itr->second)->sortField.c_str();

        pData.resultset = (itr->second)->resultset;  // not safe... copying the ptr
        pData.sessionId = (itr->second)->sessionId;
        pData.timeStamp = (itr->second)->timeStamp;
    }
    else
    {
        throw VMissingDataException (
              "CacheError:  SessionID not found", __FILE__, __LINE__, 0 );
    }
}

int 
CacheManager::deleteCacheRequest( string sessionId )
{
    ProvRequestMap::iterator itr = myCache.find ( sessionId.c_str() );

    cacheLock.WriteLock();

    if ( itr != myCache.end() )
    {
        cpLog( LOG_DEBUG, "*** Session id <%s> delete ***", sessionId.c_str() );

        // delete the DBResult
        if ( (itr->second)->resultset != 0 )
        {
            delete (itr->second)->resultset;
        };

        // delete ProvData
        delete itr->second;

        // remove from cache.
        myCache.erase( itr );

        cacheLock.Unlock();
        return 1;
    }
    else
    {
        cpLog( LOG_DEBUG, "*** Session id <%s> not in cache ***", sessionId.c_str() );
        cacheLock.Unlock();
        return 0;
    }
}


void
CacheManager::showCacheEntry(const string sessionID)
{
    cacheLock.ReadLock();
    ProvRequestMap::iterator itr = myCache.find( sessionID.c_str() );
    cacheLock.Unlock();

    if ( itr != myCache.end() )
    {
        cpLog(LOG_DEBUG, "--=--=--=--=--=--=--=--=--=--=--=--=--");
        cpLog(LOG_DEBUG, " Cache Entry");
        cpLog(LOG_DEBUG, " sessionID:     %s", sessionID.c_str() );
        cpLog(LOG_DEBUG, " domain:        %s", (itr->second)->domain.c_str() ); 
        cpLog(LOG_DEBUG, " filter:        %s", (itr->second)->filter.c_str() ); 
        cpLog(LOG_DEBUG, " sortType:      %s", (itr->second)->sortType.c_str() ); 
        cpLog(LOG_DEBUG, " sortField:     %s", (itr->second)->sortField.c_str() ); 
        cpLog(LOG_DEBUG, "--=--=--=--=--=--=--=--=--=--=--=--=--");
    }
    else
    {
        cpLog(LOG_DEBUG, "*** session ID <%s> not found ***", sessionID.c_str() );
    }
}

void
CacheManager::cleanCache()
{
    // cpLog( LOG_DEBUG, "*** clean cache ***");

    cacheLock.WriteLock();

    ProvData* pData;
    ProvRequestMap::iterator itr = myCache.begin();

    timeval now;
    int err = gettimeofday( &now, NULL);
    assert (!err);

    while( itr != myCache.end() )
    {
        timeval now;
        int err = gettimeofday( &now, NULL);
        assert (!err);

        pData = itr->second;

        if ( (now.tv_sec - pData->timeStamp.tv_sec) > CACHE_DELETE_TIME)
        {
            cpLog( LOG_DEBUG, "Clean Moo:  ** sec %ld **, Now: ** sec = %ld **", 
                              pData->timeStamp.tv_sec, now.tv_sec );

            ProvRequestMap::iterator itr2 = itr ++;

            // deletes DBResultset
            if ( (itr2->second)->resultset != 0)
            {
                delete (itr2->second)->resultset;
            }

            // deletes ProvData
            delete itr2->second;

            // remove from cache
            myCache.erase( itr2 );
            cpLog( LOG_DEBUG, "Cache size (after delete):  %d", myCache.size() );            
        }
        else
        {
            itr ++;
        }
    }

    cacheLock.Unlock();
}

