
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */





static const char* const FileStore_cxx_Version =
    "$Id: FileStore.cxx,v 1.1.1.1 2001/08/21 23:30:24 dabryan Exp $";


#include <assert.h>
#include <cerrno>
#include <cstdio>

#include "FileStore.hxx"
#include "VIoException.hxx"


FileStore::FileStore( const string &fileRootP ) :
        fileRoot(fileRootP)
{

    lock.WriteLock();
    try
    {
        // check the root file system exists
        if ( !VFileSystem::dirExists( fileRoot ) )
        {
            throw VIoException( strerror(errno),
                                __FILE__,
                                __LINE__,
                                errno );
        }

    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();
}


FileStore::~FileStore()
{}


string
FileStore::dirName( const string& category ) const
{
    string ret = fileRoot;
    ret += "/";
    ret += category;
    return ret;
}


string
FileStore::fileName(const string& category, const string& name ) const
{
    string ret = dirName(category);
    ret += "/";
    ret += name;
    return ret;
}


string
FileStore::getItem( const string& category, const string& name )
throw(VException&)
{

    string ret;
    lock.ReadLock();

    try
    {
        ret = VFileSystem::readFile( fileName(category, name) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }

    lock.Unlock();

    return ret;

}

string
FileStore::getItemRaw( const string& item )
throw(VException&)
{

    string ret;
    lock.ReadLock();

    try
    {
        // abuse name here -- item isn't a dir, but we are forcing them to
        // provide the name, no explicit paths, so....
        ret = VFileSystem::readFile( dirName(item) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }

    lock.Unlock();

    return ret;

}


TimeStamp
FileStore::getItemTimeStamp( const string& category, const string& name )
throw(VException&)
{

    TimeStamp ret;

    lock.ReadLock();
    try
    {
        ret = VFileSystem::getFileTime( fileName(category, name) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return ret;
}

int
FileStore::getItemSize( const string& category, const string& name )
throw(VException&)
{
    int ret;

    lock.ReadLock();
    try
    {
        ret = VFileSystem::getFileSize( fileName(category, name) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return ret;
}


void
FileStore::putItem( const string& category,
                        const string& name,
                        const string& data,
                        TimeStamp timeStamp)
throw(VException&)
    // timestamp defaults to 0
{
    if (!isCategory(category))
    {
        addCategory(category);
    }
    string fName( fileName(category, name) );
    lock.WriteLock();
    try
    {
        VFileSystem::writeFile( fName, data );

        if (timeStamp != 0 )
        {
            VFileSystem::setFileTime( fName, timeStamp );
        }

    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();
}


void
FileStore::removeItem( const string& category, const string& name)
throw(VException&)
{
    lock.WriteLock();
    try
    {
        VFileSystem::removeFile( fileName(category, name) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();
}


bool
FileStore::isItem( const string& category, const string& name)
throw(VException&)
{
    bool ret;
    lock.ReadLock();
    try
    {
        ret = VFileSystem::fileExists( fileName(category, name) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return ret;
}


StringList
FileStore::listItems(const string& category)
throw(VException&)
{
    StringList list;

    lock.ReadLock();
    try
    {

        StringList dirList( VFileSystem::readDir( dirName(category ) ));

        StringList::iterator i = dirList.begin();
        while ( i != dirList.end() )
        {
            string s = *i;
            i++;
            
            // filter out "hidden" items in the directory
            static string cvs("CVS");
            if ( s == cvs ) continue;
            
            list.push_back(s);
        }
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return list;
}


void
FileStore::addCategory( const string& category )
throw(VException&)
{
    lock.WriteLock();
    try
    {

        VFileSystem::createDir( dirName(category) );
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }

    lock.Unlock();
}


void
FileStore::removeCategory( const string& category)
throw(VException&)
{
    lock.WriteLock();
    try
    {

        VFileSystem::removeDir( dirName(category) );

    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();
}


bool
FileStore::isCategory( const string& category)
throw(VException&)
{
    bool ret;

    lock.ReadLock();
    try
    {

        ret = VFileSystem::dirExists( dirName(category) );

    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return ret;
}


StringList
FileStore::listCategories()
{
    StringList ret;

    lock.ReadLock();
    try
    {
        // loop over all the categories found in bin 0
        StringList dirList( VFileSystem::readDir( fileRoot ));

        StringList::iterator i = dirList.begin();
        while ( i != dirList.end() )
        {
            string s = *i;
            i++;

            // filter out "hidden" items in the directory
            static string cvs("CVS");
            if ( s == cvs ) continue;

            ret.push_back(s);
        }
    }
    catch (...)
    {
        lock.Unlock();
        throw;
    }
    lock.Unlock();

    return ret;
}






