/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisionServer_cxx_Version =
    "$Id: ProvisionServer.cxx,v 1.4 2002/02/25 05:49:40 sprajpat Exp $";

//Standard Include
#include <cstdio>
#include <errno.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <stdio.h>

#if defined(__linux__)
// #include <algo.h>
#endif

//User defined
#include "VException.hxx"
#include "VIoException.hxx"
// #include "VThreadPool.hxx"
#include "VEnvVar.hxx"
#include "VCondition.h"
#include "NetworkAddress.h"
#include "cpLog.h"
#include "LocalScopeAllocator.hxx"

#include "ProvisionServer.hxx"


#include "PSProtocolService.hxx"
#include "PSDBRequest.hxx"
#include "PSMessage.hxx"
#include "PSRspXmlContentHandlers.hxx"

#include "DBResultset.hxx"
#include "DBConnection.hxx"

#include "PSCommandLine.hxx"


// using namespace Vocal;

static int myCount = 0;

typedef map< string,  string > tableType;

// intitalize static member vars
ProvisionServer* ProvisionServer::myInstance = NULL;

// _clientListLock and _featureListLock not used
// pthread_mutex_t ProvisionServer::_clientListLock;
// pthread_mutex_t ProvisionServer::_featureListLock;

// Use VMutex instead of pthread_mutex directly
VMutex ProvisionServer::_MutexconnectionMapLock;
VMutex ProvisionServer::_MutextimerLock;

// how long to wait in the thread dispatch loop -- gives time to let
// a thread mark a connection as being serviced.
const int ProvisionServer::THREAD_DELAY(250);

// how long to wait before we abort a sync (we assume it failed if it takes
// longer than this)
const int ProvisionServer::FAIL_SLEEP_TIME(600);

#if defined(__SUNPRO_CC)
extern "C" {
void ProvisionServer_sigHupHandler(int sigNo)
{
     cerr << "sigHupHandler" << endl;
}

void ProvisionServer_sigChld(int sigNo)
{
     cerr << "sigChld" << endl;
}

void ProvisionServer_sigTerm(int sigNo)
{
     cerr << "sigTerm" << endl;
}

}

#else

#define ProvisionServer_sigHupHandler ProvisionServer::sigHupHandler
#define ProvisionServer_sigChld       ProvisionServer::sigChld
#define ProvisionServer_sigTerm       ProvisionServer::sigTerm

#endif

/*
  In all cases while calling the redundant server, remember that there are
  4 possible returns -- ERR, which means we are now in an invalid state
  (unlikely to happen, since we assert on most of these), OK, which means
  all is well and that we should send an ok message to the twin, FAIL, which
  means it failed, again send a message about this to twin, and NMSG, which
  means that we should send no message to the other server (usually when
  the transition results in no change in state and no action
*/


// signal handlers
void
ProvisionServer::sigTerm(int signo)
{
     cerr << "sigHupHandler" << endl;
    // Nothing for now...
}



void
ProvisionServer::sigChld(int signo)
{
    cerr << "sigHupHandler" << endl;
    // Nothing for now...
}



void
ProvisionServer::sigHupHandler(int sigNo)
{
    cerr << "sigHupHandler" << endl;
    // Nothing for now...
}


ProvisionServer&
ProvisionServer::instance()
{
    if (!myInstance)
    {
        cpLog(LOG_ERR, "Called instance() when none had been constructed");
        assert(0);
    }
    return *myInstance;
}


ProvisionServer&
ProvisionServer::instance( int nthreads )
{
    if (!myInstance)
    {
        // Set debug level
        cpLogSetPriority( cpLogStrToPriority(
                           PSCommandLine::instance()->getDebugLevel().c_str() ) );

        cpLog(LOG_ERR, "Starting pserver...");

        PSConfiguration::instance()->show();

        myInstance = new ProvisionServer( nthreads );

        cpLog(LOG_DEBUG, "Created instance for %s\n",
              myInstance->className().c_str() );

    }

    return *myInstance;
}


ProvisionServer::ProvisionServer(int nthreads) :
         _nthreads(nthreads)
{

#if defined(__linux__)
    if (::atexit(ProvisionServer::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    }
#endif

    cpLog(LOG_DEBUG, "Creating %d threads for pServer", _nthreads);
    _threadPool = new VThreadPool(_nthreads);

    signal(SIGHUP, ProvisionServer_sigHupHandler);
    signal(SIGCHLD, ProvisionServer_sigChld);
    signal(SIGTERM, ProvisionServer_sigTerm);

    timerPending = false;

    // seed random number generator (used by timer to decide when to try
    // to sync)
    srand((int)getpid() + time(NULL));

    try {

        _dataStore = new FileStore("/usr/local/vocal/newdata");
        
    } catch (VIoException &e) {

         cpLog( LOG_DEBUG, "I/O Exception:  %s", 
               e.getDescription().c_str());

         cpLog( LOG_ERR, "Directory /usr/local/vocal/newdata does not exist or permission error!\n");
         cpLog( LOG_ERR, "You must create this to run server -- aborting.\n");

         exit(0);
         
    }

    try 
    {
        myWebStore = new FileStore("/usr/local/vocal/www");
    } 
    catch (VIoException &e) 
    {
        cpLog( LOG_DEBUG, "I/O Exception:  %s", 
               e.getDescription().c_str());

        cpLog( LOG_ERR, "Directory /usr/local/vocal/www does not exist or permission error!\n");
        cpLog( LOG_ERR, "You must create this to run server -- aborting.\n");
        
        exit(0);
    }
}


void
ProvisionServer::run()
{

    //Loop through for Incoming messages
    cpLog(LOG_ERR, "Pserver successfully started -- Running");
    for (; ; )
    {

        // get a message, accept it, add it to connection map
        if (_serverSocket.getServerConn().isReadReady())
        {

            LockedConnection conn;
            _serverSocket.accept(conn);

            _MutexconnectionMapLock.lock();
            _connectionMap[conn.getConnId()] = conn;
            _MutexconnectionMapLock.unlock();

        }

        // for each active connection with data, dispatch a thread to
        // handle it, unless it is already being handled

        // Connection map lock (cannot close a connection till released)
        _MutexconnectionMapLock.lock();
        for (ConnectionMap::iterator itr = _connectionMap.begin();
                itr != _connectionMap.end(); itr++)
        {

            LockedConnection& conn = (*itr).second;
            if (conn.isLive() && conn.isReadReady() && conn.notServiced())
            {
                // set the connection as being serviced and pass it to a thread
                conn.beingServiced();
                VFunctor functor(ProvisionServer::processRequestFunc, &conn);
                _threadPool->addFunctor(functor);
            }
        }

        _MutexconnectionMapLock.unlock();
        // Connection map unlocked

        vusleep(THREAD_DELAY);
    }
}


ProvisionServer::~ProvisionServer()
{
    cpLog(LOG_ERR, "ProvisionServer destructor called");
    _serverSocket.close();
}


void
ProvisionServer::destroy(void)
{
    delete ProvisionServer::myInstance;
    ProvisionServer::myInstance = NULL;
}


void
ProvisionServer::processRequestFunc(void* con)
{

    // a thread will process the data when it is assigned a request to process
    LockedConnection& conn = *(static_cast < LockedConnection* > (con));
    ProvisionServer::instance().processHttpRequest(conn);

}


void
ProvisionServer::processHttpRequest(LockedConnection& conn) throw (VException&)
{

    Data httpbuf;
    char urlbuf[MAXLINE];
    urlbuf[0] = '\0';
    int nread = 0;
    char tempbuf[MAXLINE];
    nread = 0;
    int cntLength = 0;
    tempbuf[0] = 0;
    bool foundContact = false;


    // read the url from the connection.
    try
    {
        if (conn.readLine(urlbuf, sizeof(urlbuf), nread) <= 0)
        {

            _MutexconnectionMapLock.lock();
            int connTemp = conn.getConnId();
            conn.close();
            conn.doneService();
            _connectionMap.erase(connTemp);
            _MutexconnectionMapLock.unlock();

            return ;
        }

        urlbuf[nread] = 0;
        cpLog( LOG_DEBUG, "Looking at URL: %s", urlbuf);
                    
    }

    catch(VException& e)
    {
        cpLog( LOG_DEBUG, "Network Exception:  %s", 
               e.getDescription().c_str());
        return;
    }

    // add the URL line to the httpbuf
    httpbuf += urlbuf;
    
    // Read headers until we find content length or CR CR
    try
    {

        while (!foundContact)
        {
            // abort on a socket read failure
            if (conn.readLine(tempbuf, sizeof(tempbuf), nread) < 0)
            {
                _MutexconnectionMapLock.lock();
                int connTemp = conn.getConnId();
                conn.close();
                conn.doneService();
                _connectionMap.erase(connTemp);
                _MutexconnectionMapLock.unlock();
                
                return ;
            }

            cpLog( LOG_DEBUG, "Looking at: %s", tempbuf);
            cpLog( LOG_DEBUG, "First char is: %d", tempbuf[0]);
            
            // if they send a CR, end of the message
            if (tempbuf[0] == 13)
            {
                foundContact = true;
            } else {

                // see if it is a contact-length
                // if it is, assume it is followed by CR, else error

                char cntStr[256];
                sscanf(tempbuf, "%s", cntStr);

                if (strcmp(cntStr, "Content-Length:") == 0)
                {
                    foundContact = true;
                    cntLength = 0;
                    sscanf(tempbuf, "%s%d", cntStr, &cntLength);
                    cpLog(LOG_DEBUG, "Content-Length: %d", cntLength);
                    tempbuf[nread] = 0;
                    httpbuf += tempbuf;
                    
                    // read extra carriage return
                    if (conn.readLine(tempbuf, sizeof(tempbuf), nread) < 0)
                    {
                        _MutexconnectionMapLock.lock();
                        int connTemp = conn.getConnId();
                        conn.close();
                        conn.doneService();
                        _connectionMap.erase(connTemp);
                        _MutexconnectionMapLock.unlock();
                        return ;
                    }

                    if (tempbuf[0] != 13)
                    {
                        cpLog(LOG_ERR, "Malformed HTTP request while reading");
                        _MutexconnectionMapLock.lock();
                        int connTemp = conn.getConnId();
                        conn.close();
                        conn.doneService();
                        _connectionMap.erase(connTemp);
                        _MutexconnectionMapLock.unlock();
                        return ;
                    }
                    
                    tempbuf[nread] = 0;
                    httpbuf += tempbuf;            
                    nread = 0;
                }
            }

            tempbuf[nread] = 0;
            httpbuf += tempbuf;            
            
        }
            
        
    }
    catch(VException& e)
    {
        cpLog( LOG_DEBUG, "Network Exception:  %s",
               e.getDescription().c_str());
        return;
    }


    // Read the contents
    char *c = new char[cntLength + 1];
    c[0] = 0;
    nread = 0;

    try
    {
        conn.readn(c, cntLength, nread);
        if (nread == -1)
        {
            cpLog (LOG_ALERT, "Read error on socket");
            delete c;
            return ;
        }
        else if (nread != cntLength)
        {
            cpLog(LOG_ALERT, "Content length mismatch, expected %d read %d",
                  cntLength, nread);
            delete c;
            return ;
        }
    }
    catch(VException& e)
    {
        cpLog( LOG_DEBUG, "Network Exception:  %s",
               e.getDescription().c_str());
        return;
    }


    // add the content to the message
    httpbuf += c; 

    cpLog(LOG_DEBUG, "Received Message : \n%s\n", httpbuf.logData() );

    // create HttpMessage object to handle the http parsing.
    Http::Message myHttpMessage;
  
    try
    {
        LocalScopeAllocator lo;
        myHttpMessage.decode( httpbuf.getData(lo) );
    }
    catch (VException& e)
    {
        cpLog(LOG_DEBUG, "Error parsing Http Message");
        delete c;
        urlbuf[0] = '\0';
        nread = 0;
        conn.doneService();
        return;
    }

    // using if/else...  switch doesn't work when including the types
    // for a different file(?)

    if (  myHttpMessage.method().type() == Http::Method::GET )
    {
        handleGetRequest( conn, myHttpMessage );
    }
    else if ( myHttpMessage.method().type() == Http::Method::PUT )
    {
        handlePutRequest( conn, myHttpMessage );
    }
    else if ( myHttpMessage.method().type() == Http::Method::POST )
    {
        handlePostRequest( conn, myHttpMessage );
    }
    else if ( myHttpMessage.method().type() == Http::Method::DELETE )
    {
        handleDeleteRequest( conn, myHttpMessage );
    }
    else
    {
        cpLog( LOG_DEBUG, "Don't know how to handle http request");
    }

    delete c;
    urlbuf[0] = '\0';
    nread = 0;
    conn.doneService();

}


void
ProvisionServer::processSyncRequest(LockedConnection& conn) throw (VException&)
{
}

void
ProvisionServer::handleGetRequest( LockedConnection& conn,
                                   const Http::Message& httpmsg)
{
    cpLog (LOG_DEBUG, "Get Request:   URL = %s", httpmsg.url().url().c_str() );

    // Get the url for the function request.

    if ( strcmp( httpmsg.url().url().c_str(), 
                 PSCommon::GET_RECORDS_IN_DOMAIN_STR.c_str() ) == 0 )
    {
        handleGetRecordsInDomain( conn, httpmsg.body() );  
    }
    else if (strcmp( httpmsg.url().url().c_str() , 
                     PSCommon::SESSION_OPEN_STR.c_str() ) == 0)
    {
        handleSessionOpen( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(), 
                     PSCommon::SESSION_GET_DB_RESULT_STR.c_str() ) == 0)
    {
        handleSessionGetDBResult( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(),
                     PSCommon::GET_FIELDS_IN_DOMAIN_STR.c_str() ) == 0)
    {
        handleGetFieldsInDomain( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(),
                     PSCommon::GET_XML_DESC_STR.c_str() ) == 0)
    {
        handleGetXmlDescription( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(),
                     PSCommon::LIST_XML_CAT_STR.c_str() ) == 0)
    {
        handleListXmlCategories( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(),
                     PSCommon::LIST_XML_DESC_STR.c_str() ) == 0)
    {
        handleListXmlDescriptions( conn, httpmsg.body() );
    }
    else
    {
//        cpLog(LOG_DEBUG, "*** unrecoginzed GET function ***");
//        return;
        getHtmlfromFunc(conn, httpmsg.url().url());
    }
};


void
ProvisionServer::handlePostRequest( LockedConnection& conn,
                                    const Http::Message& httpmsg)
{
    cpLog (LOG_DEBUG, "Post Request:   URL = %s", httpmsg.url().url().c_str() );

    if (strcmp (httpmsg.url().url().c_str(), 
                PSCommon::SESSION_CLOSE_STR.c_str() ) == 0)
    {
        handleSessionClose( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(),  
                    PSCommon::SESSION_SEARCH_DB_STR.c_str() ) == 0)
    {
        handleSessionSearchDB( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(), 
                    PSCommon::SESSION_SORT_DB_RESULT_STR.c_str() ) == 0 )
    {
        handleSessionSortDBResult( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(),
                    PSCommon::REGISTER_DOMAIN_FOR_UPDATE_STR.c_str() ) == 0 )
    {
        handleRegisterDomainForUpdate( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(),
                    PSCommon::REGISTER_RECORD_FOR_UPDATE_STR.c_str() ) == 0 )
    {
        handleRegisterRecordForUpdate( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(),
                    PSCommon::UN_REGISTER_DOMAIN_FOR_UPDATE_STR.c_str() ) == 0 )
    {
        handleUnRegisterDomainForUpdate( conn, httpmsg.body() );
    }
    else if (strcmp(httpmsg.url().url().c_str(),
                    PSCommon::UN_REGISTER_RECORD_FOR_UPDATE_STR.c_str() ) == 0 )
    {
        handleUnRegisterRecordForUpdate( conn, httpmsg.body() );
    }
    else
    {
        cpLog(LOG_DEBUG, "*** unrecognized POST function ***");
        return;
    }
};


void
ProvisionServer::handlePutRequest( LockedConnection& conn,
                                   const Http::Message& httpmsg)
{
    cpLog (LOG_DEBUG, "Put Request:   URL = %s", httpmsg.url().url().c_str() );

    if (strcmp( httpmsg.url().url().c_str(), 
                PSCommon::SET_RECORDS_IN_DOMAIN_STR.c_str() ) == 0)
    {
        handleSetRecordsInDomain( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(), 
                     PSCommon::ADD_RECORD_IN_DOMAIN_STR.c_str() ) == 0)
    {
        handleAddRecordInDomain( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(), 
                     PSCommon::SESSION_SET_DB_RESULT_STR.c_str() ) == 0 )
    {
        handleSessionSetDBResult( conn, httpmsg.body() );
    }
    else
    {
        cpLog(LOG_DEBUG, "*** unrecognized PUT function ***");
        return;
    }
};


void
ProvisionServer::handleDeleteRequest( LockedConnection& conn,
                                      const Http::Message& httpmsg)
{
    cpLog (LOG_DEBUG, "Delete Request:   URL = %s", httpmsg.url().url().c_str() );

    if (strcmp( httpmsg.url().url().c_str(), 
                PSCommon::DELETE_RECORDS_IN_DOMAIN_STR.c_str() ) == 0 )
    {
        handleDeleteRecordsInDomain( conn, httpmsg.body() );
    }
    else if (strcmp( httpmsg.url().url().c_str(), 
                     PSCommon::SESSION_DELETE_ALL_STR.c_str() ) == 0 )
    {
        handleSessionDeleteAll( conn, httpmsg.body() );
    }
    else
    {
        cpLog(LOG_DEBUG, "*** unrecognized DELETE function ***");
    }
};


/**
 **  Get Functions
 **/

void
ProvisionServer::handleGetRecordsInDomain( LockedConnection& conn,
                                           const string& httpBody)
{
    cpLog(LOG_DEBUG, "** GetRecordsInDomain **");
    myCount ++;

    if ( myCount % 1000 == 0)
    {
        timeval myTimeVal;
        gettimeofday( &myTimeVal, NULL);
        printf("seconds <%ld> , myCount < %d >\n", myTimeVal.tv_sec, myCount);
    }


    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse Request
    try
    {
        request = proto.parseRequest(
                    PSMessage::GET_RECORDS_IN_DOMAIN_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );
        return;
    }

    // 2.  Call DBSelect to get result set.
    DBResultset* resultset = 0;
    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr <string> myXmlResponse;

    vector <string> fieldlist = request->getFieldList();

    if ( fieldlist.empty() )
    {
        // fieldlist is empty.  return only the result count.
        try
        {
            resultset = dbconn->DBSelect( request->getDomain().c_str(),
                                          request->getFilter().c_str() );
        }
        catch( VException& e )
        {
            cpLog( LOG_ALERT, "Failed call to DB , reason %s",
                   e.getDescription().c_str() );

            sendHttpFail( conn, e.getDescription().c_str() );

            if (resultset != 0 )
              delete resultset;

            return;
        }

        char recordCount [16];
        sprintf(recordCount, "%u", resultset->GetRecordCount() );

        myXmlResponse =
	  myXmlContentHandler.buildResultCountRsp( request->getDomain(),
						   recordCount );
    }
    else
    {
        // fieldlist is not empty.  return the result set.
        try
        {
            resultset = dbconn->DBSelect( request->getDomain().c_str(), 
                                          request->getFilter().c_str(), 
                                          request->getFieldList() );
        }
        catch( VException& e )
        {
            cpLog( LOG_ALERT, "Failed call to DB , reason %s",
                   e.getDescription().c_str() );

            sendHttpFail( conn, e.getDescription().c_str() );

            if (resultset != 0 )
              delete resultset;

            return;
        }

        // 3.  Build the ps response using the DBResultset

        myXmlResponse =
	  myXmlContentHandler.buildResultSetRsp( request->getDomain(),
						 resultset->GetRecords(),
						 request->getFieldList() );
    };
 
    sendHttpOk( conn, myXmlResponse->c_str() );
    delete resultset;
}                                   


void
ProvisionServer::handleGetFieldsInDomain( LockedConnection& conn,
                                          const string& httpBody)
{
    cpLog(LOG_DEBUG, "** GetFieldsInDomain **");

    // FieldListtype fieldList;
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse Request
    try
    {
        request = proto.parseRequest(
                    PSMessage::GET_FIELDS_IN_DOMAIN_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2.  Call DBSelect to get result set.

    ResultTabletype restable;

    try
    {
        restable = dbconn->DBDescribe( request->getDomain().c_str() );
    }
    catch( VException& e )
    {
        cpLog( LOG_ALERT, "Failed call to DB , reason %s",
               e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 3.  Build Xml response.
    PSRspXmlContentHandlers myXmlContentHandler;

    Sptr< string > myXmlResponse =
      myXmlContentHandler.buildGetFieldsRsp( request->getDomain(),
					     restable );
    sendHttpOk( conn, myXmlResponse->c_str() );

}

void
ProvisionServer::getHtmlfromFunc( LockedConnection& conn, string url)
{

    cpLog(LOG_DEBUG, "%s", url.c_str());
    
    if (url == "/") {

        sendHttpOk( conn, "<html><head><title>Mascarpone Server</title></head><body><h1>Mascarpone Server v. 0.1 (Alpha)</h1><hr><h2>Mascarpone Main Page. Auto generated.</h2><hr></body>");
        return;

    } else {

        // try to read from file
        string payload;
        
        try {
            payload = myWebStore->getItemRaw(url);
            
        }
        catch (VIoException& e)
        {
            // Failed to read the item
            cpLog(LOG_DEBUG, "*** HTML file not found ( name = %s ) ***\n",
                  url.c_str());
            
            sendHttpFail( conn, "Requested File not available on this server");
            return;
                  
        }

        sendHttpOk(conn, payload);
            
        return;
    }
}

void 
ProvisionServer::handleSessionOpen( LockedConnection& conn,
                                    const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionOpen **");
    Sptr<PSDBRequest> request;
    PSProtocolService& proto = PSProtocolService::getInstance();

    // 1.  Parse Request

    try
    {
        request = proto.parseRequest(
                    PSMessage::SESSION_OPEN_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2.  Get field value 
    //     o  expiretime

    string myExpireTime;
    request->getFieldValue( "expiretime", myExpireTime );

    // Don't use the expire time for now...

    // Q:  Should the client app be allowed to specify the expire time?
    // Q:  Should the expire time be determined by the pserver only?


    // 3. add into cache.
    string sessionid = myCacheManager.genSessionId();

    myCacheManager.addCacheRequest( sessionid,  // string sessionid
                                    "",         // string domain
                                    "",         // string filter
                                    "",         // string sortType
                                    "",         // string sortField
                                    0 );        // DBResultset* resultset

    // 4.  Build Xml response.
    PSRspXmlContentHandlers myXmlContentHandler;

    Sptr< string > myXmlResponse =
    myXmlContentHandler.buildSessionOpenRsp ( sessionid );

    sendHttpOk(conn, myXmlResponse->c_str() );
}


void 
ProvisionServer::handleSessionGetDBResult( LockedConnection& conn,
                               const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionGetDBResult **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse request
    try
    {
        request = proto.parseRequest(
                    PSMessage::SESSION_GET_DB_RESULT_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2.  Call DBSelect()

    ProvData pData;
    string startIndex;
    string endIndex;

    DBResultset* resultset = 0;

    try
    {
        myCacheManager.getCacheRequest( request->getSessionID(), pData);
    }
    catch (VException& e)
    {
        cpLog(LOG_DEBUG, "*** Session id not found (id = %s) ***\n", 
                         request->getSessionID().c_str() );

        sendHttpFail( conn, "Session id not found");
        return;
    }

    try
    {
        request->getFieldValue("start", startIndex);
        request->getFieldValue("end", endIndex);

        Ordertype order;

        if (pData.sortType == "ascend")
        {
            order = ASCENDING;
        }
        else if (pData.sortType == "descend")
        {
            order = DESCENDING;
        }
        else
        {
            order = UNDEFINED;
        }

        resultset = dbconn->DBSelect( pData.domain,
                                      pData.filter,
                                      startIndex,
                                      endIndex,
                                      order,
                                      pData.sortField,
                                      request->getFieldList() );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Failed call to DB, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        if (resultset == 0)
           delete resultset;

        return;
    }

    // 3.  Build Xml response. 

    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse =
      myXmlContentHandler.buildResultSetRsp( pData.domain,
					     resultset->GetRecords(),
					     request->getFieldList() );
    sendHttpOk( conn, myXmlResponse->c_str() );

    delete resultset;
}



void 
ProvisionServer::handleGetXmlDescription( LockedConnection& conn,
                                   const string& httpBody)
{
    cpLog(LOG_DEBUG, "** GetXmlDescription **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;
    
    // 1.  Parse request
    try
    {
        request = proto.parseRequest(
                    PSMessage::GET_XML_DESC_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2. Try to read from File System
    
    string payload;

    try
    {
        payload = _dataStore->getItem(request->getXmlCategory(),
                                      request->getXmlDescName());
        // Success
    }
    catch (VIoException& e)
    {
        // Failed to read the item
        cpLog(LOG_DEBUG, "*** XML Description Not found (category = %s description = %s) ***\n",
              request->getXmlCategory().c_str(),
              request->getXmlDescName().c_str());

        sendHttpFail( conn, "Requested Description Not Found");
        return;

    }

    
    // 3.  Build Xml response. 
    
    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse =
        myXmlContentHandler.buildGetXmlDescriptionRsp(payload);
    sendHttpOk( conn, myXmlResponse->c_str() );

}


void 
ProvisionServer::handleListXmlCategories( LockedConnection& conn,
                                         const string& httpBody)
{
    cpLog(LOG_DEBUG, "** ListXmlCategories **");

    // 1.  Parse request
    // don't need to do this in this case
    
    // 2. Try to read from File System
    
    StringList result;

    try
    {
        result = _dataStore->listCategories();
        // Success
    }
    catch (VIoException& e)
    {
        // Failed to read the item
        cpLog(LOG_DEBUG, "*** Unable to list catagories (cannot access file system?) ***\n");
        
        sendHttpFail( conn, "Unable to list categories");
        return;

    }
    
    // 3.  Build Xml response. 
    
    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse =
        myXmlContentHandler.buildListXmlCategoriesRsp(result);
    sendHttpOk( conn, myXmlResponse->c_str() );

}

void 
ProvisionServer::handleListXmlDescriptions( LockedConnection& conn,
                                         const string& httpBody)
{
    cpLog(LOG_DEBUG, "** ListXmlDescriptions **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;
    
    // 1.  Parse request
    try
    {
        request = proto.parseRequest(
                    PSMessage::LIST_XML_DESC_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2. Try to read from File System
    
    StringList result;

    try
    {
        result = _dataStore->listItems(request->getXmlCategory());
        // Success
    }
    catch (VIoException& e)
    {
        // Failed to read the item
        cpLog(LOG_DEBUG, "*** Unable to list descriptions in category %s ***\n", request->getXmlCategory().c_str() );
        
        sendHttpFail( conn, "Could not read requested category");
        return;

    }

    
    // 3.  Build Xml response. 
    
    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse =
        myXmlContentHandler.buildListXmlDescriptionsRsp(result);
    sendHttpOk( conn, myXmlResponse->c_str() );

}


/**
 **  Post Functions
 **/

void
ProvisionServer::handleSessionClose( LockedConnection& conn,
                                     const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionClose **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;

    try
    {
        request = proto.parseRequest(
                      PSMessage::SESSION_CLOSE_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    if ( myCacheManager.deleteCacheRequest( request->getSessionID() ) == 0)
    {
        sendHttpFail( conn, "No Token Found");
    }
    else
    {
        string responseStr = "\0";
        sendHttpOk(conn, responseStr.c_str() );
    }

    return;
}


void
ProvisionServer::handleSessionSearchDB( LockedConnection& conn,
                                        const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionSearchDB **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    try
    {
        request = proto.parseRequest(
                      PSMessage::SESSION_SEARCH_DB_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // Need to store the domain, filter, and fieldlist into the cache.

    string sortType;
    string sortField;

    cpLog( LOG_DEBUG, "*** Modifying Session (id = %s) ***\n",
                       request->getSessionID().c_str() );

    try
    {
        request->getFieldValue( PSCommon::SORTTYPE_STR.c_str(), 
                                  sortType);

        request->getFieldValue( PSCommon::SORTFIELD_STR.c_str(),
                                  sortField);

    }
    catch (VException& e)
    {
        cpLog( LOG_DEBUG, "Error getting fields");
        sendHttpFail( conn, e.getDescription().c_str() );
        return;
    }


    if ( ! myCacheManager.setCache( request->getSessionID(),
                                    request->getDomain(),
                                    request->getFilter(),
                                    sortType,
                                    sortField) )
    {
        cpLog( LOG_DEBUG,
               "*** Modify failed!  Session id not found (id = %s) ***\n", 
               request->getSessionID().c_str() );

        sendHttpFail( conn, "Session id not found");
        return;
    }

    // Call DB to get the data.
    DBResultset* resultset = 0;

    try
    {
        Ordertype order;

        if (sortType == "ascend")
        {
            order = ASCENDING;
        }
        else if (sortType == "descend")
        {
            order = DESCENDING;
        }
        else
        {
            order = UNDEFINED;
        }

        resultset = dbconn->DBSelect( request->getDomain(), 
                                      request->getFilter(),
                                      request->getFieldList(),
                                      order,
                                      sortField);
    }
    catch( VException& e)
    {
        cpLog( LOG_ALERT, "Failed call to DB , reason %s",
               e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        if (resultset != 0)
            delete resultset;

        return;
    }

    // Send ps response with result count.
    char countBuf[16];
    sprintf(countBuf, "%u", resultset->GetRecordCount() );


    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse =
      myXmlContentHandler.buildSetRecordsRsp( request->getDomain().c_str(),
					      countBuf);

    sendHttpOk( conn, myXmlResponse->c_str() );

    delete resultset;

}

void
ProvisionServer::handleSessionSortDBResult( LockedConnection& conn,
                                            const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionSortDBResult **");

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    try
    {
        request = proto.parseRequest(
                      PSMessage::SESSION_SORT_DB_RESULT_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // Need to store the domain, filter, and fieldlist into the cache.

    string sortType;
    string sortField;

    try
    {
        request->getFieldValue( PSCommon::SORTTYPE_STR.c_str(), 
                                  sortType);
        request->getFieldValue( PSCommon::SORTFIELD_STR.c_str(), 
                                  sortField);
    }
    catch (VException& e)
    {
        cpLog( LOG_DEBUG, "Error setting cache");
        sendHttpFail( conn, "Error setting cache");
        return;
    }

    cpLog( LOG_DEBUG, "*** Modifying Session (id = %s) ***\n", 
           request->getSessionID().c_str() );

    if ( !myCacheManager.modifyCache( request->getSessionID(),
                                     "",               // domain
                                     "",               // filter
                                     sortType,
                                     sortField ) )
    {
        cpLog( LOG_DEBUG, 
               "*** Modify failed!  Session id not found (id = %s) ***\n", 
               request->getSessionID().c_str() );

        sendHttpFail( conn, "Session id not found");
        return;
    }

    ProvData pData;

    myCacheManager.getCacheRequest( request->getSessionID(), pData);

    // Call DB to get the data.
    DBResultset* resultset = 0;

    try
    {
        Ordertype order;

        if (sortType == "ascend")
        {
            order = ASCENDING;
        }
        else if (sortType == "descend")
        {
            order = DESCENDING;
        }
        else
        {
            order = UNDEFINED;
        }
            
        FieldListtype fieldlist; 
        fieldlist.push_back("*");

        resultset = dbconn->DBSelect( pData.domain,
                                      pData.filter,
                                      fieldlist, 
                                      order,
                                      pData.sortField);
    }
    catch( VException& e)
    {
        cpLog( LOG_ALERT, "Failed call to DB , reason %s",
               e.getDescription().c_str() );
  
        sendHttpFail( conn, e.getDescription().c_str() );

        if (resultset != 0)
            delete resultset;

        return;
    }
    
    // Send ps response with result count.

    char countBuf[16];
    sprintf(countBuf, "%u", resultset->GetRecordCount() );


    PSRspXmlContentHandlers myXmlContentHandler;
    Sptr< string > myXmlResponse = 
      myXmlContentHandler.buildSetRecordsRsp( pData.domain,
					      countBuf );

    sendHttpOk( conn, myXmlResponse->c_str() );

    delete resultset;
}

void
ProvisionServer::handleRegisterDomainForUpdate( LockedConnection& conn,
                                                const string& httpBody)
{
    cpLog(LOG_DEBUG, "** RegisterDomainForUpdate **");

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    try
    {
        request = proto.parseRequest(
                        PSMessage::REGISTER_DOMAIN_FOR_UPDATE_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    string responseStr = "\0";

    if ( myRegisterManager.registerDomainForUpdate( 
                                           request->getDomain().c_str(),
                                           request->getRegisterHostport().c_str() ) )
    {
        sendHttpOk( conn, responseStr.c_str() );
    }
    else
    {
        sendHttpFail( conn, myRegisterManager.errorMsg );//responseStr.c_str() );
    }
}


void
ProvisionServer::handleUnRegisterDomainForUpdate( LockedConnection& conn,
                                                  const string& httpBody)
{
    cpLog(LOG_DEBUG, "** UnRegisterDomainForUpdate **");

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    try
    {
        request = proto.parseRequest(
                        PSMessage::UN_REGISTER_DOMAIN_FOR_UPDATE_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    string responseStr = "\0"; 

    if ( myRegisterManager.unregisterDomainForUpdate( 
                                     request->getDomain().c_str(),
                                     request->getRegisterHostport().c_str() ) )
    {
        sendHttpOk( conn, responseStr.c_str() );
    }
    else
    {
        sendHttpFail( conn, myRegisterManager.errorMsg );//responseStr.c_str() );
    }
}


void 
ProvisionServer::handleRegisterRecordForUpdate( LockedConnection& conn,
                                                const string& httpBody)
{
    cpLog(LOG_DEBUG, "** RegisterRecordForUpdate **");

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    try
    {
        request = proto.parseRequest(
                        PSMessage::REGISTER_RECORD_FOR_UPDATE_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    string responseStr = "\0";

    if ( myRegisterManager.registerRecordForUpdate( 
                                               request->getDomain().c_str(),
                                               request->getFieldList(),
                                               request->getRegisterHostport() ) )
    {
        sendHttpOk(conn, responseStr.c_str() );
    }
    else
    {
        sendHttpFail( conn, myRegisterManager.errorMsg); //responseStr.c_str() );
    }
}


void
ProvisionServer::handleUnRegisterRecordForUpdate( LockedConnection& conn,
                                                  const string& httpBody)
{
    cpLog(LOG_DEBUG, "** UnRegisterRecordForUpdate **");

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    try
    {
        request = proto.parseRequest(
                        PSMessage::UN_REGISTER_RECORD_FOR_UPDATE_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    string responseStr = "\0";

    if ( myRegisterManager.unregisterRecordForUpdate( 
                                                 request->getDomain().c_str(),
                                                 request->getFieldList(),
                                                 request->getRegisterHostport() ) )
    {
        sendHttpOk(conn, responseStr.c_str() );
    }
    else
    {
        sendHttpFail(conn, myRegisterManager.errorMsg); //responseStr.c_str() );
    }
}

/**
 **  Put Functions
 **/


void
ProvisionServer::handleSetRecordsInDomain( LockedConnection& conn,
                                           const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SetRecordsInDomain **");

    FieldValuetype fieldvaluelist;

    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse Request

    try
    {
        request = proto.parseRequest(
                    PSMessage::SET_RECORDS_IN_DOMAIN_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2.  Call DBUpdate()
    PSRspXmlContentHandlers myXmlContentHandler;
    DBResultset* resultset = 0;

    try
    {
        // Need to call select for the update check.
        string primaryKey = dbconn->DBPrimaryKey(request->getDomain().c_str() );

        FieldListtype fieldList;
        fieldList.push_back(primaryKey.c_str() );


        printf ("Calling DBSelect **********************\n");

        resultset = dbconn->DBSelect( request->getDomain().c_str(),
                                      request->getFilter().c_str(),
                                      fieldList );

        printf ("Calling DBUpdate **********************\n");
        // Actual set call.
        dbconn->DBUpdate( request->getDomain().c_str(), 
                          request->getFilter().c_str(), 
                          request->getTable() );
    }
    catch( VException& e)
    {
        cpLog( LOG_ALERT, "Failed call to DB , reason: %s",
               e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        if (resultset != 0)
        {
            delete resultset;
        }

        return;
    }

    char resultCount [256];
    sprintf(resultCount, "%d", resultset->GetRecordCount() );


    Sptr<string > myXmlResponse = myXmlContentHandler.buildSetRecordsRsp( 
                                      request->getDomain(),
                                      resultCount );

    sendHttpOk(conn, myXmlResponse->c_str() );

    updateCheck( request, resultset );
}

void
ProvisionServer::handleAddRecordInDomain( LockedConnection& conn,
                                          const string& httpBody)
{
    cpLog(LOG_DEBUG, "** AddRecordInDomain **");

    FieldListtype fieldlist;
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    try
    {
        request = proto.parseRequest(
                    PSMessage::ADD_RECORD_IN_DOMAIN_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    try
    {
        dbconn->DBInsert(request->getDomain(), request->getTable());
    }
    catch ( VException& e)
    {
        cpLog( LOG_ALERT, "Failed Add to DB , reason %s",
               e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );
        return;
    }

    string responseStr = "\0";

    sendHttpOk( conn, responseStr.c_str() );

    updateCheck( request );
}

void
ProvisionServer::handleSessionSetDBResult( LockedConnection& conn,
                                           const string& httpBody)
{
    cpLog(LOG_DEBUG, "** SessionSetDBResult **");

    FieldListtype fieldlist;
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse Request

    try
    {
        request = proto.parseRequest(
                    PSMessage::SESSION_SET_DB_RESULT_REQ, httpBody );
    }
    catch (VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    ProvData pData;
  
    try
    { 
        myCacheManager.getCacheRequest( request->getSessionID(), pData);
    }
    catch (VException& e)
    {
        sendHttpFail( conn, "Session ID not found.  Cannot process request.");
        return;
    }

    DBResultset* resultset = 0;

    try
    {
        cpLog( LOG_DEBUG, "domain = %s, filter = %s", 
               pData.domain.c_str(), pData.filter.c_str() ); 


        // Need to call select for the update check.
        string primaryKey = dbconn->DBPrimaryKey(pData.domain.c_str() );

        FieldListtype fieldList;
        fieldList.push_back(primaryKey.c_str() );

        resultset = dbconn->DBSelect( pData.domain.c_str(),
                                      pData.filter.c_str(),
                                      fieldList );

        // Actual update call
        dbconn->DBUpdate( pData.domain.c_str(),
                          pData.filter.c_str(),
                          request->getTable() );

        // not great but reuse PSDBRequest to store the domain, filter.
        // this is needed for the updateCheck().
	request->clear();
        request->setDomain( pData.domain.c_str() );
        request->setFilter( pData.filter.c_str() );

    }
    catch( VException& e)
    {
        cpLog( LOG_ALERT, "Failed call to DB , reason %s",
               e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );
        return;
    }

    char resultCount [256];
    sprintf(resultCount, "%d", resultset->GetRecordCount() );

    PSRspXmlContentHandlers myXmlContentHandler;

    Sptr< string > myXmlResponse = myXmlContentHandler.buildSetRecordsRsp(
                                      pData.domain,
                                      resultCount );

    sendHttpOk(conn, myXmlResponse->c_str() );

    updateCheck( request, resultset );
}

/**
 **  Delete Functions
 **/

void
ProvisionServer::handleDeleteRecordsInDomain( LockedConnection& conn,
                                     const string& httpBody)
{
    cpLog(LOG_DEBUG, "** DeleteRecordsInDomain **");

    FieldListtype fieldlist;
    PSProtocolService& proto = PSProtocolService::getInstance();
    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    // 1.  Parse Request

    try
    {
        request = proto.parseRequest(
                    PSMessage::DEL_RECORDS_IN_DOMAIN_REQ, httpBody );
    }
    catch ( VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    // 2.  Call DBSelect to delete records

    DBResultset* resultset = 0;

    try
    {
        // Need to call select for the update check.
        string primaryKey = dbconn->DBPrimaryKey(request->getDomain().c_str() );

        FieldListtype fieldList;
        fieldList.push_back(primaryKey.c_str() );

        resultset = dbconn->DBSelect( request->getDomain().c_str(),
                                      request->getFilter().c_str(),
                                      fieldList );

        // Actual delete call.

        dbconn->DBDelete( request->getDomain(), 
                          request->getFilter() );
    }
    catch (VDbException& e)
    {
        cpLog(LOG_DEBUG, "Error deleting records, reason: %s", 
              e.getDescription().c_str() );
        sendHttpFail( conn, e.getDescription().c_str() );

        if (resultset != 0)
        {
            delete resultset;
        }

        return;
    }
    
    // 3.  Return 200 OK w/ number of deleted records.

    char resultCount [256];
    sprintf(resultCount, "%d", resultset->GetRecordCount() );

    PSRspXmlContentHandlers myXmlContentHandler;

    Sptr< string > myXmlResponse =
      myXmlContentHandler.buildSetRecordsRsp( request->getDomain(),
					      resultCount );

    sendHttpOk( conn, myXmlResponse->c_str() );

    updateCheck( request, resultset );
}

void
ProvisionServer::handleSessionDeleteAll( LockedConnection& conn,
                                         const string& httpBody)
{
    cpLog( LOG_DEBUG, "** SessionDeleteAll **");

    PSProtocolService& proto = PSProtocolService::getInstance();

    Sptr<PSDBRequest> request;

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    try
    {
        request = proto.parseRequest(
                      PSMessage::SESSION_DELETE_ALL_REQ, httpBody );
    }
    catch( VException& e)
    {
        cpLog( LOG_ALERT, "Error parsing XML, reason %s",
                   e.getDescription().c_str() );

        sendHttpFail( conn, e.getDescription().c_str() );

        return;
    }

    ProvData pData;

    try
    {
        myCacheManager.getCacheRequest( request->getSessionID(), pData);
    }
    catch (VException& e)
    {
        sendHttpFail( conn, "Session ID not found!  Unable to process request.");
        return;
    }

    DBResultset* resultset = 0;

    try
    {

        // Need to call select for the update check.
        string primaryKey = dbconn->DBPrimaryKey(pData.domain.c_str() );

        FieldListtype fieldList;
        fieldList.push_back(primaryKey.c_str() );

        resultset = dbconn->DBSelect( pData.domain.c_str(),
                                      pData.filter.c_str(),
                                      fieldList );

        // not great but reuse PSDBRequest to store the domain, filter.
        // this is needed for the updateCheck().
	request->clear();
        request->setDomain( pData.domain.c_str() );
        request->setFilter( pData.filter.c_str() );


        // actual delete call for this session.
        dbconn->DBDelete( pData.domain.c_str(),
                          pData.filter.c_str() );

    }
    catch( VDbException& e)
    {
        cpLog(LOG_DEBUG, "Error deleting records, reason: %s",
              e.getDescription().c_str() );
        sendHttpFail( conn, e.getDescription().c_str() );
        return;
    }

    char resultCount [256];
    sprintf(resultCount, "%d", resultset->GetRecordCount() );
    
    PSRspXmlContentHandlers myXmlContentHandler;

    Sptr< string > myXmlResponse =
      myXmlContentHandler.buildSetRecordsRsp( pData.domain.c_str(),
					      resultCount );

    sendHttpOk( conn, myXmlResponse->c_str() );

    updateCheck( request, resultset );
}


void
ProvisionServer::updateCheck( Sptr<PSDBRequest> request,
                              DBResultset* resultset)
{
    cpLog (LOG_DEBUG, "*** begin updateCheck ***");

    int modtype = -1;   //  0 = add, 1 = set/modify, 2 = delete, 
                        // -1 = unrecognized.

    switch ( request->getMessageType() )
    {
        case PSMessage::SET_RECORDS_IN_DOMAIN_REQ:
            modtype = 1;
        break;

        case PSMessage::ADD_RECORD_IN_DOMAIN_REQ:
            modtype = 0;
        break;

        case PSMessage::DEL_RECORDS_IN_DOMAIN_REQ:
            modtype = 2;
        break;

        case PSMessage::SESSION_SET_DB_RESULT_REQ:
            modtype = 1;
        break;

        case PSMessage::SESSION_DELETE_ALL_REQ:
            modtype = 2;
        break;

        default:
            modtype = -1;
        break;
    }

    DBConnection* dbconn = 
      myDBManager.GetDBConnection(PSConfiguration::instance()->getDBInfo());

    string primaryKey = dbconn->DBPrimaryKey(request->getDomain().c_str() );
    string filterString = request->getFilter().c_str();

    cpLog(LOG_DEBUG, "domain <%s>, filter <%s>",
                     request->getDomain().c_str(),
                     filterString.c_str() );

    if (modtype == 0)
    {
        cpLog (LOG_DEBUG, "** new record was added **");

        tableType reqTable = request->getTable();

        if ( reqTable.empty() )
        {
            cpLog(LOG_DEBUG, "** table empty.  updateCheck failed **");
            return;
        }
        else
        {
            tableType::iterator reqTableItr =
                                reqTable.find( primaryKey.c_str() );

            if (reqTableItr != reqTable.end() )
            {
                // replace filterString with record that was added.
                filterString = reqTableItr->second.c_str();
            }
            else
            {
                cpLog(LOG_DEBUG,
                "** primary key not in table.  updateCheck failed **");
                return;
            }
        }
    }
    else if (modtype == 1)
    {
        cpLog(LOG_DEBUG, "** record was modified **");
    }
    else if (modtype == 2 )
    {
        cpLog(LOG_DEBUG, "** record was deleted **");
    }
    else
    {
        cpLog(LOG_DEBUG, "** request does not require update check **");
        return;
    }

    // need to check the register list to send update.
    myRegisterManager.checkRegisterList( modtype,
                                         request->getDomain().c_str(),
                                         filterString.c_str(),
                                         resultset );
}


void
ProvisionServer::sendHttpFail(LockedConnection &conn, string retData)
{
    Http::Status errStatus("400");
    list<Http::Header> headers;
    string errBegin = "<error>";
    string errEnd   = "</error>";

    string tempData = errBegin + retData + errEnd;

    char buf[10];
    sprintf( buf, "%d", tempData.size() );

    Http::Header contentLength(Http::Header::ContentLength, buf );

    headers.push_back(contentLength);

    Http::Message myHttpResponse(errStatus, headers, tempData.c_str() );

    string tempString = myHttpResponse.encode().c_str();

    try
    {
        cpLog(LOG_DEBUG, "Send 400 Fail Message");
        conn.writeData( tempString );
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Caught exception when writing data, reason %s",
              e.getDescription().c_str());
    }
}


void
ProvisionServer::sendHttpOk(LockedConnection &conn, string retData)
{
    // Build http status message.
    Http::Status myStatus ("200");
    list<Http::Header> headers;

    char buf[30];
    sprintf( buf, "%d", retData.size());

    Http::Header contentLength(Http::Header::ContentLength, buf );

    headers.push_back(contentLength);
    Http::Message myHttpResponse(myStatus, headers, retData.c_str() );

    string tempString = myHttpResponse.encode().c_str();

    try
    {
        cpLog(LOG_DEBUG, "Send 200 OK response" );

        conn.writeData( tempString );
    }
    catch ( VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Caught exception when writing data, reason %s",
            e.getDescription().c_str());
    }
}

