/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const RegisterManager_cxx_Version =
    "$Id: RegisterManager.cxx,v 1.2 2001/09/11 00:33:02 maivu Exp $";


#include "RegisterManager.hxx"
#include "cpLog.h"
#include "Sptr.hxx"
#include "NetworkAddress.h"
#include "Tcp_ClientSocket.hxx"
#include "PSProtocolService.hxx"


VRWLock RegisterManager::mapLock;

RegisterManager::RegisterManager()
  : errorMsg("")
{
};


RegisterManager::~RegisterManager()
{
};


bool
RegisterManager::registerRecordForUpdate( const string& domain,
                                          FieldListtype keyList,
                                          const string& hostport )
{
    cpLog(LOG_DEBUG, "RegisterManager::registerRecordForUpdate");

    bool bRet = false;

    if ( keyList.empty() )
    {
        // problem... no keyList... 
        cpLog(LOG_DEBUG, "No record to register... register failed");

	// client need to get RegisterManager::errorMsg to get the error message
	errorMsg = "No record to register... register failed";
	bRet = false;
        //return false;
    }
    else
    {
        mapLock.WriteLock();
        HostMap& hMap = myRecordMap[ domain ];

        for (unsigned short i = 0; i < keyList.size(); i++ )
        {
            HostList& hList = hMap[ keyList[i].c_str() ];
            HostList::iterator hListItr = find (hList.begin(), hList.end(), hostport );

            if (hListItr == hList.end() )
            {
                cpLog(LOG_DEBUG, "register host <%s> for record <%s>",
                                 hostport.c_str(), keyList[i].c_str() );

                hList.push_back( hostport );

                hMap[ keyList[i].c_str() ]  = hList; 
            }
            else
            {
                cpLog(LOG_DEBUG, "Register failed for record.  Host exists.");
		// client need to get RegisterManager::errorMsg to get the error message
		errorMsg = "Register failed for record.  Host exists.";
		bRet = false;
            }
        }
        mapLock.Unlock();
	bRet = true;

        //return true;
    }
    return bRet;
};

bool
RegisterManager::unregisterRecordForUpdate( const string& domain, 
                                            FieldListtype keyList,
                                            const string& hostport )
{
    cpLog(LOG_DEBUG, "RegisterManager::unregisterRecordForUpdate");
    bool bRet = false;

    if ( keyList.empty() )
    {
        cpLog(LOG_DEBUG, "No record to unregister... unregister failed");

	// client need to get RegisterManager::errorMsg to get the error message
	errorMsg = "No record to unregister... unregister failed";
	bRet = false;
        //return false;
    }
    else
    {
        mapLock.WriteLock();
        HostMap& hMap = myRecordMap[ domain ];

        for (unsigned short i = 0; i < keyList.size(); i++ )
        {
            HostList& hList = hMap[ keyList[i].c_str() ];
            HostList::iterator hListItr = find (hList.begin(), hList.end(), hostport );

            if (hListItr != hList.end() )
            {
                cpLog(LOG_DEBUG, "UnRegister host <%s> for record <%s>",
                                 hostport.c_str(), keyList[i].c_str() );

                hList.erase( hListItr );
                hMap[ keyList[i].c_str() ] = hList;
            }
            else
            {
                cpLog(LOG_DEBUG, "UnRegister failed for record.  Host not found.");
		// client need to get RegisterManager::errorMsg to get the error message
		errorMsg = "UnRegister failed for record.  Host not found.";
		bRet = false;
            }
        }

        mapLock.Unlock();
        //return true;
	bRet = true;
    }

    return bRet;
};


bool
RegisterManager::registerDomainForUpdate( const string& domain,
                                          const string& hostport)
{
    bool bRet = false;

    mapLock.WriteLock();
    HostList& hList = myDomainMap[ domain ]; 

    HostList::iterator hListItr = find( hList.begin(), hList.end(), hostport );

    if ( hListItr == hList.end() )
    {
        cpLog(LOG_DEBUG, "Register host <%s> for domain <%s>", 
                         hostport.c_str(), domain.c_str() );
        hList.push_back( hostport );

        mapLock.Unlock();
        //return true;
	bRet = true;
    }
    else
    {
        cpLog(LOG_DEBUG, "Register failed for domain.  Host exists.");
	// client need to get RegisterManager::errorMsg to get the error message
	errorMsg = "Register failed for domain.  Host exists.";
	bRet = false;
        mapLock.Unlock();
        //return false;
    }

    return bRet;
};

bool
RegisterManager::unregisterDomainForUpdate( const string& domain,
                                            const string& hostport )
{
    mapLock.WriteLock();
    bool bRet = false;

    HostList& hList = myDomainMap[ domain ];
    HostList::iterator hListItr = find (hList.begin(), hList.end(), hostport );

    if (hListItr != hList.end() )
    {
        cpLog(LOG_DEBUG, "UnRegister host <%s> for domain <%s>",
                         hostport.c_str(), domain.c_str() );

        hList.erase( hListItr );
        mapLock.Unlock();
        //return true;
	bRet = true;
    }
    else
    {
        cpLog(LOG_DEBUG, "UnRegister failed for domain.  Host not found.");
        mapLock.Unlock();
	// client need to get RegisterManager::errorMsg to get the error message
	errorMsg = "UnRegister failed for domain.  Host not found.";
	bRet = false;
        //return false;
    }

    return bRet;
};

void
RegisterManager::checkRegisterList( const int modType,
                                    const string& domain,
                                    const string& filter,
                                    DBResultset* resultset ) 
{
    // Always check the domain list.
    checkDomainList( modType, domain, filter, resultset );

    // No need to check the record list if a record was added.
    if (modType != 0)
    {
        checkRecordList( modType, domain, filter, resultset );
    };

    if (resultset != 0)
    {
        delete resultset;
    }
};


void
RegisterManager::checkDomainList( const int modType,
                                  const string& domain,
                                  const string& filter,
                                  DBResultset* resultset )
{
    cpLog(LOG_DEBUG, "** checkDomainList **");

    mapLock.ReadLock();
    HostMap::iterator domHostMapItr = myDomainMap.find( domain );

    if ( domHostMapItr!= myDomainMap.end() )
    {
        cpLog(LOG_DEBUG, "Found list for domain update");

        // build recordList the new, modified, or deleted records
        vector <string> recordList;

        if (modType == 0) // Add
        {
            // should only be one.
            recordList.push_back( filter.c_str() );
        }            
        else if (modType == 1 || modType == 2)  // 1 = Set, 2 = Delete
        {
            // need to iterate through the resultset to generate list
            
            if ( resultset != 0)
            {
                // need to iterate through the resultset to generate list
                ResultTabletype temptable = resultset->GetRecords();

                if (temptable.size() == 0)
                {
                    cpLog(LOG_DEBUG, "table is null.  nothing to do.");
                    mapLock.Unlock();
                    return;
                }

                for (unsigned short i = 0; i < temptable.size(); i++)
                {
                    ResultRecordtype rectype = temptable[i];

                    for (unsigned short j = 0; j < rectype.size(); j++)
                    {
                        recordList.push_back( rectype[j].c_str() );
                    }
                }
            }
            else
            {
                cpLog(LOG_DEBUG, "resultset is null.  nothing to do.");
                mapLock.Unlock();
                return;
            }
        }     
        else
        {
            cpLog(LOG_DEBUG,
                  "** unrecognized modtype.  checkDomainList failed .**");
        }

        // send the update to all the registered hosts.
        HostList domHostList = domHostMapItr->second;

        for ( HostList::iterator domHostListItr = domHostList.begin();
              domHostListItr != domHostList.end();
              domHostListItr++)
        {
            cpLog(LOG_DEBUG, "*** Send domain update to host: %s ***",
                             (*domHostListItr).c_str() );

            sendUpdateDomain( domain,
                              recordList,
                              (*domHostListItr).c_str() );
        }
    }
    else
    {
        cpLog(LOG_DEBUG, "No update for this domain");
    }

    mapLock.Unlock();
}

void
RegisterManager::checkRecordList( const int modType,
                                  const string& domain,
                                  const string& filter,
                                  DBResultset* resultset )
{
    cpLog(LOG_DEBUG, "** checkRecordList **");

    mapLock.ReadLock();

    RecordMap::iterator recMapItr = myRecordMap.find( domain );

    if ( recMapItr == myRecordMap.end() )
    {
        cpLog(LOG_DEBUG, "No update for this record");
        mapLock.Unlock();
        return;
    }

    if ( modType == 1 || modType == 2)  // 1 = Set, 2 = Delete
    {
        if (resultset != 0)
        {
            ResultTabletype restable = resultset->GetRecords();

            // Go through the map to find the registered record.
            // This is ugly...  Need to clean up and make more readable.

            // iterate through the result set:   vector<vector<string>>

            // NOTE:  If the modType is Delete, then it is possible that
            //        the table is null ( restable.size() = 0 ).
            // 
            //        Why?  The table can be null if the record being deleted 
            //        does not exist in the database.  A resultset is still
            //        returned but the table in the resultset would be null.

            for (unsigned short i = 0; i < restable.size(); i++ )
            {
                ResultRecordtype myRecordtype = restable[i];

                for (unsigned short j = 0; j < myRecordtype.size(); j++)
                {
                    cpLog( LOG_DEBUG, "result set: %s", myRecordtype[j].c_str() );

                    // search the map for the modified record.
                    // NOTE:  It's possible that are no updates for this record.

                    HostMap::iterator recHostMapItr =
                            (recMapItr->second).find( myRecordtype[j].c_str() );

                    // if the record is found, then send update.
                    if (recHostMapItr != (recMapItr->second).end() )
                    {
                        cpLog(LOG_DEBUG, "Found host for record update");
     
                        HostList recHostList = recHostMapItr->second;
    
                        for (HostList::iterator recHostListItr = recHostList.begin();
                             recHostListItr != recHostList.end();
                             recHostListItr ++)
                        {
                            string temp = *recHostListItr;
                            cpLog( LOG_DEBUG, "*** Send record update to host: %s ***",
                                              (*recHostListItr).c_str() );

                            sendUpdateRecord( domain,
                                              myRecordtype[j].c_str(),
                                              (*recHostListItr).c_str() );
                        }
                    }
                    else
                    {
                        cpLog(LOG_DEBUG, "Cannot find record in record map"); 
                    }
                }
            }
        }
        else
        {
            cpLog(LOG_DEBUG, "result set is null.  nothing to do.");
            mapLock.Unlock();
            return; 
        }
    }
    else
    {
        cpLog( LOG_DEBUG, 
               "** unrecognized modtype.  checkRecordList failed .**");
    }
    mapLock.Unlock();
}

void
RegisterManager::sendUpdateRecord( const string& domain,
                                   const string& key,
                                   const string& hostport )
{
    NetworkAddress* netAddress = new NetworkAddress( hostport );

    TcpClientSocket* clientSocket = 
                      new TcpClientSocket( *netAddress, true, false);

    try
    {
        clientSocket->connect();
        Connection& conn = clientSocket->getConn();

        // build http message
        PSProtocolService& proto = PSProtocolService::getInstance();

        Sptr<Http::Message> httpMsg;

        vector<string> keyList;
        keyList.push_back( key );

        httpMsg = proto.buildUpdateReq( PSCommon::UPDATE_RECORD_STR.c_str(),
                                        domain.c_str(), keyList ); 

        string tempString = httpMsg->encode().c_str();

        conn.writeData( (void *) tempString.c_str(), tempString.size() );

        // clientSocket->close();
        // conn.close();
    }
    catch(VException& e)
    {
        cpLog(LOG_DEBUG, "Caught something... reason: %s ", 
              e.getDescription().c_str());
    }

    delete netAddress;
    delete clientSocket;
};

void
RegisterManager::sendRecordDelete( const string& domain,
                                   const string& key,
                                   const string& hostport )
{
    // NOTE:  This is the same as an updaterecord message.  There's
    //        no recorddelete message at the moment.

    NetworkAddress* netAddress = new NetworkAddress( hostport );

    TcpClientSocket* clientSocket =
                      new TcpClientSocket( *netAddress, true, false);

    try
    {
        clientSocket->connect();
        Connection& conn = clientSocket->getConn();

        // build http message
        PSProtocolService& proto = PSProtocolService::getInstance();

        Sptr<Http::Message> httpMsg;

        vector<string> keyList;
        keyList.push_back( key );

        httpMsg = proto.buildUpdateReq( PSCommon::UPDATE_RECORD_STR.c_str(),
                                        domain.c_str(), keyList );

        string tempString = httpMsg->encode().c_str();

        conn.writeData( (void *) tempString.c_str(), tempString.size() );

        // clientSocket->close();
        // conn.close();
    }
    catch(VException& e)
    {
        cpLog(LOG_DEBUG, "Caught something... reason: %s ",
              e.getDescription().c_str());
    }

    delete netAddress;
    delete clientSocket;
};

void
RegisterManager::sendUpdateDomain( const string& domain,
                                   const vector<string>& keyList,
                                   const string& hostport )
{
    NetworkAddress* netAddress = new NetworkAddress( hostport );

    TcpClientSocket* clientSocket = 
                         new TcpClientSocket( *netAddress, true, false);

    try
    {
        clientSocket->connect();
        Connection& conn = clientSocket->getConn(); 

        // build http message
        PSProtocolService& proto = PSProtocolService::getInstance();

        Sptr<Http::Message> httpMsg;

        httpMsg = proto.buildUpdateReq( PSCommon::UPDATE_DOMAIN_STR.c_str(), 
                                        domain.c_str(), keyList );

        string tempString = httpMsg->encode().c_str();

        conn.writeData( (void *) tempString.c_str(), tempString.size() );

        // clientSocket->close();
        // conn.close();
    }
    catch( VException& e)
    {
        cpLog(LOG_DEBUG, "Caught something... reason: %s ",
              e.getDescription().c_str());
    }

    delete netAddress;
    delete clientSocket;
};

void
RegisterManager::sendDomainDelete( const string& domain,
                                   const vector<string>& keyList,
                                   const string& hostport )
{
    // NOTE:  This is the same as an updatedomain message.  There's
    //        no domaindelete message at the moment.

    NetworkAddress* netAddress = new NetworkAddress( hostport );

    TcpClientSocket* clientSocket =
                         new TcpClientSocket( *netAddress, true, false);

    try
    {
        clientSocket->connect();
        Connection& conn = clientSocket->getConn();

        // build http message
        PSProtocolService& proto = PSProtocolService::getInstance();

        Sptr<Http::Message> httpMsg;


        httpMsg = proto.buildUpdateReq( PSCommon::UPDATE_DOMAIN_STR.c_str(),
                                        domain.c_str(), keyList );

        string tempString = httpMsg->encode().c_str();

        conn.writeData( (void *) tempString.c_str(), tempString.size() );

        // clientSocket->close();
        // conn.close();
    }
    catch( VException& e)
    {
        cpLog(LOG_DEBUG, "Caught something... reason: %s ",
              e.getDescription().c_str());
    }

    delete netAddress;
    delete clientSocket;
};
