#ifndef REGISTERMANAGER_HXX
#define REGISTERMANAGER_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const RegisterManager_hxx_Version =
    "$Id: RegisterManager.hxx,v 1.2 2001/09/11 00:33:03 maivu Exp $";


#include "DBResultset.hxx"
#include "PSMessage.hxx"
#include "VRWLock.hxx"

#include <list>

typedef list < string > HostList;


#if defined(__linux__)
#define PROV_USE_HASHMAP 1
#else
#define PROV_USE_HASHMAP 0
#endif


#if (PROV_USE_HASHMAP == 1)
// use hash map if possible... it is faster than map.
#include "StringHash.hxx"
#include <hash_map>
typedef hash_map <string, HostList, StringHash > HostMap;
typedef hash_map <string, HostMap, StringHash >  RecordMap;
#else
#include <map>
typedef map < string, HostList > HostMap;
typedef map < string, HostMap > RecordMap;
#endif



/** RegisterManager manages the registrations and updates for the clients
 *  of the provisioning server.
 */

class RegisterManager
{
    public:
        /** Default Constructor. Does nothing for now.
         */
        RegisterManager();

        /** Default Destructor.  Does nothing for now.
	 */ 
        ~RegisterManager();

        /** Registers a host to receive updates on a given list of records
         *  in a given domain.
         */
        bool registerRecordForUpdate( const string& domain,
                                      FieldListtype keyList,
                                      const string& hostport );

        /** UnRegisters a host from receiving updates on a given list of 
         *  records in a given domain. 
         */
        bool unregisterRecordForUpdate( const string& domain,
                                        FieldListtype keyList,
                                        const string& hostport );

        /** Registers a host to receive updates on a given domain.
         */
        bool registerDomainForUpdate( const string& domain,
                                      const string& hostport );

        /** UnRegisters a host from receiving updates on a given domain.
         */
        bool unregisterDomainForUpdate( const string& domain,
                                        const string& hostport );

        /** Checks the domain and record maps to see whether if a registered
         *  host needs to be updated. 
         */
        void checkRegisterList( const int modType,
                                const string& domain,
                                const string& filter,
                                DBResultset* resultset );

  //
  string errorMsg;

    private:

        /** Stores registered domain updates.
         *  map< string, list<string> >
         */
        HostMap myDomainMap;

        /** Stores registered record updates.
         *  map< string, map< string, list<string> > >
         */
        RecordMap myRecordMap;

        /** Read/Write lock for the maps.
         */
        static VRWLock mapLock;


        /** Checks the domain map to see whether if a registered host needs
         *  to be updated.
         */
        void checkDomainList( const int modType,
                              const string& domain,
                              const string& filter,
                              DBResultset* resultset );

        /** Checks the record map to see whether if a registered host needs
         *  to be updated.
         */
        void checkRecordList( const int modType,
                              const string& domain,
                              const string& filter,
                              DBResultset* resultset );

        /** Sends an UpdateRecord request to a registered host.
         */
        void sendUpdateRecord (const string& domain, 
                               const string& key,
                               const string& hostport );

        /** Sends an UpdateRecord request to a registered host.
         */
        void sendRecordDelete (const string& domain,
                               const string& key,
                               const string& hostport );

        /** Sends an UpdateDomain request to a registered host.
         */
        void sendUpdateDomain (const string& domain,
                               const vector<string>& keyList,
                               const string& hostport );

        /** Sends an UpdateDomain request to a registered host.
         */
        void sendDomainDelete( const string& domain,
                               const vector<string>& keyList,
                               const string& hostport );
};

#endif
