#include <iostream>
#include <string>
#include "Filterscanner.hxx"

using namespace std;

// Character Code Map shared between Scanner and Token classes
CharCode charCodeMap[128];

// Token constructor
Token::Token(void)
{
    myTokencode = tcDummy;
    myTokenstring.erase();
};

// Get the Token Code
TokenCode
Token::GetCode() const 
{ 
    return myTokencode;   
}

// Get the Token String
string&
Token::GetTokenString()
{ 
    return myTokenstring; 
}
 
// Word Token
void
WordToken::Get(istrstream& buffer)
{
    char ch; 
    char tokenbuff[MAXTOKENSIZE];
    int i = 0;
 
    //--Extract the word.
    while ( ch = buffer.peek(),
          ( charCodeMap[ch] == ccLetter) || (charCodeMap[ch] == ccDigit) )
    {
        tokenbuff[i++] = buffer.get();
    } 
    tokenbuff[i] = '\0';

    myTokenstring = tokenbuff; 
    myTokencode = tcWord;
}

// String Token
void
StringToken::Get(istrstream& buffer)
{
    // Read to pass the 1st (" or ') quote.
    char ch = buffer.get();
    char tokenbuff[MAXTOKENSIZE];
    unsigned int i = 0;
 
    // Extract a string, read until another quote (" or ') is found
    while (ch = buffer.get(), true)
    {
        if ((charCodeMap[ch] == ccDQuote) || (charCodeMap[ch] == ccSQuote))
	  break;
	
        if (i >= MAXTOKENSIZE)
	  {
	    tokenbuff[i] = '\0';
	    myTokenstring += tokenbuff;
	    memset(tokenbuff, '\0', MAXTOKENSIZE);
	    i = 0;
	  }
        tokenbuff[i++] = ch;
    } 
    tokenbuff[i] = '\0';
 
    myTokenstring = tokenbuff; 
    myTokencode = tcString;
}

//  Extract one, or two character special symbols
void
SpecialToken::Get(istrstream &buffer)
{
    char ch = buffer.peek();
    char tokenbuff[MAXTOKENSIZE];
    int i = 0;

    tokenbuff[i++] = ch;

    switch (ch) 
    {
    case '(':
        myTokencode = tcLParen;
        buffer.get();  
        break;

    case ')':
        myTokencode = tcRParen;
        buffer.get();
        break;

    case '&':
        myTokencode = tcAnd;
        buffer.get();
        break;

    case '|':
        myTokencode = tcOr;
        buffer.get();
        break;

    case '!':
        myTokencode = tcNot;
        buffer.get();
        break;

    case '*':
        myTokencode = tcStar;
        buffer.get();
        break;

    case '=':
        myTokencode = tcEqual;
        buffer.get();
        break;

    case '<':   
        ch = buffer.get(); // < or <= 
        ch = buffer.peek();
        if (ch == '=') 
        {
            tokenbuff[i++] = '=';
            myTokencode  = tcLe;
            buffer.get();
        }
        else
        {
            myTokencode = tcLt;
        }
        break;

    case '>':
        ch = buffer.get();  // > or >=
        ch = buffer.peek(); 
        if (ch == '=') 
        {
            tokenbuff[i++] = '=';
            myTokencode  = tcGe;
            buffer.get();
        }
        else 
        {
            myTokencode = tcGt;
        }
        break;

    default:    
        myTokencode = tcError; // error
        buffer.get();
        break;
    }
    tokenbuff[i] = '\0';
    myTokenstring = tokenbuff; 
}

// EOS - End of String
void
EOSToken::Get(istrstream &buffer)
{
    myTokenstring.erase();
    myTokencode = tcEOS;
}

// ErrorToken -  Anything not recognized
void
ErrorToken::Get(istrstream &buffer)
{
    char tokenbuff[MAXTOKENSIZE];

    tokenbuff[0] = buffer.get();
    tokenbuff[1] = '\0';

    myTokenstring = tokenbuff;
}

// Scanner Constructor
Scanner::Scanner(const string& buffer):myInbuffer(buffer.c_str())
{
    myInbuffer.unsetf(ios::skipws); // Let Scanner class handle spaces
    int i;
 
    //Initialize the character code map
    for (i = 0; i <= 127; ++i) 
        charCodeMap[i] = ccError;
    for (i = 'a'; i <= 'z'; ++i) 
        charCodeMap[i] = ccLetter;
    for (i = 'A'; i <= 'Z'; ++i) 
        charCodeMap[i] = ccLetter;
    for (i = '0'; i <= '9'; ++i) 
        charCodeMap[i] = ccDigit;

    charCodeMap['+'] = ccSpecial;
    charCodeMap['-'] = ccSpecial;
    charCodeMap['/'] = ccSpecial;
    charCodeMap['^'] = ccSpecial;
    charCodeMap[','] = ccSpecial;
    charCodeMap['['] = ccSpecial;
    charCodeMap[']'] = ccSpecial;
    charCodeMap['{'] = ccSpecial;
    charCodeMap['}'] = ccSpecial;
    charCodeMap[':'] = ccSpecial;
    charCodeMap[';'] = ccSpecial;
    charCodeMap['.'] = ccSpecial;
    charCodeMap['*'] = ccSpecial;
    charCodeMap['!'] = ccSpecial;
    charCodeMap['&'] = ccSpecial;
    charCodeMap['|'] = ccSpecial;
    charCodeMap['='] = ccSpecial;
    charCodeMap['<'] = ccSpecial;
    charCodeMap['>'] = ccSpecial;
    charCodeMap['('] = ccSpecial;
    charCodeMap[')'] = ccSpecial;
    charCodeMap['*'] = ccSpecial;
    charCodeMap['\''] = ccSQuote;
    charCodeMap['\"'] = ccDQuote;
    charCodeMap[' ' ] = ccWhitespace;
    charCodeMap['\t'] = ccWhitespace;
    charCodeMap['\n'] = ccWhitespace;
    charCodeMap['\0'] = ccWhitespace;
    charCodeMap[-1] = ccEOS;    // End of String
}

// GetToken - Return ptr to new Token
Token* Scanner::GetToken(void)
{
    SkipWhitespace();
    char ch = myInbuffer.peek();
    Token* tp;

    // Determine the token class, based on the current character.
    switch (charCodeMap[ch]) 
    {
        case ccLetter:
            tp = &myWordToken;
            break;
        case ccDQuote:
            tp = &myStringToken;  
            break;
        case ccSQuote:
            tp = &myStringToken;  
            break;
        case ccSpecial:
            tp = &mySpecialToken; 
            break;
        case ccEOS:
            tp = &myEOSToken;
            break;
        default:
            tp = &myErrorToken;   
            break;
    }
 
    // Extract a token of that class, and return a pointer to it
    tp->Get(myInbuffer);

    return tp;
}

// SkipWhitespace
void
Scanner::SkipWhitespace(void)
{
    char ch;

    while (ch = myInbuffer.peek(),
           charCodeMap[ch] == ccWhitespace)
    {
        ch = myInbuffer.get();  // Skip whitespaces
    }
}
