#ifndef FILTERPARSER_HXX
#define FILTERPARSER_HXX

#include <vector>
#include <stack>
#include <string>
#include <strstream>
 
/// Type Definitions
/// Character Codes
enum CharCode
{
    ccDummy,
    ccLetter,
    ccDigit,
    ccSpecial,
    ccSQuote,
    ccDQuote,
    ccWhitespace,
    ccEOS,
    ccError,
};

/// Token Codes
enum TokenCode
{
    tcDummy,
    tcWord,
    tcString,
    tcLParen, 
    tcRParen,
    tcEqual,
    tcLt,
    tcGt,
    tcLe,
    tcGe,
    tcAnd,
    tcOr,
    tcNot,
    tcStar,
    tcEscape,
    tcEOS,
    tcError,
};

// Character Code Map shared between Scanner and Token classes
extern CharCode charCodeMap[128];

// Maximum size of a single Token
const unsigned int MAXTOKENSIZE = 64;

/// Token - Base class
class Token
{
    protected:
        ///
        TokenCode myTokencode;
        ///
        string myTokenstring;

    public:
        ///
        Token(void);

        ///
        virtual ~Token(){};

        ///
        TokenCode GetCode() const; 

        ///
        string& GetTokenString();

        /// Get- Abstract function implemented in each derived Token class
        virtual void Get(istrstream &buffer) = 0;
};

/// WordToken- associated with the key in key/value pairs
class WordToken : public Token
{
    public:
        ///
        WordToken(){};
        ///
        virtual ~WordToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// StringToken- associated with a value in key/value pairs
class StringToken : public Token
{
    public:
        ///
        StringToken(){};
        ///
        virtual ~StringToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// EscapeToken- Backslash '\'
class EscapeToken : public Token
{
    public:
        ///
        EscapeToken(){};
        ///
        virtual ~EscapeToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// SpecialToken- i.e. ~ ! @ # $ % ^ & * ( ) _ + ? > < | } { ... 
class SpecialToken : public Token
{
    public:
        ///
        SpecialToken(){};
        ///
        virtual ~SpecialToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// EOSToken- End of String
class EOSToken : public Token
{
    public:
        ///
        EOSToken(){};
        ///
        virtual ~EOSToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// ErrorToken - Anything else
class ErrorToken : public Token
{
    public:
        ///
        ErrorToken(){};
        ///
        virtual ~ErrorToken(){};

        ///
        virtual void Get(istrstream &buffer);
};

/// Scanner class, used to scan and call correct Token class
class Scanner
{
    private:
        ///
        void SkipWhitespace(void);
        ///
        istrstream    myInbuffer;
        ///
        WordToken     myWordToken;
        ///
        StringToken   myStringToken;
        ///
        SpecialToken  mySpecialToken;
        ///
        EOSToken      myEOSToken;
        ///
        ErrorToken    myErrorToken;

    public:
        ///
        Scanner(const string& buffer);

        ///
        virtual ~Scanner(){};
        
        ///
        Token* GetToken(void);
};
#endif
