#include <iostream>
#include <string>
#include "VDbException.hxx"
#include "LDAPFilterparser.hxx"

using namespace std;

// LDAPFilterparser constructor
LDAPFilterparser::LDAPFilterparser(const string& filterbuffer):
myScanner(filterbuffer),
myLDAPsearchclausestrm(myLDAPsearchclause, sizeof(myLDAPsearchclause))
{
    // Parse when constructed
    Parse();
}

// GetToken- Ask myScanner for the next Token
void
LDAPFilterparser::GetToken(void)
{
    myTokenP = myScanner.GetToken();
    myTokencode = myTokenP->GetCode();
    myTokenstring = myTokenP->GetTokenString();
}

// Parse-
void
LDAPFilterparser::Parse(void)
{
    // Get First Token
    GetToken();
    
    // Parse Search Filter
    ParseFilter();

    // Check that the last token is the EndOfString
    if (myTokencode != tcEOS)
    {
         string err = "Error in Parse: End of String expected";
         throw VDbException(err, __FILE__, __LINE__, 0);
    }
    myLDAPsearchclausestrm << ends;
}

// ParseFilter-
void
LDAPFilterparser::ParseFilter(void)
{

    // Grammar specifies that filters must always begin w/ a Left Paren
    if ( myTokencode != tcLParen )
    {
        string err = "Error in ParseFilter: Filter must begin with Left Paren";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        myLDAPsearchclausestrm << myTokenstring ;
    }

    GetToken();
    ParseFilterComposition();
    
    if ( myTokencode != tcRParen )
    {
        string err = "Error in ParseFilter: Filter must end with Right Paren";
        throw VDbException(err, __FILE__, __LINE__, 0);                         
    }
    else
    {
        myLDAPsearchclausestrm << myTokenstring ;
        GetToken();
    }

}

//
void
LDAPFilterparser::ParseFilterComposition(void)
{

    // Indirect recursion to allow nested filters.
    // ParseFilter -> 
    //      ParseFilterComposition 
    //              -> ParseFilterList
    //                      -> ParseFilter
    switch ( myTokencode )
    {
        case tcAnd:
            // LDAPsearch uses same prefix grammar
            // So just save the Token
            myLDAPsearchclausestrm << myTokenstring ;
            // Parse 1 or more filters
            ParseFilterList();
            break;

        case tcOr:
            // LDAPsearch uses same prefix grammar
            // So just save the Token
            myLDAPsearchclausestrm << myTokenstring ;
            myLDAPsearchclausestrm << myTokenstring;
            // Parse 1 or more filters
            ParseFilterList();
            break;

        case tcNot:
            // LDAPsearch uses same prefix grammar
            // So just save the Token
            myLDAPsearchclausestrm << myTokenstring;
            // Parse 1 filter
            GetToken();
            ParseFilter();
            break;

        case tcWord:
            ParseItem();
            break;

        default:
            string err = "Error in ParseFilterComposition: Token not recognized";
            throw VDbException(err, __FILE__, __LINE__, 0);
    }
}

// ParseFilterList- Parse one or more filters
void
LDAPFilterparser::ParseFilterList(void)
{
    string currentoperator;

    GetToken();

    do
    {
       ParseFilter();
    }

    while (myTokencode == tcLParen);
}

// ParseItem- Items are [ Fieldname <op> "Value" ] sequences
void
LDAPFilterparser::ParseItem(void)
{
    // Current Token is Field name
    myLDAPsearchclausestrm << myTokenstring;
    
    // Next Token should be operator
    GetToken();
    switch ( myTokencode )
    {
        case tcLe:
        case tcGe:
            myLDAPsearchclausestrm << myTokenstring;
            GetToken();
            // After parsing operator, next Token should be Value
            ParseValue();
            break;
        case tcEqual:
            myLDAPsearchclausestrm << myTokenstring;
            GetToken();
            // Next Token should be StarValue
            ParseStarValue();
            break;
        default:
            string err = "Error in ParseItem: Token not recognized";
            throw VDbException(err, __FILE__, __LINE__, 0);
   }
}

// ParseValue- This value should not be a substring. This Token
// always comes after a >, <, <=, >=, operator and should be a literal
// value. Basically, the grammar states you can't say 
// (vocalphonenumber < "200*"). If you do this, the backed DB
// will probably give you incorrect results.
void
LDAPFilterparser::ParseValue(void)
{
    if (( myTokencode != tcString) && (myTokencode != tcWord ))
    {
         string err = "Error in ParseValue: Token not recognized";
         throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        // LDAPsearch doesn't like double quoted string values
        myLDAPsearchclausestrm << myTokenstring;

    }
    GetToken();
}

// ParseStarValue- This value should be a substring. This Token
// always comes after an = operator.
void
LDAPFilterparser::ParseStarValue(void)
{
    if ( myTokencode != tcString )
    {
        string err = "Error in ParseValue: Token must be string";
        throw VDbException(err, __FILE__, __LINE__, 0);
    }
    else
    {
        // LDAPsearch doesn't like double quoted string values
        // So leave the quotes off
        myLDAPsearchclausestrm << myTokenstring;
    }
    GetToken();
}

// GetLDAPsearchclause- Return LDAPsearch clause
const char* 
LDAPFilterparser::GetParsedresult(void)
{
    return myLDAPsearchclause;
}
