#include <iostream>
#include "LDAPResultset.hxx"
#include "VDbException.hxx"

using namespace std;

// LDAPResultset constructor
LDAPResultset::LDAPResultset(LDAPMessage* result, LDAP* handle, 
        const string& searchfilter):
myLDAPresult(result),
myLDAPhandle(handle),
mySearchFilter(searchfilter)
{
    myRecordCount = 0;

    // Count Records to get myRecordCount
    for (LDAPMessage* e = ldap_first_entry(myLDAPhandle, myLDAPresult);
         e != NULL;
         e = ldap_next_entry(myLDAPhandle, e))
    {
        myRecordCount++;
    }

    myFieldList = GetFields();
};
 
// GetRecords[offset] - returns the record at offset
const ResultRecordtype&
LDAPResultset::operator[](unsigned short offset)
{
    LDAPMessage*     entry;
    char*            attr;
    char**           vals;
    unsigned short   entrycount = 0;
    unsigned short   i;

    // Check offsets are within range of results
    if (offset > myRecordCount) 
    {
        throw VDbException("Record Offset out of range", __FILE__, __LINE__, 0);
    }

    // Clear contents of myResultTable for now
    myResultRecord.clear();

    // Move entry pointer to match offset
    for (entry = ldap_first_entry(myLDAPhandle, myLDAPresult);
        (entry != NULL) & (entrycount != offset);
         entry = ldap_next_entry(myLDAPhandle, entry), entrycount++);

    // Fill record buffer by iterating through each field in myFieldList
    for (i = 0; i < myFieldList.size(); i++)
    {
        attr = strdup(myFieldList[i].c_str());
        if ((vals = ldap_get_values(myLDAPhandle, entry, attr)) != NULL)
        {
            // There should only be one value for a given attribute,
            // therefore only push_back vals[0]
            myResultRecord.push_back(vals[0]);
            ldap_value_free(vals);
        }
    }

    return myResultRecord;
}

// GetRecords(offset) - returns the record at offset
const ResultRecordtype&
LDAPResultset::GetRecord(unsigned short offset)
{
    LDAPMessage*     entry;
    char*            attr;
    char**           vals;
    unsigned short   entrycount = 0;
    unsigned short   i;

    // Check offsets are within range of results
    if (offset > myRecordCount) 
    {
        throw VDbException("Record Offset out of range", __FILE__, __LINE__, 0);
    }

    // Clear contents of myResultRecord for now
    myResultRecord.clear();

    // Move entry pointer to match offset
    for (entry = ldap_first_entry(myLDAPhandle, myLDAPresult);
        (entry != NULL) & (entrycount != offset);
         entry = ldap_next_entry(myLDAPhandle, entry), entrycount++);

    // Fill record buffer by iterating through each field in myFieldList
    for (i = 0; i < myFieldList.size(); i++)
    {
        attr = strdup(myFieldList[i].c_str());
        if ((vals = ldap_get_values(myLDAPhandle, entry, attr)) != NULL)
        {
            // There should only be one value for a given attribute,
            // therefore only push_back vals[0]
            myResultRecord.push_back(vals[0]);
            ldap_value_free(vals);
        }
    }

    return myResultRecord;
};

// GetRecords(offset1, offset2) - returns all records between offset1
// and offset2
const ResultTabletype&
LDAPResultset::GetRecords(unsigned short offset1, unsigned short offset2)
{
    LDAPMessage*      entry;
    ResultRecordtype  recordbuff;
    char*             attr;
    char**            vals;
    unsigned short    entrycount = 0;
    unsigned short    i;

    // Check offsets are within range of results
    if ((offset1 > offset2) || (offset1 > myRecordCount) 
            || (offset2 > myRecordCount))
    {
        throw VDbException("Record Offsets out of range", __FILE__, __LINE__, 0);
    }

    // Clear contents of myResultTable for now
    myResultTable.clear();

    // Move entry pointer to match offset1
    for (entry = ldap_first_entry(myLDAPhandle, myLDAPresult);
        (entry != NULL) & (entrycount != offset1 - 1);
         entry = ldap_next_entry(myLDAPhandle, entry), entrycount++);

    // Iterate through entries between offset1 and offset2
    for (entry = ldap_next_entry(myLDAPhandle, entry);
        (entry != NULL) & (entrycount != offset2);
         entry = ldap_next_entry(myLDAPhandle, entry), entrycount++)
    {
        // Fill record buffer by iterating through each field in myFieldList
        for (i = 0; i < myFieldList.size(); i++)
        {
            attr = strdup(myFieldList[i].c_str());
            if ((vals = ldap_get_values(myLDAPhandle, entry, attr)) != NULL)
            {
                // There should only be one value for a given attribute,
                // therefore only push_back vals[0]
                recordbuff.push_back(vals[0]);
                ldap_value_free(vals);
            }
        }
        // Insert record buffer into result table
        myResultTable.push_back(recordbuff);
 
        // Clear recordbuff for next record
        recordbuff.clear();
    }
    return myResultTable;
};

// GetRecords - returns all records from a search
const ResultTabletype&
LDAPResultset::GetRecords()
{
    ResultRecordtype  recordbuff;
    LDAPMessage*      entry;
    char*             attr;
    char**            vals;
    unsigned short    i;

    // Clear contents of myResultTable for now
    myResultTable.clear();

    for (entry = ldap_first_entry(myLDAPhandle, myLDAPresult); 
         entry != NULL;
         entry = ldap_next_entry(myLDAPhandle, entry))
    {
        // Fill record buffer by iterating through each field in myFieldList
        for (i = 0; i < myFieldList.size(); i++)
        {
            attr = strdup(myFieldList[i].c_str());
            if ((vals = ldap_get_values(myLDAPhandle, entry, attr)) != NULL)
            {
                // There should only be one value for a given attribute,
                // therefore only push_back vals[0]
                recordbuff.push_back(vals[0]);
                ldap_value_free(vals);
            }
        }
        // Insert record buffer into result table
        myResultTable.push_back(recordbuff);
 
        // Clear recordbuff for next record
        recordbuff.clear();
    }
    return myResultTable;
};

// GetRecordCount- Return Record count
const unsigned int
LDAPResultset::GetRecordCount(void) const
{
    return myRecordCount;
};

// GetFields- Return fields on a resultset
const FieldListtype&
LDAPResultset::GetFields()
{
    LDAPMessage*   entry;
    BerElement*    ber;
    char*          attr;

    // Clear contents of myResultRecord for now
    myResultRecord.clear();

    // If Records exist, fill myFieldList
    if (myRecordCount)
    {
        entry = ldap_first_entry(myLDAPhandle, myLDAPresult);
        for (attr = ldap_first_attribute(myLDAPhandle, entry, &ber);
             attr != NULL;
             attr = ldap_next_attribute(myLDAPhandle, entry, ber))
        {
             myResultRecord.push_back(attr);
        }
    }

    return myResultRecord;
}

// GetmySearchFilter- Return searchfilter that was used to create this
//    resultset.  This is normally used when a resultset is passed back
//    to DBUpdate() or DBSelect()
const string&
LDAPResultset::GetSearchFilter(void) const
{
    return mySearchFilter;
};
