// ********************************************************************
//  Database creation tool
// ********************************************************************


#include <iostream>
#include <string>
#include <vector>
#include "DBConnection.hxx"
#include "DBResultset.hxx"
#include "ConnectionManager.hxx"
#include "VDbException.hxx"
#include <sax2/SAX2XMLReader.hpp>
#include <sax2/XMLReaderFactory.hpp>
#include <sax2/DefaultHandler.hpp>
#include <util/PlatformUtils.hpp>
#include "ParsedDef.hxx"
#include <unistd.h>
#include "DBInfo.h"


using namespace std;

/// Handler class for the XML definition
class MySAX2Handler : public DefaultHandler {
private:
    bool foundTable;


public:
    
    MySAX2Handler() {}

    void startElement(
        const   XMLCh* const    uri,
        const   XMLCh* const    localname,
        const   XMLCh* const    qname,
        const   Attributes&     attrs
        );

    void endElement (const XMLCh *const uri,
                     const XMLCh *const localname,
                     const XMLCh *const qname);

    void fatalError(const SAXParseException&);

    ParsedDef myDef;
    ParsedDef getMyDef() { return myDef; }
    
};



void MySAX2Handler::endElement (const XMLCh *const uri,
                     const XMLCh *const localname,
                     const XMLCh *const qname)
{
  string element = XMLString::transcode(qname);
 
  if (strcmp(element.c_str(), "DBEntry") == 0)
    {
      if (!foundTable)
      {
	//if the table type was not found within this data definition
	//make this column to be of the default type
        myDef.addFieldType(STRINGTYPE);
      }
      else
      {
        //if the tabletype was found within this data definition, make 
        //make this column to be of the text blob type
	myDef.addFieldType(TEXTBLOBTYPE);
      }
      foundTable = false; //reset
    }    
}


/** Called when a new element is seen. We only care if it is DataDef (begining
    of a new table) or DBEntry (a column in that table) */
void MySAX2Handler::startElement(const   XMLCh* const    uri,
                            const   XMLCh* const    localname,
                            const   XMLCh* const    qname,
                            const   Attributes&     attrs)
{

    string element = XMLString::transcode(qname);
    const XMLCh *key, *value;
    string sValue;

    
    if (strcmp(element.c_str(), "DataDefn") == 0) {
        
        key = XMLString::transcode("name");
        value = attrs.getValue(key);
        if (value != NULL)
        {
            sValue = XMLString::transcode(value);
            myDef.setDomain(sValue);
        } else
        {
            cerr << "Attribute \"name\" not found for element DataDefn" << endl;
            exit(0);
        }

        key = XMLString::transcode("key");
        value = attrs.getValue(key);
        if (value != NULL)
        {
            sValue = XMLString::transcode(value);
            myDef.setPrimaryKey(sValue);
        } else
        {
            cerr << "Attribute \"key\" not found for element DataDefn" << endl;
            exit(0);
        }
     }

    
    if (strcmp(element.c_str(), "DBEntry") == 0) {

        key = XMLString::transcode("name");
        value = attrs.getValue(key);
        if (value != NULL)
        {
            sValue = XMLString::transcode(value);
            myDef.addField(sValue);
        } else
        {
            cerr << "Attribute \"name\" not found for element DBEntry" << endl;
            exit(0);
        }
    }

    else if (strcmp(element.c_str(), "TableType") == 0)
    {
        foundTable = true;
    }
    
}




/// Handle a fatal error while parsing the description file.   
void MySAX2Handler::fatalError(const SAXParseException& exception)
{
    cout << "Fatal Error: " << XMLString::transcode(exception.getMessage())
         << " at line: " << exception.getLineNumber()
         << endl;
    exit(0);
}

void printHelp() {

    cerr << "\n  Usage: dbcreator [-hsfv] <XML description filename>" << endl;
    cerr << endl;
    cerr << "  Creates an entry in the provisioning database from an" << endl;
    cerr << "  XML description file." << endl;
    cerr << endl;
    cerr << "  Options: " << endl;
    cerr << "      -h     Prints this help message" << endl;
    cerr << "      -s     Script mode. Run non-interactive. Defaults to not deleting any" << endl;
    cerr << "             data or database entries." << endl;
    cerr << "      -f     Force mode. Run non-interactive, forcing the new XML to be used." << endl;
    cerr << "             DATA AND DATABASE ENTRIES MAY BE SILENTLY DELETED THIS WAY!" << endl;
    cerr << "      -v     Verbose mode. Prints aditional diagnostic messages. Primarily" << endl;
    cerr << "             applies to script mode" << endl;

    cerr << "\n\n" << endl;
    cerr << "NOTE: The config file \"dbcreate.cfg\" is needed to provide DB information.\n\n\n";
    cerr << "#######################################\n";
    cerr << "#   Sample entries in dbcreate.cfg    #\n";
    cerr << "#######################################\n";
    cerr << "DBConnection    string          MYSQL" << endl;
    cerr << "database        string          vocal" << endl;
    cerr << "host            string          vvs-lion" << endl;
    cerr << "user            string          vocaluser" << endl;
    cerr << "password        string          vocaluser" << endl;
    cerr << endl << endl << endl;


    cerr << endl;

}


// Main driver program
int
main(int argc, char **argv )
{

    char c;

    bool force, script, verbose, print;

    force = script = verbose = false;
    
    while (1)
    {

        c = getopt(argc, argv, "svfh");

        if (c == -1) break;

        switch(c) {

            case 's':
                // script mode
                script = true;
                break;
            case 'v':
                // verbose mode
                verbose = true;
                break;
            case 'f':
                // force mode
                force = true;
                break;
            case 'h':
                printHelp();
                return -1;
                break;

            case '?':
            default:
                printHelp();
                cerr << " ******** CREATION FAILED! ********" << endl;
                return -1;
                break;

        }

    }

    if (optind != (argc - 1))
    {
        cerr << "\n  Error! - need to provide precisely 1 filename" << endl;
        printHelp();
        cerr << " ******** CREATION FAILED! ********" << endl;
        return -1;
    }

    if (force && script) {
        cerr << "\n  Cannot select both \"force\" and \"script\" modes"
             << endl;
        printHelp();
        cerr << " ******** CREATION FAILED! ********" << endl;
        return -1;
    }
    
    // Do we print diagnostics? If interactive we do, if script or force,
    // no, unless they provide verbose switch
    print = ( (!(script || force)) || verbose);
    
    // Initialize the parser, read in the file and parse it
    try {
        XMLPlatformUtils::Initialize();
    }
    catch (const XMLException& toCatch) {
        cerr << "\n  Error during XML initialization! :\n"
             << "  " << toCatch.getMessage() << "\n";
        cerr << " ******** CREATION FAILED! ********" << endl;
        return -1;
    }

    char* xmlFile = argv[optind];

    if (print) {
        cout << "Attempting to open description file " << xmlFile << endl;
    }
    
    SAX2XMLReader* parser = XMLReaderFactory::createXMLReader();
    
    MySAX2Handler *myHandler = new MySAX2Handler();
    ContentHandler* contentHandler = myHandler;
    ErrorHandler* errHandler = (ErrorHandler*) contentHandler;
    parser->setContentHandler(contentHandler);
    parser->setErrorHandler(errHandler);

    if (print) {
        cout << "Attempting to parse XML content of description file" << endl;
    }

    try {
        parser->parse(xmlFile);
    }
    catch (const XMLException& toCatch) {
        cerr << "\n  File not found: '" << xmlFile << "'\n"
             << "  Exception message is: \n"
             << "  " << toCatch.getMessage() << "\n" ;
        cerr << " ******** CREATION FAILED! ********" << endl;
        return -1;
    }

    // get the data from the parsed file
    ParsedDef result = myHandler->getMyDef();
    string myDomain = result.getDomain();
    DBInfo dbInfo = parseConfigFile();

    ConnectionManager dbmanager;
    DBConnection* dbconn = dbmanager.GetDBConnection(dbInfo);

    if (print) {
        cout << "Connected to database." << endl;
        cout << "Looking for domain: " << myDomain << endl;
    }
        
    // See if this table already exists in the database
    bool exists = dbconn->DBVerify(myDomain);

    if (exists) {

        if (print) {
            cout << "Domain already exists. " << endl;
            cout << "  Need to compare new definition to existing definition."
                 << endl;
        }

        string newTableKey;
        ResultTabletype newTableDesc;
        
        try {
            newTableDesc = dbconn->DBDescribe(myDomain);
            newTableKey = dbconn->DBPrimaryKey(myDomain);
        }
        catch (VDbException &e) {
            cerr << "\n    Error! Domain exists but unable to get description"
                 << "\n" << endl;
            cerr << " ******** CREATION FAILED! ********" << endl;
        }
            
        ParsedDef newDesc;

        newDesc.setDomain(myDomain);
        newDesc.setPrimaryKey(newTableKey);

        if (print) {
            cout << "    Existing Primary Key : " << newTableKey << endl;
        }
        
        ResultTabletype::const_iterator resultIterator;
        
        for (resultIterator = newTableDesc.begin();
        resultIterator != newTableDesc.end(); resultIterator++)
        {
            if (print) {
                cout << "      Existing Column    : " <<
                    (*resultIterator)[0] << endl;
            }
            newDesc.addField((*resultIterator)[0]);
        }

        DifferenceSet diffs = result.compareOld(newDesc);
        
        FieldListtype fn = diffs.getNewFields();
        FieldListtype fd = diffs.getDelFields();
        
        FieldListtype::iterator itr;
        
        // check for same primary key handle different depending on options
        if (diffs.changedPrimary()) {

            // In script mode, we don't delete, so we can't change primary key
            if (script) {

                cerr << "\n  Error! Unable to change primary key in database."
                     << "\n" << endl;
                cerr << "  If you want to change primary key, the"
                     << " contents of the domain need to be" << endl;
                cerr << "  removed. If you want to do this, try again"
                     << " without the -s option.\n" << endl;
                cerr << " ******** CREATION FAILED! ********" << endl;
                return -1;
                
            }

            // in force mode we go nuke the db
            if (force) {

                if (print) {
                    cout << "  New primary key is : "
                         << result.getPrimaryKey() << endl;
                    cout << "  Because primary key has changed, must "
                         << "remove old domain." << endl;
                    cout << "  Domain " << myDomain
                         << " being removed." << endl;
                }
                
                try {
                    dbconn->DBDrop(myDomain, false);
                }
                catch (VDbException &e) {
                    cerr << "\n    Error! Unable to remove domain " <<
                        myDomain << "from database.\n" << endl;
                    cerr << "    Domain may be in unstable state.\n" << endl;
                    cerr << " ******** CREATION FAILED! ********" << endl;
                    return -1; 
                }

                try {
                    dbconn->DBCreate(myDomain, result.getFieldList(), 
                                     result.getFieldTypes(), 
                                     result.getPrimaryKey(), IGNOREEXISTENCE);
                }
                catch (VDbException &e) {
                    cerr << "\n  Error! Unable to create domain in database."
                         << "\n" << endl;
                    cerr << "  Perhaps the description was incorrect or you"
                         << " don't have permission to create?\n" << endl;
                    cerr << " ******** CREATION FAILED! ********" << endl;
                    return -1;
                }

                if (print) {
                    cout << "The new database domain has been created.\n"
                         << endl;
                    cout << " ******** SUCCESS ********" << endl;
                }
                    
                return 0;
            }
            
            // otherwise we need to prompt the user to ask what to do
            string answer = "";
            bool valid = false;
            
            while (!valid) {
                
                // handle asking for permission
                cout << "\n    The primary key for this database must be "
                     << "changed to process this request." << endl;
                cout << "    New primary key is : " << result.getPrimaryKey()
                     << endl;
                cout << "    This requires REMOVING THE ENTIRE DOMAIN!"
                     << endl;
                cout << "    ALL DATA IN FOR ALL RECORDS IN THIS DOMAIN " 
                     << "WILL BE LOST!!!"
                     << endl;
                cout << "    Delete this domain? (yes/no): ";
                cin >> answer;
                
                if (answer == "yes" || answer == "YES" || answer == "Yes") {
                    

                    cout << "    Domain " << myDomain
                         << " being removed." << endl;
                
                    try {
                        dbconn->DBDrop(myDomain, false);
                    }
                    catch (VDbException &e) {
                        cerr << "\n    Error! Unable to remove domain " <<
                        myDomain << "from database.\n" << endl;
                        cerr << "    Domain may be in unstable state.\n"
                             << endl;
                        cerr << " ******** CREATION FAILED! ********" << endl;
                        return -1; 
                    }

                    try {
                        dbconn->DBCreate(myDomain, result.getFieldList(),
                                         result.getFieldTypes(),
                                         result.getPrimaryKey(),
                                         IGNOREEXISTENCE);
                    }
                    catch (VDbException &e) {
                        cerr << "\n  Error! Unable to create domain in "
                             << "database.\n" << endl;
                        cerr << "  Perhaps the description was incorrect or "
                             << " you don't have permission to create?\n"
                             << endl;
                        cerr << " ******** CREATION FAILED! ********" << endl;
                        return -1;
                    }

                    cout << "    The new database domain has been created.\n"
                         << endl;
                    cout << " ******** SUCCESS ********" << endl;
                    
                    // don't really need to set this since we exit...
                    valid = true;
                    return 0;
                
                } else if (answer == "no" || answer == "N0" ||
                               answer == "No") {
                        
                        cout << "    Domain left in place. Database will not"
                             << " match XML description!" << endl;

                        valid = true;
                        
                } else {
                    
                    cout << "\n    Please answer yes or no." << endl;

                }

            }
            
            cout << endl;

        }
        
        // check for removed fields, handle different depending on options
        for (itr = fd.begin(); itr != fd.end(); itr++) {
            
            // if script mode, don't delete fields (we will abort before the
            // first field gets deleted)
            if (script) {

                cerr << "\n  Error! Unable to remove fields from database."
                     << "\n" << endl;
                cerr << "  This description requires that fields"
                     << " be removed from the database. If you" << endl;
                cerr << "  want to do this, try again"
                     << " without the -s option.\n" << endl;
                cerr << "  First field encountered that needs to be removed: "
                     << *itr << endl;
                cerr << "  There may be others.\n" << endl;
                cerr << " ******** CREATION FAILED! ********" << endl;
                return -1;

            }
                
            if (force) {

                if (print) {
                    cout << "  Removing field " << *itr << " from database"
                         << endl;
                }

                try {
                    dbconn->DBAlter(myDomain, *itr);
                }
                catch (VDbException &e) {
                    cerr << "\n  Error! Unable to remove field " << *itr
                         << " from database." << endl;
                    cerr << "  Perhaps the description was incorrect or you"
                         << " don't have permission to delete?\n" << endl; 
                    cerr << " ******** CREATION FAILED! ********" << endl;
                    return -1;
                }
                
            }

            string answer = "";
            bool valid = false;
            
            while (!valid) {

                // handle asking for permission
                cout << "\n    The following field needs to be removed from "
                     << "the database to process" << endl;
                cout << "    this request: " << *itr << endl;
                cout << "    ALL DATA IN THIS FIELD FOR ALL RECORDS WILL BE "
                     << "LOST!!!"
                     << endl;
                cout << "    Delete this field? (yes/no): ";
                cin >> answer;
                
                if (answer == "yes" || answer == "YES" || answer == "Yes") {

                    cout << "    Removing field " << *itr << " from database"
                         << endl;
                    try {
                        dbconn->DBAlter(myDomain, *itr);
                    }
                    catch (VDbException &e) {
                        cerr << "\n  Error! Unable to remove field " << *itr
                             << " from database." << endl;
                        cerr << "  Perhaps the description was incorrect or"
                             << " you don't have permission to delete?\n"
                             << endl; 
                        cerr << " ******** CREATION FAILED! ********" << endl;
                        return -1;
                    }

                    cout << "    Removed field " << *itr << " successfully."<< endl;
                    valid = true;
                    
                } else if (answer == "no" || answer == "N0" ||
                               answer == "No") {
                        
                        cout << "    Field left in place. Database will not"
                             << " match XML description!" << endl;

                        valid = true;
                        
                } else {
                    
                    cout << "\n    Please answer yes or no." << endl;

                }

            }
            
            cout << endl;
            
        }

        // check for new fields. In all cases, these are just added
        for (itr = fn.begin(); itr != fn.end(); itr++) {
            
            if (print) {
                cout << "  Adding field " << *itr << " to database"
                     << endl;
            }
            
            try {
                dbconn->DBAlter(myDomain, *itr, STRINGTYPE);
            }
            catch (VDbException &e) {
                cerr << "\n  Error! Unable to add field " << *itr
                     << " to database." << endl;
                cerr << "  Perhaps the description was incorrect or you"
                     << " don't have permission to add?\n" << endl; 
                cerr << " ******** CREATION FAILED! ********" << endl;
                return -1;
            }

        }
        
    }
    // if the domain doesn't exist, we add it
    else
    {
        if (print) {
            cout << "Domain does not exist. Will create new domain." << endl;
        }
        try {
            dbconn->DBCreate(myDomain, result.getFieldList(),
			     result.getFieldTypes(),
                             result.getPrimaryKey(), IGNOREEXISTENCE);
        }
        catch (VDbException &e) {
            cerr << "\n  Error! Unable to create domain in database." << endl;
            cerr << "  Perhaps the description was incorrect or you don't" 
                 << "  have permission to create?\n" << endl;
            cerr << " ******** CREATION FAILED! ********" << endl;
            return -1;
        }
        if (print) {
            cout << "The new database domain has been created.\n" << endl;
            cout << " ******** SUCCESS ********" << endl;
        }
    }
    
    return 0;

        
    
    // See if there is an existing entry for this domain in the database
//    ResultTabletype oldResult = dbconn->DBDescribe
    
    // No way to check for result?

/*    

    ParsedDef a, b;

    a.setPrimaryKey("prim");
    try {
        a.addField("field");
        a.addField("fielda");
        a.addField("fieldb");
        a.addField("fieldc");
        a.addField("fieldd");
        a.addField("fielde");
    } catch (VBadDataException &e) {
        cerr << "Error while adding a field. Error: " << endl;
        cerr << e.getDescription() << endl;
    }

    b.setPrimaryKey("prim");
    b.addField("fielda");
    b.addField("fieldb");
    b.addField("fieldc");
    b.addField("fieldd");

*/
    
} // main






