//********************************************************************
//  Test driver for DB Modules
//********************************************************************

#include <iostream>
#include <string>
#include <vector>
#include "DBConnection.hxx"
#include "DBResultset.hxx"
#include "ConnectionManager.hxx"
#include "VDbException.hxx"

using namespace std;

// Overload << to print vectors
template <typename Storable>
ostream& operator<<(ostream& os, const vector<Storable>& v)
{
    vector<Storable>::const_iterator p;
    for ( p = v.begin(); p != v.end(); p++) 
    {
        os << *p ;
        // Don't print a delimiter after the last vector element
        if ((p + 1) != v.end()) os << "   ";
    }
    return os;
}

// Print a resultset, which is a STL vector<vector<string>>
// 'Not sure if this belongs in DBResultset, probably not.
void printresultset(ResultTabletype& searchresult)
{
    ResultTabletype::const_iterator resultiterator;
    
    for (resultiterator = searchresult.begin();
        resultiterator != searchresult.end(); resultiterator++)
    {
        cout << *resultiterator << endl;
    }
}

// Hard code searchfilters  - For testing, copy various search filters in 
// each Testcase function.  Note certain filters only work for certain domains
//   
//   These filters only work w/ vocaluser as your domain
// const string filter = "(vocalname=\"100*\")";
// const string filter = "(!(vocalname=\"120*\"))";
// const string filter = "(&(vocalcfb=\"3333\")(vocalipaddr=\"192.168.1.12\"))";
// const string filter = "(!(|(vocalname=\"10*\")(vocalname=\"11*\")(vocalname=\"12*\")(vocalname=\"13*\")(vocalname=\"14*\")(vocalname=\"15*\")(vocalname=\"16*\")(vocalname=\"17*\")))";
//   
//   These filters only work w/ vocalserver as your domain
// const string filter = "(&(vocalhostname=\"host10\")(vocaltype=\"feature1\"))";
// const string filter = "(&(vocalhostname=\"host10\")(vocaltype=\"feature1\"))";
// const string filter = "(&(column5>\"50\")(column6<\"800\"))";

// DBSelect(Domain, SearchFilter, FieldList)
void
Testcase1(DBConnection* dbconn)
{
    // ResultTabletype is STL vector<vector<string>> to store n records
    ResultTabletype restable;
    // ResultRecordtype is STL vector<string> to store one record
    ResultRecordtype resrecord;

    DBResultset* resultset;

    // Hard code filter for DBSelect w/ Order test
    const string filter = "(|(vocalname=\"100*\")(vocalipaddr=\"192.0.0.1\"))";

    // Hard code fieldlist, FieldListype is vector<string>
    FieldListtype fieldlist;  // fields from vocaluser domain
    fieldlist.push_back("vocalname");
    fieldlist.push_back("vocalipaddr");
    fieldlist.push_back("vocalmarshalgroup");
    fieldlist.push_back("vocalauthenticationtype");
    fieldlist.push_back("vocalcfa");
    fieldlist.push_back("vocalfnab");
    fieldlist.push_back("vocalusertimestamp");

    // To get all fields in a domain, do this
    // FieldListtype fieldlist; 
    // fieldlist.push_back("*");

    cout << "\nDBSelect(vocaluser, "<< filter << ", fieldlist):\n";

    resultset = dbconn->DBSelect("vocaluser", filter, fieldlist);

    // Test various Resultset accessors
    // Print RecordCount
    cout << "\nDBResultset::GetRecordCount() \n";
    cout << resultset->GetRecordCount() << endl;

    // Get Field list from resultset
    cout << "\nDBResultset::GetFields() \n";
    resrecord = resultset->GetFields();
    cout << resrecord << endl;

    // Get all records from DBResultset
    cout << "\nDBResultset::GetRecords() \n";
    restable = resultset->GetRecords();
    printresultset(restable);

    // Get record at position [offset]
    cout << "\nDBResultset[3] \n";
    resrecord = (*resultset)[3];
    cout << resrecord << endl;

    // Get records between offset1 and offset2
    cout << "\nDBResultset::GetRecords(2, 4) \n";
    restable = resultset->GetRecords(2, 4);
    printresultset(restable);

    delete resultset;
}

// DBSelect(Domain, SearchFilter, FieldList, Order, OrderField)
void
Testcase2(DBConnection* dbconn)
{
    ResultTabletype restable;
    ResultRecordtype resrecord;

    DBResultset* resultset;

    // Hard code filter for DBSelect w/ Order test
    const string filter = "(&(column5>\"90\")(column6<=\"200\")(!(column7=\"30\")))";
/*
 * const string filter = "(vocalname=\"102*\")";

    // Hard code fieldlist, FieldListype is vector<string>
    FieldListtype fieldlist;  // fields from vocaluser domain
    fieldlist.push_back("vocalname");
    fieldlist.push_back("vocalipaddr");
    fieldlist.push_back("vocalmarshalgroup");
    fieldlist.push_back("vocalauthenticationtype");
    fieldlist.push_back("vocalcfa");
    fieldlist.push_back("vocalfnab");
    fieldlist.push_back("vocalusertimestamp");

    cout << "\nDBSelect(vocaluser, " << filter << ", fieldlist, DESCENDING, vocalname):\n";

    resultset = dbconn->DBSelect("vocaluser", filter, fieldlist, 
            DESCENDING, "vocalname");
*/

    // Hard code fieldlist, FieldListype is vector<string>
    FieldListtype fieldlist;  // fields from vocaluser domain
    fieldlist.push_back("column1");
    fieldlist.push_back("column2");
    fieldlist.push_back("column3");
    fieldlist.push_back("column4");
    fieldlist.push_back("column5");
    fieldlist.push_back("column6");
    fieldlist.push_back("column7");

    cout << "\nDBSelect(feature2, " << filter << ", fieldlist):\n";

    resultset = dbconn->DBSelect("feature2", filter, fieldlist);

    // Get all records from DBResultset
    cout << "\nDBResultset::GetRecords() \n";
    restable = resultset->GetRecords();
    delete resultset;

    printresultset(restable);

}

// DBUpdate(Domain, SearchFilter, FieldValueSet)
void
Testcase3(DBConnection* dbconn)
{
    ResultTabletype restable;
    DBResultset* resultset;

    // Hard code filter for DBSelect w/ Order test
    const string filter = "(|(vocalname=\"100*\")(vocalipaddr=\"192.0.0.1\"))";

    // Hard code fieldlist, FieldListype is vector<string>
    FieldListtype fieldlist;  // fields from vocaluser domain
    fieldlist.push_back("vocalname");
    fieldlist.push_back("vocalipaddr");
    fieldlist.push_back("vocalmarshalgroup");
    fieldlist.push_back("vocalauthenticationtype");
    fieldlist.push_back("vocalcfa");
    fieldlist.push_back("vocalfnab");
    fieldlist.push_back("vocalusertimestamp");

    // Hard code fieldvalue set map<string,string> for DBUpdate test
    FieldValuetype fieldvaluelist;
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalcfa"), "aaaa"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalfnab"), "bbbb"));

    cout << "\nDBSelect(vocaluser, " << filter 
        << ",fieldlist) before DBUpdate:\n";
    resultset = dbconn->DBSelect("vocaluser", filter, fieldlist);
    restable = resultset->GetRecords();
    delete resultset;
    printresultset(restable);

    cout << "\nDBUpdate(vocaluser, " << filter << ", fieldvaluelist):\n";
    dbconn->DBUpdate("vocaluser", filter, fieldvaluelist);

    cout << "\nDBSelect(vocaluser, " << filter 
        << ",fieldlist) after DBUpdate:\n";
    resultset = dbconn->DBSelect("vocaluser", filter, fieldlist);
    restable = resultset->GetRecords();
    delete resultset;
    printresultset(restable);

    // Change fields back to original values
    // So it's easier to test next time
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalcfa"), "1111"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalfnab"), "2222"));

    dbconn->DBUpdate("vocaluser", filter, fieldvaluelist);
}

// DBUpdate(Domain, *Resultset, FieldValueSet)
void
Testcase4(DBConnection* dbconn)
{
    ResultTabletype restable;
    DBResultset* resultset1;
    DBResultset* resultset2;

    // Hard code filter for DBSelect w/ Order test
    const string filter = "(vocalhostname=\"host7*\")";

    // Hard code fieldlist, FieldListype is vector<string>
    FieldListtype fieldlist;  // fields from vocalserver domain
    fieldlist.push_back("vocalhostname");
    fieldlist.push_back("vocalport");
    fieldlist.push_back("vocaltype");

    cout << "\nDBSelect(vocalserver, " << filter 
        << ", fieldlist) before DBUpdate:\n";
    resultset1 = dbconn->DBSelect("vocalserver", filter, fieldlist);
    restable = resultset1->GetRecords();
    printresultset(restable);

    // Hard code fieldvalue set map<string,string> for DBUpdate test
    FieldValuetype fieldvaluelist;
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocaltype"), "newtype"));

    cout << "\nDBUpdate(vocalserver, resultset1, fieldvaluelist):\n";
    dbconn->DBUpdate("vocalserver", resultset1, fieldvaluelist);

    cout << "\nDBSelect(vocalserver, " << filter 
        << ", fieldlist) after DBUpdate:\n";
    resultset2 = dbconn->DBSelect("vocalserver", filter, fieldlist);
    restable = resultset2->GetRecords();
    printresultset(restable);

    // Change fields back to original value, 
    // So it's easier to test next time
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocaltype"), "feature1"));
    dbconn->DBUpdate("vocalserver", resultset1, fieldvaluelist);

    // Free both resultsets
    delete resultset1;
    delete resultset2;
}

// DBInsert(Domain, FieldValueSet)
void
Testcase5(DBConnection* dbconn)
{
    ResultTabletype restable;
    DBResultset* resultset;
    const string filter = "(vocalhostname=\"feature*\")";

    // Hard code fieldvalue set map<string,string> for DBInsert test
    FieldValuetype fieldvaluelist;
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalhostname"), "feature1host"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalport"), "1111"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalhostport"), "feature1host:1111"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocaltype"), "testtype"));

    // fields from vocalserver domain
    FieldListtype fieldlist; 
    fieldlist.push_back("vocalhostname");
    fieldlist.push_back("vocalport");
    fieldlist.push_back("vocalhostport");
    fieldlist.push_back("vocaltype");
    fieldlist.push_back("vocalservertimestamp");

    // DBInsert test
    cout << "\nDBInsert(vocalserver, fieldvaluelist):\n";
    dbconn->DBInsert("vocalserver", fieldvaluelist);

    cout << "\nDBSelect(vocalserver, "<< filter << ") after DBInsert:\n";
    resultset = dbconn->DBSelect("vocalserver", filter, fieldlist);
    restable = resultset->GetRecords();
    delete resultset;
    printresultset(restable);

}

// DBDelete(Domain, SearchFilter)
void
Testcase6(DBConnection* dbconn)
{
    const string filter = "(vocalhostname=\"feature*\")";

    FieldListtype fieldlist;  // fields from vocalserver domain
    fieldlist.push_back("vocalhostname");
    fieldlist.push_back("vocalport");
    fieldlist.push_back("vocalhostport");
    fieldlist.push_back("vocaltype");
    fieldlist.push_back("vocalservertimestamp");

    cout << "\nDBDelete(vocalserver, " << filter << "):\n";
    dbconn->DBDelete("vocalserver", filter);
}

// DBDelete(Domain, *Resultset)
void
Testcase7(DBConnection* dbconn)
{
    ResultTabletype restable;
    DBResultset* resultset;
    const string filter = "(vocalhostname=\"host5\")";

    FieldListtype fieldlist;  // fields from vocalserver domain
    fieldlist.push_back("vocalhostname");
    fieldlist.push_back("vocalport");
    fieldlist.push_back("vocalhostport");
    fieldlist.push_back("vocaltype");
    fieldlist.push_back("vocalservertimestamp");

    cout << "\nDBSelect(vocalserver, " << filter << ")\n";
    resultset = dbconn->DBSelect("vocalserver", filter, fieldlist);
    restable = resultset->GetRecords();
    printresultset(restable);

    cout << "\nDBDelete(vocalserver, resultset):\n";
    dbconn->DBDelete("vocalserver", resultset);

    cout << "\nDBSelect(vocalserver, " << filter << ")\n";
    resultset = dbconn->DBSelect("vocalserver", filter, fieldlist);
    restable = resultset->GetRecords();
    printresultset(restable);

    delete resultset;

    // Put deleted record back, so it's easier to test next time
    // Hard code fieldvalue set map<string,string> for DBInsert test
    FieldValuetype fieldvaluelist;
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalhostname"), "host5"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalport"), "5060"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocalhostport"), "host5:5060"));
    fieldvaluelist.insert(FieldValuetype::value_type(
                string("vocaltype"), "feature1"));

    dbconn->DBInsert("vocalserver", fieldvaluelist);
}

// DBDescribe(Domain)
void
Testcase8(DBConnection* dbconn)
{
    ResultTabletype restable;

    cout << "\nDBDescribe(vocaluser) \n";
    restable = dbconn->DBDescribe("vocaluser");
    printresultset(restable);

    cout << "\nDBDescribe(vocalserver) \n";
    restable = dbconn->DBDescribe("vocalserver");
    printresultset(restable);

    cout << "\nDBDescribe(testtypes) \n";
    restable = dbconn->DBDescribe("testtypes");
    printresultset(restable);
}

// DBDescribe() - Returns list of domains
void
Testcase9(DBConnection* dbconn)
{
    ResultRecordtype resrecord;

    cout << "\nDBDescribe() \n";
    resrecord = dbconn->DBDescribe();
    cout << resrecord << endl;
}

// DBCreate(Domain, FieldsList, PrimaryKey, ExistenceCheck)
void
Testcase10(DBConnection* dbconn)
{
    // Hard code fieldlist for new table
    FieldListtype fieldlist;  // fields from vocaluser domain
    fieldlist.push_back("column1");
    fieldlist.push_back("column2");
    fieldlist.push_back("column3");
    fieldlist.push_back("column4");
    fieldlist.push_back("column5");

    // Hard code Primary Key
    const string pkey = "column1";
    
    cout << "\nDBCreate(feature1, fieldlist, pkey, IGNOREEXISTENCE):\n";
    dbconn->DBCreate("feature1", fieldlist, pkey, IGNOREEXISTENCE);
}

// DBCreate(Domain, FieldsList, FieldListTypes, PrimaryKey, ExistenceCheck)
void
Testcase11(DBConnection* dbconn)
{
    FieldListtype fieldlist;  // fields from vocaluser domain
    FieldValuetype fieldvaluelist;  // fields from vocaluser domain

    // Hard code fieldlist for new table
    fieldlist.push_back("column1");
    fieldlist.push_back("column2");
    fieldlist.push_back("column3");
    fieldlist.push_back("column4");
    fieldlist.push_back("column5");
    fieldlist.push_back("column6");
    fieldlist.push_back("column7");
    fieldlist.push_back("column8");

    // Hard code datatypes for new table
    // See FielListDatatype enumerated types in DBConnection.hxx
    FieldListDatatype datatypes; 
    datatypes.push_back(STRINGTYPE);
    datatypes.push_back(STRINGTYPE);
    datatypes.push_back(TIMETYPE);
    datatypes.push_back(TEXTBLOBTYPE);
    datatypes.push_back(INTEGERTYPE);
    datatypes.push_back(INTEGERTYPE);
    datatypes.push_back(FLOATTYPE);
    datatypes.push_back(TIMESTAMPTYPE);

    cout << "\nDBCreate(feature2, fieldlist, pkey, CHECKEXISTENCE):\n";
    dbconn->DBCreate("feature2", fieldlist, datatypes, "column1", CHECKEXISTENCE);

    // Add record to new domain
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(string("column1"),"feat1"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column2"),"col2"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column3"),"12:00:00"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column4"),"col4"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column5"),"100"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column6"),"200"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column7"),"300"));
    dbconn->DBInsert("feature2", fieldvaluelist);
    // Add record to new domain
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(string("column1"),"feat2"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column2"),"col2"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column3"),"12:00:00"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column4"),"col4"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column5"),"111"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column6"),"222"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column7"),"333.1"));
    dbconn->DBInsert("feature2", fieldvaluelist);
    // Add record to new domain
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(string("column1"),"feat3"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column2"),"col2"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column3"),"12:00:00"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column4"),"col4"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column5"),"101"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column6"),"202"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column7"),"303.2"));
    dbconn->DBInsert("feature2", fieldvaluelist);
    // Add record to new domain
    fieldvaluelist.clear();
    fieldvaluelist.insert(FieldValuetype::value_type(string("column1"),"feat4"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column2"),"col2"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column3"),"12:00:00"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column4"),"col4"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column5"),"103"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column6"),"202"));
    fieldvaluelist.insert(FieldValuetype::value_type(string("column7"),"301.3"));
    dbconn->DBInsert("feature2", fieldvaluelist);

}

// DBAlter(Domain, Fieldname)
void
Testcase12(DBConnection* dbconn)
{
    ResultTabletype restable;

    cout << "\nDBDescribe(feature2) before DBAlter:\n";
    restable = dbconn->DBDescribe("feature2");
    printresultset(restable);

    cout << "\nDBAlter(feature2, column3)\n";
    dbconn->DBAlter("feature2", "column3");

    cout << "\nDBDescribe(feature2) after DBAlter:\n";
    restable = dbconn->DBDescribe("feature2");
    printresultset(restable);
}

// DBAlter(Domain, Fieldname, Action)
void
Testcase13(DBConnection* dbconn)
{
    ResultTabletype restable;

    cout << "\nDBDescribe(feature1) before DBAlter:\n";
    restable = dbconn->DBDescribe("feature1");
    printresultset(restable);

    cout << "\nDBAlter(feature1, newcolumn, TEXTBLOBTYPE)\n";
    dbconn->DBAlter("feature1", "newcolumn", TEXTBLOBTYPE);

    cout << "\nDBDescribe(feature1) after DBAlter:\n";
    restable = dbconn->DBDescribe("feature1");
    printresultset(restable);
}

// DBDrop(Domain, ExistenceCheck)
void
Testcase14(DBConnection* dbconn)
{
    cout << "\nDBDrop(feature1, IGNOREEXISTENCE):\n";
    dbconn->DBDrop("feature1", IGNOREEXISTENCE);
}

// DBDrop(Domain, ExistenceCheck)
void
Testcase15(DBConnection* dbconn)
{
    cout << "\nDBDrop(feature2, CHECKEXISTENCE):\n";
    dbconn->DBDrop("feature2", CHECKEXISTENCE);
}

// DBVerify(Domain)
void
Testcase16(DBConnection* dbconn)
{
    cout << "\nDBVerify(vocaluser):\n";

    if ( dbconn->DBVerify("vocaluser") )
        cout << "Table exists\n";
    else
        cout << "Table does not exists\n";

    cout << "\nDBVerify(nonexistenttable):\n";

    if ( dbconn->DBVerify("nonexistenttable") )
        cout << "Table exists\n";
    else
        cout << "Table does not exists\n";

    cout << "\nDBVerify(feature1):\n";
    if ( dbconn->DBVerify("feature1") )
        cout << "Table exists\n";
    else
        cout << "Table does not exists\n";

    cout << "\nDBVerify(feature2):\n";
    if ( dbconn->DBVerify("feature2") )
        cout << "Table exists\n";
    else
        cout << "Table does not exists\n";
}

void
Testcase17(DBConnection* dbconn)
{
    cout << "\nDBPrimaryKey(vocaluser):\n";
    cout << "Primary Key: " << dbconn->DBPrimaryKey("vocaluser") << endl;
    cout << "\nDBPrimaryKey(vocalserver):\n";
    cout << "Primary Key: " << dbconn->DBPrimaryKey("vocalserver") << endl;
    cout << "\nDBPrimaryKey(feature1):\n";
    cout << "Primary Key: " << dbconn->DBPrimaryKey("feature1") << endl;
}

// MAIN
int
main()
{
    // ConnectionManager class handles making the connection to 
    // various databases
    ConnectionManager dbmanager;

    // hardcode the dbconnection
    DBInfo dbInfo;
//    dbInfo.dbBackend = "MYSQL";
    dbInfo.dbBackend = "PGSQL";
    dbInfo.database = "vocal";
    dbInfo.host = "vvs-lion";
    dbInfo.user = "vocaluser";
    dbInfo.password = "vocaluser";

    // Pointer to abstract object to handle connection to DB
    DBConnection* dbconn = dbmanager.GetDBConnection(dbInfo);

    char choice;

    // Loop forever prompting for DB functions
    while ( true )
    {
        cout << endl 
             << "\n"
             << "  Choose Database Function:\n\n"
             << "  Functions that act on data:\n"
             << "     1) DBSelect ( Domain, SearchFilter, FieldList )\n"
             << "     2) DBSelect ( Domain, SearchFilter, FieldList, Order, OrderField )\n"
             << "     3) DBUpdate ( Domain, SearchFilter, FieldValueSet )\n"
             << "     4) DBUpdate ( Domain, *Resultset, FieldValueSet )\n"
             << "     5) DBInsert ( Domain, FieldValueSet )\n"
             << "     6) DBDelete ( Domain, SearchFilter )\n"
             << "     7) DBDelete ( Domain, *Resultset )\n"
             << "     8) DBDescribe ( Domain )\n"
             << "     9) DBDescribe ( )  <- List Domains\n"
             << "\n"
             << "  Functions that act on DB structure:\n"
             << "     a) DBCreate ( Domain, FieldsList, PrimaryKey, IGNOREEXISTENCE )\n"
             << "     b) DBCreate ( Domain, FieldsList, FieldsListTypes, PrimaryKey, \n"
             << "                   CHECKEXISTENCE )\n"
             << "     c) DBAlter ( Domain, Fieldname ) <- Remove a Field\n"
             << "     d) DBAlter ( Domain, Fieldname, FieldType ) <- Add a Field\n"
             << "     e) DBDrop ( Domain, IGNOREEXISTENCE )\n"
             << "     f) DBDrop ( Domain, CHECKEXISTENCE )\n"
             << "     g) DBVerify ( Domain )\n"
             << "     h) DBPrimaryKey (Domain)\n"
             << "\n"
             << "     n) New Database Connection\n"
             << "     q) Quit\n\n"
             << "     Enter Choice> ";
        cin  >> choice;

        try
        {
            switch (choice)
            {
                case '1': 
                    Testcase1(dbconn);
                    break;

                case '2': 
                    Testcase2(dbconn);
                    break;

                case '3': 
                    Testcase3(dbconn);
                    break;

                case '4': 
                    Testcase4(dbconn);
                    break;

                case '5': 
                    Testcase5(dbconn);
                    break;

                case '6': 
                    Testcase6(dbconn);
                    break;

                case '7': 
                    Testcase7(dbconn);
                    break;

                case '8': 
                    Testcase8(dbconn);
                    break;

                case '9': 
                    Testcase9(dbconn);
                    break;

                case 'a': 
                    Testcase10(dbconn);
                    break;

                case 'b': 
                    Testcase11(dbconn);
                    break;

                case 'c': 
                    Testcase12(dbconn);
                    break;

                case 'd': 
                    Testcase13(dbconn);
                    break;

                case 'e': 
                    Testcase14(dbconn);
                    break;

                case 'f': 
                    Testcase15(dbconn);
                    break;

                case 'g': 
                    Testcase16(dbconn);
                    break;
                case 'h':
                    Testcase17(dbconn);
                    break;
                case 'n': 
                    delete dbconn;
                    // Use the other version of GetDBConnection for testing.
                    // The version that takes no arguments will ask the user
                    // dbconn = dbmanager.GetDBConnection();
                    dbconn = dbmanager.GetDBConnection(dbInfo);
                    break;

                case 'q': 
                    cout << endl;
                    exit(0);
                    break;

                default:
                    cerr << "\n    * Error, Choose one of the items listed *\n";
                    cerr << "";
                    break;

            } // switch()

        } // try()
        
        catch ( VDbException &e )
        {
            cerr << e.getDescription() << endl;
        }
    } // while 
    delete dbconn;
} // MAIN
