package org.vovida.prov.authentication;

import javax.security.auth.Subject;
import javax.security.auth.callback.CallbackHandler;
import javax.security.auth.login.FailedLoginException;
import javax.security.auth.login.LoginContext;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

import java.util.Iterator;

import org.apache.log4j.Category;

import org.vovida.prov.authentication.LoginCancelledException;
import org.vovida.prov.util.AppletApp;





/**
 * The top level class which will cause the authentication to be performed and
 * then run the supplied appplication's main class as a priviledged action.<p>
 * Assumes that the name of the application class to run is the same as the name
 * of the login context configuration.<p>
 */
public class AuthTopLevel extends AppletApp
{
    private static final Category logCategory = Category.getInstance(AuthTopLevel.class.getName());


    public static final String OPT_MAIN_CLASS = "-main";
    public static final String OPT_LOGIN_CALLBACK_HANDLER = "-callb";
    public static final String OPT_HELP = "-h";
    public static final String OPT_CONF_FILE = "-conf";

    private String mainName;
    private CallbackHandler handler;
    private String [] applicationArgs;





    private void configArgs()
    {
        addArg(OPT_MAIN_CLASS, 1);
        addArg(OPT_LOGIN_CALLBACK_HANDLER, 1);
        addArg(OPT_HELP, 0);
        addArg(OPT_CONF_FILE, 1);
    }


    private static void printUnusedArgs(String [] arg)
    {
        System.out.println("Arguments unrecognized by login module: ");
        System.out.print("     ");
        for (int i=0; i<arg.length; i++)
        {
            System.out.print(arg[i] + " ");
        }
        System.out.println();
        System.out.println("They will be passed to application main class.");
    }


    public void run() throws LoginInitFailedException
    {
        logCategory.debug("Starting login with configuration name \"" + mainName
            + "\" and callback handler \"" + handler + "\".");

        //create the login context
        LoginContext context;
        try
        {
            context = new LoginContext(mainName,handler);
        }
        catch (Exception ex)
        {
            String message = "Could not instantiate login context. Cannot continue.";
            logCategory.fatal(message,ex);
            throw new LoginInitFailedException(message,ex);
        }

        logCategory.debug("Created login context.");

        //give NUM_LOGIN_RETRIES attempts to login
        int i;
        for (i=0; i<AuthConstants.NUM_LOGIN_RETRIES; i++)
        {
            try
            {
                context.login();
                break;
            }
            catch (FailedLoginException ex)
            {
                logCategory.info("Login failed " + ex);
                //give another chance
            }
            catch (Exception ex)
            {
                logCategory.error(ex);
				throw new LoginInitFailedException("context.login() method failed", ex);
            }
        }

        if (i == AuthConstants.NUM_LOGIN_RETRIES) //if tried max times
        {
            logCategory.error("Maximum number of login retries exceeded.");
            return;
        }

        try
        {
            Subject.doAsPrivileged(context.getSubject(), new PSAction(applicationArgs), null);
        }
        catch (Exception ex)
        {
            String message = "Could not execute application ";
            logCategory.fatal(message, ex);
            throw new ApplicationException(message,ex);
        }

        logCategory.warn("Client application finished. Logging out.");
        try
        {
            context.logout();
        }
        catch (Exception ex)
        {
            logCategory.error("Logout failed", ex);
        }
    }

    private class PSAction implements java.security.PrivilegedExceptionAction
    {
        private Category logCategory;

        private String [] appArgs = new String[] {};

        public PSAction(String [] args)
        {
            appArgs = args;
			logCategory = Category.getInstance(this.getClass().getName());
        }

        public Object run() throws Exception
        {
            logCategory.debug("run() starting");
            logCategory.debug("main application class name = " + mainName);

            ClassLoader cl = Thread.currentThread().getContextClassLoader();
            Class c = Class.forName(mainName,true,cl);
            Method main = c.getMethod("main",new Class [] {String [].class});
            main.invoke(null, new Object [] {appArgs}); //static method may be invoked on null

            logCategory.debug("main application class invoked");
            return null; //not used anywhere
        }
    }


    public void init() throws LoginInitFailedException
    {
        //instantiate the selected callback handler which will be by the login
        //module to obtain the account id and password to log in
        String callbackName = getParameter(OPT_LOGIN_CALLBACK_HANDLER, AuthConstants.LOGIN_CALLBACK_HANDLER);

        if ( (callbackName == null) || (callbackName.equals("")) )
        {
            String message = "No callback hander name found.";
            logCategory.fatal(message);
            throw new LoginInitFailedException(message);
        }
        else
        {
            //use reflection to instantiate the callback
            Class c;
            try
            {
                c = Class.forName(callbackName);
            }
            catch (Exception ex)
            {
                String message = "Could not load class for callback handler. Cannot continue.";
                logCategory.fatal(message,ex);
                throw new LoginInitFailedException(message);
            }

            //make sure that the supplied class does implement the correct
            //interface (CallbackHandler)
            if (!CallbackHandler.class.isAssignableFrom(c))
            {
                String message = c + " is not a " +  CallbackHandler.class;
                logCategory.fatal(message);
                throw new LoginInitFailedException(message);
            }

             //find the null argument constructor and create an instance of
             //the callback handler
             Constructor constructor;
             try
             {
                constructor = c.getConstructor(null); //get no argument constructor
                handler = (CallbackHandler)constructor.newInstance(null); //instantiate
             }
             catch (Exception ex)
             {
                String message = "Could not instantiate callback handler " + c;
                logCategory.fatal(message,ex);
                throw new LoginInitFailedException(message);
             }
             if (handler == null)
             {
                 String message = "Could not instantiate callback handler. Cannot continue";
                 logCategory.fatal(message);
                 throw new LoginInitFailedException(message);
             }
        }

        //get the name of the main application class which should also be the
        //same as the name of the configuration to use for the login module
        mainName = getParameter(OPT_MAIN_CLASS, AuthConstants.MAIN_CLASS);
        if ( (mainName == null) || (mainName.equals("")) )
        {
            String message = "No configuration name/main class supplied. Cannot continue.";
            logCategory.fatal(message);
            throw new LoginInitFailedException(message);
        }

        //get the name of the file that contains the configuration information
        String configFile = getParameter(OPT_CONF_FILE, AuthConstants.CONF_FILE);
        if ( (configFile == null) || (configFile.equals("")) )
        {
            String message = "No login configuration file name supplied. Cannot continue.";
            logCategory.fatal(message);
            throw new LoginInitFailedException(message);
        }
        else
        {
            if (logCategory.isDebugEnabled())
            {
                logCategory.debug("JAAS login configuration file = " + configFile);
                try
                {
                    logCategory.debug("Current directory = "
                        + new java.io.File(".").getCanonicalPath());
                    logCategory.debug("resource path = "
                        + getClass().getResource("."));
                }
                catch (Exception ex)
                {
                    logCategory.debug("error tyring to figure out current directory", ex);
                }
            }
            //Set the value of the property to be the resource url of the
            //config file. This should allow the file to be loaded properly
            //even if it is stored in a jar file with the rest of the code.
            java.net.URL url = getClass().getResource("/"+configFile);
            String resourcePath;
            if (url != null)
            {
                resourcePath = url.toString();
            }
            else
            {
                //if that didn't find the file, try loading it as a relative path
                resourcePath = configFile;
            }
            logCategory.debug("jaas.conf path = " + resourcePath);
            System.setProperty("java.security.auth.login.config", resourcePath);
        }
    }


    private void printUsage()
    {
        System.out.println("Usage:");
        System.out.println("  org.vovida.prov.authentication.AuthTopLevel ");
        Iterator args = argMap.keySet().iterator();
        while (args.hasNext())
        {
            System.out.print(args.next() + " ");
        }
        System.out.println();
    }

	public static void main (String [] args)
    {
		if (logCategory.isDebugEnabled())
        {
            String arguments = "";
            for (int i=0; i<args.length; i++)
            {
                arguments += args[i] + " ";
            }
            logCategory.debug("Entry arguments = " + arguments);
        }

        AuthTopLevel top = new AuthTopLevel();
        top.isStandalone = true;
        top.configArgs();

        //parse the command line. The extra arguments should be supplied to the
        //application
		top.applicationArgs = top.parseArgs(args);
   		String help = top.getParameter(OPT_HELP, null);
        if (help != null)
        {
            top.printUsage();
            return;
        }

        printUnusedArgs(top.applicationArgs);

		top.init();
        top.run();
    }
}
