/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.content;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.content.condition.Condition;
import org.vovida.prov.widget.GuiComponent;
import org.vovida.prov.widget.GuiException;
import org.vovida.prov.widget.TextField;
import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.PSDBResult;
import org.vovida.prov.psLib.OperationException;

import java.util.Vector;
import java.util.List;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.IOException;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import javax.swing.tree.TreeNode;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JButton;
import javax.swing.JTextField;
import javax.swing.JOptionPane;

import org.w3c.dom.DocumentFragment;
import org.xml.sax.Attributes;

import org.jdom.Element;

import org.apache.log4j.Category;

/**
 * This class codifies the rules for validating a piece of data.
 * Subclasses of ContentData understand the rules for validating
 * specific types of data, such as IntegerType and DateType.
 * 
 * A ContentData holds a reference to a UI object which determines the widget
 * used to display it. The ContentData holds the data that is
 * entered and displayed in the UI widget.
 *
 * @author Barbara Samson
 */
public abstract class ContentData implements Cloneable, ReasonCodes, 
        DTDConstants
{
  private String myName;

  /**
   * Default value for this entry. Note that it must be initialized
   * to prevent a NullPointerException on restoreDefault().
   */

  /**
   * @link aggregation
   * @clientCardinality 1
   * @supplierCardinality 1
   */
  private GuiComponent myGuiComponent;

  private String myDefault = "";
  private String myLastSavedValue = null;
  private boolean myIsKey = false;
  private boolean myIsRequired = false;
  private boolean myIsEditable = true;
  protected ProvisionableInstance myProvisionableInstance;
  private String myReference;
  private Vector myConditions = new Vector();
  private boolean myIsValueChanged = false;

  /**
   * Message that pops up whenever a change to this field is saved.
   */
  private String myAlert = "";

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(ContentData.class.getName());

  /**
   * Class Constructor.
   * @param typeElement could be null.
   */
  public ContentData(String name, 
                     ProvisionableInstance provisionableInstance, 
                     Element typeElement)
  {
    myProvisionableInstance = provisionableInstance;
    myName = name;

    // If there was a typeElement, check it for further restrictions
    if (typeElement != null)
    {
      List typeChildren = typeElement.getChildren();

      if (!typeChildren.isEmpty())
      {
        Element restrictionElement = (Element) typeChildren.get(0);
        String restrictionName = restrictionElement.getName();

        if (restrictionName.equals(REFERENCE))
        {
          setReference(restrictionElement.getTextTrim());
        }
      }
    }
  }

  /**
   * Initalize the ContentData by setting the lastSavedValue
   */
  public void init()
  {
    myLastSavedValue = getContentAsString();
  }

  /**
   * Cloning allows new table rows to be created.
   */
  public Object clone() throws CloneNotSupportedException
  {
    ContentData clone = (ContentData) super.clone();

    clone.setGuiComponent((GuiComponent) myGuiComponent.clone());
    clone.setDefault(getDefault());

    return clone;
  }

  public abstract void validate(String data) throws InvalidDataException;

  public void validate() throws InvalidDataException
  {
    validate(getContentAsString());
  }

  /**
   * Returns true if this data has been saved before and value is
   * different from what was saved. This method supports the check
   * to see if a keyfield has changed.
   */
  protected boolean dataChanged(String value)
  {
    if (myIsValueChanged && (!myLastSavedValue.equals(value)) )
//    if ((myLastSavedValue != null) && (!myLastSavedValue.equals(value)))
    {
      return true;
    }

    return false;
  }

  /**
   * Returns true if the current value has never been saved.
   * This method supports the ProvisionableInstance method
   * getChangedData().
   */
  public boolean dataNotSaved(String value)
  {
  	if (myIsValueChanged && (!myLastSavedValue.equals(value)) )
//    if ((myLastSavedValue == null) || (!myLastSavedValue.equals(value)))
    {
      return true;
    }

    return false;

  }

  protected void checkAlert(String value)
  {
    if (!myAlert.equals(""))
    {
      if (dataChanged(value))
      {
        JButton continueButton = new JButton("Continue with Change");
        JButton cancelButton = new JButton("Cancel Change");
        final JFrame frame = new JFrame();

        continueButton.addActionListener(new ActionListener()
        {
          public void actionPerformed(ActionEvent e)
          {
            frame.setVisible(false);
            frame.dispose();
          }

        });
        cancelButton.addActionListener(new ActionListener()
        {
          public void actionPerformed(ActionEvent e)
          {
            cancel();
            frame.setVisible(false);
            frame.dispose();
          }

        });
        JOptionPane.showOptionDialog(frame, myAlert, "Special Alert", 
                JOptionPane.DEFAULT_OPTION, JOptionPane.INFORMATION_MESSAGE, 
                null, new Object[]
        {
          continueButton, cancelButton
        }, continueButton);
      }
    }
  }

  protected void checkIfRequired(String value) throws InvalidDataException
  {

    // if field is not editable it cannot be required
    if (!myIsEditable)
    {
      return;
    }

    if (isRequired())
    {
      if ((value == null) || (value.equals("")))
      {
        String message = new String("Required field " 
                + getGuiComponent().getLabel() + " cannot be empty");

        throw new InvalidDataException(message, REQ_FIELD_BLANK);
      }
    }
  }

  public void setKey(boolean isTrue)
  {
    myIsKey = isTrue;

    setRequired(isTrue);
  }

  public void setAlert(String alert)
  {
    myAlert = alert;
  }

  public void setRequired(boolean isRequired)
  {
    myIsRequired = isRequired;

    logCategory.info("data " + (isRequired ? "is" : "is not") + " required");
  }

  public boolean isRequired()
  {
    return myIsRequired;
  }

  public boolean isKey()
  {
    return myIsKey;
  }

  public void savingData(final String data) throws InvalidDataException
  {
    if (isKey() && dataChanged(data))
    {

      // for unit testing only
      if (myProvisionableInstance.isTest())
      {
        throw new InvalidDataException("", KEYFIELD_CHANGED);
      }

      JButton continueButton = new JButton("Continue with Change");
      JButton cancelButton = new JButton("Cancel Change");
      final JFrame frame = new JFrame();

      continueButton.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          myLastSavedValue = data;

          frame.setVisible(false);
          frame.dispose();
        }

      });
      cancelButton.addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          cancel();
          frame.setVisible(false);
          frame.dispose();
        }

      });

      JOptionPane.showOptionDialog(frame, 
              getGuiComponent().getLabel() + " is a key field.\n" 
              + "Are you sure you want to change it?", "Changing Key Field", 
              JOptionPane.DEFAULT_OPTION, JOptionPane.INFORMATION_MESSAGE, 
              null, new Object[]
      {
        continueButton, cancelButton
      }, continueButton);
    }
    else
    {
      myLastSavedValue = data;
    }

    // notify the GuiComponent that data is being saved.
    myGuiComponent.savingData();
  }

  public String getName()
  {
    return myName;
  }

  protected abstract String getHint();

  protected PSDBResult callGetRecordsInDomain(Element queryElement)
  {
    ProvisioningInterface ps = ProvisioningInterface.getInstance();

    try
    {
      String domain = queryElement.getAttributeValue(DOMAIN);

      if ((domain == null) || (domain.equals("")))
      {
        logCategory.warn("missing domain attribute");

        return null;
      }

      String field = queryElement.getAttributeValue(FIELD);

      if ((field == null) || (field.equals("")))
      {
        logCategory.warn("missing field attribute");

        return null;
      }

      String filter = queryElement.getAttributeValue(FILTER);

      if ((filter == null) || (filter.equals("")))
      {
        logCategory.warn("missing filter attribute");

        return null;
      }

      if (logCategory.isInfoEnabled())
      {
        logCategory.info("domain: " + domain + " field: " + field 
                         + " filter: " + filter);
      }

      PSDBResult result = ps.getRecordsInDomain(domain, filter, new String[]
      {
        field
      });

      return result;
    }
    catch (OperationException e1)
    {
      JOptionPane.showMessageDialog(new JFrame(), e1.getMessage(), 
              "OperationException", JOptionPane.ERROR_MESSAGE);

      return null;
    }
    catch (IOException e2)
    {
      e2.printStackTrace();

      return null;
    }
  }

  protected String[] callGetDomains(Element queryElement)
  {
    ProvisioningInterface ps = ProvisioningInterface.getInstance();
    String[] domains = null;

    try
    {
      String category = queryElement.getAttributeValue(CATEGORY_ATTR);

      if ((category == null) || (category.equals("")))
      {
        logCategory.warn("missing category attribute");

        return null;
      }

      domains = ps.getListXMLDescriptions(category);
    }
    catch (OperationException e1)
    {
      JOptionPane.showMessageDialog(new JFrame(), e1.getMessage(), 
              "OperationException", JOptionPane.ERROR_MESSAGE);

      return null;
    }
    catch (IOException e2)
    {
      e2.printStackTrace();

      return null;
    }

    return domains;
  }

  public GuiComponent getGuiComponent()
  {
    return myGuiComponent;
  }

  protected void setGuiComponent(GuiComponent guiComponent)
  {
    myGuiComponent = guiComponent;
  }

  public String getContent() throws InvalidDataException
  {
    try
    {
      validate();
    }
    catch (InvalidDataException e)
    {

      // disable JOption Pane for unit testing
      if (!myProvisionableInstance.isTest())
      {
        String hint = getHint();

        JOptionPane.showMessageDialog(new JFrame(), 
                e.getMessage() + "\n" + hint, "Invalid Data", 
                JOptionPane.ERROR_MESSAGE);
        getGuiComponent().getEditorComponent().requestFocus();
      }

      throw e;
    }

    String content = getContentAsString();

    savingData(content);

    return content;
  }


  public abstract String getContentAsString();

  public void setContent(String content)
  {
    myLastSavedValue = content;
    myIsValueChanged = true;

    triggerConditions();
  }

  public void setEnabled(boolean isEditable)
  {
    myIsEditable = isEditable;

    myGuiComponent.setEnabled(isEditable);
  }

  public void setUserInputOK(boolean inputOK)
  {
    myGuiComponent.setUserInputOK(inputOK);
  }

  public void performSecondPass()
  {
    if (myReference != null)
    {
      final ContentData reference = 
        myProvisionableInstance.getContentByName(myReference);

      setContent(reference.getContentAsString());

      GuiComponent referencedGuiComponent = reference.getGuiComponent();
      JComponent component = referencedGuiComponent.getEditorComponent();

      component.addKeyListener(new KeyListener()
      {
        public void keyTyped(KeyEvent e) {}

        public void keyPressed(KeyEvent e) {}

        public void keyReleased(KeyEvent e)
        {
          setContent(reference.getContentAsString());
        }

      });
    }

    for (int i = 0; i < myConditions.size(); i++)
    {
      Condition condition = (Condition) myConditions.elementAt(i);

      condition.initialize(myProvisionableInstance);
    }

  }

  protected void triggerConditions()
  {
    for (int i = 0; i < myConditions.size(); i++)
    {
      Condition condition = (Condition) myConditions.elementAt(i);

      condition.actionPerformed(null);
    }
  }

  public void setReference(String reference)
  {
    myReference = reference;

    // the rest of this is set up in second pass.
  }

  public void addCondition(Condition condition)
  {
    myConditions.addElement(condition);
  }

  /**
   * This is used by the public API to add an external
   * listener to a gui widget.
   */
  public void addActionListener(ActionListener listener)
  {
    myGuiComponent.addActionListener(listener);
  }

  // methods that get passed through to UI.

  public void setLabel(String newLabel)
  {
    myGuiComponent.setLabel(newLabel);
  }

  public void setDefault(String defaultValue)
  {
    if (defaultValue == null)
    {
      return;
    }

    try
    {
      validate(defaultValue);
    }
    catch (InvalidDataException e)
    {
      logCategory.warn("caught error setting default: " + e.getMessage());

      return;
    }

    myDefault = defaultValue;

    // myLastSavedValue = defaultValue;
    setContent(myDefault);

//    myIsValueChanged = false;
  }

  public String getDefault()
  {
    return myDefault;
  }

  public void cancel()
  {
    setContent(myLastSavedValue == null ? "" : myLastSavedValue);
    myIsValueChanged = false;
  }

  public void clearLastSavedValue()
  {
//    myLastSavedValue = null;
    myIsValueChanged = false;
  }

  public void restoreDefault()
  {
    setContent(myDefault);
  }

  /**
   * This only has meaning for numeric types.
   * This method is overridden in NumericType.
   */
  public void setLowerLimit(String limit, boolean override) {}

  /**
   * This only has meaning for numeric types.
   * This method is overridden in NumericType.
   */
  public void setUpperLimit(String limit, boolean override) {}



}


