/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.content;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.widget.GuiComponent;
import org.vovida.prov.widget.TextField;
import org.vovida.prov.widget.HostPortField;
import org.vovida.prov.widget.ComboBox;
import org.vovida.prov.widget.GuiFactory;

import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JFrame;
import javax.swing.text.PlainDocument;
import javax.swing.text.Document;
import javax.swing.text.BadLocationException;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JOptionPane;

import org.jdom.Element;

/**
 * @author Barbara Samson
 */
public class HostPortType extends StringType implements DTDConstants
{

  public final static String HOSTLABEL = "Host";
  public final static String PORTLABEL = "Port";

  HostPortField myHostPortField;
  IntegerType myPortContent;
  HostType myHostContent;


  /**
   * Class constructor. The default ui is a text field. We need to set the
   * default ui during construction because
   * we may not get further XML processing instructions to change it.
   * @param name The name label to be displayed in the UI.
   */
  public HostPortType(String name, GuiComponent gui, 
          ProvisionableInstance instance, Element typeElement)
  {
    super(name, gui, instance, typeElement);

    if (gui instanceof HostPortField)
    {
      myHostPortField = (HostPortField) gui;

      Element portElement = null;
      Element hostElement = null;

      if (typeElement != null)
      {

        // note that these could still be null.
        portElement = typeElement.getChild(INTEGERTYPE);
        hostElement = typeElement.getChild(HOSTTYPE);
      }

      myPortContent = new IntegerType(name, myHostPortField.getPortField(), 
              instance, portElement);

      myPortContent.setLowerLimit("1", false);
      myPortContent.setLabel(PORTLABEL);

      myHostContent = new HostType(name, myHostPortField.getHostField(), 
              instance, hostElement);

      myHostContent.setLabel(HOSTLABEL);
    }
    else
    {
      myPortContent = new IntegerType(name, 
              GuiFactory.newGuiComponent(PORTLABEL, new Element(TEXTFIELD)), 
              instance, new Element(INTEGERTYPE));
      myHostContent = new HostType(name, 
              GuiFactory.newGuiComponent(HOSTLABEL, new Element(COMBOBOX)), 
              instance, new Element(HOSTTYPE));
    }

  }

  public String getContentAsString()
  {
    GuiComponent gui = getGuiComponent();

    if (gui instanceof HostPortField)
    {
      String hostString = myHostContent.getContentAsString();
      String portString = myPortContent.getContentAsString();

      return hostString + ":" + portString;
    }
    else        // gui is a TextField
    {
      String returnString = super.getContentAsString();

      return returnString;
    }
  }

  public void setContent(String content)
  {

    // for the case where ui is TextField
    super.setContent(content);

    String hostport = content;
    int separatorPos = -1;

    if (hostport != null)
    {
      separatorPos = hostport.indexOf(":");
    }

    // if there is no separator
    // or if the separator is at the end of the string,
    // then assume the whole string is the hostname
    if ((separatorPos == -1) || (separatorPos == hostport.length() - 1))
    {
      myHostContent.setContent(hostport);
      myPortContent.setContent("");

      return;
    }

    myHostContent.setContent(hostport.substring(0, separatorPos));
    myPortContent.setContent(hostport.substring(separatorPos + 1, 
            hostport.length()));
  }

  public void validate() throws InvalidDataException
  {
    myHostContent.setRequired(isRequired());
    myPortContent.setRequired(isRequired());

    if (!(getGuiComponent() instanceof HostPortField))
    {
      setContent(super.getContentAsString());
    }

    myHostContent.validate();
    myPortContent.validate();

    checkAlert(getContentAsString());
  }

  protected String getHint()
  {
    StringBuffer buf = new StringBuffer();

    buf.append(HOSTLABEL + ":\n");
    buf.append(myHostContent.getHint() + "\n");
    buf.append(PORTLABEL + ":\n");
    buf.append(myPortContent.getHint());

    return buf.toString();
  }

}


