/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.content;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.widget.GuiComponent;

import javax.swing.text.Document;
import java.text.NumberFormat;
import javax.swing.text.BadLocationException;
import java.text.ParseException;
import java.util.Locale;
import javax.swing.JOptionPane;
import javax.swing.JFrame;

import org.jdom.Element;

/**
 * @author Barbara Samson
 */
abstract public class NumericType extends StringType implements DTDConstants
{

  private NumberFormat myNumberFormat;
  private Double myLowerLimit = null;
  private Double myUpperLimit = null;

  /**
   * Class Constructor.
   * @param element could be null
   */
  public NumericType(String name, GuiComponent gui, 
                     ProvisionableInstance instance, Element element)
  {
    super(name, gui, instance, element);

    myNumberFormat = NumberFormat.getNumberInstance(Locale.US);

    if (element != null)
    {
      Element lower = element.getChild(MININCLUSIVE);

      if (lower != null)
      {
        setLowerLimit(lower.getTextTrim(), true);
      }

      Element upper = element.getChild(MAXINCLUSIVE);

      if (upper != null)
      {
        setUpperLimit(upper.getTextTrim(), true);
      }
    }
  }

  protected Document newDocument()
  {
    NumberDocument doc = new NumberDocument();

    doc.addAllowedCharacter(new Character(','));
    doc.addAllowedCharacter(new Character('-'));
    doc.addAllowedCharacter(new Character('.'));

    return doc;
  }

  /**
   * Set the lower limit for this number.
   * @param lowerLimit the lower limit (inclusive).
   * @param override true if existing lower limit should be overridden.
   */
  public void setLowerLimit(String lowerLimit, boolean override)
  {
    if (!override && (myLowerLimit != null))
    {
      return;
    }

    try
    {
      Double limit = new Double(lowerLimit);

      // if upper limit has been set, ensure it is not
      // lower than lower limit.
      if (myUpperLimit != null)
      {
        if (myUpperLimit.compareTo(limit) < 0)
        {
          myLowerLimit = new Double(myUpperLimit.doubleValue());

          return;
        }
      }

      myLowerLimit = limit;
    }
    catch (NumberFormatException e)
    {
      JOptionPane.showMessageDialog(new JFrame(), 
              new String(lowerLimit + " cannot be read as numeric"));
    }
  }

  public Double getLowerLimit()
  {
    return myLowerLimit;
  }

  /**
   * Set the upper limit for this number.
   * @param upperLimit the upper limit (inclusive).
   * @param override true if existing upper limit should be overridden.
   */
  public void setUpperLimit(String upperLimit, boolean override)
  {
    if (!override && (myUpperLimit != null))
    {
      return;
    }

    try
    {
      Double limit = new Double(upperLimit);

      // if lower limit has been set, ensure it is not
      // higher than upper limit.
      if (myLowerLimit != null)
      {
        if (myLowerLimit.compareTo(limit) > 0)
        {
          myUpperLimit = new Double(myLowerLimit.doubleValue());

          return;
        }
      }

      myUpperLimit = limit;
    }
    catch (NumberFormatException e)
    {
      JOptionPane.showMessageDialog(new JFrame(), 
              new String(upperLimit + " cannot be read as numeric"));
    }
  }

  public Double getUpperLimit()
  {
    return myUpperLimit;
  }

  public void validate(String value) throws InvalidDataException
  {
    checkIfRequired(value);

    if ((value == null) || (value.equals("")))
    {
      return;
    }

    Number number;

    try
    {
      number = myNumberFormat.parse(value);

      // perform range checking, if required
      if ((myLowerLimit != null) || (myUpperLimit != null))
      {
        Double doubleNumber = new Double(number.doubleValue());

        if (myLowerLimit != null)
        {
          if (doubleNumber.compareTo(myLowerLimit) < 0)
          {
            StringBuffer message = 
              new StringBuffer(getGuiComponent().getLabel());

            message.append(" is out of range");

            throw new InvalidDataException(message.toString(), OUT_OF_RANGE);
          }
        }

        if (myUpperLimit != null)
        {
          if (doubleNumber.compareTo(myUpperLimit) > 0)
          {
            StringBuffer message = 
              new StringBuffer(getGuiComponent().getLabel());

            message.append(" is out of range");

            throw new InvalidDataException(message.toString(), OUT_OF_RANGE);
          }
        }
      }
    }
    catch (ParseException e)
    {
      throw new InvalidDataException(e, BAD_FORMAT);
    }

    String numberString = number.toString();

    if (!numberString.equals(value))
    {
      setContent(numberString);
    }

    checkAlert(numberString);
  }


  public void setContent(String content)
  {
    try
    {
      myDocument.remove(0, myDocument.getLength());
      myDocument.insertString(0, content, null);
    }
    catch (BadLocationException e)
    {

      // this shouldn't be able to happen
      e.printStackTrace();
    }

    super.setContent(content);
  }

  protected String getHint()
  {
    return "All numbers are valid";
  }

  protected void setNumberFormat(NumberFormat numberFormat)
  {
    myNumberFormat = numberFormat;
  }

  protected NumberFormat getNumberFormat()
  {
    return myNumberFormat;
  }

}


