/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.content;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.widget.GuiComponent;
import org.vovida.prov.widget.TextField;
import org.vovida.prov.widget.ComboBox;
import org.vovida.prov.psLib.PSDBResult;

import java.util.List;
import java.util.Hashtable;
import java.util.Enumeration;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import javax.swing.JFrame;
import javax.swing.text.PlainDocument;
import javax.swing.text.Document;
import javax.swing.text.BadLocationException;
import javax.swing.ComboBoxModel;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JOptionPane;

import org.jdom.Element;

import org.apache.log4j.Category;

/**
 * @author Barbara Samson
 */
public class StringType extends ContentData implements DTDConstants
{

  /**
   * Used if the UI is a TextField.
   */
  protected Document myDocument = null;

  /**
   * Used for logging.
   */
  static Category logCategory = 
    Category.getInstance(StringType.class.getName());

  /**
   * Class constructor. The default ui is a text field. We need to set the default ui during construction because
   * we may not get further XML processing instructions to change it.
   * @param name The name label to be displayed in the UI.
   * @param typeElement could be null.
   */
  public StringType(String name, GuiComponent gui, 
                    ProvisionableInstance instance, Element typeElement)
  {
    super(name, instance, typeElement);

    setGuiComponent(gui);

    if (gui instanceof TextField)
    {
      setDocument(newDocument());
      setTextFieldModel((TextField) gui, myDocument);
    }

    if (typeElement != null)
    {
      List typeChildren = typeElement.getChildren();

      if (!typeChildren.isEmpty())
      {
        Element restrictionElement = (Element) typeChildren.get(0);
        String restrictionName = restrictionElement.getName();

        if (restrictionName.equals(PSERVER))
        {
          setUpPserverQuery(restrictionElement);
        }
      }
    }
  }

  /**
   * Cloning allows new table rows to be created.
   */
  public Object clone() throws CloneNotSupportedException
  {
    StringType clone = (StringType) super.clone();
    GuiComponent cloneGui = (GuiComponent) getGuiComponent().clone();

    clone.setGuiComponent(cloneGui);

    if (cloneGui instanceof TextField)
    {
      Document doc = newDocument();

      clone.setDocument(doc);
      clone.setTextFieldModel((TextField) cloneGui, doc);
    }

    return clone;
  }

  protected Document newDocument()
  {
    return new PlainDocument();
  }

  protected void setDocument(Document doc)
  {
    myDocument = doc;
  }

  protected void setTextFieldModel(TextField gui, Document doc)
  {
    gui.setModel(doc);
  }

  /**
   * Returns the text content.
   */
  public String getContent() throws InvalidDataException
  {
    try
    {
      validate();
    }
    catch (InvalidDataException e)
    {

      // disable JOptionPane for unit testing
      if (!myProvisionableInstance.isTest())
      {
        String hint = getHint();

        JOptionPane.showMessageDialog(new JFrame(), 
                e.getMessage() + "\n" + hint, "Invalid Data", 
                JOptionPane.ERROR_MESSAGE);
        getGuiComponent().getEditorComponent().requestFocus();
      }

      throw e;
    }

    String content = getContentAsString();

    savingData(content);

    return content;
  }

  protected String getHint()
  {
    return "Any string is valid";
  }

  /**
   * Returns the text content without validation.
   */
  public String getContentAsString()
  {
    String ret = null;

    if (myDocument != null)
    {
      try
      {
        ret = myDocument.getText(0, myDocument.getLength());
      }
      catch (BadLocationException e) {}

      return ret;
    }

    return "";
  }

  protected Document getDocument()
  {
    return myDocument;
  }

  public void setContent(String content)
  {
    if (myDocument != null)
    {
      try
      {
        myDocument.remove(0, myDocument.getLength());
        myDocument.insertString(0, content, null);
      }
      catch (BadLocationException e)
      {
        return;
      }

      super.setContent(content);
    }
  }

  public void validate(String value) throws InvalidDataException
  {
    checkIfRequired(value);
    checkAlert(value);
  }

  /**
   * Fill in content from PServer query.
   */
  private void setUpPserverQuery(Element pserverElement)
  {
    logCategory.info("found a PServer query");

    Element queryElement = null;

    queryElement = pserverElement.getChild(GETRECORDSINDOMAIN);

    if (queryElement != null)
    {
      logCategory.info("query is GetRecordsInDomain");

      PSDBResult result = callGetRecordsInDomain(queryElement);

      if (result == null)
      {
        logCategory.warn("null result returned from getRecordsInDomain query");

        return;
      }

      Hashtable[] rows = result.getRows();

      if (rows.length > 0)
      {
        Enumeration enum = rows[0].elements();

        if (enum.hasMoreElements())
        {
          setContent((String) enum.nextElement());
        }
      }
    }
    else
    {
      queryElement = pserverElement.getChild(GETDOMAINS);

      if (queryElement != null)
      {
        logCategory.info("query is GetDomains");

        String[] domains = callGetDomains(queryElement);

        if ((domains == null) || (domains.length == 0))
        {
          logCategory.warn("null result returned from getDomains query");

          return;
        }

        setContent(domains[0]);
      }
    }
  }


}

