/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.content;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.dynamicGui.DescriptionReader;
import org.vovida.prov.widget.GuiFactory;
import org.vovida.prov.widget.GuiException;
import org.vovida.prov.widget.GuiComponent;
import org.vovida.prov.widget.TextField;
import org.vovida.prov.widget.Table;
import org.vovida.prov.util.XMLProcessor;

import org.vovida.prov.util.XMLOutputterOverride;
import org.jdom.output.XMLOutputter;
import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.DefaultHandler;

import java.util.Vector;
import java.util.Enumeration;
import java.util.List;
import java.util.ListIterator;
import java.awt.Container;
import javax.swing.JComponent;
import javax.swing.JTable;
import javax.swing.JFrame;
import javax.swing.JOptionPane;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JCheckBox;
import javax.swing.table.TableModel;
import javax.swing.table.AbstractTableModel;
import javax.swing.event.TableModelListener;
import javax.swing.CellEditor;
import java.io.IOException;

import org.apache.log4j.Category;

/**
 * @author Barbara Samson
 */
public class TableType extends XMLType implements DTDConstants
{

  // constants used in table XML stored in database.
  private static final String TABLE = "table";
  private static final String ROW = "row";

  private MyTableModel myModel;
  private Vector myContentDatas = new Vector();

  /**
   * for logging
   */
  static Category logCategory = 
    Category.getInstance(TableType.class.getName());

  /**
   * Class Constructor.
   * @param typeElement could be null.
   */
  public TableType(String name, GuiComponent gui, 
                   ProvisionableInstance instance, Element typeElement)
  {
    super(name, instance, typeElement);

    myModel = new MyTableModel();

    setGuiComponent(gui);

    if (gui instanceof Table)
    {
      ((Table) gui).setContent(this);
    }
    else
    {
      logCategory.error("Wrong widget for TableType: " 
                        + gui.getClass().getName());

      return;
    }

    if (typeElement != null)
    {
      List columnList = typeElement.getChildren(COLUMN);
      ListIterator columnIter = columnList.listIterator();

      int position = 0;

      while (columnIter.hasNext())
      {

        // create column widget
        Element columnElement = (Element) columnIter.next();
        String columnName = columnElement.getAttributeValue(NAME);
        String columnLabel = columnElement.getChildTextTrim(LABEL);
        ContentData columnContent = DescriptionReader.addEntry(columnElement, 
                null, myProvisionableInstance);

        addColumn(columnName, columnLabel, columnContent, position++);
      }
    }
  }

  public TableModel getTableModel()
  {
    return myModel;
  }

  public void addRowAfter(int selection)
  {
    myModel.addNewEntry(selection);
  }

  public void deleteRow(int row)
  {
    myModel.deleteRow(row);
  }

  public void addColumn(String name, String label, ContentData content, 
          int position)
  {
    Table tableComponent = (Table) getGuiComponent();
    JTable table = (JTable) tableComponent.getEditorComponent();

    myModel.addColumn(name, label, content, position, table);
  }

  public void validate(String value) throws InvalidDataException
  {

    // first build a DOM out of value.
    Document dom = null;
    try
    {
      dom = XMLProcessor.buildDocument(value);
    }
    catch (JDOMException e)
    {
      throw new InvalidDataException(e.getMessage(), ReasonCodes.BAD_FORMAT);
    }

    Element root = dom.getRootElement();

    // then check myContentDatas against the proposed values in the DOM.
    for (int i = 0; i < myContentDatas.size(); i++)
    {
      ContentData content = (ContentData) myContentDatas.elementAt(i);
      String name = content.getName();
      Element thatElement = root.getChild(name);

      if (thatElement != null)
      {
        content.validate(thatElement.getTextTrim());
      }
    }
  }

  public void validate() throws InvalidDataException
  {
    for (int i = 0; i < myContentDatas.size(); i++)
    {
      ((ContentData) myContentDatas.elementAt(i)).validate();
    }
  }


  protected String getHint()
  {
    return "Content is a table";
  }

  protected ContentData[] getContentArray()
  {
    return myModel.createNewEntry();
  }

  protected void setContent(Vector data)
  {
    myModel.setData(data);
  }

  public void setContent(String xmlString)
  {
    if ((xmlString == null) || (xmlString.equals("")))
    {
      myModel.setData(new Vector());

      return;
    }

    try
    {
      XMLProcessor.doSAXParse(xmlString, new SetContentHandler());
    }
    catch (SAXException e1)
    {
      JOptionPane.showMessageDialog(new JFrame(), 
              "Bad XML content: " + e1.getMessage(), "SAXException", 
              JOptionPane.ERROR_MESSAGE);
      e1.printStackTrace();
    }

  }

  public String getContentAsString()
  {
    Element root = new Element(TableType.TABLE);
    Table table = (Table) getGuiComponent();
    CellEditor editor = ((JTable) table.getEditorComponent()).getCellEditor();

    if (editor != null)
    {

      // for some reason the editor is null sometimes
      editor.stopCellEditing();
    }

    Enumeration rows = myModel.getRows();

    while (rows.hasMoreElements())
    {
      ContentData[] row = (ContentData[]) rows.nextElement();
      Element rowElement = new Element(TableType.ROW);

      for (int i = 0; i < row.length; i++)
      {
        Element element = new Element(row[i].getName());

        String value = row[i].getContentAsString();

        if (value != null)
        {
          element.setText(value);
        }

        rowElement.addContent(element);
      }

      root.addContent(rowElement);
    }

    XMLOutputter output = new XMLOutputterOverride("  ", true);
    String s = "";

    try
    {
      s = output.outputString(root);
    }
    catch (IOException e2)
    {
      e2.printStackTrace();
    }

    return new String("<![CDATA[" + s + "]]>");
  }

  /**
   * Handles callbacks from the SAX parsing of the XML content that
   * comes from the database. Puts the XML content into a format that
   * can be held in the TableModel and displayed in the table.
   */
  class SetContentHandler extends DefaultHandler
      {

        // new data model
        private Vector data = null;
  
        // how many columns we have filled in the current row.
        private int rowCounter = 0;

        // current row
        private ContentData[] row = null;

        // true when characters should put result into current row.
        private boolean parsingColumnData = false;

        // buffer where character data is built up
        private StringBuffer buffer;

        public void startElement(String ignore1, String ignore2, 
              String rawName, Attributes attrs)
        {
          if (rawName.equals(TableType.TABLE))
          {
            data = new Vector();
          }
          else if (rawName.equals(TableType.ROW))
          {
            rowCounter = 0;
            row = myModel.createNewEntry();
          }
          else
          {
            parsingColumnData = true;
            buffer = new StringBuffer();
          }
        }

        public void endElement(String ignore1, String ignore2, String rawName)
        {
          if (rawName.equals(TableType.ROW))
          {
            data.add(row);

            row = null;
          }
          else if (rawName.equals(TableType.TABLE))
          {
            setContent(data);
          }
          else //column data ended
          {
            if (row != null)
            {
              row[rowCounter++].setContent(buffer.toString());
            }
          }
          parsingColumnData = false;
        }

        public void characters(char[] ch, int start, int end)
        {
		  if(parsingColumnData)
          {
            buffer.append(ch,start,end); //sax parses may split character data into several chunks
          }
/**
          if ((parsingColumnData) && (row != null))
          {
            row[rowCounter++].setContent(new String(ch, start, end));
          }
**/
        }

      }

}


