/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.dynamicGui;

import org.vovida.prov.content.ReasonCodes;
import org.vovida.prov.content.ContentData;
import org.vovida.prov.content.InvalidDataException;
import org.vovida.prov.widget.GuiComponent;
import org.vovida.prov.widget.GuiException;
import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;
import org.vovida.prov.psLib.PSDBResult;

import org.apache.log4j.Category;

import org.jdom.Document;
import org.jdom.Element;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.List;
import java.util.ListIterator;
import java.io.IOException;
import java.awt.Container;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JOptionPane;


/**
 * A ProvisionableInstance is a single instance of a ProvisionableType.
 * It is equivalent to a single record, or row, of a table in the database.
 *
 * @author Barbara Samson
 */
public class ProvisionableInstance implements ReasonCodes, DTDConstants
{
  private Container myContainer;
  private ProvisionableType myType;

  /**
   * @link aggregation
   * @clientCardinality 1
   * @supplierCardinality 1
   * @directed
   */
  private ContentData lnkContentData;

  /**
   * myContentDatas entries are hashed on DBEntry name
   * @link aggregationByValue
   */
  private Hashtable myContentDatas;

  /**
   * Set to true for unit testing. Disables JOptionPane pop-ups so
   * that program will not be paused during the test.
   */
  private boolean myIsTest = false;

  static Category logCategory = 
    Category.getInstance(ProvisionableInstance.class.getName());

  protected ProvisionableInstance(Container container, ProvisionableType type)
  {
    myContainer = container;
    myType = type;
    myContentDatas = new Hashtable();

    Document dom = type.getDOM();

    // rootElement must be DataDefn
    Element rootElement = dom.getRootElement();

    if (logCategory.isInfoEnabled())
    {
      logCategory.info("creating ProvisionableInstance for type " 
                       + rootElement.getAttributeValue(NAME));
    }

    recurseOnChildren(rootElement, container);

    // perform second pass for References
    Enumeration enum = myContentDatas.elements();

    while (enum.hasMoreElements())
    {
      ContentData content = (ContentData) enum.nextElement();

      content.performSecondPass();
    }

	//set the lastSavedValue to the value currently in the gui (instead of null)
    //so getChangedData works properly
	enum = myContentDatas.elements();
    while(enum.hasMoreElements())
    {
	    ContentData contentData = (ContentData)enum.nextElement();
        contentData.init();
    }
  }

  public String getName()
  {
    return myType.getName();
  }

  public Container getContainer()
  {
    return myContainer;
  }

  public ProvisionableType getType()
  {
    return myType;
  }

  /**
   * Restores all fields to their last saved value
   */
  public void cancel()
  {
    Enumeration enum = myContentDatas.elements();

    while (enum.hasMoreElements())
    {
      ContentData content = (ContentData) enum.nextElement();

      content.cancel();
    }
  }

  /**
   * Sets all fields to their default values, or blank.
   */
  public void reset()
  {
    Enumeration enum = myContentDatas.elements();

    while (enum.hasMoreElements())
    {
      ContentData content = (ContentData) enum.nextElement();

      content.restoreDefault();
    }
  }

  /**
   * Sets the keyfield to a blank value so that the instance
   * may be reused.
   */
  public void clearKeyValue()
  {
    String key = myType.getKey();
    ContentData content = (ContentData) myContentDatas.get(key);

    content.clearLastSavedValue();
  }

  /**
   * Causes all components to be redrawn inside the existing container.
   * Use this to change component visibility.
   */
  public void redraw()
  {
    myContainer.removeAll();

    Document dom = myType.getDOM();

    Element rootElement = dom.getRootElement();

    if (logCategory.isInfoEnabled())
    {
      logCategory.info("redrawing ProvisionableInstance of type " 
                       + rootElement.getAttributeValue(NAME));
    }

    redrawChildren(rootElement, myContainer);
  }


  private void redrawChildren(Element currentElement, Container container)
  {
    List children = currentElement.getChildren();
    ListIterator iter = children.listIterator();

    while (iter.hasNext())
    {
      Element child = (Element) iter.next();
      String childName = child.getName();

      if (childName.equals(DBENTRY))
      {
        GuiComponent nextWidget = 
          getWidgetByName(child.getAttributeValue(NAME));

        try
        {
          nextWidget.renderSelf(container);
        }
        catch (GuiException e)
        {
          e.printStackTrace();

          return;
        }

        redrawChildren(child, container);
      }
      else if (childName.equals(BORDER))
      {
        Container newContainer = DescriptionReader.addContainer(child, 
                container);

        redrawChildren(child, newContainer);
      }
      else if (childName.equals(PAGED_CONTAINER))
      {
        Container newContainer = DescriptionReader.addPagedContainer(child,container);
        redrawChildren(child, newContainer);
      }
      else if (childName.equals(PAGE))
      {
        Container newContainer = DescriptionReader.addPage(child,container);
        redrawChildren(child,newContainer);
      }
    }
  }


  /**
   * Sets the value of all fields from a hashtable.
   * @param data a hashtable of String values that are keyed on
   * DBField names.
   */
  public void setData(Hashtable data) throws InvalidDataException
  {
    if (data == null)
    {
      throw new InvalidDataException("null dataset", BAD_FORMAT);
    }

    String[] fields = myType.getDBFields();

    for (int i = 0; i < fields.length; i++)
    {
      ContentData content = (ContentData) myContentDatas.get(fields[i]);

      if (content == null)
      {
        continue;
      }

      String value = (String) data.get(fields[i]);

      content.setContent(value);
    }
  }


  /**
   * Same as <code>getChangedData()</code> but does not mark the data as
   * saved.
   */
  public Hashtable readChangedData() throws InvalidDataException
  {
    Hashtable hash = new Hashtable();
    String[] keys = this.getType().getDBFields();

    for (int i = 0; i < keys.length; i++)
    {
      ContentData content = (ContentData) myContentDatas.get(keys[i]);
      String currentValue = content.getContentAsString();
      content.validate();
      if (content.dataNotSaved(currentValue))
      {
         hash.put(keys[i], currentValue);
      }
    }
  
    return hash;
  }


  /**
   * Gets the values of all fields that have changed since the last
   * save. Side effect is to mark all fields as saved.
   * @return a hashtable of String values keyed on field names.
   */
  public Hashtable getChangedData() throws InvalidDataException
  {
    Hashtable hash = new Hashtable();
    String[] keys = this.getType().getDBFields();

    for (int i = 0; i < keys.length; i++)
    {
      ContentData content = (ContentData) myContentDatas.get(keys[i]);
      String currentValue = content.getContentAsString();

      if (content.dataNotSaved(currentValue))
      {

        // need to call getContent to activate validation.
        hash.put(keys[i], content.getContent());
      }
    }

    return hash;
  }

  /**
   * Gets the values of all fields. Side effect is to mark all fields
   * as saved.
   * @return a hashtable of String values keyed on field names.
   */
  public Hashtable getData() throws InvalidDataException
  {

    // it might be more efficient to start with the
    // original mapping.
    Hashtable hash = new Hashtable(myContentDatas);
    Enumeration keys = myContentDatas.keys();

    while (keys.hasMoreElements())
    {
      String key = (String) keys.nextElement();
      ContentData content = (ContentData) myContentDatas.get(key);

      hash.put(key, content.getContent());
    }

    return hash;
  }

  /**
   * Writes all data to the database.
   */
  public void addEntryToDB()
  {
    Hashtable fieldValuePairs = null;

    try
    {
      fieldValuePairs = getData();
    }
    catch (InvalidDataException e1)
    {
      JOptionPane.showMessageDialog(new JFrame(), e1.getMessage(), 
              "Error writing data to DB", JOptionPane.ERROR_MESSAGE);

      return;
    }

    if (fieldValuePairs == null)
    {
      logCategory.error("getData returned null fieldValuePairs");

      return;
    }

    ProvisioningInterface ps = ProvisioningInterface.getInstance();
    boolean result = false;

    try
    {
      result = ps.addRecordInDomain(getName(), fieldValuePairs);
    }
    catch (IOException e2)
    {
      JOptionPane.showMessageDialog(new JFrame(), e2.getMessage(), 
              "IOException writing data to DB", JOptionPane.ERROR_MESSAGE);

      return;
    }
    catch (OperationException e3)
    {
      JOptionPane.showMessageDialog(new JFrame(), e3.getMessage(), 
              "OperationException writing data to DB", 
              JOptionPane.ERROR_MESSAGE);
    }

    if (!result)
    {
      JOptionPane.showMessageDialog(new JFrame(), "write data failed");
    }
  }

  /**
   * Reads data for all fields from the database.
   */
  public void readDataFromDB()
  {
    String keyField = myType.getKey();
    ContentData keyContent = (ContentData) myContentDatas.get(keyField);
    String keyValue = keyContent.getContentAsString();

    if ((keyValue == null) || (keyValue.equals("")))
    {

      // this should never happen
      logCategory.error("Null key value in readDataFromDB");

      return;
    }

    String filter = "(" + keyField + "=\"" + keyValue + "\")";

    ProvisioningInterface ps = ProvisioningInterface.getInstance();

    try
    {
      PSDBResult result = ps.getRecordsInDomain(getName(), filter, 
              myType.getDBFields());
      ;

      if (result == null)
      {
        logCategory.error("Null result returned from getRecordsInDomain");

        return;
      }

      try
      {
        setData(result.getRow(0));
      }
      catch (InvalidDataException e2)
      {
        e2.printStackTrace();

        return;
      }
    }
    catch (IOException e3)
    {
      JOptionPane.showMessageDialog(new JFrame(), e3.getMessage(), 
              "IOException writing data to DB", JOptionPane.ERROR_MESSAGE);

      return;
    }
    catch (OperationException e4)
    {
      JOptionPane.showMessageDialog(new JFrame(), e4.getMessage(), 
              "OperationException writing data to DB", 
              JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * Given the name of a database field, returns a reference
   * to the ContentData that holds the GUI value of that field.
   * @param name ContentData to be returned.
   * @return ContentData with given name.
   */
  public ContentData getContentByName(String name)
  {
    return (ContentData) myContentDatas.get(name);
  }

  /**
   * Given the name of a database field, returns a reference to the
   * widget that displays that field on the screen.
   * @param name of database field
   * @return GuiComponent that renders field on screen.
   */
  public GuiComponent getWidgetByName(String name)
  {
    ContentData content = (ContentData) myContentDatas.get(name);

    if (content == null)
    {
      return null;
    }

    GuiComponent widget = content.getGuiComponent();

    return widget;
  }

  /**
   * Given the name of a database field, returns a reference to the JComponent
   * that is used to set the value for this field.
   * The caller will typically use this call to set an ActionListener on that
   * JComponent.
   * @param name of database field
   * @return JComponent used to set value of field
   */
  public JComponent getComponentByName(String name)
  {
    GuiComponent widget = getWidgetByName(name);

    return widget.getEditorComponent();
  }

  private void recurseOnChildren(Element currentElement, Container container)
  {
    List children = currentElement.getChildren();
    ListIterator iter = children.listIterator();

    while (iter.hasNext())
    {
      Element child = (Element) iter.next();
      String childName = child.getName();

      if (childName.equals(DBENTRY))
      {
        ContentData newContent = DescriptionReader.addEntry(child, container, 
                this);

        myContentDatas.put(newContent.getName(), newContent);
        recurseOnChildren(child, container);
      }
      else if (childName.equals(BORDER))
      {
        Container newContainer = DescriptionReader.addContainer(child, container);
        recurseOnChildren(child, newContainer);
     }
      else if (childName.equals(PAGED_CONTAINER))
      {
        Container newContainer = DescriptionReader.addPagedContainer(child,container);
        recurseOnChildren(child, newContainer);
      }
      else if (childName.equals(PAGE))
      {
        Container newContainer = DescriptionReader.addPage(child,container);
        recurseOnChildren(child, newContainer);
      }
    }
  }

  /**
   * Used for automated testing.
   */
  public void setTest(boolean isTest)
  {
    myIsTest = isTest;
  }

  /**
   * Used for automated testing.
   */
  public boolean isTest()
  {
    return myIsTest;
  }

}



