/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.dynamicGui;

import org.jdom.Document;
import org.jdom.Element;

import java.util.Vector;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.awt.Container;

/**
 * A ProvisionableType is built from an XMLDescription file.
 * It represents a table in the database.
 * Each ProvisionableType must have a unique name, which is the name
 * attribute of DataDefn. It is stored in a hash table, keyed on its name
 * for later retrieval. ProvisionableTypes are created and hashed by
 * the TypeBuilder class.
 *
 * @author Barbara Samson
 */
public class ProvisionableType implements DTDConstants
{

  /**
   * @link dependency
   * @stereotype instantiate
   */

  /* #ProvisionableInstance lnkProvisionableInstance; */

  private String myName;
  private String myLabel;
  private String myTooltip;
  private Document myDOM;

  /**
   * Holds the names of all fields that make up the Primary Key for database
   * references to this Type.
   */
  private String myKey = "";

  /**
   * Returns the name of the database table that represents this Type.
   */
  public String getName()
  {
    return myName;
  }

  /**
   * Returns the name of the suggested label for this Type.
   * This will be the same as name if no label is given.
   */
  public String getLabel()
  {
    return myLabel;
  }

  /**
   * Returns the suggeted tooltip for this Type.
   * Blank if there is no tooltip.
   */
  public String getTooltip()
  {
    return myTooltip;
  }

  public String getKey()
  {
    return myKey;
  }

  public boolean hasKey(String name)
  {
    if (myKey.equals(name))
    {
      return true;
    }

    return false;
  }

  ProvisionableType(String name, Document DOM)
  {
    myName = name;
    myDOM = DOM;

    // find all key values
    Element rootElement = myDOM.getRootElement();

    myKey = rootElement.getAttributeValue(KEY);
    myTooltip = rootElement.getChildText(TOOLTIP);

    String label = rootElement.getChildText(LABEL);

    if ((label == null) || (label.equals("")))
    {
      myLabel = myName;
    }
    else
    {
      myLabel = label;
    }
  }

  public Document getDOM()
  {
    return myDOM;
  }

  /**
   * Returns all of the field names associated with this type.
   */
  public String[] getDBFields()
  {
    Vector fields = new Vector();
    Element rootElement = myDOM.getRootElement();

    // this should not happen
    if (rootElement.getName().equals(DBENTRY))
    {
      fields.add(rootElement.getAttributeValue(NAME));
    }

    recurseOnChildren(rootElement, fields);

    // convert fields vector to array of strings
    int numFields = fields.size();

    if (numFields == 0)
    {
      return null;
    }

    String[] retArray = new String[numFields];

    for (int i = 0; i < numFields; i++)
    {
      retArray[i] = (String) fields.elementAt(i);
    }

    return retArray;
  }


  /**
   * Creates a new ProvisionableInstance which is of this type.
   */
  public ProvisionableInstance buildInstance(Container container)
  {
    return new ProvisionableInstance(container, this);
  }


  private void recurseOnChildren(Element currentElement, Vector fields)
  {
    List children = currentElement.getChildren();
    ListIterator iter = children.listIterator();

    while (iter.hasNext())
    {
      Element child = (Element) iter.next();

      if (child.getName().equals(DBENTRY))
      {
        fields.add(child.getAttributeValue(NAME));
      }

      recurseOnChildren(child, fields);
    }
  }

}




