/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.dynamicGui;

import org.vovida.prov.util.XMLProcessor;

import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;

import org.jdom.Document;
import org.jdom.JDOMException;

import java.util.Hashtable;
import java.io.File;
import java.net.URL;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.FileReader;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.FileNotFoundException;
import javax.swing.JOptionPane;
import javax.swing.JFrame;

import org.apache.log4j.Category;

/**
 * This class creates ProvisionableTypes and stores them in a hashtable
 * keyed on their names.
 *
 * @author Barbara Samson
 */
public class TypeBuilder
{

  /**
   * @link dependency
   * @stereotype instantiate
   */

  /* #ProvisionableType lnkProvisionableType; */

  /**
   * @link aggregationByValue
   * @directed
   */
  private static Hashtable myTypes = new Hashtable();

  private static String dtdString;
  private static int dtdLineCount = 0;

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(TypeBuilder.class.getName());

  /**
   * The DTD that we validate against is type.dtd. This is stored as a resource
   * in the jarfile. Unfortunately, the only way to get the parser to validate
   * an XML file against type.dtd is to prepend it to the XML file before
   * parsing. Therefore, we check to make sure it is present before doing
   * anything else.
   */
  static
  {
    URL resourceURL = TypeBuilder.class.getResource("resources/type.dtd");

    if (resourceURL == null)
    {
      logCategory.error("resource URL is null");
    }
    else
    {
      try
      {
        InputStream stream = resourceURL.openStream();
        BufferedReader reader = 
          new BufferedReader(new InputStreamReader(stream));
        StringBuffer buf = new StringBuffer();
        String line = "";
        int linesRead = 0;

        while ((line = reader.readLine()) != null)
        {
          linesRead++;

          buf.append(line + "\n");
        }

        dtdString = new String(buf);
        dtdLineCount = linesRead;
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }
  }

  /**
   * Create a ProvisionableType and store it in the hashtable for future
   * reference. Gets the type name from the name attribute of DataDefn
   * in xmlString.
   * @param xmlString XML conforming to type.dtd that describes the
   * ProvisionableType.
   */
  public static void buildType(final String xmlString)
  {

    // first dig the typeName out of the file
    // someday we should do this without parsing twice.
    String typeName = null;

    try
    {
      XMLProcessor.doSAXParse(xmlString, new DefaultHandler()
      {
        public void startElement(String ignore1, String ignore2, 
                String rawName, Attributes attrs)
        {
          if (rawName.equals("DataDefn"))
          {
            String typeName = attrs.getValue("name");

            if ((typeName == null) || (typeName.equals("")))
            {
              JOptionPane.showMessageDialog(new JFrame(), 
                      "Data Definition has no name attribute");

              return;
            }

            getType(typeName, xmlString);
          }
        }

      });
    }
    catch (SAXException e)
    {
      e.printStackTrace();
    }
  }

  /**
   * Get a ProvisionableType by name.
   * @return the ProvisionableType named typeName, or null if no
   * such entry exists in the hashtable.
   */
  public static ProvisionableType getType(String typeName)
  {
    Object type = myTypes.get(typeName);

    return (type == null) ? null : (ProvisionableType) type;

    // when we get Pserver queries going we will use them to find the type.
  }

  /**
   * Get a ProvisionableType by name, or create it by parsing xml
   * if it does not already exist.
   * @param typeName key on which type is ProvisionableType is hashed.
   * @param xml XML string to build ProvisionableType from if it does
   * not already exist.
   * @return the ProvisionableType named typeName, or null if no
   * such entry exists in the hashtable and there is an error parsing
   * xml.
   * @param typeName name of the type.
   * @param xmlString XML conforming to type.dtd that describes the
   * ProvisionableType.
   */
  public static ProvisionableType getType(String typeName, String xmlString)
  {

    // first check to see if type exists
    if (typeName != null)
    {
      Object type = myTypes.get(typeName);

      if (type != null)
      {
        return (ProvisionableType) type;
      }
    }

    // if type.dtd is missing, there is nothing we can do.
    if (dtdString == null)
    {
      logCategory.error("Cannot locate dtd. Check to make sure type.dtd is present");
      JOptionPane.showMessageDialog(new JFrame(), 
              "Check to make sure type.dtd is present", "Error Parsing DTD", 
              JOptionPane.ERROR_MESSAGE);

      return null;
    }

    int declEndIdx = 0;

    if (xmlString.trim().startsWith("<?"))
    {
      declEndIdx = xmlString.indexOf("?>");
    }

    StringBuffer buf = new StringBuffer(xmlString);

    buf.insert(declEndIdx == 0 ? 0 : declEndIdx + 2, 
               "\n<!DOCTYPE DataDefn [\n" + dtdString + "\n]>");

    Document dom = null;
    try
    {
      dom = XMLProcessor.buildDocument(new String(buf));
    }
    catch (JDOMException e)
    {
      String message = e.getMessage();

      // because we prepended the DTD to the XML, we must
      // adjust the line number in the error message.
      int indexOfColon = message.lastIndexOf(":");
      int indexOfSpace = message.lastIndexOf(" ", indexOfColon);
      String lineNumberString = message.substring(indexOfSpace + 1, 
              indexOfColon);
      int lineNumber = Integer.parseInt(lineNumberString);

      // the addition of 3 to dtdLineCount comes from
      // the 3 newlines added in the declaration above.
      lineNumber -= (dtdLineCount + 3);

      StringBuffer msgBuf = new StringBuffer(message.substring(0, 
              indexOfSpace + 1));

      msgBuf.append(String.valueOf(lineNumber));
      msgBuf.append(message.substring(indexOfColon));
      JOptionPane.showMessageDialog(new JFrame(), 
              "Error parsing " + typeName + ":\n" + new String(msgBuf), 
              "XML Description Error", JOptionPane.ERROR_MESSAGE);

      return null;
    }

    if (dom == null)
    {
      return null;
    }

    ProvisionableType pType = new ProvisionableType(typeName, dom);

    if (typeName != null)
    {
      myTypes.put(typeName, pType);
    }

    return pType;
  }

  /**
   * Returns the names of all ProvisionableTypes that are stored in the
   * hashtable.
   */
  public static String[] getAllTypeNames()
  {
    return (String[]) myTypes.keySet().toArray(new String[0]);
  }

}




