package org.vovida.prov.group;

import org.vovida.prov.util.assert.*;

import javax.swing.JTree;
import javax.swing.tree.TreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.tree.TreePath;

import java.awt.Point;

import java.awt.dnd.DragSource;
import java.awt.dnd.DropTarget;
import java.awt.dnd.DragSourceDragEvent;
import java.awt.dnd.DragSourceDropEvent;
import java.awt.dnd.DragSourceEvent;
import java.awt.dnd.DragGestureListener;
import java.awt.dnd.DragGestureEvent;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DragSourceListener;
import java.awt.dnd.DropTargetListener;
import java.awt.dnd.DropTargetDropEvent;
import java.awt.dnd.DropTargetDragEvent;
import java.awt.dnd.DropTargetEvent;
import java.awt.dnd.DragGestureRecognizer;
import java.awt.datatransfer.Transferable;

import java.util.Enumeration;



/**
 * A tree where the nodes may be drag-and-dropped. <p>
 * It is assumed that the tree will have nodes which are marked as not allowing
 * children and so may not be drop targets. If a drop operation is performed
 * on one of these nodes, its parent will be retrieved and the dropped node will
 * be added as a child to that node instead.
 */
public abstract class DragAndDropTree extends JTree implements DragGestureListener,
    DragSourceListener, DropTargetListener, TreeSelectionListener
{
    private DragSource dragSource;
    private TreePath sourcePath = null;

    public DragAndDropTree (DefaultTreeModel model)
    {
        super(model);
        dragSource = DragSource.getDefaultDragSource();
        DragGestureRecognizer recognizer = dragSource.createDefaultDragGestureRecognizer(this,DnDConstants.ACTION_MOVE,this);
        DropTarget dropTarget = new DropTarget(this,this);
        this.addTreeSelectionListener(this);
    }

    //--------------start DragGestureListener interface------------------------
	public void dragGestureRecognized(DragGestureEvent event)
    {
		Assert.assert("node is a DefaultMutableTreeNode", DefaultMutableTreeNode.class.isInstance(getSelectionPath().getLastPathComponent()));

        //determine if the drag gesture is valid:
        //see if a node is selected
        TreePath path = this.getSelectionPath();
        if (path == null)
        {
            return;
        }
        DefaultMutableTreeNode node = (DefaultMutableTreeNode)path.getLastPathComponent();


		//groups may not be dragged - they should be the only thing that allows
        //children
        if (node.getAllowsChildren())
        {
            return;
        }

        //this doesn't support copy or link operations
        if (event.getDragAction() != DnDConstants.ACTION_MOVE)
        {
            return;
        }

        //Get the transferable object (now know this is a leaf)
		Transferable transferable = new java.awt.datatransfer.StringSelection(node.toString());

        event.startDrag(DragSource.DefaultCopyDrop, transferable, this);
     }
    //--------------end DragGestureListener interface------------------------

    //-------------start DragSourceListener interface------------------------
    public void dragEnter(DragSourceDragEvent event)
    {
//		System.out.println("DragSourceListener.dragEnter");
    }

    public void dragOver(DragSourceDragEvent event)
    {
//        System.out.println("DragSourceListener.dragOver");
    }

    public void dragActionChanged(DragSourceDragEvent event)
    {
//        System.out.println("DragSourceListener.dragSourceChanged");
    }

    public void dropActionChanged(DragSourceDragEvent event)
    {
//        System.out.println("DragSourceListener.dropActionChanged");
    }

    public void dragDropEnd(DragSourceDragEvent event)
    {
//        System.out.println("DragSourceListener.dragDropEnd");
    }

    public void dragExit(DragSourceEvent event)
    {
//        System.out.println("DragSourceListener.dragExit");
//        event.getDragSourceContext().setCursor(DragSource.DefaultMoveNoDrop);
    }

    public void dragDropEnd(DragSourceDropEvent event)
    {
//        System.out.println("DragSourceListener.dragDropEnd");
    }
    //---------------end DragSourceListener interface------------------------

    //-------------start DropTargetListener interface------------------------

     public void dragEnter(DropTargetDragEvent event)
     {
//        System.out.println("DropTargetDragEvent.dragEnter");
     }

	 public void dragExit(DropTargetEvent event)
     {
//        System.out.println("DropTargetDragEvent.dragExit");
     }

	 public void dragOver(DropTargetDragEvent event)
     {
//        System.out.println("DropTargetDragEvent.dragOver");
//        Point location = event.getLocation();
		//accept all drops, even if they fall on a leaf (user). In that case,
        //need to figgure out the user's group and add the user being dragged
        //to it
		event.acceptDrag(DnDConstants.ACTION_MOVE);
     }

	 public void drop(DropTargetDropEvent event)
     {
//        System.out.println("DropTargetDragEvent.drop");

        Assert.assert("sourcePath != null", sourcePath != null);

        //determine whether the drop is valid:

        //check the data flavour
        Transferable transferable = event.getTransferable();
        if (!isDataFlavourSupported(transferable))
        {
            event.rejectDrop();
            return;
        }

        //find the new parent node
        Point location = event.getLocation();
        TreePath destinationPath = this.getPathForLocation(location.x,location.y);

        //see if this is a valid destionation:

        if (destinationPath == null)
        {
            event.rejectDrop();
            return;
        }

        //need to figgure out whether the destination is a folder or not
        //it is assumed that all nodes that are not folders will have
        //allowsChildren set to false.
        //The behaviour will be that if the destination is not a floder but some
        //other leaf, then the source will be moved to the folder that is the
        //parent of the destination node that it was dropped onto
        //So ....

        //first, get the destination node
        DefaultMutableTreeNode destination = (DefaultMutableTreeNode)destinationPath.getLastPathComponent();

        //see if this destination node allows children
        while (!destination.getAllowsChildren())
        {
            //find the parent of this destination and make that the new
            //destination
            destination = (DefaultMutableTreeNode)destination.getParent();
     	}

        DefaultMutableTreeNode source = (DefaultMutableTreeNode)sourcePath.getLastPathComponent();

        if (destination.equals(source))
        {
            event.rejectDrop();
            return;
        }

        event.acceptDrop(DnDConstants.ACTION_MOVE);

        //remove the source node from its parent
        DefaultMutableTreeNode sourceParent = (DefaultMutableTreeNode)source.getParent();
        sourceParent.remove(source);

        //find out whether the new parent already contains this node
        Enumeration children = destination.children();
        boolean okToAdd = okToDropNode(destination,source);
        while(children.hasMoreElements())
        {
        	DefaultMutableTreeNode child = (DefaultMutableTreeNode)children.nextElement();
            if (child.getUserObject().toString().equals(source.getUserObject().toString()))
            {
                okToAdd = false;
            }
        }
		if (okToAdd)
        {
			//add the source to its new parent
			destination.add(source);
        }

        this.dragEnded(sourceParent,destination,source);

        //repaint the tree
        DefaultTreeModel model = (DefaultTreeModel)this.getModel();
		model.reload(sourceParent);
        model.reload(destination);
     }

	 public void dropActionChanged(DropTargetDragEvent event)
     {
//        System.out.println("DropTargetDragEvent.event");
     }

    //--------------end DropTargetListener interface------------------------

    //---------------start TreeSelectionListener interface----------------------

     public void valueChanged(TreeSelectionEvent event)
     {
//		System.out.println("treeSelection - value changed");
		sourcePath = event.getNewLeadSelectionPath();
     }

    //-----------------end TreeSelectionListener interface----------------------

     protected abstract boolean isDataFlavourSupported(Transferable transferable);
     protected abstract void dragEnded(DefaultMutableTreeNode oldParent,
        DefaultMutableTreeNode newParent, DefaultMutableTreeNode movedNode);

     protected abstract boolean okToDropNode(DefaultMutableTreeNode parent,
       DefaultMutableTreeNode newChild);
}
