package org.vovida.prov.group;

import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.UnsupportedFlavorException;

/**
 * This is stored as the user object in the group tree and represents the
 * group itself. Members should be children of this node. It stores a flag
 * indicating whether the list of members for this group has been loaded from
 * the server or not.
 **/
public class GroupNode implements Transferable
{
    private Object target = null;
    private boolean dataLoaded = false;
	private String groupName;

    /**
     * Defines a custom type of data (the flavour) that this transferable object
     * may be represented as.
     */
    public static final DataFlavor GROUP_NODE_FLAVOR = new DataFlavor(GroupNode.class, "Group node");

    /**
     * A list of the types of data that a GroupNode may be represented as.
     */
    private static final DataFlavor [] flavors = new DataFlavor[] {GROUP_NODE_FLAVOR, DataFlavor.stringFlavor};


    public GroupNode(Object obj, String aGroupName)
    {
        target = obj;
        groupName = aGroupName;
    }

    /**
     * Returns the "nice" name of the group. <p>
     * This is used for the root of the tree. When displayed on the screen,
     * that node should be labeled "groups" or something similar. However, when
     * passed to the provisioning server, its name should be an empty string in
     * order to retrieve and add to it all accounts that don't belong to any
     * group.
     */
    public String getGroupName()
    {
        return groupName;
    }


    /**
     * Return value of flag indicating whether the group members for this node
     * have been loaded from the provisioning server.<p>
     */
    public boolean isDataLoaded()
    {
    	return dataLoaded;
    }

	/**
     * Flag set to indicate whether the data for this group (its children) has
     * been loaded from the server yet.
     */
    public void setDataLoaded()
    {
        dataLoaded = true;
    }

    /**
     * This should return the group name as understood by the provisioning
     * server.<p>
     */
    public String toString()
    {
    	return target.toString();
    }

    //------------start implementation of Transferable interface----------------

    public DataFlavor [] getTransferDataFlavors()
    {
        return flavors;
    }

    public boolean isDataFlavorSupported(DataFlavor flavor)
    {
        for (int i=0; i<flavors.length; i++)
        {
            if (flavors[i].equals(flavor))
            {
                return true;
            }
        }
        return false;
    }

    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException
    {
        if (flavor.equals(DataFlavor.stringFlavor))
        {
            return this.toString();
        }
        else if (flavor.equals(GROUP_NODE_FLAVOR))
        {
            return this;
        }
        else
        {
            throw new UnsupportedFlavorException(flavor);
        }

    }

}

