package org.vovida.prov.group;

import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.JScrollPane;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreePath;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.InputEvent;
import java.awt.BorderLayout;
import javax.swing.JPopupMenu;
import java.util.Arrays;

import org.vovida.prov.util.ProvisioningServerWrapper;
import org.vovida.prov.util.assert.*;
import org.vovida.prov.util.task.TaskListener;
import org.vovida.prov.util.task.Task;

import org.vovida.prov.group.action.ActionGroup;
import org.vovida.prov.group.action.ActionDeleteGroup;
import org.vovida.prov.group.action.ActionNewGroup;
import org.vovida.prov.group.task.TaskLoadGroupMembers;

import org.apache.log4j.Category;


/**
 * Contains a panel which displays the groups defined by an account in a tree. <p>
 * The list of groups defined by a particular administrative account are
 * displayed as a folder in a tree while the group members are its children.
 * The list of members is loaded as each group node is expanded.
 **/
public class GroupScreen extends JPanel
{
    private static Category logCategory = Category.getInstance(GroupScreen.class.getName());

    private DragAndDropTree groupTree;
    private DefaultTreeModel model;

	/**
     * Stores the actions which respond to clicks in the popup menu. <p>
     * These are saved so that they may be accessed each time the menu is poped
     * up so that the tree path the user clicked at may be saved.
     **/
    private ActionGroup [] actions;


    /**
     * The identifier of the domain (db table) that the account whose groups
     * are being displayed is stored in. <p>
     */
    private String parentDomain;

    private String childDomain;

    /**
     * The fully quallified identifer of the acount whose groups are being
     * displayed.
     */
    private String accountIdentifier;

    /**
     * Used to communicate with the provisioning server.
     **/
	private ProvisioningServerWrapper server;

    /** @link dependency */
    /*#GroupSelectionListener lnkGroupSelectionListener;*/

    /** @link dependency */
    /*#GroupNode lnkGroupNode;*/

    public GroupScreen (String parent, String child, String accountId,
        ProvisioningServerWrapper wrapper)
    {
		Assert.assert("server != null", wrapper != null);
//        Assert.assert("accountId != null && accountId != \"\" ", accountId != null && accountId != "");
        Assert.assert("accountId != null", accountId != null);
		accountIdentifier = accountId; //the parent account
        parentDomain = parent;
        childDomain = child;
        server = wrapper;
        initGroups();
		loadMembersWithNoGroup();
        initLayout();
        initExpansionListener();
        initPopupMenu();
    }


    private void initPopupMenu()
    {
        Assert.assert("groupTree != null", groupTree != null);

        final JPopupMenu menu = new JPopupMenu();
        actions = new ActionGroup[2];
        actions[0] = new ActionNewGroup("Create group", accountIdentifier, parentDomain, server, groupTree);
        actions[1] = new ActionDeleteGroup("Remove group", accountIdentifier, parentDomain, childDomain, server, groupTree);
        for (int i=0; i<actions.length; i++)
        {
            menu.add(actions[i]);
        }

        groupTree.addMouseListener(new MouseAdapter()
        {
        	public void mouseReleased(MouseEvent event)
	        {
                if ((event.getModifiers() & InputEvent.BUTTON3_MASK)
                    == InputEvent.BUTTON3_MASK)
                {
                    int x = event.getX();
                    int y = event.getY();
                    TreePath path = groupTree.getPathForLocation(x,y);
                    if (path != null) // the user clicked on a path in the tree
                    {
                        for (int i=0; i<actions.length; i++)
                        {
                            actions[i].setClickedPath(path);
                        }

						//check whether the path is a group
                        DefaultMutableTreeNode node = (DefaultMutableTreeNode)path.getLastPathComponent();
                        if (!GroupNode.class.isInstance(node.getUserObject()))
                        {
                            actions[1].setEnabled(false); //the user clicked on an account
                        }
                        else
                        {
	                        actions[1].setEnabled(true); //the user clicked on a group
                        }
                    }
                    else
                    {
                        actions[1].setEnabled(false); //only enable "remove group" option if the user clicked on a group
                    }
                    menu.show(groupTree,x,y);
                }
	        }
        });
    }

     private void initExpansionListener()
    {
        Assert.assert("groupTree != null", groupTree != null);
        Assert.assert("tree model not null", model != null);
        groupTree.addTreeExpansionListener(new GroupSelectionListener(childDomain,
            accountIdentifier, server, model));
    }

    /**
     * Layout the components which are to be shown in this panel. <p>
     */
    private void initLayout()
    {
		Assert.assert("groupTree != null", groupTree != null);
		this.setLayout(new BorderLayout());
        JScrollPane scroller = new JScrollPane(groupTree);
        this.add(scroller,BorderLayout.CENTER);

        groupTree.setCellRenderer(new GroupTreeRenderer());
    }


    /**
     * Get the list of groups that are defined for the given account identifier
     * and create a top level node in the tree for each group. <p>
     * Group members will be loaded as the tree is expanded.
     **/
    private void initGroups()
    {
		//create a node for the root of the tree
        GroupNode rootNode = new GroupNode("Groups","");
        rootNode.setDataLoaded();
        DefaultMutableTreeNode root = new DefaultMutableTreeNode(rootNode); //the display name is "groups" while the group name is empty string

        String [] groups = new String[]{};
        try
        {
			//get the list of groups this account defines
            groups = server.getDefinedGroups(parentDomain, accountIdentifier);
	
	    	//sort the list of groups alphabetically
	        Arrays.sort(groups);
        }
        catch (Exception ex)
        {
            String message = "Could not retrieve list of defined groups for account "
                + accountIdentifier + " in domain " + parentDomain;
            logCategory.error(message,ex);
            org.vovida.prov.util.logging.Util.showException(message,ex);
        }

        //add a top level node to the tree for each group
        for (int i=0; i<groups.length; i++)
        {
			DefaultMutableTreeNode node = new DefaultMutableTreeNode(new GroupNode(groups[i],groups[i]));
            root.add(node);
			//create a temporary node as a child of the group node which will
            //be a placeholder while the list of the group's members is
            //retrieved from the ps
            node.add(new DefaultMutableTreeNode("loading data ..."));
        }

		//create the tree
        model = new DefaultTreeModel(root);
        groupTree = new GroupDragAndDropTree(model, server, childDomain, accountIdentifier);
    }

    /**
     * Start a task to load the account names which belong to this administrator
     * but which do not belong to any group. <p>
     * These will be added to the tree at the same level as the group folders.
     */
    private void loadMembersWithNoGroup()
    {
        new TaskLoadGroupMembers((DefaultMutableTreeNode)model.getRoot(),
            childDomain, accountIdentifier, new TaskListener()
        {
            public void taskFinished(Task task, boolean status, Object returnValue)
            {
                if (status)
                {
	                model.reload((DefaultMutableTreeNode)model.getRoot());
                }
            }
        }, server).start();

    }
}
