/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.navigation;


import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.Hashtable;
import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;
import javax.swing.JFrame;
import javax.swing.JSplitPane;
import javax.swing.JPanel;
import javax.swing.JOptionPane;
import java.awt.CardLayout;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.apache.log4j.Category;

import org.vovida.prov.util.logging.LogClient;
import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;
import org.vovida.prov.psLib.SocketProvisioningInterface;
import org.vovida.prov.util.XMLProcessor;

/**
 * @author Barbara Samson
 */
public class TopLevel extends JFrame implements ConfigConstants
{

  /**
   * @link aggregationByValue
   */
  private JSplitPane mySplitPane;

  /**
   * @clientRole right panel
   */
  private JPanel myRightPanel;

  private CardLayout myRightPanelLayout;

  /**
   * @clientRole left panel
   */
  private NavBar myNavBar;

  /**
   * @clientRole right panel
   */
  private Hashtable myClientHash = new Hashtable();

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(TopLevel.class.getName());

  public TopLevel()
  {
    mySplitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT, true);

    getContentPane().add(mySplitPane);

    myRightPanel = new JPanel();
    myRightPanelLayout = new CardLayout();

    myRightPanel.setLayout(myRightPanelLayout);
    mySplitPane.add(NavBar.getInstance());
    mySplitPane.add(myRightPanel);

    setDefaultCloseOperation(EXIT_ON_CLOSE);
  }

  /**
   * Uses reflection to create clients from the file NavBar.xml,
   * which is delivered by PServer. Each entry in the file has
   * a name attribute, which contains the display name of the
   * client, and a class attribute which contains the fully qualified
   * classname of the client. This is done so that we can use a list
   * of these strings in a user profile file to set up the main
   * navigation panel. Both the order and the type of screens available
   * to a particular user will be determined by the contents of the file.
   */
  public void createClients()
  {
    ProvisioningInterface ps = ProvisioningInterface.getInstance();
    NavBar navBar = NavBar.getInstance();

    String xmlString = null;

    try
    {
      xmlString = ps.getXMLFile(CONFIG, NAVBAR_XML);
    }
    catch (OperationException e2)
    {
      JOptionPane.showMessageDialog(this, 
              "Error in NavBar file: " + e2.getMessage(), 
              "OperationException", JOptionPane.ERROR_MESSAGE);

      return;
    }
    catch (IOException e3)
    {
      JOptionPane.showMessageDialog(this, 
              "Error getting NavBar file.\n" 
              + "Check connection to PServer\n" + e3.getMessage(), 
              "IOException", JOptionPane.ERROR_MESSAGE);

      return;
    }

    try
    {
      XMLProcessor.doSAXParse(xmlString, new DefaultHandler()
      {
        public void startElement(String ignore1, String ignore2, 
                String rawName, Attributes attrs)
        {
          if (rawName.equals(CLIENT))
          {
            String name = attrs.getValue(NAME);
            String classDef = attrs.getValue(CLASS);

            createNavBarClient(name, classDef);
          }
        }

      });
    }
    catch (SAXException e4)
    {
      JOptionPane.showMessageDialog(this, e4.getMessage());
      e4.printStackTrace();

      return;
    }
  }


  private void createNavBarClient(String name, String classDef)
  {
    String dummyString = new String();
    NavBarClient client = null;

    try
    {
      Class navBarClass = Class.forName(classDef);
      Constructor navBarConstructor = navBarClass.getConstructor(new Class[]
      {
        dummyString.getClass(), this.getClass()
      });

      client = (NavBarClient) navBarConstructor.newInstance(new Object[]
      {
        name, this
      });
    }
    catch (ClassCastException e1)
    {
      logCategory.error("ClassCastException for:\n" + name + " (" + classDef 
                        + ")" + " is not of type NavBarClient");
    }
    catch (ClassNotFoundException e2)
    {
      logCategory.error("ClassNotFoundException for:\n" + name + " (" 
                        + classDef + ")" + " cannot be found");
    }
    catch (InstantiationException e3)
    {
      logCategory.error("InstantiationException for:\n" + name + " (" 
                        + classDef + ")" + " cannot be instantiated");
    }
    catch (IllegalAccessException e4)
    {
      logCategory.error("IllegalAccessException for:\n" + name + " (" 
                        + classDef + ")\n" + e4.getMessage());
    }
    catch (NoSuchMethodException e5)
    {
      logCategory.error("NoSuchMethodException for:\n" + name + " (" 
                        + classDef + ")\n" + e5.getMessage());
    }
    catch (InvocationTargetException e6)
    {
      logCategory.error("InvocationTargetException for:\n" + name + " (" 
                        + classDef + ")\n" + e6.getMessage());
      e6.printStackTrace();
    }
    catch (IllegalArgumentException e7)
    {
      logCategory.error("IllegalArgumentException for:\n" + name + " (" 
                        + classDef + ")\n" + e7.getMessage());
    }

    if (client != null)
    {
      myRightPanel.add(client, name);
      myClientHash.put(name, client);
    }
  }

  protected void setTopScreen(String screenName)
  {
    myRightPanelLayout.show(myRightPanel, screenName);

    NavBarClient client = (NavBarClient) myClientHash.get(screenName);

    if (client != null)
    {
      client.showSelf();
    }
  }

  public static void main(String[] args)
  {
    if (args.length != 2)
    {
      System.out.println("Usage: TopLevel <host> <port>");

      return;
    }

    try
    {
      SocketProvisioningInterface.initialize(args[0], 
              Integer.parseInt(args[1]));
    }
    catch (NumberFormatException e)
    {
      System.out.println("Usage: TopLevel <host> <port>");

      return;
    }

    LogClient.configureFromFile();

    TopLevel topLevel = new TopLevel();

    topLevel.createClients();
    topLevel.pack();
    topLevel.setVisible(true);

  }


}




