/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.psLib;

import java.net.Socket;
import java.io.OutputStreamWriter;
import java.io.InputStreamReader;
import java.io.BufferedWriter;
import java.io.BufferedReader;
import java.io.IOException;
import java.util.Vector;

import org.apache.log4j.Category;

/**
 * Implements a Http TCP persistant connection.  If we want to support
 * persistant connections, unfortunately some HTTP logic must be wound tightly
 * with the actual transport (TCP socket).
 *
 * @author Matt Naish
 */
public class HttpPSConnection implements PSConnection
{

  private String host;
  private int port;

  private Socket socket;
  private BufferedWriter out;
  private BufferedReader in;

  /**
   * If a send fails (from timeout, persistant connection closes or other
   * error, resend request a maximum of this many times
   */
  private static final int MAX_SEND_ATTEMPTS = 2;

  /**
   * If no response from the server after this time elaspes, abort waiting
   * for reponse
   */
  private static final int SOCKET_TIMEOUT = 5000;

  private static final String CRLF = "\r\n";

  // private static final String CRLF = "\n";

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(HttpPSConnection.class.getName());

  /**
   * The parameters used to attached to a remote computer
   * @param host the hostname of the remote computer
   * @param port the port of the remote computer
   */
  public HttpPSConnection(String host, int port)
  {
    this.host = host;
    this.port = port;
  }

  /**
   * Send request and receive a response.  Synchronous.
   * @throws IOException if any Input/Output exception occurs while
   * creating the socket, sending the request or reading the response.
   */
  public synchronized PSMessage sendRequest(PSMessage request) 
    throws IOException
  {
    Vector respHeader = new Vector();
    String respBody = null;

    if (socket == null)
    {

      // first time sendRequest() is run, create new socket
      createNewSocket();
    }

    String[] reqHeader = request.getHeader();
    boolean sendSuccess = false;
    String inputLine;
    int contentLength = 0;
    String contentLengthStr;

    // save the output into a buffer for logging
    StringBuffer logBuf = new StringBuffer();

    for (int i = 0; i < MAX_SEND_ATTEMPTS &&!sendSuccess; i++)
    {
      for (int j = 0; j < reqHeader.length; j++)
      {
        if (logCategory.isInfoEnabled())
        {
          logBuf.append(reqHeader[j] + CRLF);
        }

        out.write(reqHeader[j] + CRLF);
      }

      if (logCategory.isInfoEnabled())
      {
        logBuf.append(CRLF);
      }

      out.write(CRLF);

      if (request.getBody() != null)
      {
        if (logCategory.isInfoEnabled())
        {
          logBuf.append(request.getBody());
        }

        out.write(request.getBody());
      }

      out.flush();

      if (logCategory.isInfoEnabled())
      {
        logCategory.info("SENT TO PSERVER:\n" + new String(logBuf));
      }

      inputLine = in.readLine();

      sendSuccess = inputLine != null;

      if (sendSuccess)
      {
        logCategory.info("RETURNED FROM PSERVER:");
        logCategory.info(inputLine);

        // read until there is no more to read from the socket (null) or
        // read until the end of the header is reached (blank line)
        while (inputLine != null &&!inputLine.equals(""))
        {
          respHeader.add(inputLine);

          // find the content length if a content length has not been found yet
          if (contentLength == 0 
              && inputLine.startsWith(HttpXmlProtocolService.CONTENT_LENGTH))
          {
            contentLengthStr = 
              inputLine.substring(HttpXmlProtocolService.CONTENT_LENGTH.length(), 
                    inputLine.length()).trim();

            contentLength = Integer.parseInt(contentLengthStr);
          }

          inputLine = in.readLine();

          logCategory.info(inputLine);
        }

        // now start reading the message body
        if (contentLength > 0)
        {
          char[] messageBody = new char[contentLength];
          int bytesRead = 0;
          int totalBytesRead = 0;
          StringBuffer response = new StringBuffer();

          while (totalBytesRead < contentLength)
          {
            bytesRead = in.read(messageBody, 0, contentLength);

            logCategory.info("bytes read = " + bytesRead);

            if (bytesRead == -1)
            {
              throw new IOException("Input stream closed unexpectedly.");
            }

            totalBytesRead += bytesRead;

            response.append(messageBody, 0, bytesRead);
          }

          respBody = response.toString();

          logCategory.info(respBody);
          logCategory.info("total bytes read = " + totalBytesRead);
        }
        else
        {
          respBody = null;
        }
      }
      else
      {

        // The send failed (because the first in.readLine() returned null).
        // Most likely because the persistant connection was closed by the
        // server.  Open a new socket and try again.
        logCategory.warn("Response failed");
        createNewSocket();
      }
    }

    return new PSMessage((String[]) respHeader.toArray(new String[0]), 
            respBody);
  }

  private void createNewSocket() throws IOException
  {
    close();

    socket = new Socket(host, port);

    socket.setSoTimeout(SOCKET_TIMEOUT);

    out = 
      new BufferedWriter(new OutputStreamWriter(socket.getOutputStream()));
    in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
  }

  public void close() throws IOException
  {
    if (in != null)
    {
        in.close();
        in = null;
    }

    if (out != null)
    {
        out.close();
        out = null;
    }

    if (socket != null)
    {
        socket.close();
        socket = null;
    }
  }

  public static void main(String[] xxx) throws Exception
  {
    PSConnection conn = new HttpPSConnection("mnaish-lnx", 80);
    PSMessage request = new PSMessage(new String[]
    {
      "GET /index.html HTTP/1.1", "Host: "
    }, null);

    PSMessage response;
    String responseStr;

    logCategory.info("Sending request...");

    response = conn.sendRequest(request);
    responseStr = response.getBody();

    logCategory.info(responseStr.substring(responseStr.length() - 20, 
            responseStr.length()));
    logCategory.info("");

    logCategory.info("Sending request...");

    response = conn.sendRequest(request);
    responseStr = response.getBody();

    logCategory.info(responseStr.substring(responseStr.length() - 20, 
            responseStr.length()));
    logCategory.info("");

    logCategory.info("Sleeping for 10sec.");

    try
    {
      Thread.sleep(10000);
    }
    catch (InterruptedException ie)
    {
      ie.printStackTrace();
    }

    logCategory.info("Sending request...");

    response = conn.sendRequest(request);
    responseStr = response.getBody();

    logCategory.info(responseStr.substring(responseStr.length() - 20, 
            responseStr.length()));
    logCategory.info("");

    logCategory.info("Sleeping for 40sec.");

    try
    {
      Thread.sleep(40000);
    }
    catch (InterruptedException ie)
    {
      ie.printStackTrace();
    }

    logCategory.info("Sending request...");

    response = conn.sendRequest(request);
    responseStr = response.getBody();

    logCategory.info(responseStr.substring(responseStr.length() - 20, 
            responseStr.length()));
    logCategory.info("");

    logCategory.info("End");

  }

}

