/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.psLib;

import java.io.IOException;
import java.net.ProtocolException;
import org.jdom.Element;
import org.jdom.Document;
import org.jdom.JDOMException;
import org.jdom.output.XMLOutputter;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.List;
import java.util.Iterator;
import java.util.Vector;
import java.util.Hashtable;

import org.vovida.prov.util.XMLOutputterOverride;
import org.vovida.prov.util.XMLProcessor;

import org.apache.log4j.Category;

/**
 * <P>Provides the implementation of a PSProtocolService.
 * The protocol used for transport is HTTP with an XML payload.</P>
 * <P>Defines the two methods buildRequest(...) and parseResponse(..).</P>
 *
 * @author Matt Naish
 */
public class HttpXmlProtocolService extends PSProtocolService
{

  // request HTTP header keywords
  protected static final String GET = "GET";
  protected static final String PUT = "PUT";
  protected static final String POST = "POST";
  protected static final String DELETE = "DELETE";
  protected static final String HTTP_VER = "HTTP/1.1";
  protected static final String HOST = "Host:";
  protected static final String CONTENT_LENGTH = "Content-Length:";

  // request methods
  private static final String GET_LIST_XML_DESCRIPTIONS_STR = 
    "ListXMLDescriptions";
  private static final String GET_XML_DESCRIPTION_STR = "GetXMLDescription";
  private static final String GET_RECORDS_IN_DOMAIN_STR = 
    "GetRecordsInDomain";
  private static final String SET_RECORDS_IN_DOMAIN_STR = 
    "SetRecordsInDomain";
  private static final String ADD_RECORD_IN_DOMAIN_STR = "AddRecordInDomain";
  private static final String DELETE_RECORDS_IN_DOMAIN_STR = 
    "DeleteRecordsInDomain";
  private static final String SESSION_OPEN_STR = "SessionOpen";
  private static final String SESSION_CLOSE_STR = "SessionClose";
  private static final String SESSION_SEARCH_DB_STR = "SessionSearchDB";
  private static final String SESSION_GET_DB_RESULT_STR = 
    "SessionGetDBResult";
  private static final String SESSION_SORT_DB_RESULT_STR = 
    "SessionSortDBResult";
  private static final String SESSION_SET_DB_RESULT_STR = 
    "SessionSetDBResult";
  private static final String SESSION_DELETE_ALL_STR = "SessionDeleteAll";
  private static final String GET_FIELDS_IN_DOMAIN_STR = "GetFieldsInDomain";

  // request XML keywords
  private static final String ARGUMENTS = "arguments";
  private static final String XMLCATEGORY = "xmlcategory";
  private static final String XMLDESCRIPTION = "xmldescription";
  private static final String FILTER = "filter";
  private static final String FIELD_LIST = "fieldlist";
  private static final String LIST_ITEM = "listitem";
  private static final String FIELD_VALUE = "fieldvalue";
  private static final String EXPIRE_TIME = "expiretime";
  private static final String START = "start";
  private static final String END = "end";
  private static final String SORT_TYPE = "sorttype";
  private static final String SORT_FIELD = "sortfield";

  // request and response XML keywords
  private static final String DOMAIN = "domain";
  private static final String TOKEN = "token";


  // response HTTP codes
  private static final int HTTP_OK = 200;
  private static final int HTTP_NO_CONTENT = 204;
  private static final int HTTP_FAIL = 400;

  // response XML keywords
  private static final String RESPONSE = "response";
  private static final String RESULTCOUNT = "resultcount";
  private static final String MATCH = "match";
  private static final String FIELDVALUE = "fieldvalue";
  private static final String FIELDLIST = "fieldlist";

  /**
   * @link
   * @shapeType PatternLink
   * @pattern Singleton
   * @supplierRole Singleton factory
   */

  /* # private HttpXmlProtocolService _httpxmlProtocolService; */

  private static HttpXmlProtocolService instance = null;

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(HttpXmlProtocolService.class.getName());

  // more here

  private HttpXmlProtocolService() {}

  public static HttpXmlProtocolService getInstance()
  {
    if (instance == null)
    {
      instance = new HttpXmlProtocolService();
    }

    return instance;
  }

  /**
   * Build a request message.
   */
  public PSMessage buildRequest(int type, Object arg0, Object arg1, 
          Object arg2, Object arg3, Object arg4, Object arg5)
  {
    switch (type)
    {

      case PSProtocolService.GET_RAW_FILE:
        return buildGetRawFileRequest((String) arg0);

      case PSProtocolService.GET_LIST_XML_DESCRIPTIONS:
        return buildGetListXMLDescriptionsRequest((String) arg0);

      case PSProtocolService.GET_XML_DESCRIPTION:
        return buildGetXMLFileRequest((String) arg0, (String) arg1);

      case PSProtocolService.GET_RECORDS_IN_DOMAIN:
        return buildGetRecordsInDomainRequest((String) arg0, (String) arg1, 
                (String[]) arg2);

      case PSProtocolService.SET_RECORDS_IN_DOMAIN:
        return buildSetRecordsInDomainRequest((String) arg0, (String) arg1, 
                (Hashtable) arg2);

      case PSProtocolService.ADD_RECORDS_IN_DOMAIN:
        return buildAddRecordInDomainRequest((String) arg0, (Hashtable) arg1);

      case PSProtocolService.DELETE_RECORDS_IN_DOMAIN:
        return buildDeleteRecordsInDomainRequest((String) arg0, 
                (String) arg1);

      case PSProtocolService.SESSION_OPEN:
        return buildSessionOpenRequest(((Long) arg0).longValue());

      case PSProtocolService.SESSION_CLOSE:
        return buildSessionCloseRequest((String) arg0);

      case PSProtocolService.SESSION_SEARCH_DB:
        return buildSessionSearchDBRequest((String) arg0, (String) arg1, 
                (String) arg2, (String[]) arg3, (String) arg4, (String) arg5);

      case PSProtocolService.SESSION_GET_DB_RESULT:
        return buildSessionGetDBResultRequest((String) arg0, 
                ((Integer) arg1).intValue(), ((Integer) arg2).intValue(), 
                (String[]) arg3);

      case PSProtocolService.SESSION_SORT_DB_RESULT:
        return buildSessionSortDBResultRequest((String) arg0, (String) arg1, 
                (String) arg2);

      case PSProtocolService.SESSION_SET_DB_RESULT:
        return buildSessionSetDBResultRequest((String) arg0, 
                (Hashtable) arg1);

      case PSProtocolService.SESSION_DELETE_ALL:
        return buildSessionDeleteAllRequest((String) arg0);

      case PSProtocolService.GET_FIELDS_IN_DOMAIN:
      {
        return buildGetFieldsInDomainRequest((String) arg0);
      }

      default:
        throw new IllegalArgumentException("No matching 'type' found.");
    }
  }

  private PSMessage buildGetListXMLDescriptionsRequest(String category)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, XMLCATEGORY, category);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      GET + " " + GET_LIST_XML_DESCRIPTIONS_STR + " " + HTTP_VER, 
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildGetXMLFileRequest(String category, String filename)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, XMLCATEGORY, category);
    addTextElement(argumentsElem, XMLDESCRIPTION, filename);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      GET + " " + GET_XML_DESCRIPTION_STR + " " + HTTP_VER, 
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildGetRawFileRequest(String filename)
  {
    String[] header = 
    {
      GET + " " + filename + " " + HTTP_VER, CONTENT_LENGTH + " " + 0
    };

    return new PSMessage(header, "");
  }

  private PSMessage buildGetFieldsInDomainRequest(String domain)
  {
    checkDomain(domain);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, DOMAIN, domain);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      GET + " " + GET_FIELDS_IN_DOMAIN_STR + " " + HTTP_VER, 
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildGetRecordsInDomainRequest(String domain, 
          String filter, String[] fieldList)
  {
    checkDomain(domain);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, DOMAIN, domain);
    addTextElement(argumentsElem, FILTER, filter);
    addFieldList(argumentsElem, fieldList);

    String body = createXMLString(argumentsElem);

    String[] header = 

    // {GET + " " + GET_RECORDS_IN_DOMAIN_STR + " " + HTTP_VER,
    // HOST,
    // CONTENT_LENGTH + " " + body.length()};
    {
      GET + " " + GET_RECORDS_IN_DOMAIN_STR + " " + HTTP_VER, 
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSetRecordsInDomainRequest(String domain, 
          String filter, Hashtable fieldValuePairs)
  {
    checkDomain(domain);
    checkFieldValuePairs(fieldValuePairs);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, DOMAIN, domain);
    addTextElement(argumentsElem, FILTER, filter);
    addFieldValuePairs(argumentsElem, fieldValuePairs);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      PUT + " " + SET_RECORDS_IN_DOMAIN_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }


  private PSMessage buildAddRecordInDomainRequest(String domain, 
          Hashtable fieldValuePairs)
  {
    checkDomain(domain);
    checkFieldValuePairs(fieldValuePairs);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, DOMAIN, domain);
    addFieldValuePairs(argumentsElem, fieldValuePairs);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      PUT + " " + ADD_RECORD_IN_DOMAIN_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildDeleteRecordsInDomainRequest(String domain, 
          String filter)
  {
    checkDomain(domain);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, DOMAIN, domain);
    addTextElement(argumentsElem, FILTER, filter);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      DELETE + " " + DELETE_RECORDS_IN_DOMAIN_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildRegisterDomainForUpdateRequest(String domain, 
          Object huh)
  {
    throw new Error("buildRegisterDomainForUpdateRequest not implemented yet");
  }

  private PSMessage registerRecordForUpdate(String domain, Object huh)
  {
    throw new Error("buildRegisterRecordForUpdateRequest not implemented yet");
  }

  private PSMessage buildGetCurrentRecordsInDomainRequest(String domain, 
          String filter, String[] fieldList)
  {
    throw new Error("buildGetCurrentRecordsInDomianRequest not implemented yet");
  }

  private PSMessage buildSessionOpenRequest(long timeout)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, EXPIRE_TIME, String.valueOf(timeout));

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      GET + " " + SESSION_OPEN_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionCloseRequest(String token)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      POST + " " + SESSION_CLOSE_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionSearchDBRequest(String token, String domain, 
          String filter, String[] fieldList, String sortType, 
          String sortField)
  {
    checkDomain(domain);

    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);
    addTextElement(argumentsElem, DOMAIN, domain);
    addTextElement(argumentsElem, FILTER, filter);
    addFieldList(argumentsElem, fieldList);
    addTextElement(argumentsElem, SORT_TYPE, sortType);
    addTextElement(argumentsElem, SORT_FIELD, sortField);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      POST + " " + SESSION_SEARCH_DB_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionGetDBResultRequest(String token, int start, 
          int end, String[] fieldList)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);
    addTextElement(argumentsElem, START, String.valueOf(start));
    addTextElement(argumentsElem, END, String.valueOf(end));
    addFieldList(argumentsElem, fieldList);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      GET + " " + SESSION_GET_DB_RESULT_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionSortDBResultRequest(String token, 
          String sortType, String sortField)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);
    addTextElement(argumentsElem, SORT_TYPE, sortType);
    addTextElement(argumentsElem, SORT_FIELD, sortField);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      POST + " " + SESSION_SORT_DB_RESULT_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionSetDBResultRequest(String token, 
          Hashtable fieldValuePairs)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);
    addFieldValuePairs(argumentsElem, fieldValuePairs);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      PUT + " " + SESSION_SET_DB_RESULT_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  private PSMessage buildSessionDeleteAllRequest(String token)
  {
    Element argumentsElem = new Element(ARGUMENTS);

    addTextElement(argumentsElem, TOKEN, token);

    String body = createXMLString(argumentsElem);

    String[] header = 
    {
      DELETE + " " + SESSION_DELETE_ALL_STR + " " + HTTP_VER, 

      // HOST,
      CONTENT_LENGTH + " " + body.length()
    };

    return new PSMessage(header, body);
  }

  /**
   * Ensure the string argument is not null or zero length.
   * @throws IllegalArgumentException is the above is true.
   * Does nothing otherwise.
   */
  private void checkDomain(String domain)
  {
    if (domain == null || domain.length() == 0)
    {
      throw new IllegalArgumentException("Argument 'domain' cannot be null and must not be zero length");
    }
  }

  /**
   * Ensure the Hashtable argument is not null and contains at least one
   * key-value mapping.
   * @throws IllegalArgumentException is the above is true. Does nothing
   * otherwise.
   */
  private void checkFieldValuePairs(Hashtable fieldValuePairs)
  {
    if (fieldValuePairs == null || fieldValuePairs.size() == 0)
    {
      throw new IllegalArgumentException
        ("Argument 'fieldValuePairs' cannot be null and must contain data");
    }
  }

  /**
   * Add a text element - a xml element with only text as a child node.
   * @param parent add the new text element to this element.
   * @param name the name of the new element
   * @param text the actual text of the new element
   */
  private void addTextElement(Element parent, String name, String text)
  {
    Element element = new Element(name);

    element.setText(text);
    parent.addContent(element);
  }

  /**
   * Adds a field list.  A field list looks like: <pre>
   * &lt fieldList &gt
   * &lt listitem &gt 1000 &lt /listitem &gt
   * &lt listitem &gt smith &lt /listitem &gt
   * ...
   * &lt /fieldList &gt
   * </pre>
   * @param parent add the fieldList to this parent element.
   * @param fieldList the fieldList
   */
  private void addFieldList(Element parent, String[] fieldList)
  {
    Element fieldListElem = new Element(FIELD_LIST);

    if (fieldList != null)
    {
      for (int i = 0; i < fieldList.length; i++)
      {
        fieldListElem.addContent(new Element(LIST_ITEM).setText(fieldList[i]));
      }
    }

    parent.addContent(fieldListElem);
  }

  /**
   * Adds a field pair list.  A field list looks like: <pre>
   * &lt fieldValue &gt
   * &lt uid &gt 1000 &lt /uid &gt
   * &lt name &gt smith &lt /name &gt
   * ...
   * &lt /fieldValue &gt
   * </pre>
   * @param parent add the fieldList to this parent element.
   * @param fieldValuePairs the field value pairs
   */
  private void addFieldValuePairs(Element parent, Hashtable fieldValuePairs)
  {
    Element fieldValuePairsElem = new Element(FIELD_VALUE);
    Element fieldValueElem;
    Enumeration keys = fieldValuePairs.keys();
    Enumeration values = fieldValuePairs.elements();

    while (keys.hasMoreElements())
    {
      fieldValueElem = new Element((String) keys.nextElement());

      fieldValueElem.setText((String) values.nextElement());
      fieldValuePairsElem.addContent(fieldValueElem);
    }

    parent.addContent(fieldValuePairsElem);
  }

  /**
   * Create a string of xml starting with the supplied root element.
   * This method currently uses the jdom XMLOutputter, which
   * converts all XML content to entity notation. In order to get
   * XML content as XML, create a subclass of XMLOutputter and
   * override the method escapeElementEntities.
   */
  private String createXMLString(Element root)
  {
    XMLOutputter outputter = new XMLOutputterOverride("", true);
    String xmlStr = "";

    try
    {
      xmlStr = outputter.outputString(root) + "\r\n";
    }
    catch (IOException ioe)
    {
      ioe.printStackTrace();
    }

    return xmlStr;
  }

  /**
   * Create a String of the path to an Element, starting from the Root.
   * @param lastElement The Element that terminates the path.
   */
  private String createPathString(Element lastElement, Vector path)
  {
    if (path == null)
    {
      path = new Vector();
    }

    path.add(0, lastElement.toString());

    if (!lastElement.isRootElement())
    {
      return createPathString(lastElement.getParent(), path);
    }

    StringBuffer buf = new StringBuffer();

    for (int i = 0; i < path.size(); i++)
    {
      buf.append("--" + path.elementAt(i) + "\n");
    }

    return new String(buf);
  }

  /**
   * Parse a response message.
   * @param message the message to parse
   * @return true on HTTP_OK
   */
  public Boolean parseBooleanResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    {
      if (returnCode == HTTP_OK)
      {
        return new Boolean(true);
      }
      else
      {

        // not sure what else to do
        return new Boolean(false);
      }
    }
  }

  public String parseStringResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    if (returnCode == HTTP_NO_CONTENT)
    {

      // the response has no content.  Normal operation - used when no
      // result from a query, etc.
      return null;
    }

    if (returnCode == HTTP_OK)
    {
      try
      {
        Document doc = XMLProcessor.buildDocument(message.getBody());
        Element root = doc.getRootElement();

        if (root.getName().equals(RESPONSE))
        {
          List list = root.getChildren();
          Iterator iterator = list.iterator();
          Element firstElem = (Element) iterator.next();

          if (firstElem.getName().equals(TOKEN))
          {

            // success for SessionOpen
            return firstElem.getText();
          }
          else
          {
            throw new ProtocolException("Expecting 'domain' or 'token', found '" 
                    + firstElem.getName() + "'\n" 
                    + createPathString(firstElem, null));
          }
        }
        else
        {
          throw new ProtocolException("response element missing");
        }
      }
      catch (JDOMException e)
      {
        e.printStackTrace();

        throw new ProtocolException("Exception in response message handling: " 
                + e.getMessage());
      }
    }

    return null;
  }



  /**
   * Parse a response message.
   * @param message the message to parse
   * @return result count
   */
  public Integer parseIntegerResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    if (returnCode == HTTP_NO_CONTENT)
    {

      // the response has no content.  Normal operation - used when no
      // result from a query, etc.
      return null;
    }

    Integer returnValue = null;

    if (returnCode == HTTP_OK)
    {
      try
      {
        Document doc = XMLProcessor.buildDocument(message.getBody());
        Element root = doc.getRootElement();

        if (root.getName().equals(RESPONSE))
        {
          List list = root.getChildren();
          Iterator iterator = list.iterator();
          Element firstElem = (Element) iterator.next();

          if (firstElem.getName().equals(DOMAIN))
          {
            if (iterator.hasNext())
            {
              Element secondElem = (Element) iterator.next();

              if (secondElem.getName().equals(RESULTCOUNT))
              {

                // sucess for SetRecordsInDomain, DeleteRecordsInDomain,
                // SessionSearchDB, SessionSortDBResult, SessionSetDBResult,
                // SessionDeleteAll
                returnValue = new Integer(secondElem.getTextTrim());
              }
              else
              {

                throw new ProtocolException
                  ("Expecting 'resultcount' or 'match', found '" 
                        + secondElem.getName() + "'\n" 
                        + createPathString(secondElem, null));
              }
            }
            else
            {

              // no resultcount
              throw new ProtocolException("resultcount element missing");
            }
          }
          else if (firstElem.getName().equals(TOKEN))
          {

            // success for SessionOpen
            returnValue = new Integer(firstElem.getText());
          }
          else
          {
            throw new ProtocolException("Expecting 'domain' or 'token', found '" 
                    + firstElem.getName() + "'\n" 
                    + createPathString(firstElem, null));
          }
        }
        else
        {
          throw new ProtocolException("response element missing");
        }
      }
      catch (JDOMException e)
      {
        e.printStackTrace();

        throw new ProtocolException("Exception in response message handling: " 
                + e.getMessage());
      }
    }

    return returnValue;
  }

  /**
   * Parse an incoming response message
   * @return the XML string inside the result tags.
   */
  public String parseXMLResultResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    if (returnCode == HTTP_NO_CONTENT)
    {

      // the response has no content.  Normal operation - used when no
      // result from a query, etc.
      return null;
    }

    try
    {

      // we need to strip any declarations from the XML
      // before it can be parsed.
      String xmlString = message.getBody();

      // find a declaration that looks like <?xml version ...?>
      int declIdx = xmlString.indexOf("<?xml");
      int declEndIdx = -1;

      if (declIdx > -1)
      {
        declEndIdx = xmlString.indexOf(">", declIdx);

        if (declEndIdx == -1)
        {
          logCategory.error("Found XML declaration without end marker");

          return null;
        }
      }

      // find a declaration that looks like <!DOCTYPE ...>
      int docTypeIdx = xmlString.indexOf("<!DOCTYPE");
      int docTypeEndIdx = -1;

      if (docTypeIdx > -1)
      {
        docTypeEndIdx = xmlString.indexOf(">", docTypeIdx);

        if (docTypeEndIdx == -1)
        {
          logCategory.error("Found DOCTYPE declaration without end marker");

          return null;
        }
      }

      // remove declarations from the XML
      if ((declIdx > -1) || (docTypeIdx > -1))
      {
        StringBuffer buf = new StringBuffer(xmlString);
        int beginReplacement = 0;
        int endReplacement = 0;

        if (declIdx > -1)
        {
          beginReplacement = declIdx;
        }
        else
        {
          beginReplacement = docTypeIdx;
        }

        if (docTypeEndIdx > -1)
        {
          endReplacement = docTypeEndIdx;
        }
        else
        {
          endReplacement = declEndIdx;
        }

        buf.replace(beginReplacement, endReplacement + 1, "");

        xmlString = new String(buf);
      }

      Document doc = XMLProcessor.buildDocument(xmlString);
      Element root = doc.getRootElement();

      if (!root.getName().equals(RESPONSE))
      {
        logCategory.error("Error: response tag is " + root.getName() 
                          + " instead of " + RESPONSE);

        return null;
      }

      List list = root.getChildren();

      if (list.isEmpty())
      {

        // shouldn't happen
        logCategory.error("parseXMLResponse returned empty string");

        return null;
      }

      Element internalElement = (Element) list.remove(0);
      XMLOutputter outputter = new XMLOutputter();

      return outputter.outputString(internalElement);
    }
    catch (JDOMException e1)
    {
      e1.printStackTrace();

      throw new ProtocolException("Exception in response message handling: " 
              + e1.getMessage());
    }
    catch (IOException e2)
    {
      e2.printStackTrace();

      throw new ProtocolException("Exception in response message handling: " 
              + e2.getMessage());
    }
  }

  public String[] parseGetDescriptionList(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    if (returnCode == HTTP_NO_CONTENT)
    {

      // the response has no content.  Normal operation - used when no
      // result from a query, etc.
      return null;
    }

    String count = "0";

    try
    {
      Document doc = XMLProcessor.buildDocument(message.getBody());
      Element root = doc.getRootElement();

      if (!root.getName().equals(RESPONSE))
      {
        logCategory.error("Error: response tag is " + root.getName() 
                          + " instead of " + RESPONSE);

        return null;
      }

      count = root.getChildTextTrim(RESULTCOUNT);

      String[] result = new String[Integer.parseInt(count)];
      List list = root.getChildren(XMLDESCRIPTION);
      Iterator iter = list.iterator();
      int i = 0;

      while (iter.hasNext())
      {
        Element item = (Element) iter.next();

        if (i >= result.length)
        {
          logCategory.error("Error: number of children greater " 
                            + "than resultcount");

          return result;
        }

        result[i++] = item.getTextTrim();
      }

      return result;
    }
    catch (NumberFormatException e1)
    {
      throw new ProtocolException("bad resultcount: " + count);
    }
    catch (JDOMException e2)
    {
      throw new ProtocolException(e2.getMessage());
    }
  }


  /**
   * Parse a response message.
   * @param message the message to parse
   * @return the resultset from last query.
   */
  public PSDBResult parsePSDBResultResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {
    int returnCode = parseResponse(message);

    {
      if (returnCode == HTTP_NO_CONTENT)
      {

        // the response has no content.  Normal operation - used when no
        // result from a query, etc.
        return null;
      }

      try
      {
        Document doc = XMLProcessor.buildDocument(message.getBody());
        Element root = doc.getRootElement();

        if (root.getName().equals(RESPONSE))
        {
          List list = root.getChildren();
          Iterator iterator = list.iterator();
          Element firstElem = (Element) iterator.next();

          if (firstElem.getName().equals(DOMAIN))
          {
            if (iterator.hasNext())
            {
              Element secondElem = (Element) iterator.next();

              if (secondElem.getName().equals(RESULTCOUNT))
              {
                int resultCount = Integer.parseInt(secondElem.getTextTrim());

                if (resultCount < 1)
                {
                  return new PSDBResult(null, null);
                }
              }

              // if resultCount > 0 then there will be more elements
              Element thirdElem = (Element) iterator.next();

              if (thirdElem.getName().equals(MATCH))
              {

                // success (with matches) for
                // GetRecordsInDomain, SessionGetDBResult
                Vector matches = new Vector();
                Vector fieldNames = new Vector();
                Hashtable fieldValues;
                boolean firstTime = true;
                Element matchElem = thirdElem;

                // Element fieldValue;
                Element field;
                Element fieldValuePair;
                Iterator fieldIterator;

                while (matchElem != null)
                {

                  // fieldValue = matchElem.getChild(FIELDVALUE);
                  // if (fieldValue != null)
                  if (matchElem.getChild(FIELDVALUE) != null)
                  {
                    field = matchElem.getChild(FIELDVALUE);
                    fieldIterator = field.getChildren().iterator();
                    fieldValues = new Hashtable();

                    while (fieldIterator.hasNext())
                    {
                      fieldValuePair = (Element) fieldIterator.next();

                      // we need a special case for XML content such as table.
                      if (fieldValuePair.hasChildren())
                      {

                        // we know that the content is XML, not text.
                        try
                        {
                          List xmlChildren = fieldValuePair.getChildren();
                          StringBuffer buf = new StringBuffer();
                          XMLOutputter outputter = new XMLOutputter("", 
                                  false);

                          // there should be a single root element,
                          // but we won't make this assumption.
                          Iterator xmlIter = xmlChildren.iterator();

                          while (xmlIter.hasNext())
                          {
                            Element xmlElement = (Element) xmlIter.next();
                            String s = outputter.outputString(xmlElement);

                            buf.append(s);
                          }

                          String value = new String(buf);

                          fieldValues.put(fieldValuePair.getName(), value);

                        }
                        catch (IOException e)
                        {
                          e.printStackTrace();
                        }
                      }
                      else
                      {
                        fieldValues.put(fieldValuePair.getName(), 
                                        fieldValuePair.getText());
                      }

                      // if this is the first time through the outer loop do..
                      if (firstTime)
                      {
                        fieldNames.add(fieldValuePair.getName());
                      }
                    }

                    firstTime = false;

                    matches.add(fieldValues);

                    if (iterator.hasNext())
                    {
                      matchElem = (Element) iterator.next();

                      if (!matchElem.getName().equals(MATCH))
                      {
                        throw new ProtocolException("Expecting 'match', found '" 
                                + matchElem.getName() + "'\n" 
                                + createPathString(matchElem, null));
                      }
                    }
                    else
                    {
                      matchElem = null;
                    }
                  }

                  // for getFieldsInDomain
                  else if (matchElem.getChild(FIELDLIST) != null)
                  {
                    field = matchElem.getChild(FIELDLIST);
                    fieldIterator = field.getChildren().iterator();

                    while (fieldIterator.hasNext())
                    {
                      fieldValuePair = (Element) fieldIterator.next();

                      fieldNames.add(fieldValuePair.getText());
                    }

                    if (iterator.hasNext())
                    {

                      // something wrong here, but who cares
                    }
                    else
                    {
                      matchElem = null;
                    }
                  }
                  else
                  {
                    throw new ProtocolException("Expecting \"" + FIELDVALUE 
                            + "\" or \"" + FIELDLIST + "\", not found\n" 
                            + createPathString(matchElem, null));
                  }
                }

                return new PSDBResult((String[]) fieldNames.toArray
                  (new String[0]), 
                        (Hashtable[]) matches.toArray(new Hashtable[0]));
              }
              else
              {
                throw new ProtocolException("Expecting 'match', found '" 
                        + thirdElem.getName() + "'\n" 
                        + createPathString(thirdElem, null));
              }
            }
            else
            {

              // success (but not matches) for GetRecordsInDomain,
              // SessionGetDBResult
              return null;
            }
          }
          else
          {
            throw new ProtocolException("Expecting 'domain' found '" 
                    + firstElem.getName() + "'\n" 
                    + createPathString(firstElem, null));
          }
        }
        else
        {
          throw new ProtocolException("Expecting 'response', found '" 
                  + root.getName() + "'" + " at root");
        }
      }
      catch (JDOMException e)
      {
        e.printStackTrace();

        throw new ProtocolException("Exception in response message handling: " 
                + e.getMessage());
      }
    }
  }


  /**
   * Parse a response message.
   * @return the HTTP result code of the response.
   */
  private int parseResponse(PSMessage message) 
    throws ProtocolException, OperationException
  {

    String[] header = message.getHeader();
    String body = message.getBody();

    int returnCode;
    String returnStatement;
    StringTokenizer tokenizer;

    if (header == null || header.length == 0)
    {
      throw new ProtocolException("No response header!");
    }
    else
    {

      // there is at least one header line, parse the first line of the header
      try
      {
        tokenizer = new StringTokenizer(header[0]);

        returnCode = 0;

        while (tokenizer.hasMoreTokens())
        {
          String token = tokenizer.nextToken();

          // find the first token that does not start with "HTTP"
          // and set the returnCode to that value
          if (token.startsWith("HTTP"))
          {
            continue;
          }

          returnCode = Integer.parseInt(token);

          break;
        }

        if (returnCode == 0)
        {
          throw new ProtocolException("Header format problem: " + header[0]);
        }

        returnStatement = tokenizer.nextToken();
      }
      catch (NumberFormatException e)
      {
        logCategory.error("NumberFormatException: " + e.getMessage());

        throw new ProtocolException("Header format problem: " + header[0]);
      }

      if (returnCode == HTTP_FAIL)
      {
        try
        {
          Document doc = XMLProcessor.buildDocument(body);
          Element root = doc.getRootElement();

          // throw an OperationException.  This is normal operation on the 
          // client side but something has gone wrong on the server side
          throw new OperationException("Response error: " + root.getText());
        }
        catch (JDOMException e)
        {
          throw new ProtocolException(e.getMessage());
        }
      }
    }

    if ((returnCode == HTTP_OK) || (returnCode == HTTP_FAIL) 
        || (returnCode == HTTP_NO_CONTENT))
    {
      return returnCode;
    }
    else
    {
      throw new ProtocolException("Expecting HTTP_OK, HTTP_FAIL"
        + " or HTTP_NO_CONTENT, " 
              + "Found '" + returnCode + "'");
    }
  }

  public static void main(String[] args)
  {
    HttpXmlProtocolService ps = HttpXmlProtocolService.getInstance();
    Hashtable hash = new Hashtable();

    hash.put("uid", "1004");
    hash.put("fname", "bob");
    hash.put("lname", "twiggles");

    if (logCategory.isDebugEnabled())
    {
      printPSMessage(ps.buildSetRecordsInDomainRequest("domainSet", 
              "uid=1000*", hash));
    }
  }

  private static void printPSMessage(PSMessage mesg)
  {
    for (int i = 0; i < mesg.getHeader().length; i++)
    {
      logCategory.debug(mesg.getHeader()[i]);
    }

    logCategory.debug("");
    logCategory.debug(mesg.getBody());
  }

}



