/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.psLib;

import java.net.ProtocolException;

/**
 * Offers message building and parsing services.  It is an abstract class as
 * the protocol specific implementation is left up to subclasses.
 *
 * @author Matt Naish
 */
public abstract class PSProtocolService
{

  public static final int GET_RAW_FILE = 1;
  public static final int GET_LIST_XML_DESCRIPTIONS = 2;
  public static final int GET_XML_DESCRIPTION = 3;
  public static final int GET_RECORDS_IN_DOMAIN = 4;
  public static final int GET_FIELDS_IN_DOMAIN = 5;
  public static final int SET_RECORDS_IN_DOMAIN = 6;
  public static final int ADD_RECORDS_IN_DOMAIN = 7;
  public static final int DELETE_RECORDS_IN_DOMAIN = 8;
  public static final int SESSION_OPEN = 9;
  public static final int SESSION_CLOSE = 10;
  public static final int SESSION_SEARCH_DB = 11;
  public static final int SESSION_GET_DB_RESULT = 12;
  public static final int SESSION_SORT_DB_RESULT = 13;
  public static final int SESSION_SET_DB_RESULT = 14;
  public static final int SESSION_DELETE_ALL = 15;

  /**
   * Build a one arguement outgoing message.
   * Internally calls buildRequest(type, arg0, null, null, null, null, null);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public PSMessage buildRequest(int type, Object arg0)
  {
    return buildRequest(type, arg0, null, null, null, null, null);
  }

  /**
   * Build a two arguement outgoing message
   * Internally calls buildRequest(type, arg0, arg1, null, null, null, null);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public PSMessage buildRequest(int type, Object arg0, Object arg1)
  {
    return buildRequest(type, arg0, arg1, null, null, null, null);
  }

  /**
   * Build a three arguement outgoing message
   * Internally calls buildRequest(type, arg0, arg1, arg2, null, null, null);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public PSMessage buildRequest(int type, Object arg0, Object arg1, 
          Object arg2)
  {
    return buildRequest(type, arg0, arg1, arg2, null, null, null);
  }

  /**
   * Build a four arguement outgoing message
   * Internally calls buildRequest(type, arg0, arg1, arg2, arg3, null, null);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public PSMessage buildRequest(int type, Object arg0, Object arg1, 
          Object arg2, Object arg3)
  {
    return buildRequest(type, arg0, arg1, arg2, arg3, null, null);
  }

  /**
   * Build a five arguement outgoing message
   * Internally calls buildRequest(type, arg0, arg1, arg2, arg3, arg4, null);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public PSMessage buildRequest(int type, Object arg0, Object arg1, 
          Object arg2, Object arg3, Object arg4)
  {
    return buildRequest(type, arg0, arg1, arg2, arg3, arg4, null);
  }

  /**
   * Build a six argument outgoing message
   * Internally calls buildRequest(type, arg0, arg1, arg2, arg3, arg4, arg5);
   * @param type the the type of message
   * (eg. one of GET_RECORDS_IN_DOMAIN, SET_RECORDS_IN_DOMAIN, etc.)
   */
  public abstract PSMessage buildRequest(int type, Object arg0, Object arg1, 
          Object arg2, Object arg3, Object arg4, Object arg5);

  /**
   * Parse an incoming response message
   */
  public abstract Boolean parseBooleanResponse(PSMessage response) 
    throws ProtocolException, OperationException;

  /**
   * Parse an incoming response message
   */
  public abstract Integer parseIntegerResponse(PSMessage response) 
    throws ProtocolException, OperationException;

  public abstract String parseStringResponse(PSMessage message) 
    throws ProtocolException, OperationException;

  /**
   * Parse an incoming response message
   */
  public abstract PSDBResult parsePSDBResultResponse(PSMessage response) 
    throws ProtocolException, OperationException;

  /**
   * Parse an incoming response message
   * @return the XML string inside the result tags.
   */
  public abstract String parseXMLResultResponse(PSMessage response) 
    throws ProtocolException, OperationException;

  /**
   * Parse an incoming response message.
   * @return the list of xmldesc elements as a string array.
   */
  public abstract String[] parseGetDescriptionList(PSMessage response) 
    throws ProtocolException, OperationException;

}

