/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.psLib;

import java.util.Hashtable;
import java.io.IOException;

/**
 * <P>This interface defines the API for a client application to
 * communicate with the Provisioning Server.</P>
 * <P>Commands are separated into two groups: Transaction and Session. <UL>
 * <LI>Transaction - used to do a single database lookup or manipulation</LI>
 * <LI>Session - used for more complicated database manipulation that may
 * involve multiple query transactions such as search and sort. A session
 * is identified by a server generated token.</LI>
 * </UL> </P> <P>Other terms used in ProvisioningInterface are: <UL>
 * <LI>Domain -   The type or category of information. Examples would be users,
 * a particular type of server, etc. Analogous to the idea of a table in
 * SQL.</LI>
 * <LI>Record -   A particular entry in the database and that data associated
 * with that -- a user and all the information about a user. Analogous to a
 * row in SQL.</LI>
 * <LI>Field -    A particular piece of data, of which one exists for each
 * record. An example would be a phone number -- there would be one
 * associated with each user. Analogous to a column in
 * SQL.</LI> <LI>Value -     The contents of a particular field
 * -- a particular phone number in the case described above.</LI>
 * <LI>Filter -    A filter using the grammar defined in ???.</LI>
 * <LI>FieldList - A list of fields, usually specifying which fields one
 * wants to work with</LI> </UL> </P>
 */

/**
 * @author Matt Naish
 */
public abstract class ProvisioningInterface
{

  // Remove this when testing is complete
  public static boolean templateTesting = false;

  /**
   * Sort a query by ascending values.
   */
  public static final String SORT_TYPE_ASC = "ascend";

  /**
   * Sort a query by descending values.
   */
  public static final String SORT_TYPE_DESC = "descend";

  /**
   * The reference to the instance of this class
   */
  private static ProvisioningInterface instance = null;

  /**
   * Returns an instance of this class.  All calls to this method receive the same object
   */
  public static ProvisioningInterface getInstance()
  {
    if (instance == null)
    {
      instance = new SocketProvisioningInterface();
    }

    return instance;
  }


  /**
   * Close the single connection to the data source or server.
   * @throws IOException if the close operation fails
   */
  public abstract void close() throws IOException;


  /**
   * Send request to the PServer to authenticate the given account id and
   * password.
   * @param accountId the user name
   * @param password the password for that account
   * @returns true if the authentication succeeded
   */
  public abstract boolean login(String accountId, String password) throws
    IOException, OperationException;

  /**
   * Send request to the PServer to close an authenticated session. <p>
   */
  public abstract boolean logout() throws IOException, OperationException;

  /**
   * Send a request to PServer to retrieve all the matching records
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon.  It is a string of boolean logic (uid=100 & name!='John')
   * @param fieldList used to identify which fields to return for each record
   */
  public abstract PSDBResult getRecordsInDomain(String domain, String filter, 
          String[] fieldList) throws IOException, OperationException;

  /**
   * Get the field names within a domain (corresponding to the column names
   * of, say, a SQL record).
   */
  public abstract String[] getFieldsInDomain(String domain) 
    throws IOException, OperationException;

  /**
   * Get an XML data description. This describes both a table in the database
   * and a GUI for entering data into that table.
   * @return an XML string describing the requested data.
   */
  public abstract String getXMLFile(String category, 
          String filename) throws IOException, OperationException;

  /**
   * Get a text file "as is" from the www directory.
   */
  public abstract String getRawFile(String filename) 
    throws IOException, OperationException;

  /**
   * Get the list of all XML data descriptions of a particular category.
   */
  public abstract String[] getListXMLDescriptions(String category) 
    throws IOException, OperationException;

  /**
   * Given a list of fields within a domain, for each matching field,
   * replace its value with that in the fieldValuePairs.
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon.  It is a string of boolean logic (uid=1000 & name!='John')
   * @param fieldValuePairs a list of key (eg uid) values (eg 1000)
   */
  public abstract boolean setRecordsInDomain(String domain, String filter, 
          Hashtable fieldValuePairs) throws IOException, OperationException;

  /**
   * Add a new record to a domain.
   * @param domain the domain to act upon
   * @param fieldValuePairs a list of key (eg uid) values (eg 1000)
   */
  public abstract boolean addRecordInDomain(String domain, 
          Hashtable fieldValueList) throws IOException, OperationException;

  /**
   * Remove all records within a domain that match the filter
   * @param domain the domain to act upon
   * @param filter identifies which records within the supplied domain to act
   * upon.   It is a string of boolean logic (uid=100 & name!='John')
   */
  public abstract boolean deleteRecordsInDomain(String domain, 
          String filter) throws IOException, OperationException;

  /**
   * Instruct PServer that the client is interested in monitoring and
   * data change wihtin a domain.  ProvisionigInterface listens for update
   * notification and updates the client by invoking the call back function.
   * <B>NOT IMPLEMENTED</B>
   */
  public abstract boolean registerDomainForUpdate(String domain, 
          Object huh) throws IOException, OperationException;

  /**
   * Similar to registerDomainForUpdate, except the client is interested in
   * monitoring a specific filed within a domain. <B>NOT IMPLEMENTED</B>
   */
  public abstract boolean registerRecordForUpdate(String domain, 
          Object huh) throws IOException, OperationException;

  /**
   * Similar to getRecordsInDomain(), except it caches a snapshot of the data
   * locally for the client and automatically monitors and updates the data.
   * The first invocation of this fucntion sends a request to PServer to look
   * up the value(s) associated with the field.  Subsequent calls to this
   * function will result in accessing the resultatnt data locally instead
   * of doing a "true" get via PServer.
   */
  public abstract PSDBResult getCurrentRecordsInDomain(String domain, 
          String filter, 
          String[] fieldList) throws IOException, OperationException;

  /**
   * Open a query session with Provisioning Server which returns a unique
   * indentifier that to which subsequent request can refer.
   * @param timeout how long in milliseconds the server should leave this
   * session open.  After the time has elapsed the server automatically
   * invalidates the session.
   */
  public abstract String sessionOpen(long timeout) 
    throws IOException, OperationException;

  /**
   * Terminate a query session.
   * @param token the query session indentifier.
   */
  public abstract boolean sessionClose(String token) 
    throws IOException, OperationException;

  /**
   * Query PServer to lookup all records satisfying the supplied constraints
   * within a domain.  PServer returns the number of matching records.
   * @param token the query session indentifier
   * @param filter identifies which records within the supplied domain to act
   * upon.   It is a string of boolean logic (uid=100 & name!='John')
   * @param fieldList used to identify which fields to return for each record
   * @param sortType one of SORT_TYPE_ASC | SORT_TYPE_DESC
   * @param sortField the field to on which to sort.
   */
  public abstract int sessionSearchDB(String token, String domain, 
          String filter, String[] fieldList, String sortType, 
          String sortField) throws IOException, OperationException;

  /**
   * Query PServer to retrieve a range of records based on the result of
   * the previous sessionSearchDB()
   * @param token the query session indentifier
   * @param start the start number of the range to retrieve.  Inclusive.
   * @param end the end number of the range to retrieve.  Inclusive.
   */
  public abstract PSDBResult sessionGetDBResult(String token, int start, 
          int end, String[] fieldList) throws IOException, OperationException;

  /**
   * Query PServer to sort the result from a previous call to sessionSearchDB()
   * @param token the query session indentifier
   * @param sortType one of SORT_TYPE_ASC | SORT_TYPE_DESC
   * @param sortField the field to on which to sort.
   */
  public abstract boolean sessionSortDBResult(String token, String sortType, 
          String sortField) throws IOException, OperationException;

  /**
   * Instruct PServer to update all the records from a previous call to
   * sessionSearchDB() with the supplied fieldValuePairs
   * @param token the query session indentifier
   * @param fieldValuePairs a list of key (eg uid) values (eg 1000)
   */
  public abstract boolean sessionSetDBResult(String token, 
          Hashtable fieldValuePairs) throws IOException, OperationException;

  /**
   * Remove all the records found by a previous call to sessionSearchDB()
   * @param token the query session indentifier
   */
  public abstract boolean sessionDeleteAll(String token) 
    throws IOException, OperationException;
}



