/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.psLib;

import java.util.Hashtable;
import java.io.IOException;

import org.apache.log4j.Category;

/**
 * An implementation of the ProvisioningInterface interface.  Right now
 * SocketProvisioningInterface uses HttpXMLProtocolService to generate and
 * parse the HTTP/XML messages and HttpPSConnection to send and receive
 * the messages.
 *
 * @author Matt Naish
 */
public class SocketProvisioningInterface extends ProvisioningInterface
{

  /**
   * The hostname of the socket connection.
   */
  private static String myHostname = "localhost";

  /**
   * The port of the socket connection.
   */
  private static int myPort = 80;

  /**
   * Does the actual sending and receiving of the messages
   * @supplierCardinality 1
   */
  private PSInterface psInterface;

  /**
   * Does message generation and parsing
   * @supplierCardinality 1
   */
  private PSProtocolService psProtocolService;

  /**
   * @supplierCardinality 1
   */
  private PSListener psListener;

  /**
   * @supplierCardinality 0..1
   */
  private PSDataObservable PSDataObservable;

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(SocketProvisioningInterface.class.getName());

  /**
   * Private default constructor
   */
  protected SocketProvisioningInterface()
  {
    psInterface = new PSInterface(myHostname, myPort);
    psProtocolService = HttpXmlProtocolService.getInstance();
  }

  public void close() throws IOException
  {
    myHostname = null;
    myPort = -1;

    psInterface.close();
    psInterface = null;

    psProtocolService = null;
  }

  /**
   * Sets up hostname and port. This must be called before the constructor.
   */
  public static void initialize(String hostname, int port)
  {
    logCategory.info("Connection initialized with host " + hostname 
                     + " and port " + port);

    myHostname = hostname;
    myPort = port;
  }

  public String getRawFile(String filename) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_RAW_FILE, 
            filename);
    PSMessage response = psInterface.sendRequest(request);
    String result = response.getBody();

    return result;
  }

  public String[] getListXMLDescriptions(String category) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest
        (PSProtocolService.GET_LIST_XML_DESCRIPTIONS, category);
    PSMessage response = psInterface.sendRequest(request);
    String[] result = psProtocolService.parseGetDescriptionList(response);

    return result;
  }

  public String getXMLFile(String category, 
          String filename) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_XML_DESCRIPTION, 
            category, filename);
    PSMessage response = psInterface.sendRequest(request);
    String result = psProtocolService.parseXMLResultResponse(response);

    return result;
  }

  public PSDBResult getRecordsInDomain(String domain, String filter, 
          String[] fieldList) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_RECORDS_IN_DOMAIN, 
            domain, filter, fieldList);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);

    return result;
  }

  public String[] getFieldsInDomain(String domain) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.GET_FIELDS_IN_DOMAIN, 
            domain);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);
    String[] fieldNames = result.getFieldNames();

    if (logCategory.isDebugEnabled())
    {
      logCategory.debug("getFieldsInDomain returning field names: ");

      for (int i = 0; i < fieldNames.length; i++)
      {
        logCategory.debug("  " + fieldNames[i]);
      }
    }

    return fieldNames;
  }

  public boolean setRecordsInDomain(String domain, String filter, 
          Hashtable fieldValuePairs) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SET_RECORDS_IN_DOMAIN, 
            domain, filter, fieldValuePairs);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean addRecordInDomain(String domain, 
          Hashtable fieldValuePairs) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.ADD_RECORDS_IN_DOMAIN, 
            domain, fieldValuePairs);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }


  public boolean login(String accountId, String password) throws
    IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean logout() throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean registerDomainForUpdate(String domain, 
          Object huh) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public boolean registerRecordForUpdate(String domain, 
          Object huh) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public PSDBResult getCurrentRecordsInDomain(String domain, String filter, 
          String[] fieldList) throws IOException, OperationException
  {
    throw new IllegalStateException("Not Implemented yet.");
  }

  public String sessionOpen(long timeout) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_OPEN, 
            new Long(timeout));
    PSMessage response = psInterface.sendRequest(request);

    return psProtocolService.parseStringResponse(response);

  }

  public boolean deleteRecordsInDomain(String domain, 
          String filter) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.DELETE_RECORDS_IN_DOMAIN, 
            domain, filter);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean sessionClose(String token) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_CLOSE, token);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public int sessionSearchDB(String token, String domain, String filter, 
          String[] fieldList, String sortType, 
          String sortField) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SEARCH_DB, 
            token, domain, filter, fieldList, sortType, sortField);
    PSMessage response = psInterface.sendRequest(request);
    Integer result = psProtocolService.parseIntegerResponse(response);

    return result.intValue();
  }

  public PSDBResult sessionGetDBResult(String token, int start, int end, 
          String[] fieldList) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_GET_DB_RESULT, 
            token, new Integer(start), new Integer(end), fieldList);
    PSMessage response = psInterface.sendRequest(request);
    PSDBResult result = psProtocolService.parsePSDBResultResponse(response);

    return result;
  }

  public boolean sessionSortDBResult(String token, String sortType, 
          String sortField) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SORT_DB_RESULT, 
            token, sortType, sortField);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  public boolean sessionSetDBResult(String token, 
          Hashtable fieldValuePairs) throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_SET_DB_RESULT, 
            token, fieldValuePairs);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }


  public boolean sessionDeleteAll(String token) 
    throws IOException, OperationException
  {
    PSMessage request = 
      psProtocolService.buildRequest(PSProtocolService.SESSION_DELETE_ALL, 
            token);
    PSMessage response = psInterface.sendRequest(request);
    Boolean result = psProtocolService.parseBooleanResponse(response);

    return result.booleanValue();
  }

  private static void printPSDBResult(PSDBResult result)
  {
    String[] fieldNames = result.getFieldNames();

    for (int i = 0; i < result.getNumRecords(); i++)
    {
      for (int j = 0; j < fieldNames.length; j++)
      {
        logCategory.debug(fieldNames[j] + "=" 
                          + result.getFieldValue(i, fieldNames[j]) + ", ");
      }

      logCategory.debug("");
    }
  }

}


