/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.server;

import java.util.Vector;
import java.util.List;
import java.util.ListIterator;
import java.util.Hashtable;
import java.io.InputStream;
import java.io.IOException;
import java.net.URL;
import java.awt.BorderLayout;
import java.awt.FlowLayout;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.Container;
import javax.swing.JFrame;
import javax.swing.JDialog;
import javax.swing.JPanel;
import javax.swing.JButton;
import javax.swing.JScrollPane;
import javax.swing.JOptionPane;
import javax.swing.JTree;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.TreeSelectionModel;
import javax.swing.tree.DefaultTreeSelectionModel;
import javax.swing.tree.TreeNode;
import javax.swing.tree.TreePath;

import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;

import org.jdom.Document;
import org.jdom.Element;
import org.jdom.JDOMException;
import org.jdom.input.SAXBuilder;
import org.jdom.output.XMLOutputter;

import org.apache.log4j.Category;

import org.vovida.prov.dynamicGui.DTDConstants;
import org.vovida.prov.dynamicGui.TypeBuilder;
import org.vovida.prov.dynamicGui.ProvisionableType;
import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.content.InvalidDataException;
import org.vovida.prov.navigation.ConfigConstants;
import org.vovida.prov.navigation.NavBarClient;
import org.vovida.prov.navigation.TopLevel;
import org.vovida.prov.server.TableDialog;
import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;
import org.vovida.prov.util.XMLProcessor;

/**
 * The main panel for Server provisioning. The center of the panel
 * contains a JTree of available server types. In order to be listed
 * anywhere but the top level in the JTree, a server type must have a
 * description file that lists "Server" as its top-level Category.
 *
 * @author Barbara Samson
 */
public class ServerClient extends NavBarClient implements DTDConstants
{
  private JTree myTree;
  private Document myDocumentModel;
  private Element myRoot;
  private DefaultMutableTreeNode myTreeTop;

  /**
   * Used to associate TreeNodes with JdomNodes.
   */
  private Hashtable myNodeHashtable;

  /**
   * Displays a table of servers of the currently selected type.
   * @link aggregationByValue
   */
  private TableDialog myTableDialog;

  /**
   * Used as a parent for EditServerDialogs.
   */
  private JFrame myParent;

  /**
   * for logging
   */
  private static Category logCategory = 
    Category.getInstance(ServerClient.class.getName());

  public ServerClient(String name, TopLevel parent)
  {
    super(name, parent);

    buildTree();

    myTableDialog = new TableDialog(parent);
    myParent = parent;
  }


  private void buildTree()
  {
    if (myDocumentModel != null)
    {
      return;
    }

    ProvisioningInterface ps = ProvisioningInterface.getInstance();

    myRoot = new Element("Servers");
    myDocumentModel = new Document(myRoot);

    Document listOfServers = null;
    String serverFile = "";

    try
    {
      String[] allServers = null;

      try
      {
        allServers = ps.getListXMLDescriptions(ConfigConstants.SERVERS);
      }
      catch (Exception e1)
      {
        JOptionPane.showMessageDialog(this, 
                "error in list of XML Descriptions" + ": " + e1.getMessage(), 
                e1.getClass().toString(), JOptionPane.ERROR_MESSAGE);

        return;
      }

      if (allServers == null)
      {
        return;
      }

      for (int i = 0; i < allServers.length; i++)
      {
        serverFile = allServers[i];

        String xmlString = null;

        try
        {
          xmlString = ps.getXMLFile(ConfigConstants.SERVERS, serverFile);
        }
        catch (Exception e2)
        {
          JOptionPane.showMessageDialog(this, 
                  "error in " + serverFile + ": " + e2.getMessage(), 
                  e2.getClass().toString(), JOptionPane.ERROR_MESSAGE);

          continue;
        }

        XMLProcessor.doSAXParse(xmlString, new BuildTreeHandler(xmlString));
      }
    }
    catch (SAXException e3)
    {
      e3.printStackTrace();
    }

    if (logCategory.isInfoEnabled())
    {
      XMLOutputter outputter = new XMLOutputter("  ", true);

      try
      {
        logCategory.info(outputter.outputString(myDocumentModel));
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }

    renderScreen();
  }

  private void renderScreen()
  {

    // set up bottom buttons
    JButton showButton = new JButton("Show Servers");

    showButton.setToolTipText("Show servers of this type");
    showButton.addActionListener(new ShowServersListener());

    JPanel buttonPanel = new JPanel();

    buttonPanel.setLayout(new FlowLayout(FlowLayout.LEFT));

    buttonPanel.add(showButton);
    add(buttonPanel, BorderLayout.SOUTH);

    myNodeHashtable = new Hashtable();

    JScrollPane scrollPane = new JScrollPane();

    add(scrollPane, BorderLayout.CENTER);

    myTreeTop = new DefaultMutableTreeNode("All Servers");

    DefaultMutableTreeNode treeNodeParent = myTreeTop;

    myNodeHashtable.put(myTreeTop, myRoot);

    myTree = new JTree(myTreeTop);

    iterateJdom(myRoot, myTreeTop);

    TreeSelectionModel selectionModel = new DefaultTreeSelectionModel();

    selectionModel.setSelectionMode
      (TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
    myTree.setSelectionModel(selectionModel);
    myTree.setShowsRootHandles(true);

    for (int i = 0; i < myTree.getRowCount(); i++)
    {
      myTree.expandRow(i);
    }

    scrollPane.getViewport().add(myTree);
  }

  private void iterateJdom(Element jdomParent, 
          DefaultMutableTreeNode treeNodeParent)
  {
    List jdomChildren = jdomParent.getChildren();
    ListIterator iter = jdomChildren.listIterator();

    while (iter.hasNext())
    {
      Element jdomChild = (Element) iter.next();

      if ((jdomChild.getName().equals(BuildTreeHandler.CATEGORY)) 
          || (jdomChild.getName().equals(BuildTreeHandler.TYPE)))
      {
        String categoryName = jdomChild.getAttributeValue(BuildTreeHandler.NAME);
        DefaultMutableTreeNode treeNodeChild = 
          new DefaultMutableTreeNode(categoryName);

        treeNodeParent.add(treeNodeChild);
        myNodeHashtable.put(treeNodeChild, jdomChild);
        iterateJdom(jdomChild, treeNodeChild);
      }
    }
  }

  /**
   * Get the JDOM Elements that correspond to the currently selected nodes
   * in myTree.
   */
  protected Element[] getSelectedElements()
  {
    TreePath[] paths = myTree.getSelectionPaths();

    if ((paths == null) || (paths.length == 0))
    {
      return new Element[]
      {
        myRoot
      };
    }

    Vector elements = new Vector();

    for (int i = 0; i < paths.length; i++)
    {
      Object key = paths[i].getLastPathComponent();
      Object node = myNodeHashtable.get(key);

      elements.add(node);
    }

    return (Element[]) elements.toArray(new Element[elements.size()]);
  }

  /**
   * Traverse the documentModel from the currently selected node on down
   * and find all the server types in that subtree.
   */
  protected String[] getSelectedTypes()
  {
    Vector selectedTypes = new Vector();
    Element[] selectedElements = getSelectedElements();

    for (int i = 0; i < selectedElements.length; i++)
    {
      traverseDocumentModel(selectedElements[i], selectedTypes);
    }

    String[] typeArray = 
      (String[]) selectedTypes.toArray(new String[selectedTypes.size()]);

    return typeArray;
  }

  private void traverseDocumentModel(Element parent, Vector vector)
  {
    if (parent.getName().equals(BuildTreeHandler.TYPE))
    {
      String name = parent.getAttributeValue(BuildTreeHandler.NAME);

      if ((name == null) || (name.equals("")))
      {
        logCategory.error("parent element has no name: " + parent);

        return;
      }

      vector.addElement(name);
    }

    List children = parent.getChildren();
    ListIterator iter = children.listIterator();

    while (iter.hasNext())
    {
      Element child = (Element) iter.next();

      traverseDocumentModel(child, vector);
    }

  }

  /**
   * The ActionListener for the Show button.
   */
  protected class ShowServersListener implements ActionListener
  {

    /**
     * Make the ServerTable build a table of servers of all selected types.
     */
    public void actionPerformed(ActionEvent e)
    {
      String[] selectedTypes = getSelectedTypes();

      myTableDialog.buildTable(selectedTypes);
      myTableDialog.showWindow();
    }

  }


  /**
   * This SAX handler builds up a vector of Categories, then finds a place
   * for them in myTree.
   */
  protected class BuildTreeHandler extends DefaultHandler
  {

    static final String CATEGORY = "category";
    static final String TYPE = "type";
    static final String NAME = "name";
    static final String SERVER = "Server";

    private String myXML;
    private String myName = "";
    private Vector myCategories = new Vector();

    public BuildTreeHandler(String xmlString)
    {
      super();

      myXML = xmlString;
    }

    public void startElement(String ignore1, String ignore2, String rawName, 
            Attributes attrs)
    {
      if (rawName.equals(DTDConstants.DATADEFN))
      {
        myName = attrs.getValue(DTDConstants.NAME);
      }
      else if (rawName.equals(DTDConstants.CATEGORY))
      {
        myCategories.addElement(attrs.getValue(DTDConstants.NAME));
      }
    }

    public void endDocument()
    {
      if (myCategories.isEmpty())
      {
        myCategories.addElement(BuildTreeHandler.SERVER);
      }

      String firstCategory = (String) myCategories.firstElement();

      if (!firstCategory.equalsIgnoreCase(BuildTreeHandler.SERVER))
      {
        myCategories.insertElementAt(BuildTreeHandler.SERVER, 0);
      }

      Element parent = myRoot;

      /*
       * Take the vector of categories and build them into myDocumentModel.
       * If a node for this category has already been created, add to it.
       * Otherwise create the node.
       */

      for (int i = 1; i < myCategories.size(); i++)
      {
        String category = (String) myCategories.elementAt(i);
        List children = parent.getChildren();
        ListIterator iter = children.listIterator();
        Element childOfThisCategory = null;

        while (iter.hasNext())
        {
          Element child = (Element) iter.next();

          if (!child.getName().equals(BuildTreeHandler.CATEGORY))
          {
            continue;
          }

          String childCategory = child.getAttributeValue(BuildTreeHandler.NAME);

          if ((childCategory != null) && (childCategory.equals(category)))
          {
            childOfThisCategory = child;

            break;
          }
        }

        if (childOfThisCategory == null)
        {

          // a matching child was not found, so create one.
          childOfThisCategory = new Element(BuildTreeHandler.CATEGORY);

          childOfThisCategory.addAttribute(BuildTreeHandler.NAME, category);
          children.add(childOfThisCategory);
          parent.setChildren(children);
        }

        parent = childOfThisCategory;
      }

      Element typeElement = new Element(BuildTreeHandler.TYPE);

      typeElement.addAttribute(BuildTreeHandler.NAME, myName);

      // ProvisionableInstances will be created as necessary.
      // Ensure the ProvisionableType is there in case we need it.
      ProvisionableType type = TypeBuilder.getType(myName, myXML);
      String[] sArray = type.getDBFields();

      for (int i = 0; i < sArray.length; i++)
      {
        logCategory.debug(myName + " type: " + sArray[i]);
      }

      List types = parent.getChildren();

      types.add(typeElement);
      parent.setChildren(types);
    }

  }



}




