package org.vovida.prov.table;

import javax.swing.JTable;
import javax.swing.JMenu;
import javax.swing.JCheckBoxMenuItem;
import javax.swing.table.TableColumn;
import javax.swing.event.TableModelEvent;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;

import org.apache.log4j.Category;


/**
 * An extension of the JTable which also creates a JMenu which allows columns
 * to be added and removed from the table at run time. <p>
 * Note that the menu is not automatically displayed. It is up to the calling
 * class to retrieve it using the <code>getMenu()</code> method and displaying
 * as appropriate.
 */
public class RemovableColumnTable extends JTable
{
    private JMenu menu = new JMenu("Select columns");

    private static Category logCategory = Category.getInstance(RemovableColumnTable.class.getName());


    /**
     * Creates a checkboxmenuitem for the given column and adds it to the menu.
     */
    private void addMenuItem(TableColumn column)
    {
        String columnName = dataModel.getColumnName(column.getModelIndex());
        JCheckBoxMenuItem item = new JCheckBoxMenuItem(columnName);
        item.setSelected(true); //assume all columns are visible at the start
        item.addActionListener(new ColumnAdder(column));
        menu.add(item);
    }

    /**
     * An instance of this class should be generated for each column added to
     * the table. It is attached as the action listener to the menu item which
     * represents the associated column and is responsible for adding and
     * removing it to/from the table when the menu item is selected.
     */
    private class ColumnAdder implements ActionListener
    {
        private TableColumn column;

        public ColumnAdder (TableColumn col)
        {
            column = col;
        }

        /**
         * Determine whether the menu item is checked or unchecked and add
         * the column to the table at the appropriate location.
         */
        public void actionPerformed(ActionEvent event)
        {
            JCheckBoxMenuItem item = (JCheckBoxMenuItem)event.getSource();
            if (item.isSelected())
            {
                logCategory.info("adding column " + RemovableColumnTable.this.dataModel.getColumnName(column.getModelIndex()) + " " + column.getModelIndex());
                addColumn(column);

                //the column gets added at the end of the table. Now move it
                //to its position specified by its index in the model
                if (columnModel.getColumnCount()-1>column.getModelIndex())
                {
                    int index = 0;
                    for (int i=0; i<columnModel.getColumnCount(); i++)
                    {
                        //if the model index of the ith column is greater than
                        //the model index of this column, then this column needs
                        //to be placed before the ith column
                        if (columnModel.getColumn(i).getModelIndex() > column.getModelIndex())
                        {
                            index = i;
                            break;
                        }
                    }
                    //move the column into its correct position
                    moveColumn(columnModel.getColumnCount()-1,index);
                }
            }
            else
            {
                logCategory.info("removing column " + RemovableColumnTable.this.dataModel.getColumnName(column.getModelIndex()) + " " + column.getModelIndex());
                removeColumn(column);
            }
            tableChanged(new TableModelEvent(dataModel));
            repaint();
        }
    }

    /**
     * Determine if the menu already contains a menu item for the given column.
     */
    private boolean menuItemExists(TableColumn aColumn)
    {
        Component [] components = menu.getMenuComponents();
        for (int i=0; i<components.length; i++)
        {
            if (JCheckBoxMenuItem.class.isInstance(components[i])) //sometimes, it could be a JPopupMenu$Separator
            {
	            JCheckBoxMenuItem item = (JCheckBoxMenuItem)components[i];
	            if (item.getText().equals(dataModel.getColumnName(aColumn.getModelIndex())))
	            {
	                return true;
	            }
            }
        }
        return false;
    }

    /**
     * Add a column to the table and at the same time add a menu item for that
     * column to the menu which allows columns to be added and removed from the
     * table.
     * This method first check whether a menu item exists for a column with
     * the same column name as the one being added.
     */
    public void addColumn(TableColumn aColumn)
    {
        super.addColumn(aColumn);
        if (!menuItemExists(aColumn))
        {
            addMenuItem(aColumn);
        }
    }

    /**
     * Returns the menu which may be used to add or remove columns in the table
     * at run time. <p>
     * Could use this as a regular menu or use <code>JMenu.getPopupMenu()</code>
     */
    public JMenu getMenu()
    {
        return menu;
    }
}
