package org.vovida.prov.table;

import org.vovida.prov.util.Loop;
import org.vovida.prov.util.assert.*;

import java.util.Hashtable;
//import java.util.Vector;
import java.util.LinkedList;

import org.apache.log4j.Category;



/**
 * This class maps the data for the accounts to be displayed in the table from
 * the order they are displayed in to the order they are stored in. <p>
 * The order the data is stored in would be the whatever order it initally
 * arrives in from the server.
 * It is assumed that all sorting of data will be accomplished by the server.
 * In that case, the ordered set set of data will be retrieved from the server
 * as just the list of ids (phone numbers or whatever identifies an account).
 * This ordered list then needs to be mapped to the data for each account
 * to avoid having to reload it from the server when the table needs to be
 * redrawn to reflect the new order.
 */
public class RowModel
{
    private static Category logCategory = Category.getInstance(RowModel.class.getName());


    /**
     * Stores information about which account id is diplayed in a given row of
     * the table. <p>
     * The ruccent mapping of row to account id is stored at the top of the loop
     * and may be retrieved with <code>rowToId.getCurrent()</code>. Each time
     * the table is resorted, a new empty map is added to the top of the loop
     * while the previous ordering may be retrieved with <code>rowToId.back()
     * </code>. This way, a few previous result sets may be accessed without
     * having to re-load the full list of names and rebuild the map from the
     * provisioning server.
     */
    private Loop rowToId = null;

    /**
     * The maximum number of entries that may be added to the rowToId loop
     * without over writing any previous entries.
     */
    public static final int LOOP_SIZE = 4;


    /**
     * This map associates an account's data with its id. It is updated as
     * data is loaded from the server but is never re-mapped.
     */
    private Hashtable idToData;


    /**
     * The number of rows stored in the model.
     */
    private int rowCount;

    public RowModel()
    {
        rowToId = new Loop(LOOP_SIZE);
        Hashtable current = new Hashtable();
        rowToId.add(current);
        idToData = new Hashtable();
    }

    /**
     * Get the data set which corresponds to the given data id.<p>
	 * @return null if the given id does not map to any data set, a hashtable of
     * the data otherwise.
     */
    public Hashtable getDataForId(String id)
    {
        return (Hashtable) idToData.get(id);
    }


    /**
     * Return the current data set at the given row. <p>
     * This accounts for the current sort order of the data.
     * @return null if the row to account id mapping does not exist, a vector
     * containing the data for that row otherwise
     */
    public Hashtable getDataAt(int row)
    {
        String id = (String)((Hashtable)rowToId.getCurrent()).get(new Integer(row));

        //The account id at this row is not known (it has not been loaded or the
        //id to row map has been invalidated). In either case, return null.
        //(The loading of the account data will occur from the scrolling or
        //resize event handlers, not here).
        if (id == null)
        {
            return null;
        }

        return (Hashtable) idToData.get(id);
    }

    public void setDataAt(int row, Hashtable value)
    {
        String id = (String)((Hashtable)rowToId.getCurrent()).get(new Integer(row));
		Assert.assert("Row is mapped to an id", id != null);

        idToData.put(id,value);
    }


    /**
     * Shift all the entries in the row to id map with a row number higher than
     * the given row up by one (and remove the last entry in the map).<p>
     */
    private void shiftRowsUp(int row)
    {
		//all the other entries with a row number lower than the one removed
        //must have their row shifted up by one
        Hashtable currentRowToId = (Hashtable)rowToId.getCurrent();
        for (int i=row; i<rowCount-1; i++)
        {
    	    String id = (String)currentRowToId.get(new Integer(i+1));
            currentRowToId.put(new Integer(i),id);
        }

        //remove the last row
		currentRowToId.remove(new Integer(rowCount-1));
        rowCount = rowCount - 1;
    }

    /**
     * Remove all the data for the given row from both maps. The rows folowing
     * the given row in the rowToId map will be shifted up by one.<p>
     */
    public void removeDataAt(int row)
    {
        String id = (String)((Hashtable)rowToId.getCurrent()).get(new Integer(row));
		Assert.assert("Row " + row + " is mapped to an id", id != null);

        //remove the requested entry from both maps
        idToData.remove(id);
		((Hashtable)rowToId.getCurrent()).remove(new Integer(row));

        //shift up the rows after the removed one
        shiftRowsUp(row);
    }


    /**
     * Adds an account name to the model and maps it to the given row. <p>
     * A mapping for this row should not exist yet.
     */
    public void addAccountAt(String accountName, String columnName, int row)
    {
        Hashtable currentMap = (Hashtable)rowToId.getCurrent();

        if (AssertLevel.ASSERT)
        {
            Assert.assert("row number is <= the maximum", row <= rowCount);
        }

        //add the row and name to the map
        currentMap.put(new Integer(row),accountName);

        //if the id to data map does not contain data for this id,
        //create a data vector for the account name and add it to the map
        if (idToData.get(accountName) == null)
        {
	        Hashtable data = new Hashtable();
            data.put(columnName,accountName);
	        idToData.put(accountName,data);
        }
    }


    /**
     * Set the total number of rows the model will store.
     */
    public void setRowCount (int count)
    {
        rowCount = count;
    }

    /**
     * Return the total number of rows in the model. <p>
     */
    public int getRowCount()
    {
        return rowCount;
    }

    /**
     * Invalidates the map of account id to account data. <p>
     * This should be called in conjunction with <code>invalidateIdToDataMap</code>
     * whenever the user wants to reload data into the table without having
     * to re-instantiate it. <p>
     */
    public void invalidateIdToDataMap()
    {
        idToData = new Hashtable();
    }

    /**
     * Invalidates the map of row index to account id. <p>
     * This should be called every time the user selects to sort the table
     * based on some column. <p>
     * This will set the value in the map for all rows to null so that next time
     * there is a request for the account id at that row, it will go back to
     * the server to retrieve it.
     */
    public void invalidateRowToIdMap()
    {
        Hashtable currentMap = (Hashtable)rowToId.getCurrent();
        //if there has been some data loaded into the current map
        if (!currentMap.isEmpty())
        {
            logCategory.debug("Storing rowToId map");
            //add a new empty map to the top of the loop. The current map is
            //pushed down and  may be retrieved later with rowToId.back()
            rowToId.add(new Hashtable());
        }
    }

    public void back()
    {
        Object map = rowToId.back();
    }

    public void forward()
    {
        Object map = rowToId.forward();
    }
}
