package org.vovida.prov.table;

import org.vovida.prov.util.assert.Assert;
import org.vovida.prov.psLib.ProvisioningInterface;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import javax.swing.JTable;
import javax.swing.table.TableColumnModel;
import javax.swing.JOptionPane;

import org.apache.log4j.Category;



/**
 * This class should be registered as an event hander on the header of the
 * JTable which displays accounts. It is responsible for detecting when the user
 * clicks on a column header and contacting the server to initiate a sort on the
 * selected data. <p>
 * The sort order is initialized in this class.<p>
 */
public abstract class SortRequestListener extends MouseAdapter
{
    private static Category logCategory = Category.getInstance(SortRequestListener.class.getName());


    /**
     * The table which is being sorterd
     */
    protected JTable table;

    /**
     * The index of the last column a sort was initiated on.
     */
    protected int currentSortColumn = 0;

    /**
     * Flag to indicate whether the data is sorted in ascending or descending
     * order.
     */
    protected String sortOrder = ProvisioningInterface.SORT_TYPE_ASC;

    /**
     * Ceate a new SortRequestListener and add it as a mouse listener to the
     * header of the given table.
     */
    public SortRequestListener(JTable aTable)
    {
        table = aTable;
        table.getTableHeader().addMouseListener(this);
    }

    /**
     * Called by the ColumnOrderRenderer so it can select the correct icon to
     * display in the column header.
     * @return true if the sort is ascending, false otherwise
     */
    public String getSortOrder()
    {
		Assert.assert("Sort order is not null", sortOrder != null);
        return sortOrder;
    }

    /**
     * Called by the AccountTableScreen when the row model is being reverted.
     * In that case, the previous sorte ordering is stored by the
     * AccountTableScreen and needs will be set here so that it may be used
     * by the renderer. <p>
     * This method is not used when a column is sorted on by the user clicking
     * its header.
     */
    public void setSortOrder(String sort)
    {
        sortOrder = sort;
    }

    /**
     * Called by the ColumnOrderRenderer so it can decided whether to add an
     * icon to the column header.
     * @return the index of the column on which the table is currently sorted
     */
    public int getCurrentSortColumn()
    {
        return currentSortColumn;
    }

    /**
     * Called by the AccountTableScreen when the row model is being reverted.
     * In that case, the previous column sorted on is stored by the
     * AccountTableScreen and needs will be set here so that it may be used
     * by the renderer. <p>
     * This method is not used when a column is sorted on by the user clicking
     * its header.
     */
    public void setCurrentSortColumn(int column)
    {
        currentSortColumn = column;
    }


    /**
     * Called when the user clicks a header of the associated JTable. <p>
     * Determines the index in the data model of the column the user clicked on. <p>
     * Then determines the sort order (ascending or descending). The behaviour
     * is that if the newly clicked column is the same as the last clicked column
     * then the sort order is reversed on that column. If they are different
     * columns, the new column is sorted in ascending order. <p>
     * Then calls the server to resort the data.
     */
    public void mouseClicked(MouseEvent event)
    {
        TableColumnModel columnModel = table.getColumnModel();

        //the index of the sort column in the column model (the view)
        int columnModelIndex = columnModel.getColumnIndexAtX(event.getX());

        //the index of the sort colun in the table model (the data)
        int dataModelIndex = table.convertColumnIndexToModel(columnModelIndex);

        if (dataModelIndex == currentSortColumn)
        {
          //resort based on this column but in the opposite direction
          if (sortOrder.equals(ProvisioningInterface.SORT_TYPE_ASC))
          {
            sortOrder = ProvisioningInterface.SORT_TYPE_DESC;
          }
          else
          {
            sortOrder = ProvisioningInterface.SORT_TYPE_ASC;
          }
        }
        else
        {
            //sort ascending on the new column
            currentSortColumn = dataModelIndex;
            sortOrder = ProvisioningInterface.SORT_TYPE_ASC;
        }

        sort(currentSortColumn,sortOrder);
    }

    protected abstract void sort(int column, String order);
}
