package org.vovida.prov.table;

import org.vovida.prov.util.assert.*;
import org.vovida.prov.util.FieldNames;

import javax.swing.JTable;
import javax.swing.table.AbstractTableModel;
import java.util.Vector;
import java.util.Hashtable;

public class TableModel extends AbstractTableModel
{
    /**
     * Identifies the type of account that is being displayed in this screen.
     * May be either END_USER (when the screen is being used by a centrex
     * administrator) or ADMINISTRATOR (when the screen is being used
     * by an isp administrator).
     * This class should be the one place rsponsible for storing this
     * information.
     **/
    protected String domain;

    /**
     * The account identifier of the parent account. <p> This would be the
     * account name or phone number or whatever serves to identify the database
     * entry on the provisioning server
     */
	protected String accountIdentifier;

    /**
     * The actual TableColumns that will be displayed in the table. <p>
     * It is expected that the table will have autocreate_columns_from_model
     * set to false and that it will add these columns instead. These need to
     * be kept track of other than by the table because if columns are removed
     * on the fly from the table, a reference to them is necessary if they need
     * to be re-added.
     */
    private Vector columns;

    /**
     * Maps the rows from the order they are stored in to the order they are
     * displayed in (which is affected by sorting) to the order they are stored
     * in (which would be whatever order they were loaded in).
     */
    private RowModel rowModel;


    /** @link dependency */
    /*#NamedTableColumn lnkNamedTableColumn;*/

    public TableModel(String aDomain, String acctId)
    {
        rowModel = new RowModel();
        domain = aDomain;
        accountIdentifier = acctId;
        columns = new Vector();
    }

    /**
     * Used to determine whether the data for the account located at the given
     * row has been loaded from the server. If the account has not been loaded
     * the only data stored on the client will be the account name or (no data
     * at all - at startup).
     */
    public boolean isDataLoaded(int row)
    {
        Hashtable data = rowModel.getDataAt(row);
        if (data == null)
        {
          return false;
        }
        if (data.size() > 1)
        {
            return true;
        }
        return false;
    }

    /**
     * Store a TableColumn. <p> The table column is created by the
     * AccountTableScreen.
     */
    public void addColumn(NamedTableColumn column)
    {
        columns.addElement(column);
    }

    public NamedTableColumn [] getColumns()
    {
        NamedTableColumn [] returnValue = new NamedTableColumn[columns.size()];
        columns.copyInto(returnValue);
        return returnValue;
    }

    /**
     * Add a new account to the model. <p>
     * This method is used initially to add the account before any data for it
     * has been loaded from the server. It will only be displayed by the name
     * provided here. It will be added at the row specified.
     */
    public void addAccount(String accountName, int row)
    {
//        rowModel.addAccountAt(accountName, getColumns()[0].getColumnId(), row);
        rowModel.addAccountAt(accountName, FieldNames.userIdConlumnName, row);
    }

    //------------------------row model operations------------------------------


    /**
     * Invalidates both the rowToIdMap and the idToDataMap
     */
    public void clear()
    {
        rowModel.invalidateIdToDataMap();
        rowModel.invalidateRowToIdMap();
    }

    /**
     * This marks the row to id map as invalid so that it will be reloaded from
     * the server the next time data is requested. This should happen after
     * each search or sort of the data.
     */
    public void invalidateRowModel()
    {
        rowModel.invalidateRowToIdMap();
    }


    /**
     * Every time the row model is changed, it will keep a previous copy of its
     * row to id map. This method causes it to restore that map as the active
     * one. This would have the effect of "undo" or "previous" after a
     * search or sort has been performed on the table.
     */
    public void backRowToIdMap()
    {
        rowModel.back();
    }

    public void forwardRowToIdMap()
    {
        rowModel.forward();
    }

    /**
     * Allocates space in the data model to ensure that the given number of
     * entries (accounts) can be stored. <p>
     * The new entries will be completely empty at this point (not even an
     * account name will be present). This should be used at start up so that
     * the scroll tab on the table will be set to the correct size to accomodate
     * all data from the server.
     */
    public void ensureDataCapacity(int rowCount)
    {
        rowModel.setRowCount(rowCount);
    }

    //------------------------row model operations------------------------------

    /**
     *
     */
    public String getDomain()
    {
        return domain;
    }

    public String getAccountId()
    {
        return accountIdentifier;
    }


	public void setValueAt(int row, String columnId, String value)
    {
        Hashtable data = rowModel.getDataAt(row);
        setValue(data, columnId, value);
System.out.println("setValueAt row=" + row + " , col=" + columnId + ", value=" + value);
    }

    public void setValue(String dataId, String columnId, String value)
    {
        Hashtable data = rowModel.getDataForId(dataId);
        setValue(data, columnId, value);
    }

    private void setValue(Hashtable data, String columnId, String value)
    {
		Assert.assert("Data is not null", data != null);

        if (value == null)
        {
            data.remove(columnId);
        }
        else
        {
            data.put(columnId, value);
        }
    }

    public void setRow(int row, Hashtable value)
    {
        Assert.assert("row is within range", (row>=0) && (row<=getRowCount()));
        Assert.assert("value is not null", value!=null);
        Hashtable data = rowModel.getDataAt(row);
        Assert.assert("row exists",data!=null);

        rowModel.setDataAt(row,value);
    }



    //-----------------start AbstractTableModel methods------------------------

    public String getColumnName(int column)
    {
        if (AssertLevel.ASSERT)
        {
            Assert.assert("column is not null", columns.elementAt(column)!= null);
        }
        String name = ((NamedTableColumn)columns.elementAt(column)).getColumnName();
        return name;
    }

    public Object getValueAt(int row, int column)
    {
        Hashtable data = rowModel.getDataAt(row);
        if (data == null) //data for this account is not loaded yet
        {
            return "";
        }

        String id = getColumns()[column].getColumnId();
        String value = (String)data.get(id);

        return value;
    }

    public Class getColumnClass(int column)
    {
        return String.class;
    }

    public int getRowCount()
    {
        return rowModel.getRowCount();
    }

    public int getColumnCount()
    {
        return columns.size();
    }

    public void removeRow(int row)
    {
		rowModel.removeDataAt(row);
    }
}
