package org.vovida.prov.table;

import org.vovida.prov.table.event.*;
import org.vovida.prov.table.task.TaskLoadAccountKeys;
import org.vovida.prov.table.task.TaskLoadAccountData;
import org.vovida.prov.table.action.*;
import org.vovida.prov.util.task.TaskListener;
import org.vovida.prov.util.task.Task;
import org.vovida.prov.util.assert.*;
import org.vovida.prov.util.Coalesce;
import org.vovida.prov.util.FieldNames;
import org.vovida.prov.util.IntPair;
import org.vovida.prov.util.ProvisioningServerWrapper;
import org.vovida.prov.util.logging.Util;

import javax.swing.JComponent;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JTable;
import javax.swing.JScrollPane;
import javax.swing.JScrollBar;
import javax.swing.JPopupMenu;
import javax.swing.table.JTableHeader;
import javax.swing.table.TableCellRenderer;
import javax.swing.table.TableColumnModel;

import java.awt.event.ActionEvent;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.InputEvent;
import java.awt.event.MouseEvent;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Frame;
import java.awt.Window;

import java.util.Vector;

import org.apache.log4j.Category;



public class TableScreen extends JPanel
{
    private static Category logCategory = Category.getInstance(TableScreen.class.getName());


    protected RemovableColumnTable table;
    protected TableModel model;
    protected JScrollPane scroller;
    protected ProvisioningServerWrapper server;

    private SortInfoKeeper keeper;

    private String parentDomain;

    private Window parentComponent;

    /** @link dependency */
    /*#SortRequestListener lnkSortRequestListener;*/

    /** @link dependency */
    /*#ColumnOrderRenderer lnkColumnOrderRenderer;*/


    /**
     * Create a table based on the given column identifers. <p>
     * The column identifiers should be a set of field names that are understood
     * by the provisioning server and are valid for the given account identifier.
     * The table will create a column for each identifer, then retrieve the
     * data for the table from the provisioning server using the given server
     * interface.
     **/
    public TableScreen (String accountIdentifier, String aParentDomain, String childDomain,
	        String [] columnIdentifers, ProvisioningServerWrapper aServer, Window parentWindow)
    {
        server = aServer;
        parentDomain = aParentDomain;
        parentComponent = parentWindow;

        table = new RemovableColumnTable();
        table.setAutoCreateColumnsFromModel(false);
        table.setAutoResizeMode(table.AUTO_RESIZE_OFF);
        model = new TableModel(childDomain, accountIdentifier);
        scroller = new JScrollPane(table);

        table.setModel(model);
//        table.setSelectionMode(javax.swing.ListSelectionModel.SINGLE_SELECTION);
        initColumns(columnIdentifers);

        initColumnSorter();

        setLayout(new BorderLayout());
        add(scroller, BorderLayout.CENTER);

        table.setDefaultRenderer(String.class,new RowRenderer());

        initPopupMenu();
    }


    public void startLoadingData()
    {
		Assert.assert("SortInfo is not null", keeper.getSortInfo() != null);
        try
        {
            //first, set the scrollbars to the correct size by determining
            //how many accounts are on the server
            int count = server.getAccountCount(model.getDomain(), model.getAccountId(), keeper.getSortInfo().sortOrder);
            logCategory.info("There are " + count + " rows in the table.");

	        model.ensureDataCapacity(count);
	        model.fireTableStructureChanged();

/**
            //set the size of the panel to its preferred size so that the
            //correct number of rows is visible and data is loaded properly
            Dimension size = getPreferredSize();
            getTopLevelAncestor().setSize(size);
**/
            getTopLevelAncestor().validate();

            //next load the first screenfull of data
        	loadScreenOfData();
        }
        catch (Exception ex)
        {
            String msg = "Could not find out how many accounts are stored on the server.";
            logCategory.error(msg);
            Util.showException(msg,ex);
        }

        initEventCompactor();
    }

    private void initColumnSorter()
    {
        Assert.assert("table has at least one column", model.getColumns().length>0);

		//find the index of the column that stores the key field
        int keyIndex = -1;
		NamedTableColumn [] col = model.getColumns();
        for (int i=0; i<col.length; i++)
        {
            if (col[i].getColumnId().equals(FieldNames.keyField))
            {
				keyIndex = col[i].getModelIndex();
            }
        }

        //sort on the key field first
        SortRequestListener listener =
            new AccountSortRequestListener(table,server,this,keyIndex);
        ColumnOrderRenderer renderer = new ColumnOrderRenderer(listener);
        JTableHeader header = table.getTableHeader();
        header.setDefaultRenderer(renderer);

        keeper = new SortInfoKeeper(listener, table, model);

        table.getTableHeader().addMouseListener(keeper);
    }

    public void backupModel()
    {
        keeper.back();
    }

    public void forwardModel()
    {
        keeper.forward();
    }


    private void initColumns(String [] columns)
    {
        for (int i=0; i<columns.length; i++)
        {
            NamedTableColumn column = new NamedTableColumn(columns[i]);
            model.addColumn(column);
            column.setModelIndex(i); //?? :(
            table.addColumn(column);
        }

        //make sure the table is updated
        model.fireTableStructureChanged();
    }


    public void loadScreenOfData()
    {
        Assert.assert("There is a first column", model.getColumns()[0] != null);

		//get the rows which don't have data loaded
        int [] emptyRows = TableScreen.this.getRowsWithNoData();

        //if there are no visible rows in the table
        if (emptyRows.length == 0)
        {
            logCategory.debug("No new rows visible in the table of accounts - not loading data");
            return;
        }

        //load all the data for all the visible rows which do not
        //have data loaded yet

        final IntPair [] ranges = Coalesce.coalesce(emptyRows);
        if (ranges != null)
        {
            //get the field names of all the columns
            Vector fields = new Vector();
            NamedTableColumn [] columns = model.getColumns();
            for (int field=0; field<model.getColumnCount(); field++)
            {
               if (columns[field].isDomainColumn())
               {
                   fields.addElement(columns[field].getColumnId());
               }
            }

            final String [] fieldNames = new String [fields.size()];
            fields.copyInto(fieldNames);

//            final String [] fieldNames = new String [] {"*"};

            for (int i=0; i<ranges.length; i++)
            {
				logCategory.info("Starting to load data for row range " + ranges[i]);

                 new TaskLoadAccountKeys(ranges[i].getFirst(),
                     ranges[i].getSecond(), model.getDomain(),
                     model.getAccountId(), FieldNames.keyField,
                     keeper.getSortInfo().sortOrder, server,
                     model, new MyListener(ranges[i].getFirst(),
					ranges[i].getSecond(), model,server,fieldNames)).start();
             }
        }
    }

    private void initEventCompactor()
    {
        //The EventCompactor that will listen to scroll events
        AdjustmentEventCompactor compactor = new AdjustmentEventCompactor();
        CustomScrollBarUI customUI = new CustomScrollBarUI();
        scroller.setVerticalScrollBar(new CustomScrollBar(customUI));
        customUI.addMouseListener(compactor);
        scroller.getVerticalScrollBar().addMouseListener(compactor);
        scroller.getVerticalScrollBar().addAdjustmentListener(compactor);
        compactor.addCompactEventListener(new CompactEventListener()
        {
            //The table has finished scrolling
            public void compactEventOccurred(CompactEvent event)
            {
                logCategory.info("scroll compactEvent occurred");
				loadScreenOfData();
            }
        });


        //The event compactor that will listen for events dealing with
        //resizing the frame.
        scroller.addComponentListener(new ComponentAdapter()
        {
             public void componentResized(ComponentEvent e)
             {
                logCategory.info("resize event occurred");
                loadScreenOfData();
                sizeColumnsToFitData();
             }
        });
    }

    /**
     * Determine the index of the first visible row in the JTable contained
     * in this screen.
     */
    public int getFirstVisibleRow()
    {
        int row = table.rowAtPoint(scroller.getViewport().getViewPosition());
        logCategory.debug("First visible row = " + row);
        return row;
    }

    /**
     * Determine the index of the last visible row in the JTabel contained in
     * this screen.
     */
    public int getLastVisibleRow()
    {
        Dimension extent = scroller.getViewport().getExtentSize();
        int lastY = scroller.getViewport().getViewPosition().y + (int)extent.getHeight()-1;
        int row = table.rowAtPoint(new Point(1,lastY));
        //this would give you the index of the last visible row in the table,
        //assuming that there are enough rows to fill the entire view.
        //if there isn't a row at the end of the view, then return the last
        //row that exists in the table
        if (row == -1)
        {
            if (table.getRowCount() > 0) //if there is data in the table
            {
	            row = table.getRowCount() -1; //this is supposed to be the index, not lenght
            }
        }
        logCategory.debug("The last visible row is " + row);
        return row;
    }

    /**
     * Return an array of indices of the rows which are currently visible on
     * the screen but for which the acount data has not been loaded from the
     * provisioning server yet.
     */
    private int [] getRowsWithNoData()
    {
        //find which rows are visible
        int first = getFirstVisibleRow();
        int last = getLastVisibleRow();

        //if there are no visible rows
        if ( (first == -1) || (last == -1))
        {
			return new int []{};
        }

        Vector rowsWithNoData = new Vector();

        //of the visible rows, find out which ones don't have data
        //loaded for them yet and add these to the list of rows
        for (int i=first; i<=last; i++)
        {
            if (!model.isDataLoaded(i))
            {
                rowsWithNoData.addElement(new Integer(i));
            }
        }

        int [] rows = new int [rowsWithNoData.size()];
        for (int i=0; i<rowsWithNoData.size(); i++)
        {
            rows[i] = ((Integer)rowsWithNoData.elementAt(i)).intValue();
        }
        return rows;
    }


    public TableModel getTableModel()
    {
        return model;
    }


    /**
     * Resize all the columns in the table to fit their data. <p>?
     * Each column is resized to the widest of either the space taken up by its
     * header or the widest piece of data currently visible in that column. <p>
     * If, after the above operation, the total width of all the columns is less
     * than the width of the scrollpane the table is displayed in, any difference
     * is split evenly among all the columns.
     */
    public void sizeColumnsToFitData()
    {
        Assert.assert("There's a table", table!=null);
        Assert.assert("The table has a model", model!=null);
        Assert.assert("There's a scroller", scroller!=null);
        Assert.assert("The table has a first column", table.getColumnModel().getColumn(0)!=null);

        //first, determine the width taken up by the headers
        TableCellRenderer renderer = table.getTableHeader().getDefaultRenderer();
        JComponent rendererComponent = (JComponent)renderer;
        int numColumns = table.getColumnCount();
        TableColumnModel columnModel = table.getColumnModel();
        for (int i=0; i<numColumns; i++)
        {
            //the width taken up by the text itself
            int width = (int)renderer.getTableCellRendererComponent(table,model.getColumnName(i),false,false,0,i).getPreferredSize().getWidth();
            //add the width taken up by the border
            width += rendererComponent.getInsets().left + rendererComponent.getInsets().right;
            //set the width of the column
            columnModel.getColumn(i).setPreferredWidth(width);
        }

        //next, determine if any of the visible data takes up more space
        int first = this.getFirstVisibleRow();
        int last = this.getLastVisibleRow();
        for (int column=0; column<numColumns; column++)
        {
            renderer = table.getDefaultRenderer(table.getColumnClass(column));
            rendererComponent = (JComponent)renderer;
            int width = columnModel.getColumn(column).getPreferredWidth();
            for (int row=first; row<=last; row++)
            {
                int newWidth = (int)renderer.getTableCellRendererComponent(table,model.getValueAt(row,column),false,false,row,column).getPreferredSize().getWidth();
                if (newWidth > width)
                {
                    width = newWidth;
                }
            }
            width += rendererComponent.getInsets().left + rendererComponent.getInsets().right;
            columnModel.getColumn(column).setPreferredWidth(width);
        }

        //finally, determine if the scroller the table is displayed in has
        //more room available than is taken up by the columns
        //if it does, share the extra space equally among all the columns
        int totalWidth = getTotalColumnWidth();
        int scrollerWidth = scroller.getViewport().getWidth();
        if (totalWidth < scrollerWidth)
        {
            int extra = scrollerWidth-totalWidth;
            int widthIncrement = extra/numColumns;
            for (int i=0; i<numColumns; i++)
            {
                columnModel.getColumn(i).setPreferredWidth(columnModel.getColumn(i).getPreferredWidth()+widthIncrement);
            }
        }

        //see if there's any space left over in the scroller
        totalWidth = getTotalColumnWidth();
        if (totalWidth < scrollerWidth)
        {
            //there's some left over due to rounding
            //ummmm ... asign it to the first column
            columnModel.getColumn(0).setPreferredWidth(columnModel.getColumn(0).getPreferredWidth()+scrollerWidth-totalWidth);
        }
    }

    /**
     * Calculate the total preferred width of all the columns in the table.<p>
     */
    private int getTotalColumnWidth()
    {
        int totalWidth = 0;
        int numColumns = table.getColumnCount();
        for (int i=0; i<numColumns; i++)
        {
            totalWidth += table.getColumnModel().getColumn(i).getPreferredWidth();
        }
        return totalWidth;
    }

    public RemovableColumnTable getTable()
    {
        return table;
    }

    /**
     * A TaskListener which will load the account data once the account keys
     * have been loaded. <p>
     * Used in the loadScreenOfData() method.
     */
    private class MyListener implements TaskListener
    {
        private String [] fields;
        private ProvisioningServerWrapper server;
        private TableModel data;
        private int startIndex;
        private int stopIndex;



        public MyListener (int from, int to, TableModel model, ProvisioningServerWrapper wrapper, String [] fieldNames)
        {
            fields = fieldNames;
            server = wrapper;
            data = model;
            startIndex = from;
            stopIndex = to;
        }

		public void taskFinished(Task task, boolean status, Object returnValue)
         {
                if (status == true)
                {
	                //once the keys have been loaded, load the data as well
			        new TaskLoadAccountData(startIndex,stopIndex,
                        data.getDomain(),data.getAccountId(),server,
			            data,fields).start();
                }
         }

    }

	/**
     * Return information on the current sort status of the table (sort column
     * index and sort order.<p>
     */
    public SortInfo getSortInfo()
    {
    	return keeper.getSortInfo();
    }


	/**
     * Add the popup menu to table and scrollpane which allows data to be edited
     * added, deleted and searched. <p>
     */
    private void initPopupMenu()
    {
		final JPopupMenu menu = new JPopupMenu();
        menu.add(new ActionSearch(server,this,"Search",(Frame)parentComponent));
        menu.add(new ActionClearSearch(server,this,"Clear search", (Frame)parentComponent));
		menu.add(new JPopupMenu.Separator());
        menu.add(new ActionEdit(server, this, "Edit account",
            model.getAccountId(), parentDomain, model.getDomain(),parentComponent));
        menu.add(new ActionAdd(server, this, "Add account",
            model.getAccountId(), parentDomain, model.getDomain(),parentComponent));
        menu.add(new ActionDelete(server,this,"Delete account",
            model.getDomain(), parentComponent));

        class PopupListener extends MouseAdapter
	    {
	      public void mouseReleased(MouseEvent event)
	      {
	        if ((event.getModifiers() & InputEvent.BUTTON3_MASK)
	            == InputEvent.BUTTON3_MASK)
	        {
	          menu.show(TableScreen.this, event.getX(), event.getY());
	        }
            if (event.getClickCount() == 2) //if this is a double click
            {
				new ActionEdit(server, TableScreen.this, "Edit account",
		            model.getAccountId(), parentDomain, model.getDomain(),
                    parentComponent).actionPerformed(new ActionEvent(this, 9999, "edit"));
            }
	      }
	    }
	
	    PopupListener listener = new PopupListener();
	    table.addMouseListener(listener);
	    scroller.addMouseListener(listener);
    }

    public void setColumnDescriptions (String xml)
    {
        final NamedTableColumn [] columns = model.getColumns();

		org.apache.xerces.parsers.SAXParser parser = new org.apache.xerces.parsers.SAXParser();

        org.xml.sax.helpers.DefaultHandler handler = new org.xml.sax.helpers.DefaultHandler()
        {
			private String currentColumnId = null;
        	private boolean inLabel = false;
            private boolean inTooltip = false;

		    public void startElement(String uri, String local, String raw, org.xml.sax.Attributes attrs)
            {
				if (local.equals("DBEntry") || local.equals("DataDefn") )
                {
                    currentColumnId = attrs.getValue("name");
                    String key = attrs.getValue("key");
                    if ( (key != null) && (!"".equals(key)) )
                    {
                        FieldNames.keyField = key;
                    }
                }
                else if (local.equals("Label"))
                {
                    inLabel = true;
                }
                else if (local.equals("Tooltip"))
                {
                    inTooltip = true;
                }
            }

            public void endElement(String uri, String local, String qName)
            {
				if (local.equals("DBEntry") || local.equals("DataDefn"))
                {
                    currentColumnId = null;
                }
                else if (local.equals("Label"))
                {
                    inLabel = false;
                }
                else if (local.equals("Tooltip"))
                {
                    inTooltip = false;
                }
            }

            public void characters(char[] ch, int start, int length)
            {
                if (inLabel)
                {
                    for (int i=0; i<columns.length; i++)
                    {
                        if (columns[i].getColumnId().equals(currentColumnId))
                        {
                            columns[i].setColumnName(new String(ch,start,length));
                            break;
                        }
                    }
                }
                else if (inTooltip)
                {
                    for (int i=0; i<columns.length; i++)
                    {
                        if (columns[i].getColumnId().equals(currentColumnId))
                        {
		                	columns[i].setTooltip(new String(ch,start,length));
                        }
                    }
                }
            }
        };

        parser.setContentHandler(handler);
        try
        {
			parser.parse(new org.xml.sax.InputSource(new java.io.StringReader(xml)));
        }
        catch (Exception ex)
        {
            System.out.println(ex);
        }
    }

    public void clear()
    {
        model.invalidateRowModel();
        model.fireTableDataChanged();
    }

}
