package org.vovida.prov.table.event;

import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.util.EventObject;
import java.util.Vector;
import java.util.ListIterator;

import org.apache.log4j.Category;


 /**
  * This class "compacts" events that occur between a mousePressed and
  * mouseReleased event. Both the mouseEvents and the other event type are
  * generated by the same class.<p>
  * What this means is that concrete implementations of this class should
  * also implement an event listener interface for some other type of event
  * (for example, an AdjustmentEvent). The concrete implementation is then
  * registered as a xxxEventListener on some class and also as a mouseListener
  * on that same class. Then when any xxxEvents are generated after a
  * pressedEvent, they will be sent to this class and discarded except for the
  * last one which occurs before a mouseReleased event. The class interested
  * in the compacted event is notified only once for the last time that event
  * occured. <p>
  * This is intended to be used in cases such as resizing a window which
  * generates a whole series of events while the window is being resized but
  * the program is only interested in the case where the window has <italic>
  * finised </italic> being resized.
  */
abstract public class EventCompactor extends MouseAdapter
{

    private static Category logCategory = Category.getInstance(EventCompactor.class.getName());

    /**
     * Flag indicating a mousePressed event has been detected. This means
     * that all events (of the type being listened for by the concrete
     * implementation of this class) which occur from now untill the
     * mouseReleased happens are to be compacted (discarded and only the last
     * one is saved.
     */
    protected boolean clicked = false;

    /**
     * The last event of the type the concrete subclass is listening for which
     * was detected. This should be replaced each time a new event is detected.
     */
    protected EventObject lastEvent = null;

    /**
     * List of listeners which want to be notified when a compacted event occurs.
     */
    private Vector eventListeners;

    public EventCompactor()
    {
        eventListeners = new Vector();
    }

    //---------- start implementation of MouseLitener interface ---------

    /**
     * Sets a flag to indicate that the mouse has been clicked down on the
     * component being monitored
     */
    public void mousePressed(MouseEvent e)
    {
        logCategory.debug("mouse pressed");
        clicked = true;
    }

    /**
     * Called when the user releases the mouse on the component being monitored.
     * If an event of the type that is being listened for has been detected,
     * it will be passed on to all listeners of this component and the clicked
     * flag will be cleared.
     */
    public void mouseReleased(MouseEvent e)
    {
        logCategory.debug("mouse released");
        if (lastEvent != null)
        {
            fireCompactEvent(lastEvent);
            lastEvent = null;
            clicked = false;
        }
    }
    //---------- end implementation of MouseLitener interface ---------

    //---------- start event listener pattern -------------------------

    protected void fireCompactEvent(EventObject source)
    {
        ListIterator listeners = eventListeners.listIterator();
        while (listeners.hasNext())
        {
            CompactEventListener listener = (CompactEventListener)listeners.next();
            CompactEvent compactEvent = new CompactEvent(source);
            listener.compactEventOccurred(compactEvent);
        }
    }

    public void addCompactEventListener(CompactEventListener listener)
    {
        eventListeners.addElement(listener);
    }

    public void removeCompactEventListener(CompactEventListener listener)
    {
        ListIterator listeners = eventListeners.listIterator();
        int count=0;
        while (listeners.hasNext())
        {
            CompactEventListener temp = (CompactEventListener)listeners.next();
            count++;
            if (temp.equals(listener))
            {
                eventListeners.remove(count);
                return;
            }
        }
    }

    //---------- end event listener pattern -------------------------

    /**
     * Called by the concrete subclasses when whatever event they are listening
     * for occurs.<p>
     * Just saves that event.
     */
    protected void eventOccurred(EventObject event)
    {
        logCategory.debug("event occurred");
        lastEvent = event;
    }
}
