/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.template;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowListener;
import java.awt.event.WindowAdapter;
import java.awt.event.MouseAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.BorderLayout;
import java.awt.GridLayout;
import java.awt.Component;
import java.awt.Dialog;
import java.awt.Frame;
import javax.swing.BorderFactory;
import javax.swing.JScrollPane;
import javax.swing.JOptionPane;
import javax.swing.JComponent;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JPanel;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.vovida.prov.dynamicGui.ProvisionableInstance;
import org.vovida.prov.dynamicGui.ProvisionableType;
import org.vovida.prov.dynamicGui.TypeBuilder;
import org.vovida.prov.widget.SmartLayoutManager;
import org.vovida.prov.content.ContentData;
import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;

/**
 * <P>This class functions as a controller for handling the addition of many
 * database entries to the Provisioning System.  It is driven by a template
 * that provides the necessary data for a database entry.  The BulkAddController
 * handles data entry from a bar code scanner.  It also facilitates adding
 * one record at a time or many records with one click of a button.</P>
 * <P>To use this class, create an instance and then add it as an ActionListener
 * to some gui object.  When that gui object fires an action event the
 * BulkAddController will jump to life.</P>
 *
 * @author Matt Naish
 */
public class BulkAddController extends KeyAdapter implements ActionListener
{
  private static final String BULK_ADD = "Bulk Add";
  private static final String QUIT = "Quit";
  private static final String SKIP = "Skip";
  private static final String ADD_ALL = "Add All";
  private static final String SAVE_NEXT = "Save & Next>";
  private static final String SAVE_FINISH = "Save & Finish";
  private static final String CANCEL = "Cancel";

  // the time, in milliseconds, to "hold down" the save button when a
  // bar code is read
  private static final int PRESS_TIME = 1000;

  private Component parent;
  private TemplateLoader tl;

  private Hashtable preFetch;

  private JDialog dialog;
  private ProvisionableType provType;
  private ProvisionableInstance provInstance;

  private Hashtable dataProviders;
  private Vector fieldNames;

  private JPanel dynamicPanel;
  private JButton quitButton;
  private JButton skipButton;
  private JButton addAllButton;
  private JButton saveButton;

  private JComponent barCodeComponent;

  private RecordUpdaterData rud;

  /**
   * Only used for display purposes
   */
  private int numIter = 0;

  /**
   * Used in addAll()
   */
  private boolean cancelClicked = false;


  /**
   * @param parent the parent Frame for the GUI Dialogs
   * @param tl provides the necessary data for addition to the database.
   */
  public BulkAddController(Frame parent, TemplateLoader tl)
  {
    this.parent = parent;
    this.tl = tl;
  }

  /**
   * @param parent the parent Dialog for the GUI Dialogs
   * @param tl provides the necessary data for addition to the database.
   */

  public BulkAddController(Dialog parent, TemplateLoader tl)
  {
    this.parent = parent;
    this.tl = tl;
  }

  /**
   * Action controller.  If the source of the ActionEvent is not one of the
   * recognized internal Swing objects (eg. Close button) the
   * BulkAddController GUI will be brought to life.
   */
  public void actionPerformed(ActionEvent ae)
  {
    if (ae.getSource() == quitButton)
    {
      quit();
    }
    else if (ae.getSource() == skipButton)
    {
      fillInNext();
    }
    else if (ae.getSource() == addAllButton)
    {
      addAll();
    }
    else if (ae.getSource() == saveButton)
    {
      save();
    }
    else
    {
      showUI();
    }
  }

  /**
   * Used to 'start' the BulkAddController.  Shows the UI.
   */
  private void showUI()
  {
    String templateName = (String) JOptionPane.showInputDialog(parent, 
            "Select a template to use.", "Select", 
            JOptionPane.QUESTION_MESSAGE, null, tl.getAvailableTemplates(), 
            null);

    // Did the user select something or did they hit cancel
    if (templateName != null)
    {

      // Selected something!

      DataProviderWrapper[] fields;
      DataProvider dp;
      ContentData content;
      String barCodeField;

      // try to load the template file
      try
      {
        tl.load(templateName);
      }
      catch (NestedException ne)
      {
        JOptionPane.showMessageDialog(parent, ne.getMessage(), "Error", 
                JOptionPane.ERROR_MESSAGE);

        return;
      }

      // get the ProvisionableType for the selected template
      provType = TypeBuilder.getType(tl.getParent());

      if (provType == null)
      {
        JOptionPane.showMessageDialog(parent, 
                "XML definition file '" + tl.getParent() + "' not found.", 
                "Error", JOptionPane.ERROR_MESSAGE);

        return;
      }

      // create the dynamic GUI portion
      dynamicPanel = new JPanel(new SmartLayoutManager());
      provInstance = provType.buildInstance(dynamicPanel);

      // parse the template file and assign DataProviders to the
      // appropriate fields.  Also assign the bar code field.
      try
      {
        dataProviders = new Hashtable();
        fieldNames = new Vector();
        fields = tl.getFields();

        for (int i = 0; i < fields.length; i++)
        {
          dp = fields[i].getDataProvider();

          // if there is a DataProvider for a field, intitalize it.
          if (dp != null)
          {
            fieldNames.add(fields[i].getDBName());
            dataProviders.put(fields[i].getDBName(), dp);
            dp.connect();

            if (dp.next())
            {
              content = provInstance.getContentByName(fields[i].getDBName());

              if (content == null)
              {
                JOptionPane.showMessageDialog(parent, 
                        "DBEntry '" + fields[i].getDBName() 
                        + "' not found in '" + tl.getParent() 
                        + "'.\nOpen template and re-save to fix problem.", 
                        "Error", JOptionPane.ERROR_MESSAGE);

                return;
              }

              content.setContent(dp.getData());
            }
            else
            {

              // If one DataProvider.next() call returns false on its first
              // call, there is no data to add.
              JOptionPane.showMessageDialog(parent, 
                      "Template returns no data.", "Error", 
                      JOptionPane.ERROR_MESSAGE);

              return;
            }
          }
        }

        if (fieldNames.size() == 0)
        {
          JOptionPane.showMessageDialog(parent, "Template returns no data.", 
                  "Error", JOptionPane.ERROR_MESSAGE);

          return;
        }

        // intialize the bar code field, if any.
        barCodeField = tl.getBarCodeField();

        if (barCodeField != null)
        {
          barCodeComponent = provInstance.getComponentByName(barCodeField);

          if (barCodeComponent == null)
          {
            JOptionPane.showMessageDialog(parent, 
                    "Bar Code Field '" + barCodeField + "' not found in '" 
                    + tl.getParent() 
                    + "'.\nOpen template and re-save to fix problem.", 
                    "Error", JOptionPane.ERROR_MESSAGE);

            return;
          }

          barCodeComponent.addKeyListener(this);
        }

        // initialize the 'Per Record Update', if any.
        rud = tl.getRecordUpdaterData();

        if (!rud.active ||!rud.dataValid())
        {
          rud = null;
        }

        // all looks good.  Create the UI.
        dialog = createUI();

        // Do a pre-fetch of data.  This is done so that we know to disable
        // the skip button and change the save button to "Save & Finish"
        doPreFetch();

        // Let's see it!
        dialog.setVisible(true);
      }
      catch (NestedException ne)
      {

        // There was some error getting the data.
        JOptionPane.showMessageDialog(parent, ne.getMessage(), "Error", 
                JOptionPane.ERROR_MESSAGE);
      }
    }
  }

  /**
   * Listens for a CTRL-C key press on a certain text field.
   * If received, simulate a click on the save button
   */
  public void keyReleased(KeyEvent ke)
  {

    // is CTRL and only CTRL pressed
    if ((ke.getModifiers() & InputEvent.CTRL_MASK) == InputEvent.CTRL_MASK)
    {

      // is this a CTRL-C key press? (The Metrologic bar code scanner
      // we have is programmed to send an ascii ETX (end of text)
      // after it sends its data.  For some reason this shows up
      // as a CTRL-C.
      if (ke.getKeyCode() == KeyEvent.VK_C)
      {
        saveButton.doClick(PRESS_TIME);
      }
    }
  }

  /**
   * Run when the user clicks "Quit"
   */
  private void quit()
  {
    int resp = JOptionPane.showConfirmDialog(dialog, 
            "Are you sure you want to quit?", "Quit?", 
            JOptionPane.YES_NO_OPTION);

    if (resp == JOptionPane.YES_OPTION)
    {
      closeAndCleanUp();
    }
  }

  /**
   * Does some object tidy up and close the dialog window.  Called when the
   * user "Quits", closes the window, or has finished adding all the data.
   */
  private void closeAndCleanUp()
  {
    dialog.dispose();

    dialog = null;
    preFetch = null;
    dynamicPanel = null;
    quitButton = null;
    skipButton = null;
    addAllButton = null;
    saveButton = null;
    provType = null;
    provInstance = null;
    fieldNames = null;
    barCodeComponent = null;
    numIter = 0;

    for (Enumeration elements = dataProviders.elements(); 
         elements.hasMoreElements(); )
    {
      try
      {
        ((DataProvider) elements.nextElement()).close();
      }
      catch (NestedException ne)
      {

        // don't do much because we are closing things anyway.
        ne.printStackTrace();
      }
    }

    dataProviders = null;
  }

  /**
   * Create the UI.  Returns a Dialog with all the UI components added.
   */
  private JDialog createUI()
  {
    JPanel contentPanel = new JPanel();
    JScrollPane scrollPane = new JScrollPane();
    JPanel buttonPanel = new JPanel();
    JPanel bottomPanel = new JPanel();
    WindowListener wListener;

    quitButton = new JButton(QUIT);
    skipButton = new JButton(SKIP);
    addAllButton = new JButton(ADD_ALL);
    saveButton = new JButton(SAVE_NEXT);

    dynamicPanel.setBorder(BorderFactory.createEmptyBorder(10, 10, 10, 10));
    scrollPane.getViewport().add(dynamicPanel);

    buttonPanel.setLayout(new GridLayout(1, 0, 5, 5));
    buttonPanel.add(quitButton);
    buttonPanel.add(skipButton);
    buttonPanel.add(addAllButton);
    buttonPanel.add(saveButton);
    bottomPanel.add(buttonPanel);

    contentPanel.setLayout(new BorderLayout());
    contentPanel.add(scrollPane, BorderLayout.CENTER);
    contentPanel.add(bottomPanel, BorderLayout.SOUTH);

    quitButton.addActionListener(this);
    skipButton.addActionListener(this);
    addAllButton.addActionListener(this);
    saveButton.addActionListener(this);

    if (parent instanceof Frame)
    {
      dialog = new JDialog((Frame) parent, BULK_ADD, true);
    }
    else
    {
      dialog = new JDialog((Dialog) parent, BULK_ADD, true);
    }

    if (barCodeComponent != null)
    {
      dialog.addWindowListener(new WindowAdapter()
      {

        // if the user closes (kills) the window
        public void windowClosing(WindowEvent we)
        {
          quit();
        }

        // when the window opens for the first time, put the focus to the bar
        // code field, if any.
        public void windowOpened(WindowEvent we)
        {
          barCodeComponent.requestFocus();
        }

      });
    }
    else
    {
      dialog.addWindowListener(new WindowAdapter()
      {
        public void windowClosing(WindowEvent we)
        {
          quit();
        }

      });
    }

    dialog.setContentPane(contentPanel);
    dialog.setTitle(BULK_ADD + " - Item # " + ++numIter);
    dialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
    dialog.pack();

    return dialog;
  }

  /**
   * Populate the UI with the next set of data
   */
  private void fillInNext()
  {
    String fieldName;

    dialog.setTitle(BULK_ADD + " - Item # " + ++numIter);

    // restore the default values
    provInstance.reset();
    provInstance.clearKeyValue();

    for (Enumeration enum = fieldNames.elements(); enum.hasMoreElements(); )
    {
      fieldName = (String) enum.nextElement();

      provInstance.getContentByName(fieldName).setContent((String) preFetch.get(fieldName));
    }

    // Put the focus to the bar code field, if any.
    if (barCodeComponent != null)
    {
      barCodeComponent.requestFocus();
    }

    try
    {

      // Do a pre-fetch of data.  This is done so that we know to disable
      // the skip button and change the save button to "Save & Finish"
      doPreFetch();
    }
    catch (NestedException ne)
    {
      JOptionPane.showMessageDialog(dialog, ne.getMessage(), "Error", 
              JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * Do a pre-fetch of data.  This is done so that we know to disable
   * the skip button and change the save button to "Save & Finish"
   */
  private void doPreFetch() throws NestedException
  {
    DataProvider dp;
    String fieldName;
    boolean hasNext = true;

    preFetch = new Hashtable();

    for (Enumeration enum = fieldNames.elements(); 
         enum.hasMoreElements() && hasNext; )
    {
      fieldName = (String) enum.nextElement();
      dp = (DataProvider) dataProviders.get(fieldName);
      hasNext = dp.next();

      if (hasNext)
      {
        preFetch.put(fieldName, dp.getData());
      }
    }

    if (!hasNext)
    {

      // JOptionPane.showMessageDialog(dialog, "No more data");
      preFetch = null;

      saveButton.setText(SAVE_FINISH);
      skipButton.setEnabled(false);
      addAllButton.setEnabled(false);
    }
  }

  /**
   * The user clicked 'Save & Next>' or 'Save & Finish'.
   */
  private void save()
  {
    if (doPerRecordUpdate())
    {
      provInstance.addEntryToDB();

      if (preFetch != null)
      {
        fillInNext();
      }
      else
      {
        closeAndCleanUp();
      }
    }
  }

  /**
   * Perform a 'Per Record Update', any exists.
   * @returns true if successful (or nothing to do), false if error occured
   * while performing update and user does not want to continue.
   */
  private boolean doPerRecordUpdate()
  {
    boolean result = true;

    if (rud != null)
    {
      ProvisioningInterface pi = ProvisioningInterface.getInstance();
      Hashtable fieldValuePairs = new Hashtable();

      fieldValuePairs.put(rud.update, rud.toBe);

      try
      {
        pi.setRecordsInDomain(rud.in, 
                rud.where + "='" 
                + provInstance.getContentByName
                  (rud.selectedIsValue).getContentAsString() 
                + "'", fieldValuePairs);
      }
      catch (Exception e)
      {
        int resp = JOptionPane.showConfirmDialog(dialog, 
                "A problem occured while performing the 'Per Record Update':\n" 
                + e.getMessage() 
                + "\n\nNo other 'Per Record Updates' will occur " 
                + "during this session.\n\n"
                + "Do you want to continue to 'add' anyway?", 
                "Error", JOptionPane.YES_NO_OPTION, 
                JOptionPane.ERROR_MESSAGE);

        result = (resp == JOptionPane.YES_OPTION);
        rud = null;
      }
    }

    return result;
  }

  /**
   * User clicked 'Add All'.  Curse through the data and start adding
   * it automatically
   */
  private void addAll()
  {
    if (JOptionPane.showConfirmDialog(dialog,
      "Are you sure you want to automatically add all data?",
      "Question?", JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE) 
        == JOptionPane.YES_OPTION)
    {

      // this is all pretty inefficient because we show all the data in the GUI
      // a better way would be to just show a status (progress indicator).

      final JOptionPane statusPane = 
        new JOptionPane("Click Cancel to stop adding data.", 
                        JOptionPane.INFORMATION_MESSAGE, 
                        JOptionPane.CANCEL_OPTION, null, new String[]
      {
        CANCEL
      });
      final JDialog statusDialog = new JDialog(dialog, "Adding Data...", 
              true);

      statusDialog.setDefaultCloseOperation(JDialog.DO_NOTHING_ON_CLOSE);
      statusDialog.getContentPane().add(statusPane);

      // add a listener to status dialog to direct the focus to Cancel
      statusDialog.addWindowListener(new WindowAdapter()
      {
        public void windowActivated(WindowEvent we)
        {

          // Once window gets focus, set initial focus
          statusPane.selectInitialValue();
        }

      });

      // add a listener to the statusPane to close the dialog
      // when cancel is clicked
      statusPane.addPropertyChangeListener(new PropertyChangeListener()
      {
        public void propertyChange(PropertyChangeEvent event)
        {
          if (statusDialog.isVisible() && event.getSource() == statusPane 
              && (event.getPropertyName().equals(JOptionPane.VALUE_PROPERTY) 
                  || event.getPropertyName().equals
                    (JOptionPane.INPUT_VALUE_PROPERTY)))
          {

            statusDialog.setVisible(false);
            statusDialog.dispose();
          }
        }

      });

      statusDialog.pack();
      statusDialog.setLocationRelativeTo(dialog);

      cancelClicked = false;

      Thread thread = new Thread()
      {
        public void run()
        {
          boolean perRecordUpdateFailed = !doPerRecordUpdate();

          while (preFetch != null &&!cancelClicked &&!perRecordUpdateFailed)
          {
            provInstance.addEntryToDB();

            /**
             * Take this out after testing is done **************
             */
            try
            {
              Thread.sleep(1000);
            }
            catch (Exception e) {}

            fillInNext();

            perRecordUpdateFailed = !doPerRecordUpdate();
          }

          if (!cancelClicked)
          {
            if (perRecordUpdateFailed)
            {
              statusDialog.dispose();
            }
            else
            {
              provInstance.addEntryToDB();
              statusDialog.dispose();
              JOptionPane.showMessageDialog(dialog, 
                      "All data has be added.\nClick OK to close the window.");

              closeAndCleanUp();
            }
          }
        }

      };

      thread.start();

      statusDialog.show();

      // Either cancel was clicked or all the data has been added
      // when we are here.  Either way set cancelClicked to be true because if
      // cancel was clicked we need to, if not to does not affect anything.
      cancelClicked = true;
    }
  }

}

