/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.template;


import javax.swing.JPanel;

/**
 * Provides an interface to objects that provide data.  Data is stored as a
 * list of elements that is accessed element by element.  A cursor is
 * maintained that points the DataProvider to the current element.
 * Elements can only be accessed sequentially and only in a forward direction.
 *
 * @author Matt Naish
 */
public abstract class DataProvider implements Cloneable
{

  /**
   * Save the values from the user interface into internal storage.
   */
  public abstract void saveUIValues() throws IllegalArgumentException;

  /**
   * Clear all the values from user interface.  Does not destory the UI but
   * just resets the UI to default values
   */
  public abstract void clearUI();

  /**
   * Populate the UI with the values from internal storage
   */
  public abstract void populateUI();

  /**
   * Return the name of this DataProvider.
   */
  public abstract String getName();

  /**
   * Return a clone of this object.  Does the default shallow copy so that
   * all primitive types are actually copied but only references to objects
   * are copied.
   */
  public Object clone()
  {
    try
    {
      return super.clone();
    }
    catch (CloneNotSupportedException e) {}

    return null;
  }

  /**
   * Advance the cursor to the next row of data.  Initially the cursor
   * is positioned before the first row so next() must be called before the
   * first row of data can be accessed.
   * @return true if there is more data (another row) of data to be read,
   * false otherwise
   */
  public abstract boolean next() throws NestedException;

  /**
   * Get the current data element.
   */
  public abstract String getData() throws NestedException;

  /**
   * Create0s and returns the user interface that is used to input
   * configuration details for the DataProvider.  For example, when using
   * a database query the user has to enter the dbname, filter, fields and
   * orderType <P>
   * Also, this user interface should have a "Test" button which will then
   * display a sample of the data, perhaps just a few rows.
   */
  public abstract JPanel getUI();

  /**
   * Connect to the source of the data (acquire resources)
   */
  public abstract void connect() throws NestedException;

  /**
   * Releases the resources acquired by this DataProvider
   */
  public abstract void close() throws NestedException;

}

