/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.template;

import javax.swing.JTextField;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JLabel;
import java.io.IOException;

import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;
import org.vovida.prov.psLib.PSDBResult;

/**
 * @author Matt Naish
 */
public class DatabaseDataProvider extends DataProvider
{

  /**
   * Return this many records at a time
   */
  private static int BUCKET_SIZE = 20;

  /**
   * Timeout for queries (ms).  Currently set to 10 minutes.
   */
  private static final long TIMEOUT = 600000;

  private String domain;
  private String filter;
  private String field;

  private ProvisioningInterface pInterface;
  private PSDBResult psDBResult;
  private int numRecords;
  private int index;
  private int bucketIndex;
  private String token;

  private JPanel panel;
  private JTextField domainTextField;
  private JTextField filterTextField;
  private JTextField fieldTextField;
  private JButton testButton;

  public DatabaseDataProvider() {}

  public DatabaseDataProvider(String domain, String filter, String field)
  {
    this.domain = domain;
    this.filter = filter;
    this.field = field;
  }

  public void saveUIValues() throws IllegalArgumentException
  {
    if (panel != null)
    {
      domain = domainTextField.getText();
      filter = filterTextField.getText();
      field = fieldTextField.getText();

      if (domain.equals("") || filter.equals("") || field.equals(""))
      {
        throw new IllegalArgumentException
          ("Neither domain, filter nor field can be blank.");
      }
    }
  }

  public void clearUI()
  {
    if (panel != null)
    {
      domainTextField.setText(null);
      filterTextField.setText(null);
      fieldTextField.setText(null);
    }
  }

  public void populateUI()
  {
    if (panel != null)
    {
      domainTextField.setText(domain);
      filterTextField.setText(filter);
      fieldTextField.setText(field);
    }
  }

  public String getDomain()
  {
    return domain;
  }

  public String getFilter()
  {
    return filter;
  }

  public String getField()
  {
    return field;
  }

  public String getName()
  {
    return "Database";
  }

  public boolean next() throws NestedException
  {
    if (psDBResult != null)
    {
      if (++index < numRecords)
      {
        if (++bucketIndex >= psDBResult.getNumRecords())
        {
          int diff = numRecords - index;
          int endIndex = diff < BUCKET_SIZE ? index + diff - 1 
                         : index + BUCKET_SIZE - 1;

          try
          {
            psDBResult = pInterface.sessionGetDBResult(token, index, 
                    endIndex, new String[]
            {
              field
            });
          }
          catch (Exception e)
          {
            throw new NestedException(e);
          }

          bucketIndex = 0;
        }

        return true;
      }

      return false;
    }

    return ++index < psDBResult.getNumRecords();
  }

  public String getData()
  {
    return (String) psDBResult.getFieldValue(bucketIndex, field);
  }

  public JPanel getUI()
  {
    if (panel == null)
    {
      panel = new JPanel();

      panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));

      JPanel domainPanel = new JPanel();

      domainPanel.add(new JLabel("Domain:"));
      domainPanel.add(domainTextField = new JTextField(domain, 15));

      JPanel filterPanel = new JPanel();

      filterPanel.add(new JLabel("Filter:"));
      filterPanel.add(filterTextField = new JTextField(filter, 15));

      JPanel fieldPanel = new JPanel();

      fieldPanel.add(new JLabel("Field:"));
      fieldPanel.add(fieldTextField = new JTextField(field, 15));

      testButton = new JButton("Test");

      // enable when a test case routine has been implemented
      testButton.setEnabled(false);

      panel.add(domainPanel);
      panel.add(filterPanel);
      panel.add(fieldPanel);
      panel.add(testButton);

    }

    return panel;
  }

  public void connect() throws NestedException
  {
    try
    {
      int endIndex;

      pInterface = ProvisioningInterface.getInstance();
      token = pInterface.sessionOpen(TIMEOUT);
      numRecords = pInterface.sessionSearchDB(token, domain, filter, 
              new String[]
      {
        field
      }, null, null);
      endIndex = numRecords > BUCKET_SIZE ? BUCKET_SIZE - 1 : numRecords - 1;

      index = -1;
      bucketIndex = -1;

      if (endIndex > 0)
      {
        psDBResult = pInterface.sessionGetDBResult(token, 0, endIndex, 
                new String[]
        {
          field
        });
      }
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }

  public void close() throws NestedException
  {
    try
    {
      pInterface.sessionClose(token);

      pInterface = null;
      psDBResult = null;
      token = "";
      index = -1;
      bucketIndex = -1;
      numRecords = -1;
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }

  public static void main(String[] args) throws NestedException
  {
    ProvisioningInterface.templateTesting = true;

    doTest(20);
    doTest(15);
    doTest(13);
    doTest(52);
    doTest(51);
    doTest(53);
  }

  private static void doTest(int bucketSize) throws NestedException
  {
    System.out.println("\n\n******Bucket Size " + bucketSize + "*******");

    BUCKET_SIZE = bucketSize;

    DatabaseDataProvider ddp = new DatabaseDataProvider("domain", "filter", 
            "field");

    ddp.connect();

    while (ddp.next());

    ddp.close();
  }

}

