/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.template;

import javax.swing.DefaultComboBoxModel;
import javax.swing.BorderFactory;
import javax.swing.JTextField;
import javax.swing.JComboBox;
import javax.swing.JCheckBox;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JLabel;
import javax.swing.JFrame;
import java.awt.event.FocusListener;
import java.awt.event.ItemListener;
import java.awt.event.ItemEvent;
import java.awt.BorderLayout;
import java.awt.GridLayout;


/**
 * <P>RecordUpdater is used to control per record updates while doing a
 * bulk add.
 * The object is used to shows a gui to handle data entry stores the results
 * in a RecordUpdaterData object.</P>
 * 
 * <P>A per record update could be used for say: if a whole wack of endpoints
 * stored by MAC address are in the database.  These endpoints have not been
 * assigned to a user as of yet.  As administrator is adding users to the
 * database and assigning endpoints (MACs) to them, we want to mark the endpoint
 * data as being assigned.  Because we are using a generic backend we cannot
 * rely on the relational functionality of a relational database, so we must
 * explicitly mark the endpoint record as being assigned.  This could be a flag
 * in the endpoint record which states 'assigned'.</P>
 *
 * @author Matt Naish
 */
public class RecordUpdater
{
  private static final int TF_SIZE = 8;

  private JPanel panel;
  private JCheckBox activeCheckBox;
  private JTextField updateTextField;
  private JTextField toBeTextField;
  private JTextField inTextField;
  private JTextField whereTextField;
  private JComboBox isComboBox;
  private FocusListener fl;

  private RecordUpdaterData data;

  /**
   * Lonely default constructor
   */
  public RecordUpdater() {}

  /**
   * Construct a object with the supplied data
   */
  public RecordUpdater(RecordUpdaterData data)
  {
    setData(data);
  }

  /**
   * Set the internal data store to the supplied value.  Set all the UI pieces
   * as well (if they are created)
   */
  public void setData(RecordUpdaterData data)
  {
    this.data = data;

    if (panel != null)
    {
      activeCheckBox.setSelected(data.active);
      updateTextField.setText(data.update);
      toBeTextField.setText(data.toBe);
      inTextField.setText(data.in);
      whereTextField.setText(data.where);
      isComboBox.setModel(new DefaultComboBoxModel(data.isValues));

      if (data.selectedIsValue != null)
      {
        boolean found = false;
        int i = 0;

        for (; !found && i < data.isValues.length; i++)
        {
          found = data.selectedIsValue.equals(data.isValues[i].getDBName());
        }

        if (found)
        {
          isComboBox.setSelectedIndex(i - 1);
        }
      }
    }
  }

  /**
   * Return the internal data store.
   */
  public RecordUpdaterData getData()
  {
    return data;
  }

  /**
   * Save the values from the UI into the internal RecordUpdaterData
   */
  public void saveUIValues()
  {
    if (panel != null)
    {
      data.active = activeCheckBox.isSelected();
      data.update = updateTextField.getText();
      data.toBe = toBeTextField.getText();
      data.in = inTextField.getText();
      data.where = whereTextField.getText();
      data.selectedIsValue = 
        ((DataProviderWrapper) isComboBox.getSelectedItem()).getDBName();

      if (data.active &&!data.dataValid())
      {
        throw new IllegalArgumentException("If 'Per Record Update' is active then no fields should be blank.");
      }
    }
  }

  /**
   * Clear all the values from the UI
   */
  public void clearUI()
  {
    updateTextField.setText(null);
    toBeTextField.setText(null);
    inTextField.setText(null);
    whereTextField.setText(null);
    ((DefaultComboBoxModel) isComboBox.getModel()).removeAllElements();
  }

  /**
   * Return the GUI.  Creates it the first time, after that returns a reference
   * to the already created one.
   */
  public JPanel getUI()
  {
    if (panel == null)
    {
      final JLabel updateLabel = new JLabel("Update");
      final JLabel toBeLabel = new JLabel(" To Be");
      final JLabel inLabel = new JLabel(" In");
      final JLabel whereLabel = new JLabel(" Where");
      final JLabel isLabel = new JLabel(" Is");
      final JPanel centerPanel = new JPanel();
      final JPanel compPanel0 = new JPanel();
      final JPanel compPanel1 = new JPanel();

      panel = new JPanel();

      panel.setBorder(BorderFactory.createTitledBorder("Per Record Update"));

      activeCheckBox = new JCheckBox("Active", false);

      activeCheckBox.addItemListener(new ItemListener()
      {
        public void itemStateChanged(ItemEvent ie)
        {
          boolean state = ie.getStateChange() == ItemEvent.SELECTED;

          updateTextField.setEnabled(state);
          updateLabel.setEnabled(state);
          toBeTextField.setEnabled(state);
          toBeLabel.setEnabled(state);
          inTextField.setEnabled(state);
          inLabel.setEnabled(state);
          whereTextField.setEnabled(state);
          whereLabel.setEnabled(state);
          isComboBox.setEnabled(state);
          isLabel.setEnabled(state);
        }

      });

      updateTextField = new JTextField(TF_SIZE);
      toBeTextField = new JTextField(TF_SIZE);
      inTextField = new JTextField(TF_SIZE);
      whereTextField = new JTextField(TF_SIZE);
      isComboBox = new JComboBox();

      activeCheckBox.setSelected(false);
      addFocusListener(fl);

      compPanel0.add(activeCheckBox);
      compPanel0.add(updateLabel);
      compPanel0.add(updateTextField);
      compPanel0.add(toBeLabel);
      compPanel0.add(toBeTextField);
      compPanel0.add(inLabel);
      compPanel0.add(inTextField);
      compPanel1.add(whereLabel);
      compPanel1.add(whereTextField);
      compPanel1.add(isLabel);
      compPanel1.add(isComboBox);

      centerPanel.setLayout(new GridLayout(2, 1));
      centerPanel.add(compPanel0);
      centerPanel.add(compPanel1);

      panel.setLayout(new BorderLayout());
      panel.add(BorderLayout.WEST, activeCheckBox);
      panel.add(BorderLayout.CENTER, centerPanel);
    }

    return panel;
  }

  /**
   * Add a focus listener to all the gui components.  Used by TemplateBuilder
   * to find out if any of the components had and then lost the focus (The
   * method of determining in the data is dirty
   */
  public void addFocusListener(FocusListener fl)
  {
    this.fl = fl;

    if (panel != null)
    {
      activeCheckBox.addFocusListener(fl);
      updateTextField.addFocusListener(fl);
      toBeTextField.addFocusListener(fl);
      inTextField.addFocusListener(fl);
      whereTextField.addFocusListener(fl);
      isComboBox.addFocusListener(fl);
    }
  }

  public static void main(String[] args)
  {
    RecordUpdater ru = new RecordUpdater();
    JFrame frame = new JFrame();

    frame.setContentPane(ru.getUI());
    frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    frame.pack();
    frame.show();
  }

}

