/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.template;

import javax.swing.DefaultComboBoxModel;
import javax.swing.ListSelectionModel;
import javax.swing.BorderFactory;
import javax.swing.JScrollPane;
import javax.swing.JOptionPane;
import javax.swing.JComboBox;
import javax.swing.BoxLayout;
import javax.swing.JDialog;
import javax.swing.JButton;
import javax.swing.JPanel;
import javax.swing.JList;
import javax.swing.JLabel;
import javax.swing.Box;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionListener;
import javax.swing.event.ListSelectionEvent;

import java.awt.Frame;
import java.awt.Dialog;
import java.awt.BorderLayout;
import java.awt.event.WindowListener;
import java.awt.event.ActionListener;
import java.awt.event.FocusListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.event.ActionEvent;
import java.awt.event.FocusEvent;

/**
 * This is the main hub for building templates.  This class is the gui for
 * template building and also the logic for creating, manipulating, saving
 * deleting templates.  It uses a TemplateLoader to load the data for a template
 * (field/DataProvider info and the bar code field).  It then acts upon
 * DataProviderWrapper objects (produced from TemplateLoader).  It then uses
 * the TemplateLoader to save the DataProviderWrapper objects back to the
 * template.
 *
 * @author Matt Naish
 */
public class TemplateBuilder extends JDialog implements ListSelectionListener, 
        ActionListener
{
  private static final String DIALOG_TITLE = "Template Builder";
  private static final String CLICK_TO_SELECT = "<click to select>";
  private static final String PROTOTYPE_TEXT = "Just some text that is nice.";
  private static final String FIELD_ASSIGNMENTS = "Field Assignments";
  private static final String SAVE = "Save";
  private static final String NEW = "New";
  private static final String COPY = "Copy";
  private static final String DELETE = "Delete";
  private static final String FIELDS = "Fields";
  private static final String BRACKETS = "<>";
  private static final String SELECT_FIELD = "Select a field to continue.";
  private static final String BAR_CODE_LABEL = "Bar Code Field";
  private static final String CLOSE = "Close";
  private static final String NOTHING = "<nothing>";

  // gui controls
  private JButton saveButton;
  private JButton newButton;
  private JButton copyButton;
  private JButton deleteButton;
  private JButton closeButton;

  private JComboBox nameComboBox;
  private JComboBox barCodeComboBox;

  private JList fieldList;

  private RecordUpdater recordUpdater;

  private JPanel assignPanel;
  private JComboBox dataEntryWidgetsComboBox;
  private JPanel nothingSelectedPanel;

  // flag used to identify if any data has been changed
  private boolean dataDirty = false;

  // the TemplateLoader used to load and save the template
  private TemplateLoader templateLoader;

  // the 'live' data fields.  These are the fields that have a DataProvider
  // assigned to them.
  private DataProviderWrapper[] fields;

  // These are mainly used to display a gui, grab the input from a user,
  // and then have that user entered data copied across to a object in 'fields'
  private DataProvider[] dataEntryWidgets;

  // Place holders
  private DataProviderWrapper lastSelectedField;
  private DataProvider lastSelectedDataEntryWidget;
  private int lastSelectedNameComboBoxIndex = 0;

  /**
   * @param parent the dialog parent for this dialog
   * @param templateLoader the TemplateLoader used to load and save templates.
   */
  public TemplateBuilder(Dialog parent, TemplateLoader templateLoader)
  {
    super(parent, DIALOG_TITLE, true);

    initialize(templateLoader);
  }

  /**
   * @param parent the frame parent for this dialog
   * @param templateLoader the TemplateLoader used to load and save templates.
   */
  public TemplateBuilder(Frame parent, TemplateLoader templateLoader)
  {
    super(parent, DIALOG_TITLE, true);

    initialize(templateLoader);
  }

  /**
   * Private helper to the constructors.  Mostly initializes the UI.
   * @param templateLoader the TemplateLoader used to load and save templates.
   */
  private void initialize(TemplateLoader templateLoader)
  {
    dataEntryWidgets = new DataProvider[]
    {
      new DefaultDataProvider(), new AutoNumberDataProvider(), 
      new DatabaseDataProvider(), 
    };
    this.templateLoader = templateLoader;

    createUI();

    /*
     * addWindowListener(new WindowAdapter() {
     * public void windowClose(WindowEvent we) {
     * if (promptUserForSave() != JOptionPane.CANCEL_OPTION) {
     * dispose();
     * }
     * }
     * });
     */
  }

  /**
   * Does, guess what? Creates the GUI!
   */
  private void createUI()
  {
    JPanel mainPanel = new JPanel();
    JPanel topPanel = new JPanel();
    JPanel buttonPanel = new JPanel();

    JPanel midPanel = new JPanel();
    JPanel listPanel = new JPanel();

    assignPanel = new JPanel();

    JScrollPane fieldScrollPane;

    JPanel bottomPanel = new JPanel();
    JPanel fieldComboPanel = new JPanel();
    JPanel closePanel = new JPanel();

    FocusListener fl;

    mainPanel.setLayout(new BorderLayout());

    // Top panel
    saveButton = new JButton(SAVE);
    newButton = new JButton(NEW);
    copyButton = new JButton(COPY);
    deleteButton = new JButton(DELETE);

    saveButton.addActionListener(this);
    newButton.addActionListener(this);
    copyButton.addActionListener(this);
    deleteButton.addActionListener(this);
    buttonPanel.add(saveButton);
    buttonPanel.add(Box.createGlue());
    buttonPanel.add(newButton);
    buttonPanel.add(Box.createGlue());
    buttonPanel.add(copyButton);
    buttonPanel.add(Box.createGlue());
    buttonPanel.add(deleteButton);

    if (templateLoader.getAvailableTemplates() != null)
    {
      nameComboBox = new JComboBox(templateLoader.getAvailableTemplates());

      nameComboBox.insertItemAt(CLICK_TO_SELECT, 0);
    }
    else
    {
      nameComboBox = new JComboBox(new Object[]
      {
        CLICK_TO_SELECT
      });
    }

    nameComboBox.setSelectedIndex(0);
    nameComboBox.addActionListener(this);

    topPanel.add(nameComboBox);
    topPanel.add(buttonPanel);
    mainPanel.add(BorderLayout.NORTH, topPanel);

    // mid panel
    midPanel.setBorder(BorderFactory.createCompoundBorder
      (BorderFactory.createEmptyBorder(2, 2, 2, 2), 
            BorderFactory.createCompoundBorder(BorderFactory.createTitledBorder
              (FIELD_ASSIGNMENTS), 
            BorderFactory.createEmptyBorder(5, 5, 5, 5))));

    midPanel.setLayout(new BoxLayout(midPanel, BoxLayout.X_AXIS));

    fieldList = new JList();

    fieldList.setPrototypeCellValue(PROTOTYPE_TEXT);
    fieldList.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
    fieldList.addListSelectionListener(this);

    fieldScrollPane = new JScrollPane(fieldList, 
            JScrollPane.VERTICAL_SCROLLBAR_ALWAYS, 
            JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);

    listPanel.setLayout(new BoxLayout(listPanel, BoxLayout.Y_AXIS));
    listPanel.add(fieldScrollPane);
    listPanel.setBorder(BorderFactory.createTitledBorder
      (BorderFactory.createEmptyBorder(), 
            FIELDS));

    // add a focus listener to fieldList.  This acts an dirty data watcher.
    fl = new FocusListener()
    {
      public void focusGained(FocusEvent e) {}

      public void focusLost(FocusEvent e)
      {
        TemplateBuilder.this.dataDirty = true;
      }

    };

    fieldList.addFocusListener(fl);

    nothingSelectedPanel = new JPanel();

    nothingSelectedPanel.add(new JLabel(SELECT_FIELD));
    assignPanel.setBorder(BorderFactory.createTitledBorder(BRACKETS));
    assignPanel.setLayout(new BorderLayout());
    assignPanel.add(nothingSelectedPanel);

    midPanel.add(listPanel);
    midPanel.add(assignPanel);
    mainPanel.add(BorderLayout.CENTER, midPanel);

    barCodeComboBox = new JComboBox();

    barCodeComboBox.setBorder(BorderFactory.createCompoundBorder
      (BorderFactory.createTitledBorder(BAR_CODE_LABEL), 
            BorderFactory.createEmptyBorder(5, 5, 5, 5)));

    fieldComboPanel.add(barCodeComboBox);

    recordUpdater = new RecordUpdater();

    recordUpdater.addFocusListener(fl);

    closeButton = new JButton(CLOSE);

    closeButton.addActionListener(this);
    closePanel.setBorder(BorderFactory.createEmptyBorder(5, 5, 5, 5));
    closePanel.setLayout(new BoxLayout(closePanel, BoxLayout.Y_AXIS));
    closePanel.add(Box.createGlue());
    closePanel.add(closeButton);


    bottomPanel.setLayout(new BorderLayout());
    bottomPanel.add(BorderLayout.NORTH, recordUpdater.getUI());
    bottomPanel.add(BorderLayout.CENTER, fieldComboPanel);
    bottomPanel.add(BorderLayout.EAST, closePanel);
    mainPanel.add(BorderLayout.SOUTH, bottomPanel);

    // the following are pre built panels that will be used as the user
    // is selecting and deselecting items from the fieldList
    dataEntryWidgetsComboBox = new JComboBox();

    dataEntryWidgetsComboBox.addItem(NOTHING);

    for (int i = 0; i < dataEntryWidgets.length; i++)
    {
      dataEntryWidgetsComboBox.addItem(dataEntryWidgets[i].getName());
    }

    dataEntryWidgetsComboBox.addActionListener(this);

    // finish up!
    setContentPane(mainPanel);
    pack();
  }

  /**
   * Called whenever the user changes a selection from the field list.
   * Handles saving any user entered data to the appropriate field.
   */
  public void valueChanged(ListSelectionEvent lse)
  {

    // only do something if this is the final ListSelectionEvent in a chain fo
    // them.
    DataProviderWrapper selectedField = 
      (DataProviderWrapper) fieldList.getSelectedValue();

    if (!lse.getValueIsAdjusting() && (selectedField != lastSelectedField))
    {

      // first, save the user entered values from the last shown dataEntryWidget
      // to the last shown field (when this method is called the new field
      // has already been selected).
      int ignoreWarning = JOptionPane.YES_OPTION;

      try
      {
        saveUIValues(lastSelectedField, lastSelectedDataEntryWidget);
      }
      catch (IllegalArgumentException iae)
      {

        // There was an error saving the values from the UI.
        String mesg = iae.getMessage() + "\nIgnore?";

        ignoreWarning = JOptionPane.showConfirmDialog(this, mesg, "Warning", 
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
      }

      if (ignoreWarning == JOptionPane.NO_OPTION)
      {
        fieldList.setSelectedValue(lastSelectedField, true);

        return;
      }

      // Is the "nothing selected" panel showning or not?
      if (fieldList.getSelectedValue() != null)
      {

        // Not

        for (int i = 0; i < dataEntryWidgets.length; i++)
        {

          // clear all the user entered data from the widgets.  Don't worry,
          // the appropriate data has already been saved
          // (look above at saveUIValues()).
          dataEntryWidgets[i].clearUI();
        }

        ((TitledBorder) assignPanel.getBorder()).setTitle
          (selectedField.toString());

        if (assignPanel.getComponentCount() > 0 
            && assignPanel.getComponent(0) == dataEntryWidgetsComboBox)
        {

          // the last panel shown already had a the dataEntryWidgetComboBox
          // showning.  Leave it and just remove the dataEntryWidget.
          if (assignPanel.getComponentCount() > 1)
          {
            assignPanel.remove(1);
          }
        }
        else
        {

          // the last panel shown showed the "Select a field to continue." label
          assignPanel.removeAll();
          assignPanel.add(BorderLayout.NORTH, dataEntryWidgetsComboBox);
        }

        if (selectedField.getDataProvider() == null)
        {
          dataEntryWidgetsComboBox.setSelectedIndex(0);
        }
        else
        {

          // copy the selected field into the array of dataEntryWidgets,
          // show its gui, and populate the gui with its data.
          int index = 
            ((DefaultComboBoxModel) dataEntryWidgetsComboBox.getModel())
              .getIndexOf(selectedField.getDataProvider().getName());

          // -1 because "<nothing>" is the first item in the combo box
          dataEntryWidgets[index - 1] = selectedField.getDataProvider();

          dataEntryWidgetsComboBox.setSelectedIndex(index);
          selectedField.getDataProvider().populateUI();
        }

        assignPanel.validate();
        assignPanel.repaint();

        lastSelectedField = selectedField;
      }
      else
      {

        // Show the "Nothing Selected" panel
        assignPanel.removeAll();
        ((TitledBorder) assignPanel.getBorder()).setTitle("<>");
        assignPanel.add(nothingSelectedPanel);
        assignPanel.validate();
        assignPanel.repaint();

        lastSelectedField = null;
      }
    }
  }

  /**
   * Saves the values from displayed ui for a field to the actual field.
   */
  private void saveUIValues(DataProviderWrapper field, 
          DataProvider dataEntryWidget) throws IllegalArgumentException
  {
    if (field != null)
    {
      if (dataEntryWidget != null)
      {
        dataEntryWidget.saveUIValues();
        field.setDataProvider((DataProvider) dataEntryWidget.clone());
      }
      else
      {
        field.setDataProvider(null);
      }
    }
  }

  /**
   * Called when a user clicks on a button or changes the selection of
   * one of the dataEntryWidgets.
   */
  public void actionPerformed(ActionEvent ae)
  {
    if (ae.getSource() == dataEntryWidgetsComboBox)
    {
      dataEntryWidgetClicked();
    }
    else if (ae.getSource() == nameComboBox)
    {
      nameComboBoxClicked();
    }
    else if (ae.getSource() == barCodeComboBox)
    {
      dataDirty = true;
    }
    else if (ae.getSource() == saveButton)
    {
      saveClicked();
    }
    else if (ae.getSource() == newButton)
    {
      newClicked();
    }
    else if (ae.getSource() == copyButton)
    {
      copyClicked();
    }
    else if (ae.getSource() == deleteButton)
    {
      deleteClicked();
    }
    else if (ae.getSource() == closeButton)
    {
      int resp = promptUserForSave();

      if (resp != JOptionPane.CANCEL_OPTION)
      {
        dispose();
      }
    }
  }

  /**
   * Called when the selection in the ComboBox showing all of the
   * possible DataEntryWidgets has changed.  DataEntryWidgets are just the
   * guis of all the availabel DataProviders.
   */
  private void dataEntryWidgetClicked()
  {

    // replace, if shown, the last dataEntryWidget with the newly selected one.
    if (lastSelectedDataEntryWidget != null)
    {
      assignPanel.remove(lastSelectedDataEntryWidget.getUI());
    }

    if (dataEntryWidgetsComboBox.getSelectedIndex() == 0)
    {
      lastSelectedDataEntryWidget = null;
    }
    else
    {
      assignPanel.add(dataEntryWidgets[dataEntryWidgetsComboBox.getSelectedIndex() - 1].getUI());

      lastSelectedDataEntryWidget = 
        dataEntryWidgets[dataEntryWidgetsComboBox.getSelectedIndex() - 1];
    }

    assignPanel.validate();
    assignPanel.repaint();
  }

  /**
   * Called when the selection in the ComboBox showing all of the available
   * templates has changed.
   */
  private void nameComboBoxClicked()
  {
    if (nameComboBox.getSelectedIndex() != lastSelectedNameComboBoxIndex)
    {
      if (promptUserForSave() != JOptionPane.CANCEL_OPTION)
      {
        if (nameComboBox.getSelectedIndex() == 0)
        {

          // the user has selecting "<nothing>"
          fields = null;

          fieldList.setListData(new Object[0]);
          ((DefaultComboBoxModel) barCodeComboBox.getModel())
            .removeAllElements();
          recordUpdater.clearUI();
        }
        else if (nameComboBox.getSelectedIndex() > 0)
        {

          // the user has selected a template
          String barCodeField;

          try
          {
            templateLoader.load(nameComboBox.getSelectedItem().toString());

            // load the field list
            fields = templateLoader.getFields();

            fieldList.setListData(fields);

            // load the bar code combo box
            barCodeComboBox.removeAllItems();
            barCodeComboBox.addItem(NOTHING);

            for (int i = 0; i < fields.length; i++)
            {
              barCodeComboBox.addItem(fields[i].toString());
            }

            // set to "Nothing" selected as a default
            barCodeField = templateLoader.getBarCodeField();

            if (barCodeField != null)
            {
              boolean found = false;
              int i = 0;

              for (; !found && i < fields.length; i++)
              {
                found = barCodeField.equals(fields[i].getDBName());
              }

              if (found)
              {
                barCodeComboBox.setSelectedIndex(i);
              }
            }

            // load the recordUpdater Info
            recordUpdater.setData(templateLoader.getRecordUpdaterData());
          }
          catch (NestedException ne)
          {
            JOptionPane.showMessageDialog(this, ne.getMessage(), "Error", 
                    JOptionPane.ERROR_MESSAGE);
          }
        }

        lastSelectedNameComboBoxIndex = nameComboBox.getSelectedIndex();
      }
      else
      {

        // the user decided to cancel switching to a new template after being
        // prompted to save.
        nameComboBox.setSelectedIndex(lastSelectedNameComboBoxIndex);
      }
    }
  }

  /**
   * A utility method that is used by several functions to possibly prompt
   * the user to save his current work before moving on.  Used when actions
   * could possibly result in loss of entered information if a save is not
   * performed.  Only prompts if the current data is dirty.
   * @returns what the user responeded (JOptionPane.YES_OPTION |
   * JOptionPane.NO_OPTION | JOptionPane.CANCEL_OPTION) or
   * JOptionPane.YES_OPTION if the data is not dirty.
   */
  public int promptUserForSave()
  {
    if (dataDirty && lastSelectedNameComboBoxIndex != 0)
    {
      int resp = JOptionPane.showConfirmDialog(this, 
              "Do you want to save your changes to " 
              + nameComboBox.getItemAt(lastSelectedNameComboBoxIndex) + "?");

      if (resp == JOptionPane.YES_OPTION)
      {

        // save the data.  saveData() sets dataDirty = false;
        saveData();
      }
      else if (resp == JOptionPane.NO_OPTION)
      {
        dataDirty = false;
      }

      return resp;
    }
    else
    {

      // the data did not need a save.  All systems go.
      return JOptionPane.YES_OPTION;
    }
  }

  /**
   * The save button was clicked
   */
  private void saveClicked()
  {
    if (nameComboBox.getSelectedIndex() == 0)
    {
      JOptionPane.showMessageDialog(this, "No template selected for save.");
    }
    else
    {
      saveData();
    }
  }


  /**
   * Save the current data.  This means save the currently visible data from
   * the UI to a its assoicated DataProvider, and then do a save on all the
   * data to the PS.
   */
  private void saveData()
  {

    // save ui values
    try
    {
      if (fieldList.getSelectedIndex() != -1)
      {
        int index = dataEntryWidgetsComboBox.getSelectedIndex();

        if (index == 0)
        {
          saveUIValues((DataProviderWrapper) fieldList.getSelectedValue(), 
                       null);
        }
        else
        {
          saveUIValues((DataProviderWrapper) fieldList.getSelectedValue(), 
             dataEntryWidgets[dataEntryWidgetsComboBox.getSelectedIndex() - 1]);
        }
      }
    }
    catch (IllegalArgumentException iae)
    {
      JOptionPane.showMessageDialog(this, iae.getMessage(), "Warning", 
              JOptionPane.WARNING_MESSAGE);
    }

    String barCodeField;

    if (barCodeComboBox.getSelectedIndex() == 0)
    {
      barCodeField = null;
    }
    else
    {
      barCodeField = 
        fields[barCodeComboBox.getSelectedIndex() - 1].getDBName();
    }

    try
    {
      recordUpdater.saveUIValues();
    }
    catch (IllegalArgumentException iae)
    {
      JOptionPane.showMessageDialog(this, iae.getMessage(), "Warning", 
              JOptionPane.WARNING_MESSAGE);
    }

    try
    {
      templateLoader.save(barCodeField);

      dataDirty = false;
    }
    catch (NestedException ne)
    {
      JOptionPane.showMessageDialog(this, "Error", ne.getMessage(), 
              JOptionPane.ERROR_MESSAGE);
    }
  }

  /**
   * The new button was clicked.
   */
  private void newClicked()
  {
    String[] definitions = templateLoader.getDefinitions();

    if (definitions.length > 0)
    {
      String definitionName = (String) JOptionPane.showInputDialog(this, 
              "Select a definition file on which to base the new template.", 
              "Select", JOptionPane.QUESTION_MESSAGE, null, definitions, 
              null);

      if (definitionName != null)
      {

        // user did not click 'cancel'.  Continue
        String newTemplateName = (String) JOptionPane.showInputDialog(this, 
                "Enter a name for the new template.");

        if (newTemplateName != null)
        {

          // user did not click 'cancel'.  Continue
          if (promptUserForSave() != JOptionPane.CANCEL_OPTION)
          {
            try
            {
              templateLoader.createNew(newTemplateName, definitionName);
              nameComboBox.setModel(new DefaultComboBoxModel
                (templateLoader.getAvailableTemplates()));
              nameComboBox.insertItemAt(NOTHING, 0);
              nameComboBox.setSelectedItem(newTemplateName);
            }
            catch (NestedException ne)
            {
              JOptionPane.showMessageDialog(this, "Error", ne.getMessage(), 
                      JOptionPane.ERROR_MESSAGE);
            }
          }
        }
      }
    }
    else
    {
      JOptionPane.showMessageDialog(this, 
              "No XML defintion files available.  " 
              + "'New' operation cannot be completed.");
    }
  }

  /**
   * The copy button was clicked
   */
  private void copyClicked()
  {
    if (nameComboBox.getSelectedIndex() == 0)
    {
      JOptionPane.showMessageDialog(this, 
              "Please select a template to copy.");
    }
    else
    {
      String newTemplateName = (String) JOptionPane.showInputDialog(this, 
              "Enter a name for the replica template.");

      if (newTemplateName != null)
      {

        // the user did not click 'cancel'
        if (promptUserForSave() != JOptionPane.CANCEL_OPTION)
        {
          try
          {
            templateLoader.copy(newTemplateName);
            nameComboBox.setModel(new DefaultComboBoxModel
              (templateLoader.getAvailableTemplates()));
            nameComboBox.insertItemAt(NOTHING, 0);
            nameComboBox.setSelectedItem(newTemplateName);
          }
          catch (NestedException ne)
          {
            JOptionPane.showMessageDialog(this, "Error", ne.getMessage(), 
                    JOptionPane.ERROR_MESSAGE);
          }
        }
      }
    }
  }

  /**
   * The delete button was clicked
   */
  private void deleteClicked()
  {
    if (nameComboBox.getSelectedIndex() == 0)
    {
      JOptionPane.showMessageDialog(this, 
              "Please select a template to delete.");
    }
    else
    {
      int resp = JOptionPane.showConfirmDialog(this, 
              "Are you sure you want to delete " 
              + nameComboBox.getSelectedItem() + "?", "Delete", 
              JOptionPane.YES_NO_OPTION);

      if (resp == JOptionPane.YES_OPTION)
      {
        try
        {
          templateLoader.delete();

          dataDirty = false;

          nameComboBox.removeItemAt(nameComboBox.getSelectedIndex());
        }
        catch (NestedException ne)
        {
          JOptionPane.showMessageDialog(this, "Error", ne.getMessage(), 
                  JOptionPane.ERROR_MESSAGE);
        }
      }
    }
  }

}

