/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.template;

import org.jdom.Element;

/**
 * Wraps a DataProvider and an XML element together
 *
 * @author Matt Naish
 */
public class XMLDataProviderWrapper implements DataProviderWrapper
{
  protected static final String LABEL = "label";

  /**
   * The display name for this object
   */
  private String prettyName;

  /**
   * The database name for this object
   */
  private String dbName;

  /**
   * The XML element.  The object holds onto this, even when the data is
   * stored in a DataProvider because the Element may contain other bits of
   * information what we are not concerned with but should leave alone.
   * We can then easily change the data we are concerned with, leave the
   * rest alone, and someone else then still has an intact Element.
   */
  private Element element;

  /**
   * flag stating whether or not the DataProvider was ever set.  Needed
   * in saveDataProvider()
   */
  private boolean dpNeverSet = true;

  /**
   * The DataProvider
   */
  private DataProvider dp;

  /**
   * Create a DataProviderWrapper using the supplied Element for data.
   * Basically this object starts using the Element for its original data, then
   * when one the internal DataProvider is "got" or "set" for the first time,
   * the true source of data should be the DataProvider (it is (could be) more
   * up to date than the Element).
   */
  public XMLDataProviderWrapper(Element element)
  {
    this.element = element;

    prettyName = element.getAttributeValue(LABEL);

    if (prettyName == null || prettyName.equals(""))
    {
      throw new IllegalArgumentException("No '" + LABEL 
              + "' attribute found!");
    }

    dbName = element.getAttributeValue(XMLTemplateLoader.NAME);

    if (prettyName == null || prettyName.equals(""))
    {
      throw new IllegalArgumentException("No '" + XMLTemplateLoader.NAME 
              + "' attribute found!");
    }
  }

  /**
   * Return the internal Element object (original source of data)
   */
  public Element getElement()
  {
    return element;
  }

  /**
   * Return a the display (pretty) name of this object.
   */
  public String toString()
  {
    return prettyName;
  }

  /**
   * Return the database name of this object
   */
  public String getDBName()
  {
    return dbName;
  }

  /**
   * Return the internal DataProvider.  Create if necessary.
   */
  public DataProvider getDataProvider()
  {
    if (dpNeverSet)
    {
      dp = XMLDataProviderUtil.createDataProviderFromElement(element);
      dpNeverSet = false;
    }

    return dp;
  }

  /**
   * Set the internal DataProvider to be this
   */
  public void setDataProvider(DataProvider dp)
  {
    this.dp = dp;
  }

  /**
   * Save the data contained within the internal DataProvider to the internal
   * Element.
   */
  public void saveDataProviderData()
  {

    // ensure the DataProvider is created, if need be.
    if (dpNeverSet)
    {
      dp = XMLDataProviderUtil.createDataProviderFromElement(element);
      dpNeverSet = false;
    }

    XMLDataProviderUtil.saveDataProviderToElement(dp, element);
  }

}

