/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */
package org.vovida.prov.template;

import org.vovida.prov.psLib.ProvisioningInterface;
import org.vovida.prov.psLib.OperationException;
import org.vovida.prov.psLib.PSDBResult;
import org.vovida.prov.dynamicGui.TypeBuilder;
import org.vovida.prov.util.XMLProcessor;

import java.io.StringReader;
import java.io.IOException;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.List;
import java.util.Arrays;
import org.jdom.Element;
import org.jdom.Document;
import org.jdom.output.XMLOutputter;

/**
 * Loads and save template information to and from the Provisioning Server
 * using ProvisioningInterface.  The templates are internally stored as XML.
 *
 * @author Matt Naish
 */
public class XMLTemplateLoader implements TemplateLoader
{
  private static final String TEMPLATES_DOMAIN = "templates";
  private static final String NAME_FIELD = "name";
  private static final String DATA_FIELD = "data";
  private static final String PARENT_FIELD = "parent";

  protected static final String DBENTRY = "DBEntry";
  protected static final String NAME = "name";
  protected static final String LABEL = "Label";
  protected static final String VALUE = "Value";
  protected static final String DEFAULT = "Default";

  protected static final String AUTO_NUMBER = "AutoNumber";
  protected static final String START = "Start";
  protected static final String END = "End";

  protected static final String DATABASE = "Database";
  protected static final String DOMAIN = "Domain";
  protected static final String FILTER = "Filter";
  protected static final String FIELD = "Field";

  private static final String TEMPLATE = "Template";
  private static final String BAR_CODE_FIELD = "BarCodeField";

  private static final String RECORD_UPDATER = "RecordUpdater";
  private static final String ACTIVE = "active";
  private static final String UPDATE = "update";
  private static final String TOBE = "toBe";
  private static final String IN = "in";
  private static final String WHERE = "where";
  private static final String IS = "is";

  /**
   * The interface to provisioning server and thus the database.  Use to
   * load and save templates.
   */
  private ProvisioningInterface psInterface;

  // The details of the current template.  These fields are set in
  // either load() or save()
  private Document document;
  private String templateName;
  private DataProviderWrapper[] fields;
  private String barCodeField;
  private RecordUpdaterData recordUpdaterData;
  private String parent;

  /**
   * The names of all the templates available to the user
   */
  private String[] availableTemplates;

  /**
   * Connects to the template source and retrieves a list of available
   * templates.
   */
  public XMLTemplateLoader() throws NestedException
  {
    psInterface = ProvisioningInterface.getInstance();

    loadAvailableTemplates();
  }

  /**
   * Load the names of all the available templates.
   */
  private void loadAvailableTemplates() throws NestedException
  {
    PSDBResult result;
    Vector availableTemplatesVec = new Vector();

    // Get all the templates
    try
    {
      String filter = "(name=\"*\")";

      result = psInterface.getRecordsInDomain(TEMPLATES_DOMAIN, filter, 
              new String[]
      {
        NAME_FIELD
      });

      // load all the available templates names
      for (int i = 0; i < result.getNumRecords(); i++)
      {
        availableTemplatesVec.add(result.getFieldValue(i, NAME_FIELD));
      }

      availableTemplates = 
        (String[]) availableTemplatesVec.toArray(new String[0]);

    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }


  /**
   * Returns a list of available templates
   */
  public String[] getAvailableTemplates()
  {
    return availableTemplates;
  }

  /**
   * Returns a list of the available definition files
   */

  // we could remove this and have apps just call TypeBuilder.getAllTypeNames()
  // directly.
  public String[] getDefinitions()
  {
    return TypeBuilder.getAllTypeNames();
  }

  /**
   * Loads a specified template.  The data from the template is then
   * available via the getFields() and getBarCodeField() methods
   */
  public void load(String templateName) throws NestedException
  {
    PSDBResult result;

    try
    {
      String filter = "(" + NAME_FIELD + "=\"" + templateName + "\")";
      String[] fieldList = new String[]
      {
        "name", "parent", "data"
      };

      result = psInterface.getRecordsInDomain(TEMPLATES_DOMAIN, filter, 
              fieldList);

      if (result.getNumRecords() == 1)
      {
        String xmlString;
        Document parentDoc;
        List templateFields;
        Element ruElem;

        this.templateName = templateName;

        xmlString = result.getFieldValue(0, DATA_FIELD);
        document = XMLProcessor.buildDocument(xmlString);

        templateFields = document.getRootElement().getChildren(DBENTRY);

        // this is the good way.  Use this when real database is used.
        parent = result.getFieldValue(0, PARENT_FIELD);

        // this line of code are the bad way.  Bad.
        // parent = document.getRootElement().getAttributeValue("parent");

        parentDoc = TypeBuilder.getType(parent).getDOM();

        if (parentDoc != null)
        {
          List parentFields;

          String[] templateFieldNames;
          String[] parentFieldNames;

          Vector dpWrappersVec;
          Hashtable dpWrappersHash;
          XMLDataProviderWrapper wrapper;
          Element defElem;
          Element templateElem;
          String labelValue;

          parentFields = new Vector();

          getChildrenRecursive(parentDoc.getRootElement(), DBENTRY, 
                  parentFields);

          templateFieldNames = new String[templateFields.size()];
          parentFieldNames = new String[parentFields.size()];

          // put all the entries from the template into a hash
          dpWrappersHash = new Hashtable();

          for (int i = 0; i < templateFields.size(); i++)
          {
            wrapper = 
              new XMLDataProviderWrapper((Element) templateFields.get(i));

            dpWrappersHash.put(wrapper.toString(), wrapper);
          }

          // now go through the definition/parent file. For each DBEntry
          // grab the template entry from the hash.  If an entry does not exist
          // in the hash create a new one.
          dpWrappersVec = new Vector();

          for (int i = 0; i < parentFields.size(); i++)
          {
            defElem = (Element) parentFields.get(i);
            labelValue = defElem.getChildText(LABEL);

            if (labelValue != null)
            {
              wrapper = 
                (XMLDataProviderWrapper) dpWrappersHash.get(labelValue);

              if (wrapper == null)
              {
                templateElem = new Element(DBENTRY);

                templateElem.addAttribute(NAME, 
                        defElem.getAttributeValue(NAME));
                templateElem.addAttribute(XMLDataProviderWrapper.LABEL, 
                        labelValue);
                document.getRootElement().addContent(templateElem);

                wrapper = new XMLDataProviderWrapper(templateElem);
              }

              dpWrappersVec.add(wrapper);
            }
          }

          fields = 
            (DataProviderWrapper[]) dpWrappersVec.toArray
              (new DataProviderWrapper[0]);
        }
        else
        {
          fields = new DataProviderWrapper[templateFields.size()];

          for (int i = 0; i < fields.length; i++)
          {
            fields[i] = 
              new XMLDataProviderWrapper((Element) templateFields.get(i));
          }
        }

        // find the bar code information
        barCodeField = document.getRootElement().getChildText(BAR_CODE_FIELD);

        if (barCodeField != null && barCodeField.length() == 0)
        {
          barCodeField = null;
        }

        ruElem = document.getRootElement().getChild(RECORD_UPDATER);

        if (ruElem != null)
        {
          recordUpdaterData = 
            new RecordUpdaterData((new Boolean(ruElem.getAttributeValue
              (ACTIVE)).booleanValue()), 
                  ruElem.getAttributeValue(UPDATE), 
                  ruElem.getAttributeValue(TOBE), 
                  ruElem.getAttributeValue(IN), 
                  ruElem.getAttributeValue(WHERE), 
                  ruElem.getAttributeValue(IS), fields);
        }
        else
        {
          recordUpdaterData = new RecordUpdaterData();
          recordUpdaterData.isValues = fields;
        }
      }
      else
      {

        // do error
        fields = null;
        templateName = null;
        barCodeField = null;
        parent = null;

        throw new NestedException(new Exception("Either more than one or no " 
                + "templates returned from database."));
      }
    }
    catch (Exception e)
    {
      e.printStackTrace();

      throw new NestedException(e);
    }
  }

  /**
   * A helper method to return all the children of the supplied Element that
   * match the name 'name'.  This method recurses all levels of children
   * (not just one deep).
   * @param Element the parent Element
   * @param name the name of children to find
   * @param list a list in which to place found children
   */
  private static void getChildrenRecursive(Element elem, String name, 
          List list)
  {
    if (elem.getName().equals(name))
    {
      list.add(elem);
    }
    else
    {
      Iterator iter = elem.getChildren().iterator();

      while (iter.hasNext())
      {
        getChildrenRecursive((Element) iter.next(), name, list);
      }
    }
  }

  /**
   * Return all the fields from the template identified by the last call to
   * load().
   * @throws IllegalStateException if no call to load has been made yet.
   */
  public DataProviderWrapper[] getFields() throws IllegalStateException
  {
    if (document == null)
    {
      throw new IllegalStateException("Must call load() first!");

    }

    return fields;
  }

  /**
   * Return the bar code field from the template identified by the last call
   * to load().
   * @throws IllegalStateException if no call to load has been made yet.
   */
  public String getBarCodeField()
  {
    if (document == null)
    {
      throw new IllegalStateException("Must call load() first!");

    }

    return barCodeField;
  }

  /**
   * Return all the information pertaining to a RecordUpdater object
   * identified by the last call to load().
   * @throws IllegalStateException if no call to load has been made yet.
   */
  public RecordUpdaterData getRecordUpdaterData()
  {
    if (document == null)
    {
      throw new IllegalStateException("Must call load() first!");

    }

    return recordUpdaterData;
  }


  /**
   * Return the name of the parent definition identified by the last call
   * to load()
   * @throws IllegalStateException if no call to load has been made yet.
   */
  public String getParent()
  {
    if (document == null)
    {
      throw new IllegalStateException("Must call load() first!");

    }

    return parent;
  }

  /**
   * Save the field information and barCodeField to the last template loaded
   * by load()
   */
  public void save(String barCodeField) throws NestedException
  {
    XMLOutputter outputter = new XMLOutputter("  ", true);

    outputter.setSuppressDeclaration(true);

    String xmlStr = "";
    Hashtable data;

    // first save the the bar code info into the document
    Element barCodeElem = document.getRootElement().getChild(BAR_CODE_FIELD);

    if (barCodeElem == null)
    {
      barCodeElem = new Element(BAR_CODE_FIELD);

      document.getRootElement().addContent(barCodeElem);
    }

    barCodeElem.setText(barCodeField);

    // now save the RecordUpdater data.
    Element ruElem = document.getRootElement().getChild(RECORD_UPDATER);

    if (ruElem == null)
    {
      ruElem = createNewRecordUpdaterElement();

      document.getRootElement().addContent(ruElem);
    }
    else
    {
      ruElem.getAttribute(ACTIVE).setValue(String.valueOf
        (recordUpdaterData.active));
      ruElem.getAttribute(UPDATE).setValue(recordUpdaterData.update);
      ruElem.getAttribute(TOBE).setValue(recordUpdaterData.toBe);
      ruElem.getAttribute(IN).setValue(recordUpdaterData.in);
      ruElem.getAttribute(WHERE).setValue(recordUpdaterData.where);
      ruElem.getAttribute(IS).setValue(recordUpdaterData.selectedIsValue);
    }

    // now save the info from the DataProviders into the document.
    for (int i = 0; i < fields.length; i++)
    {
      fields[i].saveDataProviderData();
    }

    try
    {

      // save the document to the PS
      xmlStr = outputter.outputString(document);

      String filter = "(" + NAME_FIELD + "=\"" + templateName + "\")";

      data = new Hashtable();

      data.put(DATA_FIELD, new String("<![CDATA[" + xmlStr + "]]>"));

      psInterface.setRecordsInDomain(TEMPLATES_DOMAIN, filter, data);
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }

  private Element createNewRecordUpdaterElement()
  {
    Element ruElem = new Element(RECORD_UPDATER);

    if (recordUpdaterData.update == null)
    {
      recordUpdaterData.update = "";
    }

    if (recordUpdaterData.toBe == null)
    {
      recordUpdaterData.toBe = "";
    }

    if (recordUpdaterData.in == null)
    {
      recordUpdaterData.in = "";
    }

    if (recordUpdaterData.where == null)
    {
      recordUpdaterData.where = "";
    }

    if (recordUpdaterData.selectedIsValue == null)
    {
      recordUpdaterData.selectedIsValue = "";

    }

    ruElem.addAttribute(ACTIVE, String.valueOf(recordUpdaterData.active));
    ruElem.addAttribute(UPDATE, recordUpdaterData.update);
    ruElem.addAttribute(TOBE, recordUpdaterData.toBe);
    ruElem.addAttribute(IN, recordUpdaterData.in);
    ruElem.addAttribute(WHERE, recordUpdaterData.where);
    ruElem.addAttribute(IS, recordUpdaterData.selectedIsValue);

    return ruElem;
  }

  /**
   * Create a new template with the given name and based on the supplied
   * definition file.
   */

  // this method creates a new template and then saves it back to the PS.
  // The method that calls this method (TemplateBuidler.newClicked()) then
  // reloads the 'new' template from the PS. This should be more effiicient
  // and make no network calls.  Running out of time, being laid off...
  public void createNew(String newTemplateName, 
          String definition) throws NestedException
  {
    Vector dbEntries = new Vector();
    String labelValue;
    Element dbEntryElem;
    Element defElem;
    Element templateElem;

    Element root = new Element(TEMPLATE);
    Element barCodeElem;
    Element recordUpdaterElem;

    document = new Document(root);

    recordUpdaterData = new RecordUpdaterData();
    templateName = newTemplateName;
    parent = definition;

    // add the name and parent name
    root.addAttribute(NAME_FIELD, templateName);
    root.addAttribute(PARENT_FIELD, parent);

    // create and add the bar code field (no data to add yet)
    barCodeElem = new Element(BAR_CODE_FIELD);

    root.addContent(barCodeElem);

    // create and add the RecordUpater field
    recordUpdaterElem = createNewRecordUpdaterElement();

    root.addContent(recordUpdaterElem);

    // find all the DBEntry children and add them to the document
    getChildrenRecursive(TypeBuilder.getType(definition).getDOM()
      .getRootElement(), DBENTRY, dbEntries);

    for (int i = 0; i < dbEntries.size(); i++)
    {
      defElem = (Element) dbEntries.get(i);
      labelValue = defElem.getChildText(LABEL);

      if (labelValue != null)
      {
        templateElem = new Element(DBENTRY);

        templateElem.addAttribute(NAME, defElem.getAttributeValue(NAME));
        templateElem.addAttribute(XMLDataProviderWrapper.LABEL, labelValue);
        root.addContent(templateElem);
      }
    }

    try
    {

      // add the new template to the PS.
      addTemplateToPS();
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }


  /**
   * Create a new template from a copy of the current template
   */

  // this method creates a new template and then saves it back to the PS.
  // The method that call this then reloads the 'new' template from the PS.
  // This should be more effiicient and make no network calls.  Running out of
  // time, being laid off...
  public void copy(String newTemplateName) throws NestedException
  {
    Element root = document.getRootElement();

    Element barCodeElem = root.getChild(BAR_CODE_FIELD);

    if (barCodeElem == null)
    {
      barCodeElem = new Element(BAR_CODE_FIELD);

      root.addContent(barCodeElem);
    }

    barCodeElem.setText(barCodeField);

    Element ruElem = document.getRootElement().getChild(RECORD_UPDATER);

    if (ruElem == null)
    {
      ruElem = createNewRecordUpdaterElement();

      document.getRootElement().addContent(ruElem);
    }
    else
    {
      ruElem.getAttribute(ACTIVE).setValue(String.valueOf
        (recordUpdaterData.active));
      ruElem.getAttribute(UPDATE).setValue(recordUpdaterData.update);
      ruElem.getAttribute(TOBE).setValue(recordUpdaterData.toBe);
      ruElem.getAttribute(IN).setValue(recordUpdaterData.in);
      ruElem.getAttribute(WHERE).setValue(recordUpdaterData.where);
      ruElem.getAttribute(IS).setValue(recordUpdaterData.selectedIsValue);
    }


    for (int i = 0; i < fields.length; i++)
    {
      fields[i].saveDataProviderData();
    }

    // replace the old name witht the new name.
    this.templateName = newTemplateName;

    root.removeAttribute(NAME_FIELD);
    root.addAttribute(NAME_FIELD, newTemplateName);

    try
    {

      // save the data to the PS
      addTemplateToPS();
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }

  /**
   * Add the current template to the PS
   */
  private void addTemplateToPS() throws IOException, OperationException
  {
    XMLOutputter outputter = new XMLOutputter("  ", true);

    outputter.setSuppressDeclaration(true);

    String xmlStr = "";
    Hashtable data;

    xmlStr = outputter.outputString(document);
    data = new Hashtable();

    data.put(NAME_FIELD, templateName);
    data.put(PARENT_FIELD, parent);
    data.put(DATA_FIELD, new String("<![CDATA[" + xmlStr + "]]>"));

    psInterface.addRecordInDomain(TEMPLATES_DOMAIN, data);

    // add the new template to the local list
    String[] newAvailableTemplates = 
      new String[availableTemplates.length + 1];

    System.arraycopy(availableTemplates, 0, newAvailableTemplates, 0, 
                     availableTemplates.length);

    newAvailableTemplates[newAvailableTemplates.length - 1] = templateName;
    availableTemplates = newAvailableTemplates;
  }

  /**
   * Delete the current template
   */
  public void delete() throws NestedException
  {
    Hashtable data = new Hashtable();

    document = null;
    fields = null;
    barCodeField = null;

    try
    {

      // delete the template from the PS
      String filter = "(" + NAME_FIELD + "=\"" + templateName + "\")";

      psInterface.deleteRecordsInDomain(TEMPLATES_DOMAIN, filter);

      // delete the template from the local list
      List list = Arrays.asList(availableTemplates);
      Vector vec = new Vector(list);

      vec.remove(templateName);

      availableTemplates = (String[]) vec.toArray(new String[0]);
    }
    catch (Exception e)
    {
      throw new NestedException(e);
    }
  }

}

