package org.vovida.prov.util;

import java.util.List;

import org.vovida.prov.util.assert.*;

import org.apache.log4j.Category;


/**
 * Implements a loop with a given maximum number of entries. Basically, this is
 * a doubly linked list with the head and tail linked together.
 * As new entries are added to the loop, all entries will eventually be filled,
 * at which point the first entry added to the loop will be replaced by the
 * newest added entry. The loop also has the concept of a tail (the last entry
 * added to it) and a top (oldest entry which is still in the loop). It is
 * possible to move forward and back through the loop up to the bounds defined
 * by the top and tail without removing entries. If an entry is added while
 * current position is not the tail of the loop, the new entry will be placed
 * at the current position.
 */
public class Loop
{
    private static Category logCategory = Category.getInstance(Loop.class.getName());

    private Entry top = null;
    private Entry tail = null;
    private Entry current = null;


    /**
     * Create a loop with the given number of entries.
     */
    public Loop(int maximumSize)
    {
        Assert.assert("size of LoopedStack > 0", maximumSize > 0);

        top = new Entry(null, null, null);
        tail = top;

        Entry next=null;
        Entry prev = top;
        for (int i=0; i<maximumSize-1; i++)
        {
            next = new Entry(null, null, null);
            next.previous = prev;
            prev.next = next;
            prev = next;
        }
        if (maximumSize > 1)
        {
            next.next = top;
            top.previous = next;
        }
    }


    /**
     * Add an element to the loop <p>.
     * The element is added at the position after the current position. Any
     * element located there will be ovewritten and any elements between the
     * current position and the tail of the loop will be lost as well.
     */
    public void add(Object newElement)
    {
        logCategory.debug("Adding to loop " + newElement);
        if (current == null) //there are no elements in the loop yet
        {
            current = top;
            current.element = newElement;
        }
        else
        {
            current = current.next; //advance to the next position
            current.element = newElement; //save the value
            tail = current; //mark the end of the loop
            if (tail.equals(top)) //if just overwrote the first element
            {
                top = top.next; //mark the start of the loop
            }
        }
    }


    /**
     * Check the capacity of the loop. This should be equal to the number of
     * elements specified in the constructor.
     */
    public int capacity()
    {
        Entry temp = top;
        int count = 1;
        while (!temp.next.equals(top))
        {
            count ++;
            temp = temp.next;
        }
        return count;
    }

    public Object getCurrent()
    {
        if (current == null)
        {
            return null;
        }
        return current.element;
    }


    /**
     * Get the size of the loop. This should equal the number of elements
     * currently contained in the loop.
     */
    public int size()
    {
        if (current == null)
        {
            return 0;
        }

        if (top.equals(tail) && top.equals(current))
        {
            return 1;
        }

        int size = 1;
        Entry temp = top;
        while (temp.element != null && !temp.equals(tail))
        {
            size ++;
            temp = temp.next;
        }

        return size;
    }


    /**
     * return the next element on the loop. Null is returned if the current
     * position is equal to the top of the loop.
     */
    public Object forward()
    {
        if (current == null) //no elements have been added to the loop yet
        {
            return null;
        }

        if (current.next.element != null && !current.equals(tail))
        {
            current = current.next;
            return current.element;
        }
        return null;
    }


    /**
     * Return the previous element on the loop. Null is returned if the current
     * position is equal to the tail of the loop.
     */
    public Object back()
    {
        if (current == null) //no elements have been added to the loop yet
        {
            return null;
        }

        if (current.previous.element != null && !current.equals(top))
        {
            current = current.previous;
            return current.element;
        }
        return null;
    }



    /**
     * An entry in the loop. Has pointers to the previous and next items.
     */
    private static class Entry
    {
        Object element;
        Entry next;
        Entry previous;

        Entry(Object element, Entry next, Entry previous)
        {
            this.element = element;
            this.next = next;
            this.previous = previous;
        }
    }
}
