/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 * 
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 * 
 */
package org.vovida.prov.util;

import org.xml.sax.ContentHandler;
import org.apache.xerces.parsers.SAXParser;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.Locator;
import org.xml.sax.Attributes;

import org.jdom.input.SAXBuilder;
import org.jdom.JDOMException;
import org.jdom.Document;
import org.jdom.output.XMLOutputter;

import java.io.Reader;
import java.io.StringReader;
import java.io.IOException;

import org.apache.log4j.Category;

/**
 * Contains utilities for SAX parsing and for building JDOM trees.
 * If DEBUG logging is turned on for this class, also logs all
 * important callback information and all JDOM trees built.
 */
public class XMLProcessor
{

  /**
   * for logging
   */
  static Category logCategory = 
    Category.getInstance(XMLProcessor.class.getName());

  /**
   * Parse an XML string, sending the callbacks to the given handler.
   */
  public static void doSAXParse(String xmlString, 
          ContentHandler handler) throws SAXException
  {
    String cdataStart = "<![CDATA[";
	if (xmlString.startsWith(cdataStart))
    {
        xmlString = xmlString.substring(cdataStart.length(),xmlString.length());
    }
    if (xmlString.endsWith("]]>"))
    {
       xmlString = xmlString.substring(0,xmlString.length()-3);
    }                         

    Reader reader = new StringReader(xmlString);
    InputSource source = new InputSource(reader);

    try
    {
      SAXParser parser = new SAXParser();

      // intercept callbacks and log them if debug logging is turned on
      if (logCategory.isDebugEnabled())
      {
        parser.setContentHandler(new LogHandler(handler));
      }
      else
      {
        parser.setContentHandler(handler);
      }

      parser.parse(source);
    }
    catch (SAXException e1)
    {
      throw e1;
    }
    catch (IOException e2)
    {
      e2.printStackTrace();
    }
  }

  /**
   * Build a JDom Document from an XML string
   */
  public static Document buildDocument(String xmlString) throws JDOMException
  {
    StringReader reader = new StringReader(xmlString);
    SAXBuilder builder = 
      new SAXBuilder("org.apache.xerces.parsers.SAXParser");
    Document dom = null;

    try
    {
      dom = builder.build(reader);

      if (logCategory.isDebugEnabled())
      {
        XMLOutputter outputter = new XMLOutputterOverride();

        logCategory.debug("NEW JDOM:");
        logCategory.debug(outputter.outputString(dom));
      }
    }
    catch (JDOMException e1)
    {
      e1.printStackTrace();

      throw e1;
    }
    catch (IOException e2)
    {
      e2.printStackTrace();
    }

    return dom;
  }

  /**
   * Intercepts the main ContentHandler methods to log them.
   */
  static class LogHandler implements ContentHandler
  {

    // callback handling is delegated to myRealHandler
    ContentHandler myRealHandler;

    LogHandler(ContentHandler realHandler)
    {
      super();

      myRealHandler = realHandler;
    }

    public void characters(char[] ch, int start, 
            int length) throws SAXException
    {
      logCategory.debug("CHARACTERS: " + new String(ch, start, length));
      myRealHandler.characters(ch, start, length);
    }

    public void endDocument() throws SAXException
    {
      logCategory.debug("END_DOCUMENT");
      myRealHandler.endDocument();
    }

    public void endElement(String namespaceURI, String localName, 
            String qName) throws SAXException
    {
      logCategory.debug("END_ELEMENT: " + localName);
      myRealHandler.endElement(namespaceURI, localName, qName);
    }

    public void endPrefixMapping(String prefix) throws SAXException
    {
      myRealHandler.endPrefixMapping(prefix);
    }

    public void ignorableWhitespace(char[] ch, int start, 
            int length) throws SAXException
    {
      myRealHandler.ignorableWhitespace(ch, start, length);
    }

    public void processingInstruction(String target, 
            String data) throws SAXException
    {
      logCategory.debug("PROCESSING_INSTRUCTION: " + "TARGET=" + target 
                        + " DATA=" + data);
      myRealHandler.processingInstruction(target, data);
    }

    public void setDocumentLocator(Locator locator)
    {
      myRealHandler.setDocumentLocator(locator);
    }

    public void skippedEntity(String name) throws SAXException
    {
      myRealHandler.skippedEntity(name);
    }

    public void startDocument() throws SAXException
    {
      logCategory.debug("START_DOCUMENT");
      myRealHandler.startDocument();
    }

    public void startElement(String namespaceURI, String localName, 
            String qName, Attributes atts) throws SAXException
    {
      logCategory.debug("START_ELEMENT: " + localName);

      for (int i = 0; i < atts.getLength(); i++)
      {
        logCategory.debug(atts.getLocalName(i) + "=" + atts.getValue(i));
      }

      myRealHandler.startElement(namespaceURI, localName, qName, atts);
    }

    public void startPrefixMapping(String prefix, 
            String uri) throws SAXException
    {
      myRealHandler.startPrefixMapping(prefix, uri);
    }

  }

}

