package org.vovida.prov.util.task;

/**
 * A simple implementation of a task which has its own thread and so may be
 * run in the background. <p>
 */
abstract public class Task implements Runnable
{
    private Exception exception;
    private Thread thread;
    private TaskListener callback = null;
    private Object returnValue = null;

    /**
     * If this constructor is used, there will be no callback when the task
     * finishes. <p>
     * Any return value from the task may be gotten through the getReturnValue()
     * method instead.
     */
    public Task()
    {
    }

    /**
     * If this constructor is used, the supplied TaskListener will be called
     * whenever the task finishes, normally or by throwing an exception. <p>
     * The task thread will not terminate untill the callback has finished.
     */
    public Task(TaskListener listener)
    {
        callback = listener;
    }

    public abstract void doWork() throws Exception;

    /**
     * Calls the doWork() method and notifies the callback if the task finishes
     * normally or if an exception occurs. <p>
     * Any exception generated in doWork is saved and must be retrieved
     * separately with getException.
     */
    public void run()
    {
        try
        {
            this.doWork();
            if (callback != null)
            {
                callback.taskFinished(this, true, returnValue);
            }
        }
        catch (Exception ex)
        {
            saveException(ex);
            if (callback != null)
            {
                callback.taskFinished(this, false, returnValue);
            }
        }
    }

    private void saveException (Exception ex)
    {
        exception = ex;
        ex.printStackTrace();
    }

    public Exception getException ()
    {
        return exception;
    }

    /**
     * Crate a new thread to run this task in and start it.
     */
    public void start()
    {
        if (thread == null)
        {
            String className = this.getClass().getName();
            thread = new Thread(this,className.substring(className.lastIndexOf(".")+1,className.length())+this.hashCode());
        }
        thread.start();
    }

    /**
     * Should be called from doWork() to save any value that needs to be
     * returned from the task. <p>
     * It will be passed as an argument to the callback object's taskFinished()
     * method, if a callback is set.
     */
    protected void saveReturnValue(Object value)
    {
        returnValue = value;
    }

    /**
     * Get the return value for this task. <p>
     * It is up to the concrete subclasses to set the return value. If none is
     * set, this will return null.
     */
    public Object getReturnValue()
    {
        return returnValue;
    }
}
