/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF ,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package org.vovida.prov.widget;

import org.vovida.prov.widget.GuiException;

import java.awt.Component;
import java.awt.event.ActionListener;
import java.awt.Container;
import java.awt.Font;
import java.awt.Color;
import java.awt.Dimension;
import javax.swing.JLabel;
import javax.swing.JComponent;
import javax.swing.table.DefaultTableCellRenderer;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import org.jdom.Element;

import org.apache.log4j.Category;

/**
 * This abstract class specifies how data is collected from an end-user.
 * Subclasses of UI represent the major categories of uer interaction,
 * such as GUI, TouchTone, and Voice. Only GUI is currently being
 * implemented. Subclasses of GUI represent the different types of widget
 * that could be used to enter data.
 *
 * @author Barbara Samson
 */
abstract public class GuiComponent implements Cloneable
{
  protected static final int FIELD_SIZE = 20;
  protected static final int PORT_SIZE = 6;

  private DefaultTableCellRenderer myDefaultRenderer = 
    new DefaultTableCellRenderer();

  static Category logCategory = 
    Category.getInstance(GuiComponent.class.getName());

  /**
   * Class Constructor.
   * @param widgetElement could be null.
   */
  public GuiComponent(Element widgetElement)
  {
    Font font = myDefaultRenderer.getFont();

    myDefaultRenderer.setFont(new Font(font.getName(), font.PLAIN, 
            font.getSize()));
  }

  /**
   * Cloning allows new table rows to be created.
   */
  public Object clone() throws CloneNotSupportedException
  {
    return super.clone();
  }

  public Component getTableCellRendererComponent(String value)
  {
    myDefaultRenderer.setText(value);

    return myDefaultRenderer;
  }

  public abstract void renderSelf(Container parent) throws GuiException;

  public abstract void setLabel(String newLabel);

  public abstract String getLabel();

  public abstract void setTooltip(String tooltip);

  public void setEnabled(boolean canEdit)
  {
    getEditorComponent().setEnabled(canEdit);
    getLabelComponent().setEnabled(canEdit);
    mySymbol.setEnabled(canEdit);
  }

  /**
   * This method only applies to a ComboBox.
   */
  public void setUserInputOK(boolean inputOK) {}

  protected JLabel mySymbol = new JLabel();

  /**
   * True if this is the key field.
   */
  private boolean myIsKey;

  /**
   * True if this widget does not get rendered.
   */
  protected boolean myIsHidden = false;

  /**
   * get the Component that holds data
   */
  public abstract JComponent getEditorComponent();

  /**
   * get the Component that displays the label
   */
  public abstract JComponent getLabelComponent();

  public void setRequired(boolean isRequired)
  {
    if (isRequired)
    {
      mySymbol.setText("*");
      mySymbol.setFont(new Font("Courier", Font.BOLD, 16));
      mySymbol.setForeground(Color.red);
      mySymbol.setToolTipText("required field");
      mySymbol.setPreferredSize(new Dimension(20, 20));
    }
    else
    {
      mySymbol.setText("");
      mySymbol.setToolTipText("");
    }
  }

  public JLabel getSymbol()
  {
    return mySymbol;
  }

  /**
   * Allows us to create a second JLabel with the same properties
   * as mySymbol.
   */
  protected void setSymbolProperties(JLabel symbol)
  {
    symbol.setText(mySymbol.getText());
    symbol.setFont(mySymbol.getFont());
    symbol.setForeground(mySymbol.getForeground());
    symbol.setToolTipText(mySymbol.getToolTipText());
    symbol.setPreferredSize(new Dimension(20, 20));
  }

  public void addActionListener(ActionListener listener) {}

  public void savingData() {}

  public void setHidden(boolean isHidden)
  {
    myIsHidden = isHidden;

    logCategory.info("widget " + (isHidden ? "is" : "is not") + " hidden");
  }

  public boolean isHidden()
  {
    return myIsHidden;
  }

  /**
   * Sets up the visual indicators for a key field.
   */
  public void setKey(boolean isKey)
  {
    myIsKey = true;

    setRequired(true);
  }

  public boolean isKey()
  {
    return myIsKey;
  }

}




