/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 * 
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 * 
 */

package vocal.comm;

import java.net.*;
import java.io.*;

/**
 * This class is an abstraction to communicate with the provisioning server
 * user the VPP (Vovida Provisioning Protocol) protocol.
 * 
 * Allows you open a connection, make any number of put and get operations,
 * and close the connection, abstracting away most of the details of the
 * transaction.
 */

public class VPPTransaction implements Transaction
{

  public static int WRITE_TO_FILE = 1;
  public static int WRITE_TO_PSERVER = 2;

  public static int ERROR_MEANS_NO_SUCH_FILE = 0;
  public static int ERROR_MEANS_PERM_DENIED = 1;

  /**
   * True when this connection has been closed.
   * This is used to stop reconnect attempts.
   */
  private boolean isClosed = false;

  /**
   * the size (in characters) of the buffer used to save up output from the
   * server
   */
  private static final int blockSize = 4096;

  /**
   * Indicates whether there is currently a valid connection open to the
   * server.
   */
  private boolean connected = false;

  /**
   * The timeout for read operations from a socket input stream. This is used
   * to activate non-blocking io. Units are milliseconds.
   */
  private static final int SOCKET_TIMEOUT = 12000;

  protected Socket socket;
  protected BufferedReader fromSocket;
  protected OutputStreamWriter toSocket;

  private Reconnect reconnect = null;

  private String host;
  private int port;


  public VPPTransaction(String server, int portNo) throws VPPException
  {

    host = server;
    port = portNo;

    // try and catch errors we know about, throw anything else
    try
    {
      socket = new Socket(InetAddress.getByName(server), port);

      // disable blocking I/O by setting a timeout on the
      // socketInputStream.read operation
      socket.setSoTimeout(SOCKET_TIMEOUT);
    }
    catch (IOException ex)
    {

      String message = "The program could not open a socket to " 
                       + "the server \n" + server + " on port " + port 
                       + "\n\n" + "Reason: " + ex.getMessage();

      throw new VPPLowLevelException(ex, message);

    }

    try
    {
      fromSocket = 
        new BufferedReader(new InputStreamReader(socket.getInputStream()));
    }
    catch (IOException ex)
    {

      String message = "The server was contacted, but could " 
                       + "not be read from.\n\n" + "Reason: " 
                       + ex.getMessage();

      throw new VPPLowLevelException(ex, message);

    }

    try
    {
      toSocket = new OutputStreamWriter(socket.getOutputStream());
    }
    catch (IOException ex)
    {
      String message = "The server was contacted, but could " 
                       + "not be written to.\n\n" + "Reason: " 
                       + ex.getMessage();

      throw new VPPLowLevelException(ex, message);
    }

    connected = true;   // if we got here things must be ok...

  }



  /**
   * Read the status line (200... or 400...) of our VPP transaction
   * and verify that the status says OK. It is expected that the stream is
   * at the right position to read the status.
   */
  protected String getStatusLine(int flag) throws VPPException
  {

    String statusLine = readLineFromSocket();

    System.out.println("---TRACE--- VPPTransaction:reply - " + statusLine 
                       + Thread.currentThread());

    // Our protocol demands that status line must start with the
    // return code.
    if (statusLine.startsWith("400"))
    {

      System.out.println("---TRACE--- VPPTransaction:" 
                         + "Got 400 -- Throwing exception");

      if (flag == ERROR_MEANS_NO_SUCH_FILE)
      {
        throw new VPPNoSuchFileException("File Not Found");
      }
      else
      {
        throw new VPPPermissionException("Permission Denied");
      }

    }
    else if (!statusLine.startsWith("200"))
    {

      // throw the malformed message since we should get
      // one of 200 or 400
      throw new VPPMalformedMsgException("The server did not " 
              + "respond properly" + " to the command. " + "The result is: " 
              + statusLine);

    }

    // else 200, so ok

    return statusLine;

  }



  /**
   * Read the content length line of our VPP transaction and verify
   * that it starts with the "Content-Length: " string. It assumes that the
   * underlying streams for this connection are at the right position to read
   * the content length.
   */
  protected int getContentLength() throws VPPException
  {

    String contentLengthLine = readLineFromSocket();

    // Verify that the line starts with "Content-Length: "
    if (!contentLengthLine.startsWith("Content-Length: "))
    {
      throw new VPPMalformedMsgException("The Content-Length line from the server" 
              + "was missing or malformed.\n" + "Its contents were:\n \"" 
              + contentLengthLine + "\".");
    }

    // we know the first 16 are the Content-Length, try to get the #
    String contentLengthAsString;

    contentLengthAsString = contentLengthLine.substring(16);

    int contentLength;

    try
    {
      contentLength = Integer.parseInt(contentLengthAsString);
    }
    catch (NumberFormatException e)
    {
      throw new VPPMalformedMsgException("The Content-Length line was corrupted\n" 
              + "The contents were: \n \"" + contentLengthLine + "\".");
    }

    return contentLength;

  }



  protected String getContents(int contentLength) 
    throws IOException, VPPException
  {

    char[] buffer = new char[contentLength];
    int totalCharsRead = 0;
    int charsRead = 0;
    StringBuffer response = new StringBuffer();

    while (totalCharsRead < contentLength)
    {

      try
      {
        charsRead = fromSocket.read(buffer, 0, contentLength);
      }
      catch (IOException e)
      {
        System.out.println("Reconnecting");
        reconnect();
      }

      if (charsRead == -1)
      {
        throw new VPPMalformedMsgException("There was a problem in reading" 
                + "the response:\n Content-Length was reported as " 
                + contentLength + ", \n but the stream closed after only " 
                + totalCharsRead + " characters.");
      }

      totalCharsRead += charsRead;

      response.append(buffer, 0, charsRead);

    }

    if (totalCharsRead < contentLength)
    {
      throw new VPPMalformedMsgException("There was a problem in reading the response:  " 
              + "\nContent-Length was reported as " + contentLength 
              + ", \nbut the stream closed after only " + totalCharsRead 
              + " characters.");
    }

    return response.toString();

  }

  /**
   * Send a request string exactly as given.
   * @param requestType the type of request
   * @param requestString the string to be sent
   * @param entityToPut the rest of the content, if any
   * @return the server's response
   */
  public synchronized String sendRequest(String requestType, 
          String requestString, String entityToPut) throws VPPException
  {
    System.out.println("---TRACE--- VPPTransaction:request - " 
                       + requestString + Thread.currentThread());

    if (!connected)
    {
      throw new VPPNoConnException("No connection to server");
    }

    // I get a warning about possibly throwing an undefined
    // variable below if I don't initialize this...
    VPPException exToThrow = new VPPException("No exception");

    try
    {

      // Send the request
      toSocket.write(requestString);
      toSocket.flush();

      if (requestType.equals("PUT"))
      {
        toSocket.write("Content-Length: " + entityToPut.length() + '\n');
        toSocket.flush();
        toSocket.write(entityToPut);
        toSocket.flush();
      }

      // Read the response

      String statusLine = "xxx";
      boolean isError = false;

      try
      {

        // if it fails for get, no such file, else perm. denied
        if (requestType.equals("GET"))
        {
          statusLine = getStatusLine(ERROR_MEANS_NO_SUCH_FILE);
        }
        else
        {
          statusLine = getStatusLine(ERROR_MEANS_PERM_DENIED);
        }

        System.out.println("----TRACE---- VPPTransaction:" 
                           + "sendRequest - statusLine = " + statusLine);
      }
      catch (VPPException e)
      {
        System.out.println("----TRACE---- VPPTransaction:" 
                           + "sendRequest - statusLine = EXCEPTION");
        e.printStackTrace();

        isError = true;
        exToThrow = e;
      }

      // After a put reqest, the server will respond with the length
      // of the received data.
      int contentLength = getContentLength();

      if (isError == true)
      {
        throw exToThrow;
      }

      // If there was no content (for example on a put), return
      if (contentLength == 0)
      {
        return "";
      }

      // else get and return contents
      String contents = getContents(contentLength);

      return contents;

    }
    catch (IOException e)
    {

      // We catch non-VPP exceptions of I/O type and wrap them
      // in low level error
      System.out.println("Caught " + e.getMessage() + " -- Reconnecting");
      System.out.println("The reqest was:");
      System.out.println("    " + requestString);

      try
      {
        reconnect();
      }
      catch (VPPNoConnException ex)
      {
        throw new VPPLowLevelException(ex, ex.getMessage());
      }

    }

    return "";
  }


  /**
   * Construct a request and send it across a socket.
   * @return the server's response as a string
   * @param requestType the type of request
   * @param arg1 group for requests that have a group, otherwise name.
   * @param arg2 name for requests that hava a group, otherwise empty string
   * @param entityToPut the contents of entity that should be attached to
   * the request.
   */
  protected String sendRequest(String requestType, String arg1, String arg2, 
          String entityToPut) throws VPPException
  {
    return sendRequest(requestType, 
                       requestType + ' ' + arg1 + ' ' + arg2 + '\n', 
                       entityToPut);
  }


  /**
   * Read a line of input and return it, without the newline.
   */
  protected String readLineFromSocket() throws VPPLowLevelException
  {
    StringBuffer line = new StringBuffer();
    char nextCharacter;

    try
    {
      while ((nextCharacter = (char) fromSocket.read()) != '\n')
      {
        line.append(nextCharacter);
      }
    }
    catch (IOException e)
    {
      System.out.println("Caught " + e.getMessage() + " -- Reconnecting");

      try
      {
        reconnect();
      }
      catch (VPPNoConnException ex)
      {
        throw new VPPLowLevelException(ex, ex.getMessage());
      }
    }

    return line.toString();
  }



  /**
   * Make a get request, and return the contents of the requested file if the
   * transaction was successful.  Otherwise, throw an exception.
   * @param group the virtual path of the file to request,
   * like a directory path, but with underscores instead of slashes.
   * @param file the name of the file
   * @return the contents of the file
   * @exception VPPException if the server gives a response that is
   * illegal according to this protocol
   */
  public String doGet(String group, String filename) throws VPPException
  {
    return sendRequest("GET", group, filename, null);
  }


  public String doDel(String group, String filename) throws VPPException
  {
    return sendRequest("REMOVE", group, filename, null);
  }



  public static void doPutToFile(String filePath, 
          String fileContents) throws VPPException
  {

    try
    {

      String path = filePath.substring(0, filePath.lastIndexOf("/") + 1);
      String name = filePath.substring(filePath.lastIndexOf("/") + 1);
      File outputFile = new File(path, name);

      // if can't write file, it may be that the path does not exist
      if (!outputFile.canWrite())
      {
        File pathFile = new File(path);

        // create the given directories
        boolean response = pathFile.mkdirs();
      }

      FileWriter out = new FileWriter(outputFile);

      out.write(fileContents);
      out.flush();
      out.close();
    }
    catch (IOException e)
    {
      throw new VPPLowLevelException(e, e.getMessage());
    }

  }


  public void doPut(String group, String filename, String entityToPut, 
                    int destination) throws VPPException
  {

    if (destination == WRITE_TO_FILE)
    {
      doPutToFile(group + "/" + filename, entityToPut);
    }
    else if (destination == WRITE_TO_PSERVER)
    {
      doPut(group, filename, entityToPut);
    }

  }


  /**
   * Format and send a PUT request, uploading a file.
   * @param group the virtual path to which to upload
   * @param filename name of the file to write
   * @param entityToPut the text string to write to file
   * @exception VPPException if the server gives a response that is
   * illegal according to this protocol
   */
  public void doPut(String group, String filename, 
                    String entityToPut) throws VPPException
  {
    sendRequest("PUT", group, filename, entityToPut);
  }


  /**
   * Delete a user and his associated Contact Lists.
   * @param userName the name of the user. Note that this
   * does not have any directory path prepended to it.
   */
  public void doDeleteUser(String userName) throws VPPException
  {
    sendRequest("DELETE", userName, "", null);
  }


  /**
   * Close the input stream, output stream and socket associated with this
   * connection.
   */
  public void close() throws VPPException
  {
    try
    {
      this.isClosed = true;

      toSocket.close();
      fromSocket.close();
      socket.close();
    }
    catch (IOException e)
    {
      throw new VPPLowLevelException(e, e.getMessage());
    }
  }



  /**
   * Sends a request of the given type (may be GET or PUT, or whatever) to
   * the server and returns the response as a BufferedReader. <p>
   * @param requestType and identifier of the type of request being made. Eg:
   * "GET" or "PUT"
   * @param arg1 see sendRequest arg1
   * @param arg2 see sendRequest arg2
   * @return the server's response as a BufferedReader
   */
  public BufferedReader request(String requestType, String arg1, 
          String arg2) throws VPPException
  {
    String response = sendRequest(requestType, arg1, arg2, null);

    return new BufferedReader(new StringReader(response));
  }



  public String requestString(String requestType, String arg1, 
          String arg2) throws VPPException
  {
    return sendRequest(requestType, arg1, arg2, null);
  }



  public void reconnect() throws VPPNoConnException
  {
    System.out.println("---ERROR--- reconnct started " 
                       + Thread.currentThread().toString());

    connected = false;
    reconnect = new Reconnect(this, host, port);

    Thread thread = new Thread(reconnect, "Reconnect");

    thread.start();

    throw new VPPNoConnException("Connection to server lost.");
  }



  /**
   * Callback used by the Reconnect to notify the transaction that the
   * reconnect was successful.
   */
  public void reconnected()
  {
    fromSocket = reconnect.getInputStream();
    toSocket = reconnect.getOutputStream();
    socket = reconnect.getSocket();
    connected = true;
  }



  /**
   * Callback used by the Reconnect to notify this class that the reconnect
   * failed (because the max number of retries was exceeded).
   */
  public void reconnectFailed()
  {

    // what should this do?
  }

  /**
   * Callback used by the Reconnect before each reconnect attempt is
   * started.
   */
  public boolean okToReconnect()
  {
    if (isClosed)
    {
      return false;
    }

    return true;
  }

}




