/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 *
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package vocal.data;

import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.JOptionPane;
import javax.swing.JFrame;
import java.io.StringReader;
import java.util.Stack;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Element;
import org.w3c.dom.DocumentFragment;
import org.xml.sax.Attributes;
import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.helpers.DefaultHandler;

import org.apache.xerces.parsers.SAXParser;

/**
 * This class creates a new node of the DOM tree and saves data
 * from the DOM to that new node. The new node is constructed
 * as a DataFragment, hence the name of the class.
 *
 * This class and its sublclass(es) implement the set of interfaces
 * used by the GUI to read and write data.
 *
 * The user may choose to create a new node, enter some data,
 * and then cancel the whole operation. For this reason, the
 * node is not actually inserted into the DOM tree until the
 * user has pressed the "Ok" button.
 *
 * Once this node has been saved for the first time (the user
 * pressed the Ok button) it is no longer valid to use this object
 * for saving further data changes. Thus the first save causes
 * an "invalid" flag to be set and any further calls to the object
 * will then throw an InvalidRequestException.
 */
public class DataFragment extends DataGetSet implements HostPortData,
        FeatureData, MarshalData, GroupData, CdrData, RedirectData,
        TreeNodeTypes, DataDefaults
{

  // protected fields

  protected ConfigTree configTree = null;

  /**
   * This field is used as a validity check.
   * Once a DataFragment has been saved, there are no more
   * valid operations that can be performed on it,
   * so validity becomes false.
   */
  protected boolean isValid = true;

  /**
   * The node of the ConfigTree where this fragment will be inserted.
   */
  protected DefaultMutableTreeNode parentTreeNode;

  /**
   * The DOM node where this will be inserted.
   */
  protected Node parentNode;

  /**
   * The next node up the DOM.
   */
  protected Node grandparentNode;

  /**
   * The DocumentFragment that holds the data.
   */
  protected DocumentFragment fragment;

  /**
   * String that creates the Document Fragment for a Feature Server.
   */
  private static String FEATURE_SERVER_STRING =
    new String("<featureServer host=\"none\" port=\"-1\">"
               + "<type value=\"none\"/>" + "<group value=\"none\"/>"
               + "</featureServer>");

  /**
   * String that creates the Document Fragment for a Marshal Server.
   */
  private static String MARSHAL_SERVER_STRING =
    new String("<marshalServer host=\"none\" port=\"-1\">"
               + "<type value=\"none\"/>" + "<group value=\"none\"/>"
               + "<gateway host=\"none\" port=\"-1\"/>"
               + "<noResponseTimer value=\"100\"/>"
               + "<allowUnbillableCalls value=\"true\"/>"
               + "<allowUnknownCallers value=\"false\"/>"
               + "<conferenceBridge>" + "<bridgeNumber value=\"none\"/>"
               + "<accessList/>" + "</conferenceBridge>"
               + "</marshalServer>");

  /**
   * String that creates the Document Fragment for a CDR Server.
   */
  private static String CDR_SERVER_STRING =
    new String("<cdrServer host=\"none\" port=\"-1\">"
               + "<group value=\"none\"/>"
               + "<radiusServer host=\"none\" retries=\"5\" secretKey=\"none\"/>"
               + "<billing frequency=\"300\""
               + "         dirPath=\"../billing/\""
               + "         lockFile=\".billingLock\""
               + "         dataFile=\"billing.dat\""
               + "         unsentExtension=\".unsent\""
               + "         rolloverSize=\"100000\""
               + "         rolloverPeriod=\"300\"/>"
               + "         billForRingtime=\"false\"/>" + "</cdrServer>");

  /**
   * String that creates the Document Fragment for a PDP Server.
   */
  private static String PDP_SERVER_STRING =
    new String("<pdpServer host=\"none\" port=\"3288\">"
               + "<group value=\"none\"/>" + "<maxThreads value=\"8\"/>"
               + "<keepAliveTimer value=\"5000\"/>"
               + "<responseTime value=\"10000\"/>" + "</pdpServer>");

  /**
   * String that creates the Document Fragment for a Redirect Server.
   */
  private static String REDIRECT_SERVER_STRING =
    new String("<redirectServer host=\"none\" port=\"-1\">"
               + "<group value=\"none\"/>" + "<syncPort value=\"0\"/>"
               + "</redirectServer>");

  /**
   * String that creates the Document Fragment for a Heartbeat Server.
   */
  private static String HEARTBEAT_SERVER_STRING =
    new String("<heartbeatServer host=\"none\" port=\"0\">"
               + "<group value=\"none\"/>" + "</heartbeatServer>");

  /**
   * String that creates the Document Fragment for a Group.
   */
  protected static String GROUP_STRING =
    new String("<serverGroup value=\"none\"/>");

  // Constructors

  /**
   * This constructor is used from the GUI to create a DataFragment
   * in sync with the ConfigTree.
   */
  public DataFragment(int type, ConfigTree tree, DefaultMutableTreeNode node)
  {
    super();

    configTree = tree;
    parentTreeNode = node;
    parentNode = configTree.getNode(parentTreeNode);
    grandparentNode = parentNode.getParentNode();

    Element parentElement = null;
    Element grandparentElement = null;

    try
    {
      parentElement = (Element) parentNode;
      grandparentElement = (Element) grandparentNode;
    }
    catch (ClassCastException e)
    {
      System.out.println("attempted to insert DataFragment in non-Element path: "
                         + grandparentNode.getNodeType() + " / "
                         + parentNode.getNodeType());
      e.printStackTrace();

    }

    localBuildDocumentFragment(type, parentElement, grandparentElement);
  }

  /**
   * This constructor is a last-minute add-on to support inserting
   * nodes that are not in sync with the ConfigTree. It is used for
   * processing the Deployment script.
   * The main difference is that there is no more reliance on
   * selection in the ConfigTree. Instead the constructor must
   * be pased the org.w3c.Node that is the parent of the one being built.
   */
  public DataFragment(int type, Node aParentNode)
  {
    super();

    parentNode = aParentNode;
    grandparentNode = parentNode.getParentNode();

    Element parentElement = null;
    Element grandparentElement = null;

    try
    {
      parentElement = (Element) parentNode;
      grandparentElement = (Element) grandparentNode;
    }
    catch (ClassCastException e)
    {
      System.out.println("attempted to insert DataFragment in non-Element path: "
                         + grandparentNode.getNodeType() + " / "
                         + parentNode.getNodeType());
      e.printStackTrace();

    }

    localBuildDocumentFragment(type, parentElement, grandparentElement);

  }

  /**
   *
   * @param type
   * @param parent
   * @param grandparent
   */
  protected void localBuildDocumentFragment(int type, Element parent,
          Element grandparent)
  {
    Document doc = parent.getOwnerDocument();

    switch (type)
    {

      case FEATURE_SERVER:
      {
        fragment = buildDocumentFragment(doc, FEATURE_SERVER_STRING);

        getElement("type").setAttribute("value",
                   grandparent.getAttribute("value"));
        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", FEATURE_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case MARSHAL_SERVER:
      {
        fragment = buildDocumentFragment(doc, MARSHAL_SERVER_STRING);

        getElement("type").setAttribute("value",
                   grandparent.getAttribute("value"));
        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", MARSHAL_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case CDR_SERVER:
      {
        fragment = buildDocumentFragment(doc, CDR_SERVER_STRING);

        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", CDR_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case HEARTBEAT_SERVER:
      {
        fragment = buildDocumentFragment(doc, HEARTBEAT_SERVER_STRING);

        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", HEARTBEAT_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case REDIRECT_SERVER:
      {
        fragment = buildDocumentFragment(doc, REDIRECT_SERVER_STRING);

        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", REDIRECT_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case PDP_SERVER:
      {
        fragment = buildDocumentFragment(doc, PDP_SERVER_STRING);

        getElement("group").setAttribute("value",
                   parent.getAttribute("value"));

        Element rootElement = getRootElement();

        rootElement.setAttribute("fileDir", POLICY_DIR);
        rootElement.setAttribute("fileType", Integer.toString(type));

        break;
      }

      case FEATURE_GROUP:

      case MARSHAL_GROUP:
      {
        fragment = buildDocumentFragment(doc, GROUP_STRING);

        break;
      }

      default:
      {
        System.out.println("cannot build new node of type "
                           + TreeNodeTypes.strings[type] + " ("
                           + Integer.toString(type) + ")");
        new Exception().printStackTrace();

      }
    }
  }


  // public methods

  // methods from HostPortData

  /**
   *
   */
  public String getHostData() throws InvalidRequestException
  {
    checkValidity("getHostData");

    Element rootElement = getRootElement();

    if (rootElement == null)
    {
      System.out.println("Error in getHostData: root element is null");

      return "";
    }

    return rootElement.getAttribute("host");
  }

  /**
   *
   * @param value
   *
   * @throws InvalidRequestException
   */
  public void setHostData(String value) throws InvalidRequestException
  {
    checkValidity("setHostData");

    if (configTree != null)
    {
      Element duplicate = configTree.hostPortExists(value, getPortData());
      if (duplicate != null)
      {
        JOptionPane.showMessageDialog(new JFrame(),
        value + ":" + getPortData() + " already exists", "Duplicate host:port",
        JOptionPane.ERROR_MESSAGE);
        throw new InvalidRequestException(value + ":" + getPortData()
        + " already exists in " + duplicate);
      }
    }
    getRootElement().setAttribute("host", value);
  }

  /**
   *
   */
  public String getPortData() throws InvalidRequestException
  {
    checkValidity("getPortData");
    return getRootElement().getAttribute("port");
  }

  /**
   *
   */
  public void setPortData(String value) throws InvalidRequestException
  {
    checkValidity("setPortData");

    if (configTree != null)
    {
      Element duplicate = configTree.hostPortExists(getHostData(), value);
      if (duplicate != null)
      {
        JOptionPane.showMessageDialog(new JFrame(),
        getHostData() + ":" + value + " already exists", "Duplicate host:port",
        JOptionPane.ERROR_MESSAGE);
        throw new InvalidRequestException(getHostData() + ":" + value
        + " already exists in " + duplicate);
      }
    }
    getRootElement().setAttribute("port", value);
  }

  /**
   *
   */
  public DocumentFragment getFragment() throws InvalidRequestException
  {
    isValid = false;

    return fragment;
  }

  /**
   *
   * @throws InvalidRequestException
   */
  public void saveData() throws InvalidRequestException
  {
    if (!isValid)
    {
      throw new InvalidRequestException("saveData on saved or deleted DataFragment");
    }

    System.out.println("saveData called on DataFragment");

    Element rootElement = getRootElement();

    System.out.println(new XMLUtils().buildXMLString(rootElement));
    System.out.println("Root element is " + rootElement.getNodeName());

    String dir = rootElement.getAttribute("fileDir");
    String host = null;
    String port = null;

    if (!dir.equals(""))
    {
      System.out.println("dir is not empty");

      host = rootElement.getAttribute("host");
      port = rootElement.getAttribute("port");

      rootElement.setAttribute("fileName", host + ":" + port);
    }
    else
    {
      System.out.println("dir is empty");
    }

    if (configTree != null) // null if called by DeployServers
    {
      configTree.insertDocumentFragment(fragment, parentTreeNode);
    }

    // if this is a Call Return feature, then we need to generate some
    // cpl at this point.
    if (((Element) grandparentNode).getAttribute("value").equals("CallReturn"))
    {
      String fileDir = (FEATURE_DIR + "_" + host + "_" + port + "_Called");
      String xmlString = CplGenerator.buildCallReturnServerXML();

      // write the cpl script to file.
      try
      {
        DOMToFile.writeAsIs(xmlString, fileDir, "*69@vovida.com.cpl");
      }
      catch (FailedToWriteFileException e1)
      {

        // nothing much we can do here.
      }
    }

    if (configTree != null) // null if called by DeployServers
    {
      try
      {
        DOMToFile.writeAllFilesInPath(configTree.getCurrentElement());
      }
      catch (FailedToWriteFileException e2)
      {
        configTree.deleteData();
      }
    }

    isValid = false;
  }

  /**
   *
   */
  public boolean deleteData() throws InvalidRequestException
  {
    if (!isValid)
    {
      throw new InvalidRequestException("deleteData on saved or deleted DataFragment");
    }

    // do nothing because the data fragment was never put in the tree
    isValid = false;

    return true;
  }

  // methods from GroupData

  /**
   *
   */
  public String getGroupName() throws InvalidRequestException
  {
    checkValidity("getGroupName");

    return getRootElement().getAttribute("value");
  }

  /**
   *
   * @param data
   *
   * @throws InvalidRequestException
   */
  public void setGroupName(String data) throws InvalidRequestException
  {
    checkValidity("setGroupName");
    getRootElement().setAttribute("value", data);
  }


  /**
   *
   * @param methodName
   *
   * @throws InvalidRequestException
   */
  protected void checkValidity(String methodName)
    throws InvalidRequestException
  {
    if (!isValid)
    {
      throw new InvalidRequestException(methodName
              + " on saved or deleted DataFragment");
    }
  }

  /**
   *
   */
  protected Element getRootElement()
  {
    Node rootNode = fragment.getFirstChild();

    if (rootNode == null)
    {
      System.out.println("Error: Data Fragment root element is null");
      new Exception().printStackTrace();

    }

    return (Element) rootNode;
  }

  /**
   *
   */
  protected Element getElement(String elementName)
  {
    Element rootElement = getRootElement();

    if (rootElement.getNodeName().equals(elementName))
    {
      return rootElement;
    }

    NodeList nodes = rootElement.getElementsByTagName(elementName);
    Element requestedElement = null;

    if (nodes.getLength() < 1)
    {
      System.out.println("fragment does not contain element " + elementName);
      System.out.println(new XMLUtils().buildXMLString(fragment));

      return null;
    }

    try
    {
      requestedElement = (Element) (nodes.item(0));
    }
    catch (ClassCastException e)
    {
      System.out.println(elementName + " is not an Element: "
                         + nodes.item(0).getNodeType());
      e.printStackTrace();

    }

    return requestedElement;
  }

  /**
   *
   */
  protected String getValue(String tagName)
  {
    Element element = getElement(tagName);

    if (element == null)
    {
      return null;
    }

    return element.getAttribute("value");
  }

  /**
   *
   * @param tagName
   * @param data
   */
  protected void setValue(String tagName, String data)
  {
    Element element = getElement(tagName);

    if (element == null)
    {
      return;
    }

    element.setAttribute("value", data);
  }

  /**
   * Build a Document Fragment from a string representation of the XML.
   */
  protected DocumentFragment buildDocumentFragment(final Document doc,
          String fragString)
  {
    System.out.println("building document fragment: " + fragString);

    final DocumentFragment frag = doc.createDocumentFragment();

    // read the string in with a SAX parser to build the fragment
    StringReader fragReader = new StringReader(fragString);
    XMLReader parser = new SAXParser();

    parser.setContentHandler(new DefaultHandler()
    {
      Stack elementStack = new Stack();

      /**
       *
       * @param ignore1
       * @param ignore2
       * @param rawName
       * @param attrs
       */
      public void startElement(String ignore1, String ignore2,
              String rawName, Attributes attrs)
      {
        if (rawName.equals("fragment"))
        {

          // the fragment root element is necessary for the
          // SAX parser to work.
          return;
        }

        int attrsLength = attrs.getLength();

        Element newElement = doc.createElement(rawName);

        for (int i = 0; i < attrsLength; i++)
        {
          newElement.setAttribute(attrs.getRawName(i), attrs.getValue(i));
        }

        if (!elementStack.isEmpty())
        {
          Element lastElement = (Element) (elementStack.peek());

          lastElement.appendChild(newElement);
        }

        elementStack.push(newElement);
      }

      /**
       *
       * @param ignore1
       * @param ignore2
       * @param rawName
       */
      public void endElement(String ignore1, String ignore2, String rawName)
      {
        if (!rawName.equals("fragment"))
        {
          Element lastElement = (Element) (elementStack.pop());

          if (elementStack.isEmpty())
          {
            frag.appendChild(lastElement);
          }
        }
      }

    });

    try
    {
      parser.parse(new InputSource(fragReader));
    }
    catch (Exception e)
    {
      e.printStackTrace();

    }

    return frag;
  }

}



