/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 * 
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 * 
 */

package vocal.data;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Document;

import java.util.Vector;

/**
 * This class takes care of saving GUI data to the DOM tree
 * and to file. It also takes data from the DOM tree and gives
 * it to the GUI. The concrete class DataFragment implements the
 * methods necessary for a new node, and the concrete class
 * DataToDOM implements the methods necessary for an existing node.
 * 
 * This class and its sublclass(es) implement the set of interfaces
 * used by the GUI to read and write data. The GUI screen classes
 * must request a DataManager object from the ConfigTree after
 * any significant event (ie switching context or saving the data
 * to the DOM). The ConfigTree passes back either an instance of DataToDOM
 * or an new instance of DataFragment, depending on the desired behavior.
 * Thus the only state information kept by a GUI screen is the contents
 * of its own data fields. When the GUI screen does a read or write to its
 * DataManager object, the concrete DataManager class takes care
 * of the correct implementation of that read or write method.
 */
public abstract class DataGetSet implements HostPortData, FeatureData, 
        MarshalData, GroupData, CdrData, HeartbeatData, PDPData, OSPData, 
        RedirectData, TreeNodeTypes, DataDefaults
{

  // methods from FeatureData

  /**
   * 
   */
  public String getTypeData() throws InvalidRequestException
  {
    checkValidity("getTypeData");

    return getValue("type");
  }

  /**
   * 
   */
  public void setTypeData(String data) throws InvalidRequestException
  {
    checkValidity("setTypeData");
    setValue("type", data);
  }

  /**
   * 
   */
  public String getGroupData() throws InvalidRequestException
  {
    checkValidity("getGroupData");

    return getValue("group");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setGroupData(String data) throws InvalidRequestException
  {
    checkValidity("setGroupData");
    setValue("group", data);
  }

  /**
   * Helper method that puts the uaVMServers tag
   * into the XML if it is not already there.
   * We have this method to ease the transition from
   * an old style of data where these elements did
   * not exist, to a new where they are required.
   * @return the new element inserted.
   */
  private Element insertUaVMElement()
  {
    Element featureElement = getElement("featureServer");
    Document doc = featureElement.getOwnerDocument();
    Element uaVMServers = doc.createElement("uaVMServers");

    featureElement.appendChild(uaVMServers);
    uaVMServers.setAttribute("host", "none");
    uaVMServers.setAttribute("firstPort", "0");
    uaVMServers.setAttribute("lastPort", "0");

    return uaVMServers;
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setUaVMHostData(String data) throws InvalidRequestException
  {
    System.out.println("setUaVMHostData " + data);
    checkValidity("setUaVMHostData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      System.out.println("request for uaVMServers returned null");

      uaVMServers = insertUaVMElement();
    }
    else
    {
      System.out.println("request for uaVMServers returned non-null " 
                         + uaVMServers);
    }

    uaVMServers.setAttribute("host", data);
  }

  /**
   * 
   */
  public String getUaVMHostData() throws InvalidRequestException
  {
    checkValidity("getUaVMHostData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      uaVMServers = insertUaVMElement();
    }

    return uaVMServers.getAttribute("host");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setUaVMFirstPortData(String data) throws InvalidRequestException
  {
    System.out.println("setUaVMFirstPortData " + data);
    checkValidity("setUaVMFirstPortData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      uaVMServers = insertUaVMElement();
    }

    uaVMServers.setAttribute("firstPort", data);
  }

  /**
   * 
   */
  public String getUaVMFirstPortData() throws InvalidRequestException
  {
    checkValidity("getUaVMFirstPortData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      uaVMServers = insertUaVMElement();
    }

    return uaVMServers.getAttribute("firstPort");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setUaVMLastPortData(String data) throws InvalidRequestException
  {
    System.out.println("setUaVMLastPortData " + data);
    checkValidity("setUaVMLastPortData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      uaVMServers = insertUaVMElement();
    }

    uaVMServers.setAttribute("lastPort", data);
  }

  /**
   * 
   */
  public String getUaVMLastPortData() throws InvalidRequestException
  {
    checkValidity("getUaVMLastPortData");

    Element uaVMServers = getElement("uaVMServers");

    if (uaVMServers == null)
    {
      uaVMServers = insertUaVMElement();
    }

    return uaVMServers.getAttribute("lastPort");
  }


  // methods from MarshalData

  /**
   * 
   * 
   */
  public String getGatewayHostData() throws InvalidRequestException
  {
    checkValidity("getGatewayHostData");

    Element element = getElement("gateway");

    if (element == null)
    {
      return "";
    }

    return element.getAttribute("host");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setGatewayHostData(String data) throws InvalidRequestException
  {
    checkValidity("setGatewayHostData");

    Element element = getElement("gateway");

    if (element != null)
    {
      element.setAttribute("host", data);
    }
  }

  /**
   * 
   */
  public String getGatewayPortData() throws InvalidRequestException
  {
    checkValidity("getGatewayPortData");

    Element element = getElement("gateway");

    if (element == null)
    {
      return "";
    }

    return element.getAttribute("port");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setGatewayPortData(String data) throws InvalidRequestException
  {
    checkValidity("setGatewayPortData");

    Element element = getElement("gateway");

    if (element != null)
    {
      element.setAttribute("port", data);
    }
  }

  /**
   * 
   */
  public String getNoResponseTimerData() throws InvalidRequestException
  {
    checkValidity("NoResponseTimerData");

    return getValue("noResponseTimer");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setNoResponseTimerData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setNoResponseTimerData");
    setValue("noResponseTimer", data);
  }


  /**
   * 
   */
  public String getAllowUnbillableCallsData() throws InvalidRequestException
  {
    checkValidity("AllowUnbillableCallsData");

    return getValue("allowUnbillableCalls");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setAllowUnbillableCallsData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setAllowUnbillableCallsData");
    setValue("allowUnbillableCalls", data);
  }


  /**
   * 
   */
  public String getConferenceBridgeNumberData() throws InvalidRequestException
  {
    checkValidity("ConferenceBridgeNumberData");

    return getValue("bridgeNumber");
  }

  /**
   * 
   */
  public void setConferenceBridgeNumberData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setConferenceBridgeNumberData");
    setValue("bridgeNumber", data);
  }

  /**
   * 
   */
  public Vector getAccessListData() throws InvalidRequestException
  {
    checkValidity("getAccessListData");

    Vector vector = new Vector();
    Element accessListElement = getElement("accessList");

    if (accessListElement == null)
    {
      return vector;
    }

    NodeList accessNumbers = 
      accessListElement.getElementsByTagName("accessNumber");

    for (int i = 0; i < accessNumbers.getLength(); i++)
    {
      Element accessNumber = (Element) accessNumbers.item(i);

      System.out.println("found access number: " 
                         + accessNumber.getAttribute("value"));
      vector.addElement(accessNumber.getAttribute("value"));
    }

    return vector;
  }

  /**
   * 
   * @param vector
   * 
   * @throws InvalidRequestException
   */
  public void setAccessListData(Vector vector) throws InvalidRequestException
  {
    checkValidity("setAccessListData");

    Element accessList = getElement("accessList");

    if (accessList == null)
    {

      // This will only happen when an old format is read.
      // So we'll live with the inefficiency of creating the node twice.
      Element marshalElement = getElement("marshalServer");
      Document doc = marshalElement.getOwnerDocument();

      accessList = doc.createElement("accessList");

      marshalElement.appendChild(accessList);
    }

    Node parent = accessList.getParentNode();
    Document doc = accessList.getOwnerDocument();
    Element newAccessList = doc.createElement("accessList");

    for (int i = 0; i < vector.size(); i++)
    {
      System.out.println("found access number at " + Integer.toString(i));

      Element accessNumber = doc.createElement("accessNumber");

      accessNumber.setAttribute("value", (String) vector.elementAt(i));
      newAccessList.appendChild(accessNumber);
    }

    parent.replaceChild(newAccessList, accessList);
  }

  // added to support Michael Foxworthy's work May 16, 2001
  public String getAllowUnknownCallers() throws InvalidRequestException
  {
    checkValidity("getAllowUnknownCallers");
    return getValue("allowUnknownCallers");
  }

  public void setAllowUnknownCallers(String data) 
    throws InvalidRequestException
  {
    checkValidity("setAllowUnknownCallers");
    setValue("allowUnknownCallers", data);
  }


  // methods from CdrServerData

  /**
   * 
   */
  public String getRadiusHostData() throws InvalidRequestException
  {
    checkValidity("getRadiusHostData");

    return getElement("radiusServer").getAttribute("host");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setRadiusHostData(String data) throws InvalidRequestException
  {
    checkValidity("setRadiusHostData");
    getElement("radiusServer").setAttribute("host", data);
  }

  /**
   * 
   */
  public String getRadiusRetriesData() throws InvalidRequestException
  {
    checkValidity("getRadiusRetriesData");

    return getElement("radiusServer").getAttribute("retries");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setRadiusRetriesData(String data) throws InvalidRequestException
  {
    checkValidity("setRadiusRetriesData");
    getElement("radiusServer").setAttribute("retries", data);
  }

  /**
   * 
   */
  public String getRadiusSecretKeyData() throws InvalidRequestException
  {
    checkValidity("getRadiusSecretKeyData");

    return getElement("radiusServer").getAttribute("secretKey");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setRadiusSecretKeyData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setRadiusSecretKeyData");
    getElement("radiusServer").setAttribute("secretKey", data);
  }

  /**
   * 
   */
  public String getBillingFrequencyData() throws InvalidRequestException
  {
    checkValidity("getBillingFrequencyData");

    return getElement("billing").getAttribute("frequency");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingFrequencyData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingFrequencyData");
    getElement("billing").setAttribute("frequency", data);
  }

  /**
   * 
   */
  public String getBillingDirPathData() throws InvalidRequestException
  {
    checkValidity("getBillingDirPathData");

    return getElement("billing").getAttribute("dirPath");
  }

  /**
   * 
   */
  public void setBillingDirPathData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingDirPathData");
    getElement("billing").setAttribute("dirPath", data);
  }

  /**
   * 
   */
  public String getBillingLockFileData() throws InvalidRequestException
  {
    checkValidity("getBillingLockFileData");

    return getElement("billing").getAttribute("lockFile");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingLockFileData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingLockFileData");
    getElement("billing").setAttribute("lockFile", data);
  }

  /**
   * 
   */
  public String getBillingDataFileData() throws InvalidRequestException
  {
    checkValidity("getBillingDataFileData");

    return getElement("billing").getAttribute("dataFile");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingDataFileData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingDataFileData");
    getElement("billing").setAttribute("dataFile", data);
  }

  /**
   * 
   */
  public String getBillingUnsentExtensionData() throws InvalidRequestException
  {
    checkValidity("getBillingUnsentExtensionData");

    return getElement("billing").getAttribute("unsentExtension");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingUnsentExtensionData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingUnsentExtensionData");
    getElement("billing").setAttribute("unsentExtension", data);
  }

  /**
   * 
   */
  public String getBillingRolloverSizeData() throws InvalidRequestException
  {
    checkValidity("getBillingRolloverSizeData");

    return getElement("billing").getAttribute("rolloverSize");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingRolloverSizeData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingRolloverSizeData");
    getElement("billing").setAttribute("rolloverSize", data);
  }

  /**
   * 
   */
  public String getBillingRolloverPeriodData() throws InvalidRequestException
  {
    checkValidity("getBillingRolloverPeriodData");

    return getElement("billing").getAttribute("rolloverPeriod");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setBillingRolloverPeriodData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setBillingRolloverPeriodData");
    getElement("billing").setAttribute("rolloverPeriod", data);
  }

  /**
   * 
   */
  public boolean getBillForRingtimeData() throws InvalidRequestException
  {
    checkValidity("getBillForRingtimeData");

    String truth = getElement("billing").getAttribute("billForRingtime");

    if (truth.equals("true"))
    {
      return true;
    }

    return false;
  }

  /**
   * 
   * @param truth
   * 
   * @throws InvalidRequestException
   */
  public void setBillForRingtimeData(boolean truth) 
    throws InvalidRequestException
  {
    checkValidity("setBillForRingtimeData");

    if (truth)
    {
      getElement("billing").setAttribute("billForRingtime", "true");
    }
    else
    {
      getElement("billing").setAttribute("billForRingtime", "false");
    }
  }

  // methods from RedirectData

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setSyncPortData(String data) throws InvalidRequestException
  {
    checkValidity("setSyncPortData");
    setValue("syncPort", data);
  }

  /**
   * 
   */
  public String getSyncPortData() throws InvalidRequestException
  {
    checkValidity("getSyncPortData");

    return getValue("syncPort");
  }

  // methods from PDPData

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setMaxThreadsData(String data) throws InvalidRequestException
  {
    checkValidity("setMaxThreadsData");
    setValue("maxThreads", data);
  }

  /**
   * 
   */
  public String getMaxThreadsData() throws InvalidRequestException
  {
    checkValidity("getMaxThreadsData");

    return getValue("maxThreads");
  }

  /**
   * 
   */
  public void setKeepAliveTimerData(String data) 
    throws InvalidRequestException
  {
    checkValidity("setKeepAliveTimerData");
    setValue("keepAliveTimer", data);
  }

  /**
   * 
   */
  public String getKeepAliveTimerData() throws InvalidRequestException
  {
    checkValidity("getKeepAliveTimerData");

    return getValue("keepAliveTimer");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setResponseTimeData(String data) throws InvalidRequestException
  {
    checkValidity("setResponseTimeData");
    setValue("responseTime", data);
  }

  /**
   * 
   */
  public String getResponseTimeData() throws InvalidRequestException
  {
    checkValidity("getResponseTimeData");

    return getValue("responseTime");
  }

  // methods from OSPData

  /**
   * 
   */
  public Vector getServicePointURL() throws InvalidRequestException
  {
    checkValidity("getSevicePointURL");

    Vector vector = new Vector();
    Element element = getElement("servicePoints");
    NodeList servicePoints = element.getElementsByTagName("servicePoint");

    for (int i = 0; i < servicePoints.getLength(); i++)
    {
      Element servicePoint = (Element) servicePoints.item(i);

      vector.addElement(servicePoint.getAttribute("url"));
    }

    return vector;
  }

  /**
   * 
   * @param vector
   * 
   * @throws InvalidRequestException
   */
  public void setServicePointURL(Vector vector) throws InvalidRequestException
  {
    checkValidity("setSevicePointURL");

    Element servicePoints = getElement("servicePoints");
    Node parent = servicePoints.getParentNode();
    Document doc = servicePoints.getOwnerDocument();
    Element newServicePoints = doc.createElement("servicePoints");

    for (int i = 0; i < vector.size(); i++)
    {
      Element newServicePoint = doc.createElement("servicePoint");

      newServicePoint.setAttribute("url", (String) vector.elementAt(i));
      newServicePoints.appendChild(newServicePoint);
    }

    parent.replaceChild(newServicePoints, servicePoints);
  }

  /**
   * 
   */
  public String getLocalValidation() throws InvalidRequestException
  {
    checkValidity("getLocalValidation");

    return getValue("localValidation");
  }

  /**
   * 
   */
  public void setLocalValidation(String data) throws InvalidRequestException
  {
    checkValidity("setLocalValidation");
    setValue("localValidation", data);
  }

  /**
   * 
   */
  public String getSslLifetime() throws InvalidRequestException
  {
    checkValidity("getSslLifetime");

    return getValue("sslLifetime");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setSslLifetime(String data) throws InvalidRequestException
  {
    checkValidity("setSslLifetime");
    setValue("sslLifetime", data);
  }


  /**
   * 
   */
  public String getHttpMaxConnections() throws InvalidRequestException
  {
    checkValidity("getHttpMaxConnections");

    return getValue("maxConnections");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setHttpMaxConnections(String data) 
    throws InvalidRequestException
  {
    checkValidity("setHttpMaxConnections");
    setValue("maxConnections", data);
  }

  /**
   * 
   */
  public String getHttpPersistence() throws InvalidRequestException
  {
    checkValidity("getHttpPersistence");

    return getValue("persistence");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setHttpPersistence(String data) throws InvalidRequestException
  {
    checkValidity("setHttpPersistence");
    setValue("persistence", data);
  }


  /**
   * 
   */
  public String getHttpRetryDelay() throws InvalidRequestException
  {
    checkValidity("getHttpRetryDelay");

    return getValue("retryDelay");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setHttpRetryDelay(String data) throws InvalidRequestException
  {
    checkValidity("setHttpRetryDelay");
    setValue("retryDelay", data);
  }


  /**
   * 
   */
  public String getHttpRetryLimit() throws InvalidRequestException
  {
    checkValidity("getHttpRetryLimit");

    return getValue("retryLimit");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setHttpRetryLimit(String data) throws InvalidRequestException
  {
    checkValidity("setHttpRetryLimit");
    setValue("retryLimit", data);
  }


  /**
   * 
   */
  public String getHttpTimeout() throws InvalidRequestException
  {
    checkValidity("getHttpTimeout");

    return getValue("timeout");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setHttpTimeout(String data) throws InvalidRequestException
  {
    checkValidity("setHttpTimeout");
    setValue("timeout", data);
  }


  /**
   * 
   */
  public String getOspExtension() throws InvalidRequestException
  {
    checkValidity("getOspExtension");

    return getValue("ospExtension");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setOspExtension(String data) throws InvalidRequestException
  {
    checkValidity("setOspExtension");
    setValue("ospExtension", data);
  }

  /**
   * 
   */
  public String getCustomerID() throws InvalidRequestException
  {
    checkValidity("getCustomerID");

    return getValue("customerID");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setCustomerID(String data) throws InvalidRequestException
  {
    checkValidity("setCustomerID");
    setValue("customerID", data);
  }


  /**
   * 
   */
  public String getDeviceID() throws InvalidRequestException
  {
    checkValidity("getDeviceID");

    return getValue("deviceID");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setDeviceID(String data) throws InvalidRequestException
  {
    checkValidity("setDeviceID");
    setValue("deviceID", data);
  }


  /**
   * 
   */
  public String getAuditURL() throws InvalidRequestException
  {
    checkValidity("getAuditURL");

    return getValue("auditURL");
  }

  /**
   * 
   * @param data
   * 
   * @throws InvalidRequestException
   */
  public void setAuditURL(String data) throws InvalidRequestException
  {
    checkValidity("setAuditURL");
    setValue("auditURL", data);
  }

  // abstract methods

  /**
   * 
   * @param methodName
   * 
   * @throws InvalidRequestException
   */
  protected abstract void checkValidity(String methodName) 
    throws InvalidRequestException;

  /**
   * 
   */
  protected abstract void setValue(String tagName, String data);

  /**
   * 
   */
  protected abstract String getValue(String tagName);

  /**
   * 
   */
  protected abstract Element getElement(String tagName);

}
