/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 * 
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 * 
 */


package vocal.userEditor;

/**
 * Specifies the generic interface which must be implemented by all panels which
 * are displayed in the "userEditor".<p>
 * It defines an interface for setting the data values displayed in each panel.
 */
public interface DataEditor
{

  /**
   * Get the data string associated with the given id. <P>
   * Each id is expected to be one of the integers defined in the UserTableModel
   * (these correspond to columns in the table). If the particular class that
   * implements this interface is responsible for displaying/editing the piece
   * of data associated with the given id, it should first check whether the
   * data has been modified. If it has, it should return a string representation
   * of the new value. If the data has not been changed, it should return null.
   * If the implementing class does not edit the data for the given id, it should
   * also return null.
   */
  public String getData(int dataId);

  /**
   * Clear all the editor elements. <p>
   * The implementing class is expected to reset/clear the values displayed
   * in the editor for each data element.
   */
  public void clear();

  /**
   * Set the data associated with the given id. <p>
   * Each id is expected to be one of the integers defined in the UserTableModel
   * (these correspond to columns in the table). If the particular class that
   * implements this interface is responsible for displaying/editing the piece
   * of data associated with the given id, it should display the given string
   * in the the appropriate element for that data.
   * @return true if this class was responsible for displaying this data element,
   * false otherwise
   */
  public boolean setData(int dataID, String data);

  /**
   * Set the mode for the components. <p>
   * The mode should be one of NEW_USER_MODE or EDIT_USER_MODE. This is because
   * there may be some fields which should be only available when the user's
   * data is being created for the first time as opposed to being edited. For
   * example, there needs to be a component to edit the user's name when the file
   * is first created, but it should not be changed later.
   */
  public void setMode(int mode);

  /**
   * Perform whatever checks are necessary to ensure that all data entered in
   * this component is valid. <p>
   * This would mean doing things like checking if the field can be left blank
   * or if it may contain spaces, or whatever.
   * @return true if all data is valid
   * false if at least one editor element contained invalid data
   */
  public boolean isDataValid();
}

