/*
 * ====================================================================
 * The Vovida Software License, Version 1.0
 *
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in
 * the documentation and/or other materials provided with the
 * distribution.
 *
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 * and "Vovida Open Communication Application Library (VOCAL)" must
 * not be used to endorse or promote products derived from this
 * software without prior written permission. For written
 * permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 * may "VOCAL" appear in their name, without prior written
 * permission of Vovida Networks, Inc.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 * ====================================================================
 *
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

package vocal.userEditor;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.XMLReader;
import org.xml.sax.SAXException;
import org.apache.xerces.parsers.DOMParser;
import org.apache.xerces.dom.ElementImpl;
import org.apache.xerces.dom.DocumentImpl;

import vocal.comm.VPPTransactionWrapper;
import vocal.comm.VPPNoSuchFileException;
import vocal.data.XMLUtils;
import vocal.data.NoSuchNodeException;
import vocal.data.NotTextNodeException;
import vocal.data.CplGenerator;
import vocal.data.ContactListGenerator;
import vocal.util.Logger;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.StringReader;

import java.net.ConnectException;

import java.util.*;

import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.TableModelEvent;
import javax.swing.event.TableColumnModelEvent;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Toolkit;
import java.awt.Point;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.awt.event.InputEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;


/**
 * This class displays the main table of all users. <p>
 *
 * It contains a popup menu which contains actions which allow the the table
 * to be manipulated. It is possible to delete a user, add a new user, edit
 * and existing user and view the data for a user.
 *
 * It should be possible to edit a single user account at a time or multiple
 * accounts. When multiple accounts are edited, the UserPanel which serves as
 * the editor is not loaded with data for any particular user. When the data is
 * read back from this dialog, only the data which has changed is saved for all
 * users while all other data remians unchanged. This allows for bulk
 * provisioning of users.
 *
 * In order to edit a user, a UserPanel is instantiated, the pServer is
 * queried for the user's data, then the UserPanel is instructed to load the
 * data into all its elements for display. The UserPanel is then dispalyed in
 * a modal dialog. When the dialog has been dismissed the data is collected from
 * the UserPanel and saved to the table model and the pServer.
 */
public class Editor extends JPanel
{

  /**
   * Displays the list of users and their configuration
   */
  private JTable table;

  /**
   * Contains the internal representation of the account data for all users
   */
  private UserTableModel data;

  /**
   * Menu which appears at right click to edit/delete users
   */
  private JPopupMenu popupMenu;

  private JCheckBox showUserColumns;
  private JCheckBox showAdminColumns;

  private JCheckBox showAliases;

  private JDialog userEditorDialog;
  private UserPanel userPanel;
  private JScrollPane scroller;

  private JDialog findDialog;

  VPPTransactionWrapper connection;

  ColumnModelListener columnListener = new ColumnModelListener();


  public Editor(VPPTransactionWrapper conn)
  {
    setLayout(new BorderLayout());

    connection = conn;

    table = new JTable();

    table.setAutoCreateColumnsFromModel(false);

    PServerInterface psInterface = new PServerInterface(conn);

    data = new UserTableModel(psInterface);

    table.setModel(data);

    // configure column widths
    ColumnInfo[] columns = data.getColumns();

    for (int i = 0; i < columns.length; i++)
    {
      table.addColumn(columns[i]);
    }

    table.setDefaultRenderer(String.class, new UserRenderer());

    // make the table scrollable and add it to the panel
    scroller = new JScrollPane(table);

    add(scroller, BorderLayout.CENTER);

    createPopupMenu();

    createButtons();

    table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);    // so it scrolls horizontally
  }


  /**
   * Create a new userEditor. This should include re-reading from the pserver
   * the provisioning information for the servers in the system (feature,
   * marshal). This is here so that it is possible to re-crete the
   * user editor dialog if there is any possibility that the server info
   * has changed (groups added, etc).
   */
  public void init()
  {

    // this needs to be wiped out because it stores the parsed lists of
    // feature and marshal servers
    PServerInterface psInterface = new PServerInterface(connection);

    Component parent = this.getTopLevelAncestor();

    try
    {
      if (JDialog.class.isInstance(parent))
      {
        userEditorDialog = new JDialog((JDialog) this.getTopLevelAncestor(),
                "Edit user", false);
      }
      else if (JFrame.class.isInstance(parent))
      {
        userEditorDialog = new JDialog((JFrame) this.getTopLevelAncestor(),
                "Edit user", false);
      }

      userPanel = new UserPanel(psInterface);

      // for those people with tiny monitors, as per request, make the user
      // editor panel scrollable.
      JScrollPane scroller = new JScrollPane(userPanel);

      userEditorDialog.setContentPane(scroller);
      userEditorDialog.setModal(true);
      userEditorDialog.pack();

      // This is done so that this
      // component is cleared even if the user closed this panel by clicking the
      // close icon in the title bar. (Tried an ancestorListener on the user
      // panel, but that gets called even if the dialog was dismissed via ok or
      // cancel.
      userEditorDialog.addWindowListener(userPanel);
    }
    catch (IOException e)
    {
      JOptionPane.showMessageDialog(this,
              "Could not create user panel because:\n" + e
              + "\n\nYou may browse users but you will not be able to add or edit.");
    }
    catch (VPPNoSuchFileException ex)
    {
      JOptionPane.showMessageDialog(this,
              "Could not create user panel because:\n" + ex
              + "\n\nYou may browse users but you will not be able to add or edit.");
    }

    findDialog = new JDialog();

    if (JDialog.class.isInstance(parent))
    {
      findDialog = new JDialog((JDialog) this.getTopLevelAncestor(),
              "Find user", false);
    }
    else if (JFrame.class.isInstance(parent))
    {
      findDialog = new JDialog((JFrame) this.getTopLevelAncestor(),
              "Find user", false);
    }

    findDialog.setContentPane(new UserSearchPanel(table, data));
    findDialog.setModal(true);
    findDialog.pack();

    columnListener.actionPerformed(new ActionEvent(showAdminColumns, 0,
            null));
    columnListener.actionPerformed(new ActionEvent(showUserColumns, 0, null));

    showAliases.setSelected(false);
  }


  private void createPopupMenu()
  {
    popupMenu = new JPopupMenu();

    AbstractAction action = new AbstractAction("View")
    {
      public void actionPerformed(ActionEvent e)
      {
        TaskView task = new TaskView(Editor.this);

        task.start();
      }

    };

    popupMenu.add(action);

    popupMenu.addSeparator();

    action = new AbstractAction("Edit")
    {
      public void actionPerformed(ActionEvent e)
      {
        if (!showUserColumns.isSelected() &&!showAdminColumns.isSelected())
        {
          JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                  "Select \"Show user data\" and/or \"Show admin data\" to edit an account.");
        }
        else
        {
          TaskEdit task = new TaskEdit(Editor.this);

          task.start();
        }
      }

    };

    popupMenu.add(action);

    action = new AbstractAction("Delete")
    {
      public void actionPerformed(ActionEvent e)
      {
        TaskDelete task = new TaskDelete(Editor.this);

        task.start();
      }

    };

    popupMenu.add(action);

    action = new AbstractAction("New")
    {
      public void actionPerformed(ActionEvent e)
      {
        if (userPanel == null)
        {
          return;
        }

        if (!showUserColumns.isSelected() &&!showAdminColumns.isSelected())
        {
          JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                  "Select \"Show user data\" and/or \"Show admin data\" to create a new account.");

          return;
        }

        userPanel.setMode(UserPanel.ADD_NEW_MODE);

        // Load the "default" values into the new user screen
        Vector dummyUser = data.getDummyUser();

        userPanel.setUserData(dummyUser);

        showUserEditorDialog();

        if (userPanel.getLastButtonClicked() == UserPanel.BUTTON_OK)
        {
          Vector newData = userPanel.getUserData();
          String newUsername = (String) newData.elementAt(data.USER_NAME);
          if (newUsername == null) //they didn't change the name so it must still be the default
          {
            newUsername = (String)dummyUser.elementAt(UserTableModel.USER_NAME);
          }

          // Check if another user with the same name exists already
          if (data.userExists(newUsername))
          {
            JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                    "The user named " + newUsername
                    + " already exists. New user not created.");

            return;
          }

          // Now still need to find out if there is an alias with that name:
          if (data.aliasExists(newUsername))
          {
            JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                    "An alias named " + newUsername
                    + " already exists. New user not created.");

            return;
          }

          data.addUser(dummyUser);

          int userRow = data.getRowCount() - 1;

          for (int i = 0; i <= UserTableModel.MAX_COLUMN_ID; i++)
          {
            String dataElement = (String) newData.elementAt(i);

            if (dataElement != null)
            {
              try
              {
                data.setValueAt(dataElement, userRow, i);
              }
              catch (Exception ex)
              {
                ex.printStackTrace();
                JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                        "Could not set value " + i + " for user "
                        + data.getValueAt(userRow, UserTableModel.USER_NAME)
                        + " because:\n\n " + ex);
              }
            }
          }

          data.fireTableDataChanged();

          try
          {
            data.writeBackUserAt(userRow, true, true);
          }
          catch (Exception ex)
          {
            ex.printStackTrace();
            JOptionPane.showMessageDialog(Editor.this.getTopLevelAncestor(),
                    "Could not save data for user "
                    + data.getValueAt(userRow, UserTableModel.USER_NAME)
                    + " because:\n\n " + ex);
          }

          // After the new user has been added, sort the main table
          data.sortByUserName();
          data.fireTableDataChanged();
        }
      }

    };

    popupMenu.add(action);

    RowSelectionListener listener = new RowSelectionListener();

    table.addMouseListener(listener);
    scroller.addMouseListener(listener);
  }


  private class RowSelectionListener extends MouseAdapter
  {
    public void mouseReleased(MouseEvent e)
    {
      if ((e.getModifiers() & InputEvent.BUTTON3_MASK)
          == InputEvent.BUTTON3_MASK)
      {

        // This is still not right. Needs more work. The problem is that the
        // e.getx method does not return the correct value once the table has
        // been scrolled.
        int width = Editor.this.getWidth();
        int modx = e.getX() % width;

        int height = Editor.this.getHeight();
        int mody = e.getY() % height;

        popupMenu.show(Editor.this, modx, mody);
      }
    }

  }

  private class ColumnModelListener implements ActionListener
  {
    private Vector removedUserColumns = new Vector();
    private Vector removedAdminColumns = new Vector();

    public void actionPerformed(ActionEvent e)
    {
      TableColumnModel columnModel = table.getColumnModel();
      JCheckBox source = (JCheckBox) e.getSource();

      if (source == showAdminColumns)
      {

        // The panel where users are edited may have not been created if one of
        // the system configuration files was missing from the pserver
        if (userPanel != null)
        {
          userPanel.setShowAdmin(source.isSelected());
        }

        if (source.isSelected())        // add columns
        {
          int max = removedAdminColumns.size();

          for (int i = 0; i < max; i++)
          {
            ColumnInfo column = (ColumnInfo) removedAdminColumns.elementAt(0);

            removedAdminColumns.remove(0);
            columnModel.addColumn(column);

            if (columnModel.getColumnCount() - 1 > column.getPosition())
            {
              columnModel.moveColumn(columnModel.getColumnCount() - 1,
                      column.getPosition());
            }
          }
        }
        else                            // remove columns
        {
          for (int i = 0; i < data.getColumns().length; i++)
          {
            ColumnInfo column = data.getColumns()[i];

            if (data.getAdminColumnIndices().contains(new Integer(column.getPosition())))
            {
              Enumeration columnsInModel = columnModel.getColumns();

              while (columnsInModel.hasMoreElements())
              {
                Object columnInModel = columnsInModel.nextElement();

                if (columnInModel.equals(column))
                {
                  columnModel.removeColumn(column);
                  removedAdminColumns.add(column);
                }                       // end if model contains column
              }                         // end while
            }
          }
        }
      }                                 // end source = admin
      else if (source == showUserColumns)
      {
        if (userPanel != null)
        {
          userPanel.setShowUser(source.isSelected());
        }

        if (source.isSelected())        // add columns
        {
          int max = removedUserColumns.size();

          for (int i = 0; i < max; i++)
          {
            ColumnInfo column = (ColumnInfo) removedUserColumns.elementAt(0);

            removedUserColumns.remove(0);
            columnModel.addColumn(column);

            if (columnModel.getColumnCount() - 1 > column.getPosition())
            {
              columnModel.moveColumn(columnModel.getColumnCount() - 1,
                      column.getPosition());
            }
          }
        }                               // end if add columns
        else                            // remove columns
        {
          for (int i = 0; i < data.getColumns().length; i++)
          {
            ColumnInfo column = data.getColumns()[i];

            // if this is a user column
            if (data.getUserColumnIndices().contains(new Integer(column.getPosition())))
            {
              Enumeration columnsInModel = columnModel.getColumns();

              while (columnsInModel.hasMoreElements())
              {
                Object columnInModel = columnsInModel.nextElement();

                if (columnInModel.equals(column))
                {
                  columnModel.removeColumn(column);
                  removedUserColumns.add(column);
                }                       // end if model contains column
              }                         // end while
            }                           // end if this is a user column
          }                             // end for all columns
        }                               // end remove user columns
      }
    }                                   // end method actionPerformed

  }

  private void createButtons()
  {
    showAdminColumns = new JCheckBox("Show admin data");

    showAdminColumns.setSelected(true);
    showAdminColumns.addActionListener(columnListener);

    showUserColumns = new JCheckBox("Show user data");

    showUserColumns.setSelected(false);
    showUserColumns.addActionListener(columnListener);

    JButton loadAll = new JButton("Load all users");

    loadAll.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        loadAllUsers();
      }

    });

    showAliases = new JCheckBox("Show aliases");

    showAliases.setSelected(false);
    showAliases.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        if (showAliases.isSelected())
        {
          data.showAliases(true);
        }
        else
        {
          data.showAliases(false);
        }
      }

    });

    JButton find = new JButton("Find");

    find.addActionListener(new ActionListener()
    {
      public void actionPerformed(ActionEvent e)
      {
        findDialog.setVisible(true);
      }

    });
    find.setToolTipText("Scroll to user with a given name");

    // layout the buttons in the panel
    JPanel panel = new JPanel();

    panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
    panel.add(loadAll);
    panel.add(find);
    panel.add(Box.createHorizontalGlue());
    panel.add(showAliases);
    panel.add(showAdminColumns);
    panel.add(showUserColumns);
    add(panel, BorderLayout.SOUTH);
  }

  public void loadUserNames()
  {
    TaskLoadUserNames task = new TaskLoadUserNames(this);

    task.start();
  }


  public void loadAllUsers()
  {
    TaskLoadAllUsers task = new TaskLoadAllUsers(this);

    task.start();
  }     // end method loadAllUsers


  /**
   * The renderer used to draw any cells which may have an on/off or true/false
   * value as a check box instead of a string. <p>
   * The checkbox cannot simply be used as the default renderer for the
   * concerned columns because they may not always have a value displayed in
   * them. When the data is not loaded yet from the provisioning server, the
   * value of these columns will be an empty string which will cause an
   * exception if a check box rendrer is used. So this class returns a check
   * box for cases where the value is loaded and an empty text label for cases
   * where it is not.
   */
  private class UserRenderer extends DefaultTableCellRenderer
  {
    private JCheckBox check = new JCheckBox();

    UserRenderer()
    {
      super();

      check.setHorizontalAlignment(check.CENTER);
    }

    public Component getTableCellRendererComponent(JTable table,
            Object value, boolean isSelected, boolean hasFocus, int row,
            int column)
    {
      TableModel model = table.getModel();
      Component renderer = this;

      // display the true/false and on/off values as check boxes
      int modelColumn = table.convertColumnIndexToModel(column);

      if (model.getValueAt(row, modelColumn).equals("true")
          || model.getValueAt(row, modelColumn).equals("ON"))
      {
        check.setSelected(true);

        renderer = check;
      }
      else if (model.getValueAt(row, modelColumn).equals("false")
               || model.getValueAt(row, modelColumn).equals("OFF"))

      {
        check.setSelected(false);

        renderer = check;
      }
      else if (model.getValueAt(row, UserTableModel.IS_ALIAS).equals("true"))
      {

        // make the aliases italic
        value = "<html><font color=gray size=-1><i>" + value.toString()
                + "</i></font>";
      }

      setValue(value);

      setFont(table.getFont());

      if (isSelected)
      {
        renderer.setBackground(table.getSelectionBackground());
        renderer.setForeground(table.getSelectionForeground());
      }
      else
      {
        renderer.setBackground(table.getBackground());
        renderer.setForeground(table.getForeground());
      }

      return renderer;
    }

  }


  public void showUserEditorDialog()
  {
    userEditorDialog.pack();

    Dimension d = Toolkit.getDefaultToolkit().getScreenSize();

    if (userEditorDialog.getHeight() > d.getHeight() - 80)
    {
      userEditorDialog.setSize(new Dimension(userEditorDialog.getWidth()
              + 20, (int) d.getHeight() - 80));
    }
    userEditorDialog.setVisible(true);
  }

  public JTable getTable()
  {
    return table;
  }

  public UserTableModel getData()
  {
    return data;
  }

  public UserPanel getUserPanel()
  {
    return userPanel;
  }

  public JDialog getUserEditorDialog()
  {
    return userEditorDialog;
  }

  public boolean isShowAdminColumnsSelected()
  {
    return showAdminColumns.isSelected();
  }

  public boolean isShowUserColumnsSelected()
  {
    return showUserColumns.isSelected();
  }

}       // end class JPanel

