
/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisionInterface_cxx_Version =
    "$Id: ProvisionInterface.cxx,v 1.72 2001/04/10 00:38:25 kevinmc Exp $";
#include <stdlib.h>
#include <stdio.h>
#include <strstream.h>
#include <unistd.h>
#include <errno.h>

#include "ProvisionInterface.hxx"

#include "VException.hxx"
#include "VBadDataException.hxx"
#include "VMissingDataException.hxx"
#include "VNetworkException.hxx"

#include "PSListener.hxx"
#include "cpLog.h"
#include "VPs_def.h"
#include "VXmlParser.hxx"
#include "NetworkAddress.h"

ProvisionInterface* ProvisionInterface::_instance = NULL;


ProvisionInterface&
ProvisionInterface::instance(const char* primHost, const int primPort,
                             const char* altHost, const int altPort)
throw (VException&)
{
    if (!_instance)
    {

        if (primHost == 0 || primPort == 0 || altHost == 0 || altPort == 0)
        {
            throw VNetworkException("PServer hostname and port not specified",
                                    __FILE__, __LINE__, 0);
        }

        _instance = new ProvisionInterface(primHost, primPort,
                                           altHost, altPort);

    }

    return *_instance;

}


ProvisionInterface::ProvisionInterface(const char* primHostname,
                                       const int primPort,
                                       const char *altHostname,
                                       const int altPort) :
        _interface(primHostname, primPort, altHostname, altPort)

{

    _listener = new PSListener();

#if defined(__linux__)
    //Destroy when program exits
    if (atexit(ProvisionInterface::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register ProvisionInterface::destroy");
    }
#endif

    char buf[256];
    gethostname(buf, 256);
    _hostName = buf;

}


ProvisionInterface::~ProvisionInterface()
{}



void
ProvisionInterface::destroy()
{
    delete ProvisionInterface::_instance;
    ProvisionInterface::_instance = NULL;
}


/// Reads an item from provision server
string
ProvisionInterface::readItem(const string& group, const string &name)
throw (VException&)
{

    return _interface.readItem(group, name);

}


/** Gets an item from Provision server -- called readFile for backwards comp.
    this one is called only by feature servers */
string
ProvisionInterface::readFile(const string& host, int port,
                             PContext context, const string& name)
throw (VException&)
{

    // this is a backward compat thing. name is either Called/<user>
    // or Calling/<user>, need to find out which...

    string whichList;
    string realName;
    if (name.find("Called") == string::npos)
    {
        whichList = "Calling";
    }
    else
    {
        whichList = "Called";
    }

    // needs to have a "/" since it should be as above...
    size_t loc = name.find("/");
    if (loc == string::npos)
    {
        throw VNetworkException("Invalid name given", __FILE__, __LINE__, 0);
    }

    realName = name.substr(loc + 1, name.size());

    char buf[256];
    string cntxt = dirContext(context);
    sprintf(buf, "%s_%s_%d_%s", cntxt.c_str(), host.c_str(), port,
            whichList.c_str());
    string group(buf);

    return _interface.readItem(group, realName);

}


/// Reads a list of items in a given group from the provision server
void
ProvisionInterface::getList(const string& group, vector < Data > & retList)
throw (VException&)
{

    _interface.getList(group, retList);
    return ;

}


/// Write a data item to the provision server
void
ProvisionInterface::writeItem(const string& group, const string &name,
                              const string& contents)
throw (VException&)
{

    _interface.writeItem(group, name, contents);
    return ;

}


void
ProvisionInterface::getConfigData(VGlobalConfigData &retVal)
throw(VException&)
{

    string dRead = _interface.readItem(VSYS_CONF, GLOBAL_CONFIG_ITEM);
    assert(dRead.size());

    retVal.parse(dRead);

}


void
ProvisionInterface::getOspServer(VOspServerData &retVal)
throw(VException&)
{

    string dRead = _interface.readItem(VSYS_CONF, OSP_CONFIG_ITEM);
    assert(dRead.size());

    retVal.parse(dRead);

}


void
ProvisionInterface::getJtapiServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeJtapiServerList();
    ListOfServers tList = _jtapiServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getFeatureServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeFeatureServerList();
    ListOfServers tList = _featureServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getMarshalServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeMarshalServerList();
    ListOfServers tList = _marshalServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getCdrServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeCdrServerList();
    ListOfServers tList = _cdrServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getHeartbeatServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeHeartbeatServerList();
    ListOfServers tList = _heartbeatServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getRedirectServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeRedirectServerList();
    ListOfServers tList = _redirectServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getPolicyServers(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializePolicyServerList();
    ListOfServers tList = _policyServerGrpCont.getServerList();
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getJtapiGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeJtapiServerList();
    list < string > tList;
    _jtapiServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getFeatureGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeFeatureServerList();
    list < string > tList;
    _featureServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getMarshalGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeMarshalServerList();
    list < string > tList;
    _marshalServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getCdrGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeCdrServerList();
    list < string > tList;
    _cdrServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getHeartbeatGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeHeartbeatServerList();
    list < string > tList;
    _heartbeatServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}

void
ProvisionInterface::getRedirectGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializeRedirectServerList();
    list < string > tList;
    _redirectServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getPolicyGroups(list < string > &retList)
throw(VException&)
{

    // delete anything in there when it is passed
    retList.clear();

    initializePolicyServerList();
    list < string > tList;
    _policyServerGrpCont.getGroupList(tList);
    for (ListOfServers::iterator itr = tList.begin();
            itr != tList.end(); itr++)
    {
        retList.push_back(*itr);
    }

}


void
ProvisionInterface::getJtapiHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeJtapiServerList();
    _jtapiServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getFeatureHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeFeatureServerList();
    _featureServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getMarshalHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeMarshalServerList();
    _marshalServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getCdrHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeCdrServerList();
    _cdrServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getRedirectHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeRedirectServerList();
    _redirectServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getHeartbeatHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializeHeartbeatServerList();
    _heartbeatServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getPolicyHostToGroupMap(HostToGroupMap &retList)
throw(VException&)
{

    initializePolicyServerList();
    _policyServerGrpCont.getHostToGroupMap(retList);

}


void
ProvisionInterface::getJtapiData(const string& jtapiName,
                                 VJtapiServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VJTAPIS, jtapiName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for jtapi %s",
                    jtapiName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access Jtapi server data.");
        throw VMissingDataException("Failed to access Jtapi server data.",
                                    __FILE__, __LINE__, 0);
    }

}


void
ProvisionInterface::getFeatureData(const string& featureName,
                                   VFeatureServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VFEATURES, featureName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for Feature %s",
                    featureName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access feature server data.");
        throw VMissingDataException("Failed to access feature server data.",
                                    __FILE__, __LINE__, 0);
    }

}


void
ProvisionInterface::getMarshalData(const string& marshalName,
                                   VMarshalServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VMARSHALS, marshalName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for marshal %s",
                    marshalName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access marshal server data.");
        throw VMissingDataException("Failed to access marshal server data.",
                                    __FILE__, __LINE__, 0);
    }

}

void
ProvisionInterface::getLocationData(const string& locationName,
                                   VLocationServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VLOCATIONS, locationName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for location server %s",
                    locationName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access location server data.");
        throw VMissingDataException("Failed to access location server data.",
                                    __FILE__, __LINE__, 0);
    }

}

void
ProvisionInterface::getLocationRouteData(const string& locationRouteName,
                                   VLocationRouteData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VLOCATIONS, locationRouteName);

    if (dRead.size())
    {
        try
        {                             
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for location server routes %s",
                    locationRouteName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access location server route data.");
        throw VMissingDataException("Failed to access location server route data.",
                                    __FILE__, __LINE__, 0);
    }

}




void
ProvisionInterface::getCdrData(const string& cdrName,
                               VCdrServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VCDRS, cdrName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for Cdr %s",
                    cdrName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access Cdr server data.");
        throw VMissingDataException("Failed to access Cdr server data.",
                                    __FILE__, __LINE__, 0);
    }

}


void
ProvisionInterface::getRedirectData(const string& redirectName,
                                    VRedirectServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VRSERVERS, redirectName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for redirect %s",
                    redirectName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access redirect server data.");
        throw VMissingDataException("Failed to access redirect server data.",
                                    __FILE__, __LINE__, 0);
    }

}


void
ProvisionInterface::getPolicyData(const string& policyName,
                                  VPolicyServerData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VPOLICIES, policyName);

    if (dRead.size())
    {
        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for Policy %s",
                    policyName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access policy server data.");
        throw VMissingDataException("Failed to access policy server data.",
                                    __FILE__, __LINE__, 0);
    }

}


void
ProvisionInterface::getAccountData(const string& userName,
                                   VAccountData& data)
throw (VException&)
{

    string dRead = _interface.readItem(VACCOUNTS, userName);

    if (dRead.size())
    {

        try
        {
            data.parse(dRead);
        }
        catch (VException& e)
        {
            char buf[256];
            sprintf(buf, "Failed to parse XML data for Account %s",
                    userName.c_str());
            cpLog(LOG_ALERT, buf);
            throw VBadDataException(buf, __FILE__, __LINE__, 0);
        }

    }
    else
    {

        cpLog(LOG_ALERT, "Failed to access requested account data.");
        throw VMissingDataException("Failed to access requested account data.",
                                    __FILE__, __LINE__, 0);
    }

}


string
ProvisionInterface::getUserIp(const string& userName)
throw (VException&)
{

    string retStr;
    string dRead = _interface.readItem(VACCOUNTS, userName);

    if (dRead.size())
    {
        VXmlParser parser(dRead);
        retStr = parser.getContent("user", "ipaddr");
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access user account.");
        throw VMissingDataException("Failed to access user account",
                                    __FILE__, __LINE__, 0);
    }

    return retStr;

}


string
ProvisionInterface::getMasterUser(const string& userName)
throw (VException&)
{

    string retStr;
    string dRead = _interface.readItem(VALIASES, userName);

    if (dRead.size())
    {
        VXmlParser parser(dRead);
        xmlNode *theNode = parser.findNode(NULL, "userAlias");
        retStr = parser.getAttribute(theNode, "masterUser");
    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access user account.");
        throw VMissingDataException("Failed to access user account",
                                    __FILE__, __LINE__, 0);
    }

    return retStr;

}


bool
ProvisionInterface::isAccount(const string& userName)
{

    try
    {
        string dRead = _interface.readItem(VACCOUNTS, userName);
        return true;
    }
    catch (VMissingDataException)
    {
        return false;
    }

}


void
ProvisionInterface::getAliases(const string& userName, list < string > &retVal)
throw (VException&)
{

    retVal.clear();

    string dRead = _interface.readItem(VACCOUNTS, userName);

    if (dRead.size())
    {
        VXmlParser parser(dRead);
        xmlNode *theNode = parser.findNode(NULL, "user");
        list < xmlNodePtr > listOfAliases;

        parser.findNode(theNode, "alias", listOfAliases);

        for (list < xmlNodePtr > ::iterator itr = listOfAliases.begin();
                itr != listOfAliases.end(); itr++)
        {
            xmlNodePtr aliasNode = (*itr);
            // get value attribute from node
            string val = parser.getAttribute(aliasNode, "value");
            retVal.push_back(val);
        }

    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access user account.");
        throw VMissingDataException("Failed to access user account",
                                    __FILE__, __LINE__, 0);
    }

    return ;

}


bool
ProvisionInterface::getTermContactList(const string& userName,
                                       list < string > &retVal)
throw (VException&)
{

    retVal.clear();

    string dRead = _interface.readItem(VACCOUNTS, userName);

    if (dRead.size())
    {
        VXmlParser parser(dRead);

        xmlNode *theNode = parser.findNode(NULL, "staticRegistration");
        string retStr = parser.getAttribute(theNode, "value");
        if (retStr == "true")
        {

            theNode = parser.findNode(NULL, "user");
            list < xmlNodePtr > listOfContacts;

            parser.findNode(theNode, "contact", listOfContacts);

            for (list < xmlNodePtr > ::iterator itr = listOfContacts.begin();
                    itr != listOfContacts.end(); itr++)
            {
                xmlNodePtr aliasNode = (*itr);
                // get value attribute from node
                string val = parser.getAttribute(aliasNode, "value");
                retVal.push_back(val);
            }

            return true;

        }
        else
        {

            // no list
            return false;

        }

    }
    else
    {
        cpLog(LOG_ALERT, "Failed to access user account.");
        throw VMissingDataException("Failed to access user account",
                                    __FILE__, __LINE__, 0);
    }

    return false;

}


void
ProvisionInterface::getSubscribers(vector < Data > & retList)
throw (VException &)
{

    _interface.getList(VACCOUNTS, retList);

}


void
ProvisionInterface::getAllAliases(vector < Data > & retList)
throw (VException&)
{

    _interface.getList(VALIASES, retList);

}


void
ProvisionInterface::deleteSubscriber(const string& subscriberName)
throw (VException&)
{

    _interface.deleteSubscriber(subscriberName);

}


void
ProvisionInterface::getCalledList(const string&, int,
                                  const string& acctName, vector < Data > & retData)
throw (VException&)
{

    // Read user data from the PS
    string dRead = _interface.readItem(VCALLEDLIST, acctName);
    VXmlParser parser(dRead);
    list < string > contents;
    parser.getContent("contact", "host", contents);

    for (list < string > ::iterator itr = contents.begin();
            itr != contents.end(); itr++)
    {
        retData.push_back(*itr);
    }

}


void
ProvisionInterface::getCallingList(const string&,
                                   int, const string& acctName, vector < Data > & retData)
throw (VException&)
{

    // Read user data from the PS
    string dRead = _interface.readItem(VCALLINGLIST, acctName);
    VXmlParser parser(dRead);
    list < string > contents;
    parser.getContent("contact", "host", contents);

    for (list < string > ::iterator itr = contents.begin();
            itr != contents.end(); itr++)
    {
        retData.push_back(*itr);
    }

}


void
ProvisionInterface::getCalledFeatureScripts(const string& hostName,
        int port, vector < Data > & retList)
throw (VException&)
{

    char buf[256];
    sprintf(buf, "%s_%s_%d_Called", VFEATURES, hostName.c_str(), port);
    _interface.getList(buf, retList);

}


void
ProvisionInterface::getCallingFeatureScripts(const string& hostName,
        int port, vector < Data > & retList)
throw (VException&)
{

    char buf[256];
    sprintf(buf, "%s_%s_%d_Calling", VFEATURES, hostName.c_str(), port);
    _interface.getList(buf, retList);

}


void
ProvisionInterface::getDialPlan(const string& name,
                                vector < VPlanElement > & retData)
throw (VException &)
{

    cpLog(LOG_DEBUG, "getPhoneDialPlan: %s", name.c_str());

    string xmlData = _interface.readItem(VSYS_CONF, name);
    if (!xmlData.size())
    {
        throw VMissingDataException("No dialplan data to parse",
                                    __FILE__, __LINE__, 0);
    }

    VXmlParser parser(xmlData);

    list < xmlNodePtr > listOfKeys;
    parser.findNode(NULL, "key", listOfKeys);

    for (list < xmlNodePtr > ::iterator itr = listOfKeys.begin();
            itr != listOfKeys.end(); itr++)
    {

        xmlNodePtr keyNode = (*itr);
        string regKey = parser.getAttribute(keyNode, "value");

        list < xmlNodePtr > listOfContacts;
        parser.findNode(keyNode, "contact", listOfContacts);

        VPlanElement vElem(regKey);
        for (list < xmlNodePtr > ::iterator itr2 = listOfContacts.begin();
                itr2 != listOfContacts.end(); itr2++)
        {
            xmlNodePtr contactNode = (*itr2);
            string contactStr = parser.getAttribute(contactNode,
                                                    "value");
            vElem.insertContact(contactStr);
        }

        retData.push_back(vElem);

    }

}


void
ProvisionInterface::registerForUpdateImpl(CallBackFunc func,
        const string& name)
{

    if (_callBackMap.count(name))
    {
        _callBackMap[name].push_back(func);
    }
    else
    {
        CallBackList aList;
        aList.push_back(func);
        _callBackMap[name] = aList;
    }

}


void
ProvisionInterface::registerForUpdate(CallBackFunc func,
                                      const string& name, PContext context)
{

    string group = dirContext(context);
    char hashkey[512];

    sprintf(hashkey, "ITEM_%s_%s", group.c_str(), name.c_str());
    registerForUpdateImpl(func, hashkey);

    // register with the ProvisionServer
    string hName(_hostName);
    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    hName += ":";
    char portBuf[56];
    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    _interface.registerForItem(group, name, hName);

}


void
ProvisionInterface::registerForUpdate(const string& host, int port,
                                      CallBackFunc func, PContext context)
{

    // this is actually a group registration that is called for registering
    // by asking for context (which gets mapped to group)

    string hName(_hostName);

    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    hName += ":";
    char portBuf[56];

    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    char buf[256];
    string cntxt = dirContext(context);
    char hashkey[512];
    string group;

    // If feature is mapped, there are actually two
    if (context == VFEATURE)
    {

        sprintf(buf, "%s_%s_%d_Called", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        registerForUpdateImpl(func, hashkey);
        _interface.registerForGroup(group, hName);

        sprintf(buf, "%s_%s_%d_Calling", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        registerForUpdateImpl(func, hashkey);
        _interface.registerForGroup(group, hName);

    }
    else
    {

        sprintf(buf, "%s_%s_%d", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        registerForUpdateImpl(func, hashkey);
        _interface.registerForGroup(group, hName);

    }

}


void
ProvisionInterface::registerDirForUpdate(CallBackFunc func,
        const string &group)
{

    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    string hName(_hostName);
    hName += ":";
    char portBuf[56];
    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    char hashkey[512];

    sprintf(hashkey, "GROUP_%s", group.c_str());
    registerForUpdateImpl(func, hashkey);

    _interface.registerForGroup(group, hName);

}


void
ProvisionInterface::unregisterForUpdateImpl(CallBackFunc func,
        const string& name)
{

    if (_callBackMap.count(name))
    {

        CallBackList& aList = _callBackMap[name];
        for (CallBackList::iterator i = aList.begin();
                i != aList.end(); i++)
        {
            if ((*i) == func) aList.erase(i);
            break;
        }

    }

}


void
ProvisionInterface::unregisterForUpdate(CallBackFunc func,
                                        const string& name, PContext context)
{

    string group = dirContext(context);
    char hashkey[512];

    sprintf(hashkey, "ITEM_%s_%s", group.c_str(), name.c_str());
    unregisterForUpdateImpl(func, hashkey);

    // unregister with the ProvisionServer
    string hName(_hostName);
    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    hName += ":";
    char portBuf[56];
    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    _interface.unregisterForItem(group, name, hName);

}


void
ProvisionInterface::unregisterForUpdate(const string& host, int port,
                                        CallBackFunc func, PContext context)
{

    // this is actually a group unregistration that is called for unregistering
    // by asking for context (which gets mapped to group)

    string hName(_hostName);

    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    hName += ":";
    char portBuf[56];

    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    char buf[256];
    string cntxt = dirContext(context);
    char hashkey[512];
    string group;

    // If feature is mapped, there are actually two
    if (context == VFEATURE)
    {

        sprintf(buf, "%s_%s_%d_Called", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        unregisterForUpdateImpl(func, hashkey);
        _interface.unregisterForGroup(group, hName);

        sprintf(buf, "%s_%s_%d_Calling", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        unregisterForUpdateImpl(func, hashkey);
        _interface.unregisterForGroup(group, hName);

    }
    else
    {

        sprintf(buf, "%s_%s_%d", cntxt.c_str(), host.c_str(), port);
        group = buf;
        sprintf(hashkey, "GROUP_%s", group.c_str());
        unregisterForUpdateImpl(func, hashkey);
        _interface.unregisterForGroup(group, hName);

    }

}


void
ProvisionInterface::unregisterGroupForUpdate(CallBackFunc func,
        const string &group)
{

    // Put the listener port where the notification updates would
    // be received, set the listener port that the PSListener is bound to
    string hName(_hostName);
    hName += ":";
    char portBuf[56];
    sprintf(portBuf, "%d", _listener->getListenerPort());
    hName += portBuf;

    char hashkey[512];

    sprintf(hashkey, "GROUP_%s", group.c_str());
    unregisterForUpdateImpl(func, hashkey);
    _interface.unregisterForGroup(group, hName);

}


void
ProvisionInterface::sendUpdateNotification(const string &item,
        const string &name)
{

    char hashkey[512];
    sprintf(hashkey, "ITEM_%s_%s", item.c_str(), name.c_str());

    if (_callBackMap.count(hashkey))
    {
        // Update the config update flag if the name is a list of servers
        updateConfigCallback(name);
        try
        {
            string dRead = _interface.readItem(item, name);
            CallBackList& aList = _callBackMap[hashkey];
            cpLog(LOG_DEBUG,
                  "Sending update notice to %d registrant for item (%s,%s)",
                  aList.size(), item.c_str(), name.c_str());

            for (CallBackList::iterator i = aList.begin();
                    i != aList.end(); i++)
            {
                (*i)(dRead, name, false);
            }
        }

        catch (VNetworkException& e)
        {
            cpLog(LOG_ALERT, "sendUpdateNotification failed, reason:%s",
                  e.getDescription().c_str());
        }

    }

}


void
ProvisionInterface::sendUpdateNotification(const string& group,
        const string& name,
        bool delflag)
{

    // this function gets called by listener when data is updated
    // on the server. We check if the data is a list of servers and
    // if it is, we invalidate the "cache" entry for that data.

    cpLog(LOG_DEBUG, "Send update dir notification for group (%s)",
          group.c_str());

    char hashkey[512];
    sprintf(hashkey, "GROUP_%s", group.c_str());

    if (_callBackMap.count(hashkey))
    {
        // Set the config update flag if this is a list of servers
        updateConfigCallback(name);
        try
        {
            CallBackList& aList = _callBackMap[hashkey];
            cpLog(LOG_DEBUG,
                  "Sending update notice to %d registrant for group (%s)",
                  aList.size(), group.c_str());

            for (CallBackList::iterator i = aList.begin();
                    i != aList.end(); i++)
            {
                (*i)(name, group, delflag);
            }
        }

        catch (VNetworkException& e)
        {
            cpLog(LOG_ALERT, "sendUpdateNotification failed, reason:%s",
                  e.getDescription().c_str());
        }

    }

}


void
ProvisionInterface::sendDeleteNotification(const string &item,
        const string &name)
{



    char hashkey[512];
    sprintf(hashkey, "ITEM_%s_%s", item.c_str(), name.c_str());

    if (_callBackMap.count(hashkey))
    {
        //Update the flag if we get a list of servers
        updateConfigCallback(name);
        try
        {
            CallBackList& aList = _callBackMap[hashkey];
            cpLog(LOG_DEBUG,
                  "Sending delete notice to %d registrant for item (%s,%s)",
                  aList.size(), item.c_str(), name.c_str());

            for (CallBackList::iterator i = aList.begin();
                    i != aList.end(); i++)
            {
                (*i)("", name, true);
            }
        }

        catch (VNetworkException& e)
        {
            cpLog(LOG_ALERT, "sendUpdateNotification failed, reason:%s",
                  e.getDescription().c_str());
        }

    }

}


void
ProvisionInterface::initializeJtapiServerList() throw(VException&)
{

    if (!_jtapiServerGrpCont.needsUpdate()) return ;

    /// Read the jtapi server list
    string dRead = _interface.readItem(VSYS_CONF, JTAPI_CONFIG_ITEM);
    assert(dRead.size());
    _jtapiServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializeFeatureServerList() throw(VException&)
{

    if (!_featureServerGrpCont.needsUpdate()) return ;

    /// Read the feature server list
    string dRead = _interface.readItem(VSYS_CONF, FEATURE_CONFIG_ITEM);
    assert(dRead.size());
    _featureServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializeMarshalServerList() throw(VException&)
{

    if (!_marshalServerGrpCont.needsUpdate()) return ;

    /// Read the feature server list
    string dRead = _interface.readItem(VSYS_CONF, MARSHAL_CONFIG_ITEM);
    assert(dRead.size());
    _marshalServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializeCdrServerList() throw(VException&)
{

    if (!_cdrServerGrpCont.needsUpdate()) return ;

    /// Read the cdr server list
    string dRead = _interface.readItem(VSYS_CONF, CDR_CONFIG_ITEM);
    assert(dRead.size());
    _cdrServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializeHeartbeatServerList() throw(VException&)
{

    if (!_heartbeatServerGrpCont.needsUpdate()) return ;

    /// Read the heartbeat server list
    string dRead = _interface.readItem(VSYS_CONF, HEARTBEAT_CONFIG_ITEM);
    assert(dRead.size());
    _heartbeatServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializeRedirectServerList() throw(VException&)
{

    if (!_redirectServerGrpCont.needsUpdate()) return ;

    /// Read the redirect server list
    string dRead = _interface.readItem(VSYS_CONF, REDIRECT_CONFIG_ITEM);
    assert(dRead.size());
    _redirectServerGrpCont.initializeMap(dRead);

}


void
ProvisionInterface::initializePolicyServerList() throw(VException&)
{

    if (!_policyServerGrpCont.needsUpdate()) return ;

    /// Read the policy server list
    string dRead = _interface.readItem(VSYS_CONF, POLICY_CONFIG_ITEM);
    assert(dRead.size());
    _policyServerGrpCont.initializeMap(dRead);

}


string
ProvisionInterface::dirContext(PContext context)
{
    string cntxt;
    switch (context)
    {
        case ProvisionInterface::VFEATURE:
        cntxt = VFEATURES;
        break;
        case ProvisionInterface::VREDIRECT:
        cntxt = VRSERVERS;
        break;
        case ProvisionInterface::VJTAPI:
        cntxt = VJTAPIS;
        break;
        case ProvisionInterface::VLOCATION:
        cntxt = VLOCATIONS;
        break;
        case ProvisionInterface::VMARSHAL:
        cntxt = VMARSHALS;
        break;
        case ProvisionInterface::VCDR:
        cntxt = VCDRS;
        break;
        case ProvisionInterface::VCONFIG:
        cntxt = VSYS_CONF;
        break;
        case ProvisionInterface::VPOLICY:
        cntxt = VPOLICIES;
        break;
        case ProvisionInterface::VNONE:
        cntxt = "";
        break;
        default:
        assert(0);
        break;
    }
    return cntxt;
}


/// Function call to invalidate the local cached copies of the lists.
void
ProvisionInterface::updateConfigCallback(const string& name)
{

    // This is called when an update is received to invalidate the local
    // cached copies of each of the list of servers.

    if (name.find(FEATURE_CONFIG_ITEM) != string::npos)
    {
        _featureServerGrpCont.markForUpdate();
    }
    else if (name.find(MARSHAL_CONFIG_ITEM) != string::npos)
    {
        _marshalServerGrpCont.markForUpdate();
    }
    else if (name.find(REDIRECT_CONFIG_ITEM) != string::npos)
    {
        _redirectServerGrpCont.markForUpdate();
    }
    else if (name.find(CDR_CONFIG_ITEM) != string::npos)
    {
        _cdrServerGrpCont.markForUpdate();
    }
    else if (name.find(JTAPI_CONFIG_ITEM) != string::npos)
    {
        _jtapiServerGrpCont.markForUpdate();
    }
    else if (name.find(POLICY_CONFIG_ITEM) != string::npos)
    {
        _policyServerGrpCont.markForUpdate();
    }
}
