#ifndef ProvisionInterface_hxx
#define ProvisionInterface_hxx

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisionInterface_hxx_Version =
    "$Id: ProvisionInterface.hxx,v 1.54 2001/04/10 00:38:25 kevinmc Exp $";

#include <memory>
#include <list>
#include <map>
#include <string>
#include <vector>

// class forward definitions

// server list containers
#include "VFeatureServerGroupContainer.hxx"
#include "VMarshalServerGroupContainer.hxx"
#include "VRedirectServerGroupContainer.hxx"
#include "VCdrServerGroupContainer.hxx"
#include "VHeartbeatServerGroupContainer.hxx"
#include "VJtapiServerGroupContainer.hxx"
#include "VPolicyServerGroupContainer.hxx"

// dialplan element
#include "VPlanElement.hxx"

// server data containers
#include "VRedirectServerData.hxx"
#include "VFeatureServerData.hxx"
#include "VJtapiServerData.hxx"
#include "VMarshalServerData.hxx"
#include "VLocationServerData.hxx"
#include "VLocationRouteData.hxx"
#include "VPolicyServerData.hxx"
#include "VGlobalConfigData.hxx"
#include "VCdrServerData.hxx"
#include "VSuavmServerData.hxx"
#include "VOspServerData.hxx"
#include "VAccountData.hxx"

#include "PSInterface.hxx"

#include "Data.hxx"

class NetworkAddress;
class VException;
class PSListener;

/**Function prototype, a client would define the callback function as.
   data: Contents of the item updated
   name: Name of the data item updated
*/
typedef void (*CallBackFunc)(const string& data, const string& name,
                             bool delFlag);


/** Singelton class - Only one instance is available per application.
    Implements the communications to the provision server
    and notify content update as it happens on the Provision server.
    In order to receive notification client has to register
    itself with the class with a callback function along with the item name
    for which update notification is requested. Whenever the item changes on
    the provision server, the callback function is called with the content
    of the item passed as the argument.
*/

class ProvisionInterface
{
    public:

        enum PContext {
            VNONE = 1,
            VFEATURE,
            VFEATUREX,
            VREDIRECT,
            VPOLICY,
            VMARSHAL,
            VCONFIG,
            VCDR,
            VJTAPI,
            VLOCATION,
            VMAX
    } ;

        /// Interface to get the Singelton ProvisonInterface instance
        static ProvisionInterface& instance(const char* primHost = 0,
                                            const int primPort = 0,
                                            const char* altHost = 0,
                                            const int altPort = 0)
        throw (VException&);

        /**
           Frees the memory associated the Instance.Gets register with
           atexit(), so that when program exits, it will be called automatically.
        */
        static void destroy();

        /// Destructor
        ~ProvisionInterface();


        // *****************************************************************
        // Item reading functions
        // *****************************************************************

        /// Reads an item from provision server
        string readItem(const string &group, const string &item)
        throw (VException&);

        /// Get an Item from server -- backward compatibility (hence "file")
        string readFile(const string& host, int port, PContext context,
                        const string& file)
        throw (VException&);

        /// Reads a list of items in given group from provision server
        void getList(const string& group, vector < Data > & retList)
        throw (VException&);

        /// Write an item to provision server
        void writeItem(const string& group, const string &name,
                       const string& contents)
        throw (VException&);


        // *****************************************************************
        // Server functions and data functions
        // *****************************************************************

        // All of these are by reference (rather than returning) for
        // efficency reasons (per Surendra) even though it makes things
        // uglier

        /// Get global coniguration data from server
        void getConfigData(VGlobalConfigData &retVal) throw(VException&);

        /// Get OspServer data from server
        void getOspServer(VOspServerData &retVal) throw(VException&);


        // get lists of the servers
        /**
         * Get a listing of all the Jtapi servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getJtapiServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Feature servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getFeatureServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Marshall servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getMarshalServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Cdr servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getCdrServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Heartbeat servers in the form
         * host:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getHeartbeatServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Redirect servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getRedirectServers(list < string > &retList) throw(VException&);

        /**
         * Get a listing of all the Policy servers in the form
         * host:port:port
         *
         * Pass in an empty list of strings, which will be filled
         * any elements in the list when passed in are deleted
         */
        void getPolicyServers(list < string > &retList) throw(VException&);


        // get the groups each server is in
        /** Returns all the groups defined in the Jtapi server list
         * config file.
         */
        void getJtapiGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Feature server list
         * config data.
         */
        void getFeatureGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Marshal server list
         * config data.
         */
        void getMarshalGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Cdr server list
         * config data.
         */
        void getCdrGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Heartbeat server list
         * config data.
         */
        void getHeartbeatGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Redirect server list
         * config data.
         */
        void getRedirectGroups(list < string > &retList) throw(VException&);

        /** Returns all the groups defined in the Policy server list
         * config data.
         */
        void getPolicyGroups(list < string > &retList) throw(VException&);


        // get host to group maps
        /** Return a map of host to group for the Jtapi servers
         */
        void ProvisionInterface::getJtapiHostToGroupMap(HostToGroupMap &retList)
        throw(VException&);

        /** Return a map of host to group for the Feature servers
         */
        void ProvisionInterface::getFeatureHostToGroupMap(HostToGroupMap &retList)
        throw(VException&);

        /** Return a map of host to group for the Marshal servers
         */
        void ProvisionInterface::getMarshalHostToGroupMap(HostToGroupMap &retList)
        throw(VException&);

        /** Return a map of host to group for the Cdr servers
         */
        void ProvisionInterface::getCdrHostToGroupMap(HostToGroupMap &retList)
        throw(VException&);

        /** Return a map of host to group for the Heartbeat servers
         */
        void ProvisionInterface::getHeartbeatHostToGroupMap(
            HostToGroupMap &retList) throw(VException&);

        /** Return a map of host to group for the Redirect servers
         */
        void ProvisionInterface::getRedirectHostToGroupMap(
            HostToGroupMap &retList) throw(VException&);

        /** Return a map of host to group for the Redirect servers
         */
        void ProvisionInterface::getPolicyHostToGroupMap(
            HostToGroupMap &retList) throw(VException&);


        // Given a server name, get the server data for that server
        /** Takes a jtapi server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getJtapiData(const string &jtapiName, VJtapiServerData &data)
        throw (VException &);

        /** Takes a feature server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getFeatureData(const string &featureName, VFeatureServerData &data)
        throw (VException &);

        /** Takes a marshal name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getMarshalData(const string &marshalName, VMarshalServerData &data)
        throw (VException &);

        /** Takes a location server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getLocationData(const string &locationName, VLocationServerData &data)
        throw (VException &);

        /** Takes a location server routes file name, obtains the data
         * from provision server and fills it in the data object provided.
         */
        void getLocationRouteData(const string &locationRouteName, VLocationRouteData &data)
        throw (VException &);

        /** Takes a Cdr server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getCdrData(const string &cdrName, VCdrServerData &data)
        throw (VException &);

        /** Takes a redirect server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getRedirectData(const string &redirectName, VRedirectServerData &data)
        throw (VException &);

        /** Takes a redirect server name, obtains the data
         * from provision server and fills it in the data object provied.
         */
        void getPolicyData(const string &policyName, VPolicyServerData &data)
        throw (VException &);

    ///
        void getAccountData(const string &userName, VAccountData &data)
        throw (VException &);

    ///
        string getUserIp(const string& userName) throw (VException&);

    ///
        void getAliases(const string& userName, list < string > &retVal)
        throw (VException&);

///    
        string getMasterUser(const string &userName) throw (VException &);

    ///
        bool getTermContactList(const string &userName, list < string > &retVal)
        throw (VException&);

        /// Get list of subscribers
        void getSubscribers(vector < Data > & list) throw (VException &);

        /// Get list of aliases for subscribers
        void getAllAliases(vector < Data > & list) throw (VException &);

        /// Deletes a subscriber from the system
        void deleteSubscriber(const string& subscriberName)
        throw (VException &);

        /// Get the called list for a given user account
        void getCalledList(const string& rsHost, int port,
                           const string& acctName, vector < Data > & data)
        throw (VException &);

        /// Get the calling list for a given user account
        void getCallingList(const string& rsHost, int port,
                            const string& acctName, vector < Data > & data)
        throw (VException &);

    /** See if an account name is valid. Will return true if the user name
        exists, but false if it is an alias (only returns true if the Account
        exists and is a master user) */
    bool ProvisionInterface::isAccount(const string& userName);
    
        /// Get all the called CPL scripts for feature server
        void getCalledFeatureScripts(const string& hostName,
                                     int port, vector < Data > & retList)
        throw (VException &);

        /// Get all the calling CPL scripts for feature server
        void getCallingFeatureScripts(const string& hostName,
                                      int port, vector < Data > & retList)
        throw (VException &);

        /**
         * Gets dial plan XML data, parses it and returns list of 
         * contacts in retData
         *
         * the default Phone dial plan name is defaultPhonePlan.xml
         * the default IP dial plan name is defaultIpPlan.xml
         * to receive update notification for dial plan
         * call 
         *
         * registerForUpdate(callBackFunc, "defaultPhonePlan.xml", VCONFIG);
         * registerForUpdate(callBackFunc, "defaultIpPlan.xml", VCONFIG);
         */
        void getDialPlan(const string& name, vector < VPlanElement > & retData)
        throw (VException &);


        // *****************************************************************
        // Registration functions
        // *****************************************************************

        /**
           Register client for notification for a given item
           Callback func get called whenever item is updated.
        */
        void registerForUpdate(CallBackFunc func, const string& name,
                               PContext context);

        /**
           Register client for notification for a anything new within a
           given context on the server. Callback func get called whenever
           any new information is available for server. For example
           if a new user is added and called/calling list is
           created, RS will be notified. The data sent in the callback
           function is the comma separated list of entries added for the server.
        */
        void registerForUpdate(const string& host, int port,
                               CallBackFunc func, PContext context);

        /// Register a directory for update (should be group, but backward compat)
        void registerDirForUpdate(CallBackFunc func, const string& group);

        /// Unregister for an item
        void unregisterForUpdate(CallBackFunc func, const string& name,
                                 PContext context);

        /// Unregister an a context
        void unregisterForUpdate(const string& host, int port,
                                 CallBackFunc func, PContext context);

        /// Unregister a directory for update
        // This one is called "group" but is same as dir. Since no one was using
        // the unregister I could name it properly....
        void unregisterGroupForUpdate(CallBackFunc func, const string& group);


        /// Get the network address of the currently used pServer
        const NetworkAddress& getNetAddress()
        {
            return _interface.getNetAddress();
        }

    private:

        /**
           Constructor initializes with the name of provision server
           for redundancy passed as command line argument. Intentionally
           made private so that it can not be instantiated directly.
           It also creates an instance of listener in a separate thread
           to listen for data updates (through the interface)
        */
        ProvisionInterface(const char* primHostname, const int primPort,
                           const char* altHostname, const int altPort);

        /** Register function for configuration update notification
         *  function gets called when a configuration change occures
         */
        void updateConfigCallback(const string& name);

        void registerForUpdateImpl(CallBackFunc func, const string& name);
        void unregisterForUpdateImpl(CallBackFunc func, const string& name);

        /// Sends update notification for a data item
        void sendUpdateNotification(const string &item, const string &name);

        /// Sends delete notification for a data item
        void sendDeleteNotification(const string &item, const string &name);

        /// Sends update notification is dirName contents change on PS
        void sendUpdateNotification(const string& group,
                                    const string& name,
                                    bool delFlag);


        // each of these populates the server list, which is cached. These
        // simply return if called when the cache is good. These call
        // intitializeMap on the individual container classes (see the
        // provisioning/util directory for these)
///
void initializeJtapiServerList() throw(VException&);
///
void initializeFeatureServerList() throw(VException&);
///
void initializeMarshalServerList() throw(VException&);
///
void initializeCdrServerList() throw(VException&);
///
void initializeHeartbeatServerList() throw(VException&);
///
void initializeRedirectServerList() throw(VException&);
///
void initializePolicyServerList() throw(VException&);

        /// Translates the enum of contexts to strings
        string dirContext(PContext context);

        // Containers for the Server Group containers/Server lists
        VFeatureServerGroupContainer _featureServerGrpCont;
        VMarshalServerGroupContainer _marshalServerGrpCont;
        VRedirectServerGroupContainer _redirectServerGrpCont;
        VCdrServerGroupContainer _cdrServerGrpCont;
        VHeartbeatServerGroupContainer _heartbeatServerGrpCont;
        VJtapiServerGroupContainer _jtapiServerGrpCont;
        VPolicyServerGroupContainer _policyServerGrpCont;

        /// pointer to the singleton instance
        static ProvisionInterface* _instance;

        /// Host name the library is running on
        string _hostName;

friend class PSListener;

/// listener for updates from server
        PSListener* _listener;

        typedef list < CallBackFunc > CallBackList;
        typedef map < string, CallBackList > CallBackMap;

        /// Map to contain callbacks for a given data item
        CallBackMap _callBackMap;

/// interface to the PS
        PSInterface _interface;

};

#endif

