/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisionServer_cxx_Version =
    "$Id: ProvisionServer.cxx,v 1.91 2002/04/15 20:47:54 bko Exp $";

//Standard Include
#include <cerrno>
#include <csignal>
#include <unistd.h>
#include <cstring>
#include <cstdio>
#include <algorithm>

//User defined
#include "VEnvVar.hxx"
#include "VNetworkException.hxx"
#include "VIoException.hxx"
#include "VThreadPool.hxx"
#include "VCondition.h"
#include "NetworkAddress.h"
#include "Vpp_def.h"
#include "VPs_def.h"
#include "cpLog.h"
#include "PServerContainer.hxx"
#include "LocalScopeAllocator.hxx"
#include "HeartbeatTxThread.hxx"
#include "HeartbeatRxThread.hxx"
#include "ProvisionServer.hxx"

// initialize global static vars
static const int vppOkGLBL = 1;
static const int vppFailGLBL = 0;

// intitalize static member vars
ProvisionServer* ProvisionServer::_instance = NULL;

// _clientListLock and _featureListLock not used
// pthread_mutex_t ProvisionServer::_clientListLock;
// pthread_mutex_t ProvisionServer::_featureListLock;

// Use VMutex instead of pthread_mutex directly
VMutex ProvisionServer::_MutexconnectionMapLock;
VMutex ProvisionServer::_MutextimerLock;

string ProvisionServer::SYS_CONF(VSYS_CONF);
string ProvisionServer::FEATURES(VFEATURES);
string ProvisionServer::ACCOUNTS(VACCOUNTS);
string ProvisionServer::CALLEDLIST(VCALLEDLIST);
string ProvisionServer::CALLINGLIST(VCALLINGLIST);

// how long to wait in the thread dispatch loop -- gives time to let
// a thread mark a connection as being serviced.
const int ProvisionServer::THREAD_DELAY(10000);

// how long to wait before we abort a sync (we assume it failed if it takes
// longer than this)
const int ProvisionServer::FAIL_SLEEP_TIME(600);

#if defined(__SUNPRO_CC)
extern "C" {
void ProvisionServer_sigHupHandler(int sigNo)
{
     cerr << "sigHupHandler" << endl;
}

void ProvisionServer_sigChld(int sigNo)
{
     cerr << "sigChld" << endl;
}

void ProvisionServer_sigTerm(int sigNo)
{
     cerr << "sigTerm" << endl;
}

}

#else

#define ProvisionServer_sigHupHandler ProvisionServer::sigHupHandler
#define ProvisionServer_sigChld       ProvisionServer::sigChld
#define ProvisionServer_sigTerm       ProvisionServer::sigTerm

#endif

/*
  In all cases while calling the redundant server, remember that there are
  4 possible returns -- ERR, which means we are now in an invalid state
  (unlikely to happen, since we assert on most of these), OK, which means
  all is well and that we should send an ok message to the twin, FAIL, which
  means it failed, again send a message about this to twin, and NMSG, which
  means that we should send no message to the other server (usually when
  the transition results in no change in state and no action
*/


// signal handlers
void
ProvisionServer::sigTerm(int signo)
{
     cerr << "sigHupHandler" << endl;
    // Nothing for now...
}



void
ProvisionServer::sigChld(int signo)
{
    cerr << "sigHupHandler" << endl;
    // Nothing for now...
}



void
ProvisionServer::sigHupHandler(int sigNo)
{
    cerr << "sigHupHandler" << endl;
    // Nothing for now...
}


ProvisionServer&
ProvisionServer::instance()
{

    if (!_instance)
    {
        cpLog(LOG_ERR, "Called instance() when none had been constructed");
        assert(0);
    }
    return *_instance;

}


ProvisionServer&
ProvisionServer::instance(int nbins, int nthreads,
                          const string &mcasthost, int mcastport,
                          const string &hostname, int port)
{

    if (!_instance)
    {

        cpLogSetPriority(VEnvVar::VLOG_LEVEL);
        cpLog(LOG_ERR, "Starting pserver...");

        _instance = new ProvisionServer(nbins, nthreads);

        cpLog(LOG_DEBUG, "Created instance for %s\n",
              _instance->className().c_str());

    }

    if (hostname != "")
    {

        cpLog(LOG_DEBUG, "Server has been started in redundant mode");
        
        _instance->startHeartbeatThreads(hostname, port, mcasthost, mcastport);
        _instance->_copy.initialize(hostname, port);

        _instance->setSyncTimer();

    }
    else
    {
        cpLog(LOG_DEBUG, "Server has been started in non-redundant mode");
    }
            
    return *_instance;

}


void
ProvisionServer::startHeartbeatThreads(const string &hostname, int port,
                                       const string &mcasthost, int mcastport)
{

    cpLog(LOG_ERR, "In start threads, using multicast host %s and port %d",
          mcasthost.c_str(), mcastport);
    
    // create and start all the heartbeat threads
    int localPort = VEnvVar::VPS_PORT.value();

    cpLog(LOG_ALERT, "Creating heartbeat trasnsmit thread %d", localPort);
    _itsHeartbeatTxThread = new HeartbeatTxThread(localPort, 500,
                            mcasthost.c_str(), mcastport);

    _itsHeartbeatRxThread = new HeartbeatRxThread(PServerContainer::instance(),
                            mcasthost.c_str(), mcastport);
    _itsHouseKeepingThread =
        new HouseKeepingThread(PServerContainer::instance(),
                               500);


    char buf[256];
    sprintf(buf, "%s:%d", hostname.c_str(), port);
    string remote = buf;
    PServerContainer::instance().AddServer(Data(remote));

    _itsHeartbeatTxThread->run();
    _itsHeartbeatRxThread->run();
    _itsHouseKeepingThread->run();

}


ProvisionServer::ProvisionServer(int nbins, int nthreads) :
        _nbins(nbins), _nthreads(nthreads)
{

#if defined(__linux__)
    if (::atexit(ProvisionServer::destroy) < 0)
    {
        cpLog(LOG_ALERT, "Failed to register with atexit()");
    }
#endif

    cpLog(LOG_DEBUG, "Creating %d threads for pServer", _nthreads);
    _threadPool = new VThreadPool(_nthreads);

    signal(SIGHUP, ProvisionServer_sigHupHandler);
    signal(SIGCHLD, ProvisionServer_sigChld);
    signal(SIGTERM, ProvisionServer_sigTerm);

    timerPending = false;

    // seed random number generator (used by timer to decide when to try
    // to sync)
    srand((int)getpid() + time(NULL));

    _dataStore = new FileDataStore(_nbins, getRoot());

}


void
ProvisionServer::run()
{

    //Loop through for Incoming messages
    cpLog(LOG_ERR, "Pserver successfully started -- Running");
    for (; ; )
    {

        // get a message, accept it, add it to connection map
        if (_serverSocket.getServerConn().isReadReady())
        {

            LockedConnection conn;
            _serverSocket.accept(conn);

            _MutexconnectionMapLock.lock();
            _connectionMap[conn.getConnId()] = conn;
            _MutexconnectionMapLock.unlock();

        }

        // for each active connection with data, dispatch a thread to
        // handle it, unless it is already being handled

        // Connection map lock (cannot close a connection till released)
        _MutexconnectionMapLock.lock();
        for (ConnectionMap::iterator itr = _connectionMap.begin();
                itr != _connectionMap.end(); itr++)
        {
            assert( (*itr).first == (*itr).second.getConnId());

            LockedConnection& conn = (*itr).second;
            if (conn.isLive() && conn.isReadReady() && conn.notServiced())
            {
                // set the connection as being serviced and pass it to a thread
                conn.beingServiced();
		cpLog(LOG_DEBUG, "servicing connection");
                cpLog(LOG_DEBUG, "Connection %d marked", conn.getConnId());  

		assert(conn.isLive());
		assert(conn.isReadReady());
		assert(!conn.notServiced());
		assert(conn.getConnId() != 0);

		VFunctor functor(ProvisionServer::processRequestFunc, &conn);
                _threadPool->addFunctor(functor);
            }
        }

        _MutexconnectionMapLock.unlock();
        // Connection map unlocked

        vusleep(THREAD_DELAY);

    }

}


ProvisionServer::~ProvisionServer()
{
    cpLog(LOG_ERR, "ProvisionServer destructor called");
    _serverSocket.close();
}


void
ProvisionServer::destroy(void)
{
    delete ProvisionServer::_instance;
    ProvisionServer::_instance = NULL;
}


void
ProvisionServer::processRequestFunc(void* con)
{

    // a thread will process the data when it is assigned a request to process
    LockedConnection& conn = *(static_cast < LockedConnection* > (con));
    ProvisionServer::instance().processRequest(conn);

}


void
ProvisionServer::timerFunc(void* arg)
{

    // sleep 5 to 30 seconds and send sync
    int sleep_time = 5 + ( (int) (25.0 * rand() / (RAND_MAX + 1.0)) );
    cpLog(LOG_DEBUG, "Setting Sync Timer -- Sleeping %d", sleep_time);
    sleep(sleep_time);
    ProvisionServer::instance().initiateSync();

}


void
ProvisionServer::timerFailFunc(void* arg)
{

    // sleep and check if we still syncing...
    sleep(FAIL_SLEEP_TIME);
    ProvisionServer::instance().failTimerHandler();

}


void
ProvisionServer::failTimerHandler()
{

    _MutextimerLock.lock();
    failTimerPending = false;
    int result = _copy.timedoutSync();
    _MutextimerLock.unlock();

    // we get err if we are in an illegal state, and FAIL if we are in SLAVE,
    // in which case we assume we had a sync request that timed out. If we
    // actually succeeded, then restarted, it will interupt the sync, but
    // we will eventually sync since we start the timer again. The only
    // way this can fail is if the delay is too low to allow sync
    if (result == RedundancyManager::ERR ||
            result == RedundancyManager::FAIL)
    {
        cpLog(LOG_ERR, "Sync took >600s, abort sync and reset timer.");
        setSyncTimer();
    }

}


void
ProvisionServer::initiateSync()
{

    cpLog(LOG_DEBUG, "Sync Timer expired. Attempting to sync.");

    int result;

    // set timer to false and check state machine
    _MutextimerLock.lock();
    timerPending = false;
    result = _copy.startSync();
    _MutextimerLock.unlock();

    if (result == RedundancyManager::OK)
    {

        // if we get here we are the sync master.
        map < string, TimeStamp > remoteMap, localMap;

        result = getLocalMap(localMap);
        if (result != 0)
        {
            cpLog(LOG_ERR, "Unable to generate list of local data items");
            return ;
        }

        // get the remote list of items/timestamps
        result = _copy.getRemoteMap(remoteMap);
        if (result != RedundancyManager::OK)
        {
            if (result == RedundancyManager::ERR)
            {
                _copy.connFailed();
                return ;
            }
            if (result == RedundancyManager::FAIL)
            {
                _copy.badReturn();
                setSyncTimer();
                return ;
            }
        }

        // go through local list and compare to remote 4 possibilities:
        //   1) Not on remote -- send to remote
        //   2) Local is older -- get from remote
        //   3) Local is newer -- send to remote
        //   4) Both are equal -- no data sent to remote
        // remove from the remote list if found, then we go through what is
        // left on that list -- that is stuff on remote we don't have

        const int DO_NOTHING = 0;
        const int SEND_REMOTE = 1;
        const int GET_REMOTE = 2;

        string itemData;
        TimeStamp timestamp = 0;
        string group, name;

        for (map < string, TimeStamp > ::iterator itr = localMap.begin();
                itr != localMap.end(); itr++)
        {

            int action = DO_NOTHING;

            map < string, TimeStamp > ::iterator remoteItr
            = remoteMap.find((*itr).first);

            if (remoteItr == remoteMap.end())
            {
                action = SEND_REMOTE;          // not on remote server, so put
            }

            else
            {
                if ( (*itr).second > (*remoteItr).second )
                {
                    action = SEND_REMOTE;
                }
                else if ((*itr).second < (*remoteItr).second)
                {
                    action = GET_REMOTE;
                }
                else
                {
                    action = DO_NOTHING;
                }
            } // else nothing, essentially

            if (action == SEND_REMOTE)
            { // write to remote server

                try
                {
                    if (parseMapItem((*itr).first, group, name) != 0)
                    {
                        cpLog(LOG_ERR, "Failed to parse map entry : %s",
                              (*itr).first.c_str());
                        _copy.abortSync();
                        setSyncTimer();
                        return ;
                    }
                    itemData = _dataStore->getItem(group, name);
                }
                catch (VIoException& e)
                {
                    // Failed to read the item -- either because it is
                    // not there or permissions are wrong
                    string reason = e.getDescription();
                    cpLog(LOG_ERR, "Failed to read local item (%s,%s): %s",
                          group.c_str(), name.c_str(), reason.c_str());
                    cpLog(LOG_ERR, "Cannot read local file, aborting",
                          group.c_str(), name.c_str(), reason.c_str());

                    // abort 
                    assert(0);

                    // should never get here, but just in case
                    _copy.abortSync();
                    setSyncTimer();
                    return ;
                }

                cpLog(LOG_DEBUG, "Sending remote %s %s with date %lu",
                      group.c_str(), name.c_str(), timestamp);
                result = _copy.writeItem(group, name, itemData, (*itr).second);
                if (result != RedundancyManager::OK)
                {
                    if (result == RedundancyManager::ERR)
                    {
                        _copy.connFailed();
                        return ;
                    }
                    if (result == RedundancyManager::FAIL)
                    {
                        _copy.badReturn();
                        setSyncTimer();
                        return ;
                    }
                }

            } // end of write item

            if (action == GET_REMOTE)
            {

                if (parseMapItem((*itr).first, group, name) != 0)
                {
                    cpLog(LOG_ERR, "Failed to parse map entry : %s",
                          (*itr).first.c_str());
                    _copy.abortSync();
                    setSyncTimer();
                    return ;
                }
                result = _copy.readItem(group, name, itemData, timestamp);

                if (result != RedundancyManager::OK)
                {
                    if (result == RedundancyManager::ERR)
                    {
                        _copy.connFailed();
                        return ;
                    }
                    if (result == RedundancyManager::FAIL)
                    {
                        _copy.badReturn();
                        setSyncTimer();
                        return ;
                    }
                }

                try
                {
                    cpLog(LOG_DEBUG, "Putting %s %s with date %lu from remote",
                          group.c_str(), name.c_str(), timestamp);
                    _dataStore->putItem(group, name, itemData, timestamp);
                    _registerManager.modifyItem(group, name, false);
                }
                catch (VIoException& e)
                {
                    // Failed to write the item. Log and die since this
                    // means we can't access our own datastore properly
                    string reason = e.getDescription();
                    
                    cpLog(LOG_ERR, "Failed to write local item : %s to filestore",
                          reason.c_str());
                    cpLog(LOG_ERR, "Aborting. Check file permissions on datastore");
                    assert(0);

                    // should never get here, but just in case...
                    _copy.abortSync();
                    setSyncTimer();
                    return ;
                }

            }

            // remove it from the remote map
            remoteMap.erase((*itr).first);

        } // end of local list loop

        // whatever is left in the remote list doesn't exist in the local
        // map, so grab them
        for (map < string, TimeStamp > ::iterator itr = remoteMap.begin();
                itr != remoteMap.end(); itr++)
        {

            if (parseMapItem((*itr).first, group, name) != 0)
            {
                cpLog(LOG_ERR, "Failed to parse map entry : %s",
                      (*itr).first.c_str());
                _copy.abortSync();
                setSyncTimer();
                return ;
            }

            result = _copy.readItem(group, name, itemData, timestamp);

            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                    return ;
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                    return ;
                }
            }

            try
            {
                cpLog(LOG_DEBUG, "Putting %s %s with date %lu from remote",
                      group.c_str(), name.c_str(), timestamp);
                _dataStore->putItem(group, name, itemData, timestamp);
                _registerManager.modifyItem(group, name, false);
            }
            catch (VIoException& e)
            {
                //Failed to write the item
                string reason = e.getDescription();
                cpLog(LOG_ERR, "Failed to write local item : %s",
                      reason.c_str());
                cpLog(LOG_ERR, "Check that you have permission to write files",
                      reason.c_str());

                // should die now since the server can't write its own data
                assert(0);

                // should never get here, but just in case
                _copy.abortSync();
                setSyncTimer();
                return ;
            }

        }

        // get their registration
        cpLog(LOG_DEBUG, "Getting remote registration list");

        result = _copy.getRemoteRegistration(_registerManager);
        if (result != RedundancyManager::OK)
        {
            if (result == RedundancyManager::ERR)
            {
                _copy.connFailed();
                return ;
            }
            if (result == RedundancyManager::FAIL)
            {
                _copy.badReturn();
                setSyncTimer();
                return ;
            }
        }

        // send our registration
        result = _copy.putLocalRegistration(_registerManager);
        if (result != RedundancyManager::OK)
        {
            cpLog(LOG_ERR, "Send registration failed -- Sync failed!");
            if (result == RedundancyManager::ERR)
            {
                _copy.connFailed();
                return ;
            }
            if (result == RedundancyManager::FAIL)
            {
                _copy.badReturn();
                setSyncTimer();
                return ;
            }
        }

        // finally, send SYNCGOOD
        int result = _copy.sendGood();
        if (result != RedundancyManager::OK)
        {
            if (result == RedundancyManager::ERR)
            {
                _copy.connFailed();
                return ;
            }
            if (result == RedundancyManager::FAIL)
            {
                _copy.badReturn();
                setSyncTimer();
                return ;
            }
        }

        cpLog(LOG_DEBUG, "Synchronization Succeeded!");

    }
    else
    {

        // no action (expired but we are already syched or something

    }


    return ;

}

void
ProvisionServer::processRequest(LockedConnection& conn) throw (VException&)
{
    cpLog(LOG_DEBUG, "%d: processing request %d", pthread_self(), 
	  conn.getConnId());
    char buf[MAXLINE];

    buf[0] = '\0';
    int nread = 0;
    if (conn.readLine(buf, sizeof(buf) - 1, nread) <= 0)
    {
	cpLog(LOG_DEBUG, "%d: closing %d", pthread_self(), 
	      conn.getConnId());

        _MutexconnectionMapLock.lock();
        int tmp = conn.getConnId();
        cpLog(LOG_DEBUG, "processRequest: %p", &_MutexconnectionMapLock);
        conn.close();
        conn.doneService();
        _connectionMap.erase(tmp);
        _MutexconnectionMapLock.unlock();

        return ;
    }

    buf[nread] = '\0';

    char 
	req[MAXLINE], 
	arg[MAXLINE], 
	arg2[MAXLINE],
	arg3[MAXLINE], 
	rest[MAXLINE];

    sscanf(buf, "%s%s%s%s%s", req, arg, arg2, arg3, rest);
    string vppReq;
    vppReq = req;
    cpLog(LOG_DEBUG, "%d: %d: Received message : %s, command is %s",
          pthread_self(), conn.getConnId(), buf, vppReq.c_str());

    // figure out what we are doing and get it...
    if ( vppReq == GET_REQ)
    {
        handleGetRequest(conn, arg, arg2);
    }
    else if (vppReq == PUT_REQ)
    {
        handlePutRequest(conn, arg, arg2);
    }
    else if (vppReq == LIST_REQ)
    {
        handleListRequest(conn, arg, false);
    }
    else if (vppReq == NLIST_REQ)
    {
        handleListRequest(conn, arg, true);
    }
    else if (vppReq == REMOVE_REQ )
    {
        handleRemoveRequest(conn, arg, arg2);
    }
    else if (vppReq == DELETE_USER_REQ )
    {
        handleDeleteUserRequest(conn, arg);
    }
    else if (vppReq == REGISTER_ITEM_REQ )
    {
        handleItemRegisterRequest(arg, arg2, arg3);
    }
    else if (vppReq == REGISTER_GROUP_REQ )
    {
        handleGroupRegisterRequest(arg, arg2);
    }
    else if (vppReq == UNREGISTER_ITEM_REQ )
    {
        handleItemUnregisterRequest(arg, arg2, arg3);
    }
    else if (vppReq == UNREGISTER_GROUP_REQ )
    {
        handleGroupUnregisterRequest(arg, arg2);
    }
    else if (vppReq == SYNC_DEL_REQ)
    {
        handleSyncDelRequest(conn, arg, arg2);
    }

    else if (vppReq == SYNC_PUT_REQ)
    {
        handleSyncPutRequest(conn, arg, arg2);
    }
    else if (vppReq == SYNC_START_REQ)
    {
        handleSyncStartRequest(conn);
    }
    else if (vppReq == SYNC_GET_REQ)
    {
        handleSyncGetRequest(conn, arg, arg2);
    }
    else if (vppReq == SYNC_STAT_REQ)
    {
        handleSyncStatRequest(conn);
    }
    else if (vppReq == SYNC_GOOD_REQ )
    {
        handleSyncGoodRequest(conn);
    }
    else if (vppReq == SYNC_REGISTER_ITEM_REQ )
    {
        handleSyncRegisterItemRequest(arg, arg2, arg3);
    }
    else if (vppReq == SYNC_REGISTER_GROUP_REQ )
    {
        handleSyncRegisterGroupRequest(arg, arg2);
    }
    else if (vppReq == SYNC_UNREGISTER_ITEM_REQ )
    {
        handleSyncUnregisterItemRequest(arg, arg2, arg3);
    }
    else if (vppReq == SYNC_UNREGISTER_GROUP_REQ )
    {
        handleSyncUnregisterGroupRequest(arg, arg2);
    }
    else if (vppReq == SYNC_GRLIST_REQ )
    {
        handleSyncGetRListRequest(conn);
    }
    else if (vppReq == SYNC_PRLIST_REQ )
    {
        handleSyncPutRListRequest(conn);
    }
    else
    {
    }

    buf[0] = '\0';
    if (vppReq == PUT_REQ) req[0] = '\0';
    nread = 0;

    conn.doneService();
    cpLog(LOG_DEBUG, "done processing request %d", conn.getConnId());
}


void
ProvisionServer::handleGetRequest(LockedConnection& conn,
                                  const string &group,
                                  const string &name)
{
    string gData;

    try
    {
	gData = _dataStore->getItem(group, name);
        // Success
        string retData = makeVPPHeader(vppOkGLBL, gData.size());
        retData += gData;
        cpLog(LOG_DEBUG, "Got file (%s)", retData.c_str());
        conn.writeData(retData);
    }
    catch (VIoException& e)
    {
        // Failed to read the item
        sendVPPFail(conn);
        string reason = "Failed to get data: " + e.getDescription();
        cpLog(LOG_DEBUG, reason.c_str());
    }

}


void
ProvisionServer::handlePutRequest(LockedConnection& conn,
                                  const string &group,
                                  const string &name)
{

    // Get the content length
    char buf[MAXLINE];
    int nread = 0;

    // Read content length
    nread = 0;
    if (conn.readLine(buf, sizeof(buf), nread) < 0)
    {
        // Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
        return ;
    }

    int cntLength = 0;
    char cntStr[56];
    sscanf(buf, "%s%d", cntStr, &cntLength);
    cpLog(LOG_DEBUG, "Content-Length: %d", cntLength);

    // Read the contents
    char *c = new char[cntLength + 1];
    nread = 0;
    conn.readn(c, cntLength, nread);
    if (nread == -1)
    {
        cpLog (LOG_ALERT, "Read error on socket");
        delete c;
        return ;
    }
    else if (nread != cntLength)
    {
        cpLog(LOG_ALERT, "Content length mismatch, expected %d read %d",
              cntLength, nread);
        delete c;
        return ;
    }

    c[cntLength] = 0;
    cpLog(LOG_DEBUG, "Data for PUT request: (%s)", c);

    // Write data to the data store
    try
    {

        // Write to the date store and try to send to redundant server
        _dataStore->putItem(group, name, c);

        if (_copy.inUse() && _copy.isAlive())
        {
            int result;
            result = _copy.writeItem(group, name, c,
                                     _dataStore->getItemTimeStamp(group,
                                                                  name));
            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                }
            }
        }

        // send updates
        _registerManager.modifyItem(group, name);
        sendVPPOk(conn);

    }

    catch (VIoException& e)
    {
        cpLog(LOG_ALERT, "Failed to put (%s,%s) in data store. Reason : %s ",
              group.c_str(), name.c_str(), strerror(errno));
        sendVPPFail(conn);
    }

    delete c;
}


void
ProvisionServer::handleListRequest(LockedConnection& conn,
                                   const string& group,
                                   bool useHeader)
{

    cpLog(LOG_DEBUG, "ProvisionServer::handleListRequest for (%s)",
          group.c_str());

    string gData;

    try
    {

        StringList items;
        string retData, finalData;
        bool val = _dataStore->isGroup(group);

        if (val)
        {
            items = _dataStore->listItems(group);
        }

        // if not there, just send empty list (not fail, since the
        // behaviour on a create for a bin that doesn't exist is to create it
        if (!val || items.size() == 0)
        {
            retData = "\n";
        }
        else
        {

            StringList::iterator itr;
            for (itr = items.begin(); itr != items.end(); itr++)
            {
                retData += (*itr);
                retData += ",";
            }
            retData += "\n";
        }

        // use header?
        if (useHeader == true)
        {
            finalData = makeVPPHeader(vppOkGLBL, retData.size());
        }
        else
        {
            finalData = "";
        }

        finalData += retData;
        conn.writeData(finalData);

    }

    catch (VIoException& e)
    {
        //Failed to get the list of items
        cpLog(LOG_ERR, "dataStore listItem() failed for (%s)",
              group.c_str());
        sendVPPFail(conn);
    }
}


void
ProvisionServer::handleRemoveRequest(LockedConnection& conn,
                                     const string &group, const string &name)
{

    cpLog(LOG_DEBUG, "ProvisionServer::handleRemoveRequest for (%s,%s)",
          group.c_str(), name.c_str());
    string gData;

    try
    {

        // if the item doesn't exist, don't bother trying to remove it,
        // but do still do all normal processing (and return ok)
        bool exists = _dataStore->isItem(group, name);

        if (exists)
        {
            _dataStore->removeItem(group, name);
        }
        
        cpLog(LOG_DEBUG, "Removed item (%s,%s)", group.c_str(), name.c_str());

        _registerManager.deleteItem(group, name);

        if (_copy.inUse() && _copy.isAlive())
        {
            int result = _copy.sendDeleteItem(group, name);
            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                    sendVPPOk(conn);
                    return ;
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                    sendVPPOk(conn);
                    return ;
                }
            }
        }

        sendVPPOk(conn);

    }
    catch (VIoException& e)
    {

        // Failed to remove the item for whatever reason (likely permissions)
        cpLog(LOG_ERR, "Caught exception when trying to delete, reason %s",
              e.getDescription().c_str());
        sendVPPFail(conn);

    }

    return ;

}


void
ProvisionServer::handleDeleteUserRequest(LockedConnection& conn,
        const string& userName)
{

    // this is a multifile delete -- essentially we delete the called,
    // calling, and account items for a user.

    try
    {

        bool wasCalled = false;
        bool wasCalling = false;
        if (_dataStore->isItem(CALLEDLIST, userName))
        {
            _dataStore->removeItem(CALLEDLIST, userName);
            wasCalled = true;
        }
        if (_dataStore->isItem(CALLINGLIST, userName))
        {
            _dataStore->removeItem(CALLINGLIST, userName);
            wasCalling = true;
        }
        _dataStore->removeItem(ACCOUNTS, userName);

        cpLog(LOG_DEBUG, "Removed user (%s)", userName.c_str());

        if (wasCalled)
        {
            _registerManager.deleteItem(CALLEDLIST, userName);
        }

        if (wasCalling)
        {
            _registerManager.deleteItem(CALLINGLIST, userName);
        }

        _registerManager.deleteItem(ACCOUNTS, userName);

        if (_copy.inUse() && _copy.isAlive())
        {

            // send all three removes to the redundant, just in case
            // the copy has one.
            int result = _copy.sendDeleteItem(CALLEDLIST, userName);
            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                    return ;
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                    return ;
                }
            }

            result = _copy.sendDeleteItem(CALLINGLIST, userName);
            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                    return ;
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                    return ;
                }
            }

            result = _copy.sendDeleteItem(ACCOUNTS, userName);
            if (result != RedundancyManager::OK)
            {
                if (result == RedundancyManager::ERR)
                {
                    _copy.connFailed();
                    return ;
                }
                if (result == RedundancyManager::FAIL)
                {
                    _copy.badReturn();
                    setSyncTimer();
                    return ;
                }
            }
        }

        sendVPPOk(conn);

    }
    catch (VIoException& e)
    {

        // Failed to remove the item for whatever reason
        cpLog(LOG_ERR,
              "Caught exception when trying to delete user, reason %s",
              e.getDescription().c_str());
        sendVPPFail(conn);

    }

    return ;

}


void
ProvisionServer::handleItemRegisterRequest(const string &group,
        const string &name,
        const string &hostPortPair)
{

    // check all arguments exist and final one has a : (trivial check for
    // host port pair)
    if (group == "" || name == "" || hostPortPair == "" ||
            (hostPortPair.find(":") == string::npos))
    {
        cpLog(LOG_ERR, "Received malformed register request.");
        return ;
    }

    _registerManager.registerItem(group, name, hostPortPair);

    if (_copy.inUse() && _copy.isAlive())
    {

        int result = _copy.sendRegisterItem(group, name, hostPortPair);
        if (result == RedundancyManager::ERR)
        {
            _copy.connFailed();
        }
    }

}


void
ProvisionServer::handleGroupRegisterRequest(const string &group,
        const string &hostPortPair)
{

    // check all arguments exist and final one has a : (trivial check for
    // host port pair)
    if (group == "" || hostPortPair == "" ||
            (hostPortPair.find(":") == string::npos))
    {
        cpLog(LOG_ERR, "Received malformed register request.");
        return ;
    }


    _registerManager.registerGroup(group, hostPortPair);

    if (_copy.inUse() && _copy.isAlive())
    {
        int result = _copy.sendRegisterGroup(group, hostPortPair);
        if (result == RedundancyManager::ERR)
        {
            _copy.connFailed();
        }
    }
}


// not used for now
void
ProvisionServer::handleItemUnregisterRequest(const string &group,
        const string &name,
        const string &hostPortPair)
{

    // check all arguments exist and final one has a : (trivial check for
    // host port pair)
    if (group == "" || name == "" || hostPortPair == "" ||
            (hostPortPair.find(":") == string::npos))
    {
        cpLog(LOG_ERR, "Received malformed unregister request.");
        return ;
    }

    _registerManager.unregisterItem(group, name, hostPortPair);
    //FIX! Send to copy

}


// not used for now

void
ProvisionServer::handleGroupUnregisterRequest(const string &group,
        const string &hostPortPair)
{

    // check all arguments exist and final one has a : (trivial check for
    // host port pair)
    if (group == "" || hostPortPair == "" ||
            (hostPortPair.find(":") == string::npos))
    {
        cpLog(LOG_ERR, "Received malformed unregister request.");
        return ;
    }

    _registerManager.unregisterGroup(group, hostPortPair);
    // FIX! Send to copy

}



void
ProvisionServer::sendVPPFail(LockedConnection &conn)
{

    string retData = makeVPPHeader(vppFailGLBL, 0);
    try
    {
        conn.writeData(retData);
        cpLog(LOG_DEBUG, "Send Fail (400) Message");
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Caught exception when writing data, reason %s",
              e.getDescription().c_str());
    }

}


void
ProvisionServer::sendVPPOk(LockedConnection &conn)
{

    string retData = makeVPPHeader(vppOkGLBL, 0);
    try
    {
        conn.writeData(retData);
        cpLog(LOG_DEBUG, "Send Ok (200) Message");
    }
    catch (VNetworkException& e)
    {
        cpLog(LOG_DEBUG, "Caught exception when writing data, reason %s",
              e.getDescription().c_str());
    }

}


void
ProvisionServer::handleSyncGetRListRequest(LockedConnection& conn)
{


    cpLog(LOG_DEBUG, "Got into the handle sync");
    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    int state = _copy.syncRegList();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCGRLIST when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }

    string gData;

    try
    {
        _registerManager.getList(gData);
        //Success
        string retData = makeVPPHeader(vppOkGLBL, gData.size());
        retData += gData;
        retData += "\n";
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
    }
    catch (VIoException& e)
    {
        //Failed to read the data
        string reason = e.getDescription();
        string retData = makeVPPHeader(vppFailGLBL, reason.size());
        retData += reason;
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
    }


}


void
ProvisionServer::handleSyncPutRListRequest(LockedConnection& conn)
{

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    int state = _copy.syncRegList();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCPRLIST when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }

    //Get the content length
    char buf[MAXLINE];
    int nread = 0;

    try
    {

        //Read content length
        nread = 0;
        if (conn.readLine(buf, sizeof(buf), nread) < 0)
        {
            //Bad VPP header
            string retData = makeVPPHeader(vppFailGLBL, 0);
            cpLog(LOG_DEBUG, retData.c_str());
            conn.writeData(retData);
            return ;
        }

        char msg[256];
        int len;

        // get content length
        sscanf(buf, "%s%d", msg, &len);

        char *bigbuf;

        LocalScopeAllocator lAllocator(&bigbuf, len+1);

        conn.readn((void*)bigbuf, len, nread);

        if (nread < len)
        {
            sendVPPFail(conn);
            cpLog(LOG_ERR, "Insufficent data on socket - closing");
            return ;
        }

        bigbuf[len] = 0;

        char *token;
        char item[512], host[256];
        int mode = 0;

        token = strtok(bigbuf, ",");
        if (token == NULL)
        {
            sendVPPFail(conn);
            cpLog(LOG_ERR, "Invalid data sent for list of reg.");
            return ;
        }
        sscanf(token, "%s", item);

        if (string(item) != "++ITEM++")
        {
            sendVPPFail(conn);
            cpLog(LOG_ERR, "Parse error on registration data.");
            return ;
        }

        while ((token = strtok(NULL, ",")) != NULL)
        {

            cpLog(LOG_DEBUG, "token (%s)", token);
            sscanf(token, "%s", item);
            if (string(item) == "++GROUP++")
            {
                mode = 1;
                continue;
            }
            token = strtok(NULL, ",");
            if (token == NULL)
            {
                sendVPPFail(conn);
                cpLog(LOG_ERR, "Parse error on socket - closing");
                return ;
            }
            sscanf(token, "%s", host);
            if (mode == 0)
            {
                _registerManager.registerItemCat(item, host);
            }
            else
            {
                _registerManager.registerGroup(item, host);
            }
        }

        sendVPPOk(conn);
    }
    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to parse incoming reg list.");
    }
}


void
ProvisionServer::setSyncTimer()
{

    // see if a timer is pending
    _MutextimerLock.lock();
    if (timerPending)
    {
        // yep, nothing to do
        _MutextimerLock.unlock();
        return ;
    }

    // no. Set it to true and dispatch thread
    timerPending = true;
    _MutextimerLock.unlock();
    VFunctor functor(ProvisionServer::timerFunc, NULL);
    _threadPool->addFunctor(functor);

}


void
ProvisionServer::setSyncFailTimer()
{

    // see if a timer is pending
    _MutextimerLock.lock();
    if (failTimerPending)
    {
        // yep, nothing to do
        _MutextimerLock.unlock();
        return ;
    }

    // no. Set it to true and dispatch thread
    failTimerPending = true;
    _MutextimerLock.unlock();
    VFunctor functor(ProvisionServer::timerFailFunc, NULL);
    _threadPool->addFunctor(functor);

}


void
ProvisionServer::handleSyncGetRequest(LockedConnection& conn,
                                      const string& group,
                                      const string &name)
{

    string gData;
    TimeStamp timestamp;
    int state;

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    state = _copy.syncExchange();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCGET when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }

    try
    {
        gData = _dataStore->getItem(group, name);
        timestamp = _dataStore->getItemTimeStamp(group, name);
        //Success
        string retData = makeVPPHeader(vppOkGLBL, gData.size(), timestamp);
        retData += gData;
        conn.writeData(retData);
    }
    catch (VIoException& e)
    {
        //Failed to read the item
        sendVPPFail(conn);
        string reason = e.getDescription();
        cpLog(LOG_DEBUG, reason.c_str());
    }

}


void
ProvisionServer::handleSyncStartRequest(LockedConnection& conn)
{

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    cpLog(LOG_DEBUG, "ProvisionServer::handleSyncStartRequest");
    string gData;

    int state;
    state = _copy.syncStart();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCSTART when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }

    try
    {

        map < string, TimeStamp > localMap;
        getLocalMap(localMap);
        string retData;
        string finalData;
        list < string > tosend;
        char timestamp[32];
        int charcnt = 0;
        int cnt = 0;

        for (map < string, TimeStamp > ::iterator itr = localMap.begin();
                itr != localMap.end(); itr++)
        {
            retData += (*itr).first;
            sprintf(timestamp, ",%lu", (*itr).second);
            retData += timestamp;
            charcnt += ((*itr).first).size();
            charcnt += string(timestamp).size();
            charcnt += 1;
            if (cnt == 1000)
            {
                retData += ",\n";
                cpLog(LOG_DEBUG, retData.c_str());
                tosend.push_back(retData);
                cnt = 0;
                retData = "";
            }
            else
            {
                retData += ",";
            }
            cnt++;
        }

        ///Finish any remaining in the list
        if (cnt)
        {
            retData += "\n";
            cpLog(LOG_DEBUG, retData.c_str());
            tosend.push_back(retData);
        }
        finalData = makeVPPHeader(vppOkGLBL, charcnt);
        list < string > ::iterator itr2 = tosend.begin();

        // no data, so send empty 200
        if (itr2 == tosend.end())
        {
            conn.writeData(finalData);
        }

        for (itr2 = tosend.begin(); itr2 != tosend.end(); itr2++)
        {
            finalData += (*itr2);
            conn.writeData(finalData);
            finalData = "";
        }

        setSyncFailTimer();

    }

    catch (VIoException& e)
    {

        //Failed to get the dir list
        string retData = makeVPPHeader(0, 0);
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
    }

}



void
ProvisionServer::handleSyncPutRequest(LockedConnection& conn,
                                      const string& group, const string &name)
{

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    char buf[MAXLINE];
    int nread = 0;
    int cntLength = 0;
    char cntStr[56];
    TimeStamp timestamp;
    int state;

    state = _copy.syncExchange();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCPUT when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }


    // Read content length
    nread = 0;
    if (conn.readLine(buf, sizeof(buf), nread) < 0)
    {
        // Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
        return ;
    }
    sscanf(buf, "%s%lu", cntStr, &timestamp);

    // Read Timestamp
    if (conn.readLine(buf, sizeof(buf), nread) < 0)
    {
        // Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, retData.c_str());
        conn.writeData(retData);
        return ;
    }
    sscanf(buf, "%s%d", cntStr, &cntLength);

    //Read the contents
    char *c = new char[cntLength + 1];
    nread = 0;
    conn.readn(c, cntLength, nread);

    if (nread == -1)
    {
        cpLog (LOG_ALERT, "Read error on socket");
        delete c;
        return ;
    }

    else if (nread != cntLength)
    {
        cpLog(LOG_ALERT, "Content length mismatch, expected %d read %d",
              cntLength, nread);
        delete c;
        return ;
    }

    c[cntLength] = 0;
    cpLog(LOG_DEBUG, "Data for PUT request:%s", c);
    //Write data to the data store
    try
    {

        _dataStore->putItem(group, name, c, timestamp);

        // updates are sent in the PsProcess module, so no need to do it

        string retData = makeVPPHeader(vppOkGLBL, cntLength);
        try
        {
            conn.writeData(retData);
        }
        catch (VNetworkException& e)
        {
            cpLog(LOG_DEBUG, "Caught exception when writing data, reason %s",
                  e.getDescription().c_str());
        }
    }
    catch (VIoException& e)
    {
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_ALERT, "Failed to write item (%s,%s), reason %s",
              group.c_str(), name.c_str(), strerror(errno));
        conn.writeData(retData);
    }

    delete c;

}



void
ProvisionServer::handleSyncStatRequest(LockedConnection& conn)
{

    string data;

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        data += "This server is not using redundancy.\n";
    }
    else
    {

        cpLog(LOG_DEBUG, "Received SYNCSTAT Request");

        data += "Remote Connection Status : ";
        if (_copy.isAlive())
        {
            data += "Alive\n";
        }
        else
        {
            data += "Not Alive\n";
        }

        data += "Syncronization Status    : ";
        if (_copy.isSync())
        {
            data += "Synched\n";
        }
        else
        {
            data += "Not Synched\n";
        }

        data += "Mastering Status         : ";
        if (_copy.isMaster())
        {
            data += "Master\n";
        }
        else
        {
            data += "Not Master\n";
        }

        data += "Slave Status             : ";
        if (_copy.isSlave())
        {
            data += "Slave\n";
        }
        else
        {
            data += "Not Slave\n";
        }

    }

    string retData = makeVPPHeader(vppOkGLBL, data.size());
    retData += data;
    cpLog(LOG_DEBUG, retData.c_str());
    conn.writeData(retData);

    return ;

}


void
ProvisionServer::handleSyncGoodRequest(LockedConnection& conn)
{

    if (!_copy.inUse())
    {
        // We aren't syncing, so ignore this
        cpLog(LOG_ERR, "Sync request but not using sync");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        return ;
    }

    cpLog(LOG_DEBUG, "Received SYNCGOOD");

    int state;

    state = _copy.syncGood();
    if (state == RedundancyManager::ERR)
    {
        // this is bad
        cpLog(LOG_ERR, "Server went into ERROR sync state.");
        string retData = makeVPPHeader(vppFailGLBL, 0);
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }
    if (state == RedundancyManager::FAIL)
    {
        //Bad VPP header
        string retData = makeVPPHeader(vppFailGLBL, 0);
        cpLog(LOG_DEBUG, "Got a SYNCGOOD when not expecting it.");
        conn.writeData(retData);
        setSyncTimer();
        return ;
    }

    string retData = makeVPPHeader(vppOkGLBL, 0);
    conn.writeData(retData);

    cpLog(LOG_DEBUG, "Synchronization complete and succeeded!");

    return ;

}


string
ProvisionServer::makeVPPHeader(int success, int contLength, TimeStamp timestamp)
{

    string retData;
    char buf[256];

    if (success)
    {
        retData += "200 OK\n";
    }
    else
    {
        retData += "400 FAILED\n";
    }

    // if they gave us a timestamp, use it.
    if (timestamp != 0)
    {
        sprintf(buf, "Timestamp: %lu\n", timestamp);
        retData += buf;
    }

    // add the content length
    sprintf(buf, "%s %d\n", CONT_LENGTH, contLength);
    retData += buf;

    return (retData);

}


void
ProvisionServer::handleSyncRegisterItemRequest(const string& group,
        const string &name,
        const string &host)
{

    _registerManager.registerItem(group, name, host);

}


void
ProvisionServer::handleSyncRegisterGroupRequest(const string& group,
        const string &host)
{

    _registerManager.registerGroup(group, host);

}


void
ProvisionServer::handleSyncUnregisterItemRequest(const string& group,
        const string &name,
        const string &host)
{

    _registerManager.unregisterItem(group, name, host);

}


void
ProvisionServer::handleSyncUnregisterGroupRequest(const string& group,
        const string &host)
{

    _registerManager.unregisterGroup(group, host);

}




int
ProvisionServer::getLocalMap(map < string, TimeStamp > &local)
{

    local.clear();

    cpLog(LOG_DEBUG, "ProvisionServer::generating local map for sync");

    try
    {

        StringList listOfGroups = _dataStore->listGroups();

        if (listOfGroups.size() == 0)
        {
            return 0;
        }

        StringList::iterator itr;

        // get just the items, and just their names. Make sure
        // we exclude CVS
        for (itr = listOfGroups.begin(); itr != listOfGroups.end(); itr++)
        {

            StringList listOfItems = _dataStore->listItems(*itr);

            StringList::iterator itr2;

            for (itr2 = listOfItems.begin(); itr2 != listOfItems.end(); itr2++)
            {
                char hashkey[256];
                sprintf(hashkey, "%s|%s", (*itr).c_str(), (*itr2).c_str());
                local[hashkey] = _dataStore->getItemTimeStamp((*itr), (*itr2));

            }
        }
    }

    catch (VIoException& e)
    {

        //Failed to get the dir list
        return -1;
        cpLog(LOG_DEBUG, "Failed to generate localmap.");
    }

    return 0;
}



void ProvisionServer::gotBeat()
{

    int val = _copy.heartBeat();
    if (val == RedundancyManager::OK)
    {
        setSyncTimer();
    }
}


int
ProvisionServer::parseMapItem(const string &input, string &group, string &name)
{

    size_t loc = input.find("|");

    if (loc == string::npos)
    {
        return -1;
    }

    group = input.substr(0, loc);
    name = input.substr(loc + 1, input.size());

    return 0;

}


string
ProvisionServer::getRoot()
{

    // if we have asked before, return the result, else generate first time
    if (_root.size())
    {
        return _root;
    }

    string retStr(VEnvVar::VPS_FILESYSTEM.value());

    // ensure root ends in a slash and strip any extra (double, triple etc.)
    if (retStr.size() && retStr[retStr.size() - 1] != '/')
    {
        retStr += "/";
    }
    while (retStr.find("//") != string::npos)
    {
        retStr.replace(retStr.find("//"), 2, "/");
    }

    _root = retStr;
    return _root;

}


void
ProvisionServer::handleSyncDelRequest(LockedConnection& conn,
                                      const string &group, const string &name)
{

    cpLog(LOG_DEBUG, "ProvisionServer::handleRemoveRequest for (%s,%s)",
          group.c_str(), name.c_str());
    string gData;

    try
    {

        _dataStore->removeItem(group, name);
        cpLog(LOG_DEBUG, "Removed item (%s,%s)", group.c_str(), name.c_str());
        sendVPPOk(conn);


    }
    catch (VIoException& e)
    {

        // Failed to remove the item for whatever reason
        cpLog(LOG_ERR, "Caught exception when trying to delete, reason %s",
              e.getDescription().c_str());
        sendVPPFail(conn);

    }

    return ;

}


/* Local Variables: */
/* c-file-style: "stroustrup" */
/* indent-tabs-mode: nil */
/* c-file-offsets: ((access-label . -) (inclass . ++)) */
/* c-basic-offset: 4 */
/* End: */
