#ifndef PROVISIONSERVER_HXX
#define PROVISIONSERVER_HXX

/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const ProvisionServer_hxx_Version =
    "$Id: ProvisionServer.hxx,v 1.53 2001/06/29 21:02:42 bko Exp $";

#include <sys/time.h>
#include <sys/types.h>
#include <map>
#include <string>

#include "Tcp_ServerSocket.hxx"
#include "LockedConnection.hxx"
#include "RedundancyManager.hxx"
#include "HouseKeepingThread.hxx"
#include "VMutex.h"
#include "VIoException.hxx"

class VThreadPool;
class HeartbeatTxThread;
class HeartbeatRxThread;

/**
 *  Uses VPP to serve the client request to get and put data items
 *  on the ProvisionServer. VPP is Vocal Provisioning Protocol, the
 *  HTTP like protocol used to communicate with the provisioning server
 */ 
class ProvisionServer
{
    public:

        /// Interface to get reference to singelton object.
        static ProvisionServer& instance();

        /// Interface to get reference to singelton object.
        //  static ProvisionServer& instance(int nusers, int nthreads);

        /// Interface to get reference to singelton with redundancy
        static ProvisionServer& instance( int nusers, int nthreads,
                                      const string &mcasthost = "",
                                      int mcastport = 0,
                                      const string &hostname = "",
                                      int port = 0);
        /// Return the name of the class
        string className()
        {
            return "ProvisionServer";
        }

        /// Main process loop. Waits on connections and handles them
        void run();

        /// Destructor
        ~ProvisionServer();

        /// Get file system root
        string getRoot();

        // Symbolic string aliases for common group names
        static string SYS_CONF;
        static string FEATURES;
        static string ACCOUNTS;
        static string CALLEDLIST;
        static string CALLINGLIST;

        /// Called when we see the heartbeat from the redundant server
        void gotBeat();

        /// Called when we notice that the heartbeat is gone
        void lostBeat()
        {
            _copy.heartBeatGone();
        }


    protected:

        /// Smart pointer to the heartbeat transmit thread
        Sptr < HeartbeatTxThread > _itsHeartbeatTxThread;

        /// Smart pointer to the heartbeat receiver thread
        Sptr < HeartbeatRxThread > _itsHeartbeatRxThread;
    
        /// Smart pointer to the heartbeat process/housekeeping thread
        Sptr < HouseKeepingThread > _itsHouseKeepingThread;

    private:

        static const int THREAD_DELAY;
        static const int FAIL_SLEEP_TIME;
    
        /** 
         *  Private constructor to build the class (public access using
         *  the instance() operator since this is a singleton) 
         */
        ProvisionServer(int nusers, int nthreads);

        /** 
         *  Basically a big set of if's to decide how to handle an incoming
         *  message.
         */
        void processRequest(LockedConnection& conn) throw (VException&);

        /**
         *  Frees the memory associated with singelton instance.
         *  gets register to atexit() function at the time of creation.
         */
        static void destroy(void);

        /// A thread from the pool is passed this with a conenction to service
        static void processRequestFunc(void* con);

        /// Timer function passed to a thread when we need to sync. Sync on expire
        static void timerFunc(void* arg);

        /// Timer function passed to a thread while synching. Abort on expire
        static void timerFailFunc(void* arg);

        /**
         *  Signal handler for SIGHUP signal, causes the client list to
         *  be reinitialized.
         */
        static void sigHupHandler(int sigNo);

        /// Signal handler for SIGTERM
        static void sigTerm(int signo);

        /// Signal handler to cleanup the child
        static void sigChld(int signo);

        /// Processes the GET request
        void handleGetRequest(LockedConnection& conn, const string &group,
                              const string &item);

        /// Processes the PUT request
        void handlePutRequest(LockedConnection& conn, const string &group,
                              const string &item);

        /// Handles request to return directory listing w/ content-length header
        void handleListRequest(LockedConnection& conn, const string &group,
                               bool useHeader);

        /// Handles request to remove an item
        void handleRemoveRequest(LockedConnection& conn, const string& group,
                                 const string &item);

        /// Handles request to remove an item
        void handleDeleteUserRequest(LockedConnection& conn,
                                     const string& userName);

        /// Handle request to register an item for notification
        void handleItemRegisterRequest(const string &group,
                                       const string &item,
                                       const string &hostPortPair);

        /// Handle request to register a group for notification
        void handleGroupRegisterRequest(const string &group,
                                        const string &hostPortPair);

        /// Handle request to unregister an item for notification
        void handleItemUnregisterRequest(const string &group,
                                         const string &item,
                                         const string &hostPortPair);

        /// Handle request to unregister a group for notification
        void handleGroupUnregisterRequest(const string &group,
                                          const string &hostPortPair);

        /// Handles request to return timestamped directory listing for sync
        void handleSyncStartRequest(LockedConnection& conn);

        /// Processes the SYNCPUT request (put with timestamp)
        void handleSyncPutRequest(LockedConnection& conn, const string& group,
                                  const string &name);

        /// Processes the SYNCGET request (get with timestamp)
        void handleSyncGetRequest(LockedConnection& conn, const string& group,
                                  const string &name);

        /** 
         *  Process a SYNCSTAT request (can call this from GUI or
         *  other outside location to see if we are synched) 
         */
        void handleSyncStatRequest(LockedConnection &conn);

        /// Handle a SYNCGOOD from other redundant server
        void handleSyncGoodRequest(LockedConnection& conn);

        /// Handle a REMOVE request
        void handleRemoveRequest(LockedConnection &conn, const string &arg);

        /// Handle a SYNCGLIST request
        void handleSyncGetRListRequest(LockedConnection &conn);

        /// Handle a SYNCPLIST request
        void handleSyncPutRListRequest(LockedConnection &conn);

        /// Handle request to register a host for notification
        void handleSyncRegisterItemRequest(const string &group,
                                           const string &name,
                                           const string &host);

        /// Handle an incoming request to add a register from another server
        void handleSyncRegisterGroupRequest(const string &group,
                                            const string &host);

        /// Handle request to unregister a host for notification
        void handleSyncUnregisterItemRequest(const string &group,
                                             const string &name,
                                             const string &host);

        /// Handle an incoming request to remove a register from another server
        void handleSyncUnregisterGroupRequest(const string &group,
                                              const string &host);


        /// Process a delete from the other server
        void handleSyncDelRequest(LockedConnection &conn, const string &name,
                                  const string &group);

        /** 
         *  Creates a VPP header, that is set out with response
         *  based on the success or failure of the request
         *  conLength is the length of data following the
         *  header.
         */
                                  
        /// Make a 200 or 400 header and return it
        string makeVPPHeader(int success, int contLength,
                             TimeStamp timestamp = 0);

        /// Make a 400 and send on the provided connection
        void sendVPPFail(LockedConnection &conn);

        /// Make a 200 and send on the provided connection
        void sendVPPOk(LockedConnection &conn);

        /// generate a map of items/timestamps of local items for sync
        int getLocalMap(map < string, TimeStamp > &local);

        // the word "file" appears in these as a backward compatability thing...

        /// Reads the content of item from the data store
        string readFile(const string& fileName, TimeStamp *timestamp = 0)
        throw (VIoException&);

        /// Writes the contents of item on the data store, returns timestamp
        void writeFile(const string& fileName, const string& contents,
                       TimeStamp *timestamp)
        throw (VIoException&);

        /// If a new server register for update, update the client list
        void updateClientList();

        /**
         *  Initialises the list of hosts, to which,
         *  notification has to be sent.Gets called when
         *  object is first created or a HUP signal is received.         
         */
        void initializeClientList() throw (VIoException&);

        /// 
        void initializeFeatureServerList() throw (VIoException&);

        /// Create the heartbeat and other threads
        void startHeartbeatThreads (const string &hostname, int port, 
                                    const string &mcasthost, int mcastport);

        /** 
         *  These two timers are used for syncing. The sync timer is
         *  set when we see another server -- at expiry, we try to sync.
         *  Once syncing starts, the SyncFail timer is set, and if it
         *  expires and the sync is still going on, the sync is aborted
         *  (timeout)
         */
    
        /// Set timer, on expire, tries to sync
        void setSyncTimer();

        /// Set timer, if expires and still synching, took too long -- abort
        void setSyncFailTimer();

        /// Called if we start sync. Manages the entire sync process.
        void initiateSync();

        /// Cleanup func if a sync fails (for example if one dies during sync)
        void failTimerHandler();

        /// Parse and incoming Map item
        int parseMapItem(const string &input, string &group, string &name);

        /* Various locks */
        // _clientListLock and _featureListLock not used
        // static pthread_mutex_t _clientListLock;
        // static pthread_mutex_t _featureListLock;

	// Use VMutex instead of pthread_mutex directly
        static VMutex _MutexconnectionMapLock;
        static VMutex _MutextimerLock;

        typedef map < int, LockedConnection > ConnectionMap;

        /// Map of int to LockedConnection, tracks the active connections
        ConnectionMap _connectionMap;

        /// pointer to the singleton instance
        static ProvisionServer* _instance;

        /// Socket to listen on
        TcpServerSocket _serverSocket;

        /// Holds the PorvisionServer port
        int _pServerPort;

        /// Pool of worker threads
        Sptr < VThreadPool > _threadPool;

        /// File system root (location of the data)
        string _root;

        /// The redundant server manages all contact with the other server
        RedundancyManager _copy;

        /// Timer to track if we are waiting for a sync timer to expire
        bool timerPending;

        /// Timer to track if we have a timeout set for sync 
        bool failTimerPending;

        /// Smart pointer to the DataStore where the files reside
        Sptr < DataStore > _dataStore;

        /// Register manager to track client registrations
        RegisterManager _registerManager;

        /// Number of bins in the DataStore
        int _nbins;

        /// Number of threads in thread pool
        int _nthreads;

};

#endif
