/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const RedundancyManager_cxx_Version =
    "$Id: RedundancyManager.cxx,v 1.5 2001/06/29 21:02:42 bko Exp $";

#include <string.h>
#include <stdio.h>
#include <map>

#include "NetworkAddress.h"
#include "Tcp_ClientSocket.hxx"
#include "Vpp_def.h"
#include "cpLog.h"
#include "Sptr.hxx"
#include "VNetworkException.hxx"
#include "LocalScopeAllocator.hxx"
#include "RedundancyManager.hxx"

RedundancyManager::RedundancyManager()
{

    _state = DNSYNC;
    _use = false;

    _host = "";
    _port = 0;

}


void
RedundancyManager::initialize(const string &host, int port)
{

    LockHelper lockAid( this->_aMutex );

    _use = true;

    _host = host;
    _port = port;

    _networkAddress = new NetworkAddress(_host);
    _networkAddress->setPort(_port);

    cpLog(LOG_DEBUG, "Using redundant server at %s:%i", _host.c_str(), _port);

}


// All of these are state transition functions
int
RedundancyManager::heartBeatGone()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    _state = DNSYNC;  // now dead

    return NMSG;

}


int
RedundancyManager::heartBeat()
{

    LockHelper lockAid( this->_aMutex );

    if (_state != DNSYNC)
    {
        assert(checkState());
        return NMSG;
    }
    _state = ANSYNC;  // now alive, not synced

    return OK;

}


int
RedundancyManager::connFailed()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    if (_state == DNSYNC)
    {
        return ERR;
    }
    _state = DNSYNC;

    return NMSG;

}


int
RedundancyManager::abortSync()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    _state = ANSYNC;

    return NMSG;

}


int
RedundancyManager::timedoutSync()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    // this is the abort
    if (_state == ASLAVE)
    {
        _state = ANSYNC;
        return FAIL;
    }

    return NMSG;

}

int
RedundancyManager::badReturn()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    if (_state == DNSYNC)
    {
        return ERR;
    }
    _state = ANSYNC;

    return NMSG;

}


int
RedundancyManager::syncStart()
{

    LockHelper lockAid( this->_aMutex );

    if (_state == ANSYNC)
    {
        _state = ASLAVE;  // switch to slave and start sync
        return OK;
    }
    if ((_state == AMASTER) || (_state == ASLAVE) || (_state == ASYNC))
    {
        _state = ANSYNC;  // something weird, abort and resync
        return FAIL;
    }
    if (_state == DNSYNC)
    {
        return FAIL;  // stay dead, send fail (this is a really strange state)
    }

    _state = DNSYNC;

    return ERR;

}


int
RedundancyManager::startSync()
{

    LockHelper lockAid( this->_aMutex );

    // if we are alive and unsynched, we want to initiate a sync,
    // else do nothing
    if (_state == ANSYNC)
    {
        _state = AMASTER;  // switch to master and start sync
        return OK;
    }
    return FAIL;

}


int
RedundancyManager::syncGood()
{

    LockHelper lockAid( this->_aMutex );

    if (_state == ASLAVE)
    {
        _state = ASYNC;  // all is good -- go to synched
        return OK;
    }
    if ((_state == AMASTER) || (_state == ANSYNC) || (_state == ASYNC))
    {
        _state = ANSYNC;  // fail, start sync process over
        return FAIL;
    }
    if (_state == DNSYNC)
    {
        return FAIL;  // message from a "dead" partner -- ignore
    }

    _state = DNSYNC;

    return ERR;

}



int
RedundancyManager::syncExchange()
{

    LockHelper lockAid( this->_aMutex );

    // we got a message from the other server
    if ((_state == AMASTER) || (_state == ASYNC) || (_state == ASLAVE))
    {
        return OK;
    }

    if ((_state == ANSYNC))
    {
        // we should not get messages for sync until we are syncing or
        // we are synched
        return FAIL;
    }

    if (_state == DNSYNC)
    {
        return FAIL;  // the dead speak!
    }


    _state = DNSYNC;
    return ERR;

}


int
RedundancyManager::syncRegList()
{

    LockHelper lockAid( this->_aMutex );

    assert(checkState());

    if (_state == DNSYNC)
    {
        _state = DNSYNC;
        return ERR;
    }

    // we got a message from the other server
    if (_state == ASLAVE)
    {
        return OK;
    }

    _state = ANSYNC;
    return ERR;

}


bool
RedundancyManager::isAlive ()
{

    LockHelper lockAid( this->_aMutex );

    return (_state & ALIVE);

}


bool
RedundancyManager::isSync ()
{

    LockHelper lockAid( this->_aMutex );

    return (_state & SYNC);

}


bool
RedundancyManager::isMaster ()
{

    LockHelper lockAid( this->_aMutex );

    return (_state & MASTER);

}


bool
RedundancyManager::isSlave ()
{

    LockHelper lockAid( this->_aMutex );

    return (_state & SLAVE);

}


bool
RedundancyManager::checkState()
{

    return (_state == ASYNC ||
            _state == DNSYNC ||
            _state == ANSYNC ||
            _state == ASLAVE ||
            _state == AMASTER);

}


int
RedundancyManager::getRemoteMap(map < string, TimeStamp > &remote)
{

    char buf[256], msg[256];
    string sendData;
    int status, len, nread = 0;

    LockHelper lockAid( this->_aMutex );

    // want it empty, 'cause if the other server returns nothing we
    // want an empty list
    remote.clear();

    try
    {

        // send request for the informaion
        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        sprintf(buf, "SYNCSTART %s\n", VPP_VERSION);
        sendData += buf;
        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            buf[nread] = 0;
            cpLog(LOG_INFO, "Remote Map read : (%s)", buf);

            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {

                if (conn.readLine(buf, sizeof(buf), nread) <= 0)
                {
                    cpLog(LOG_ERR,
                          "Failed to read content len. -- Closing connection");
                    return ERR;
                }

                cpLog(LOG_ERR, "Data is: (%s)", buf);
                
                // get the content length
                sscanf(buf, "%s%d", msg, &len);

                if (len == 0)
                {
                    // no data on remote server, so just return OK
                    return OK;
                }

                // allocate buffer for contents
                char *bigbuf;
                LocalScopeAllocator lAllocator(&bigbuf, len+1);

                conn.readn((void*)bigbuf, len, nread);
                cpLog(LOG_ERR, "Data is: (%s)", bigbuf);
                
                if (nread < len)
                {
                    cpLog(LOG_ERR, "Insufficent data on socket - closing");
                    return ERR;
                }

                bigbuf[len] = 0;

                char *token;
                char file[256];
                TimeStamp timestamp;

                // return list is filename,timestamp,..., (ends with ,)
                // should ALWAYS be in pairs.
                token = strtok(bigbuf, ",");
                sscanf(token, "%s", file);
                token = strtok(NULL, ",");
                if (token == NULL)
                {
                    cpLog(LOG_ERR, "Parse error on socket - closing");
                    return ERR;
                }
                sscanf(token, "%lu", &timestamp);

                // add timestamp to map indexed by filename
                remote[string(file)] = timestamp;

                // got first pair, no do all the others...
                while ((token = strtok(NULL, ",")) != NULL)
                {

                    sscanf(token, "%s", file);
                    cpLog(LOG_DEBUG, "File is %s", file);
                    token = strtok(NULL, ",");
                    if (token == NULL)
                    {
                        cpLog(LOG_ERR, "Parse error on socket - closing");
                        return ERR;
                    }
                    sscanf(token, "%lu", &timestamp);
                    remote[file] = timestamp;

                }

                return OK;

            }
            else
            {

                cpLog(LOG_DEBUG, "SYNCSTART File returned FAILED");
                return FAIL;

            }

        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send SYNCSTART to %s:%d.",
              _host.c_str(), _port);
        return ERR;
    }

}


int
RedundancyManager::writeItem(const string &group, const string &name,
                           const string &contents,
                           TimeStamp timestamp)
{

    char buf[256], msg[256];
    string sendData;
    int status, nread = 0;

    LockHelper lockAid( this->_aMutex );

    try
    {

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        // format and send message
        sprintf(buf, "SYNCPUT %s %s %s\n", group.c_str(), name.c_str(),
                VPP_VERSION);
        sendData += buf;
        sprintf(buf, "Timestamp: %lu\n", timestamp);
        sendData += buf;
        sprintf(buf, "%s %d\n", CONT_LENGTH, contents.size());
        sendData += buf;
        sendData += contents;

        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            // terminate string and read status
            buf[nread] = 0;
            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {
                // if ok, just return ok
                return OK;
            }
            else
            {
                cpLog(LOG_DEBUG, "SYNCPUT File returned FAILED");
                return FAIL;
            }

        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to write to redundant server %s:%d.",
              _host.c_str(), _port);
        return ERR;
    }

}


int
RedundancyManager::readItem(const string &group, const string &name,
                          string &contents,
                          TimeStamp &timestamp)
{

    cpLog (LOG_DEBUG, "Asked twin for (%s,%s)", group.c_str(), name.c_str());

    char buf[256], msg[256];
    int len, status, nread = 0;
    string sendData;

    LockHelper lockAid( this->_aMutex );

    try
    {

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        // send the SYNCGET message to our opposite
        sprintf(buf, "SYNCGET %s %s %s\n", group.c_str(),
                name.c_str(), VPP_VERSION);
        sendData += buf;
        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            // terminate string and copy data
            buf[nread] = 0;

            sscanf(buf, "%d %s", &status, msg);

            if (status != 200)
            {
                cpLog(LOG_DEBUG, "SYNCGET File returned FAILED");
                return FAIL;
            }

        }

        // get the data
        // get the content length and timestamp
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "Premature EOF on socket -- Closing connection");
            return ERR;

        }
        else
        {
            // parse out the timestamp
            sscanf(buf, "%s%lu", msg, &timestamp);
        }

        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {
            cpLog(LOG_ERR, "Premature EOF on socket -- Closing connection");
            return ERR;

        }
        else
        {
            // parse out the content length
            sscanf(buf, "%s%d", msg, &len);
        }

        // allocate a buffer for the content
        char *bigbuf;

        LocalScopeAllocator lAllocator(&bigbuf, len+1);

        conn.readn((void*)bigbuf, len, nread);
        if (nread < len)
        {
            cpLog(LOG_ERR,
                  "Insufficent data on socket ( < CONTENT_LENGTH ) - closing");
            return ERR;
        }

        // eunsure the string is terminated and copy to string object
        bigbuf[len] = 0;
        contents = bigbuf;

        return OK;

    }

    catch (VNetworkException& e)
    {
        cpLog(LOG_DEBUG,
              "Failed to write to redundant server %s:%d, reason %s.",
              _host.c_str(), _port, e.getDescription().c_str());
        return ERR;
    }

}


int
RedundancyManager::sendGood()
{

    // lock the redundant server
    LockHelper lockAid( this->_aMutex );

    try
    {

        char buf[256];
        string sendData;
        char msg[256];
        int status, nread;

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();
        // send a sync good message, and check for a valid (200)
        // return code from the server
        sprintf(buf, "SYNCGOOD %s\n", VPP_VERSION);
        sendData += buf;
        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response sending SYNCGOOD");
            return ERR;

        }
        else
        {

            // status will be the text version which we ignore
            buf[nread] = 0;
            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {
                _state = ASYNC;  // if it accepted it, we are all synched!
                return OK;
            }
            else
            {
                cpLog(LOG_ERR, "Got a 400 FAIL while sending SYNCGOOD");
                return FAIL;
            }
        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send SYNCSTART to %s:%d.",
              _host.c_str(), _port);
        return ERR;
    }

}


int
RedundancyManager::sendRegisterItem(const string &group, const string &name,
                                  const string &host)
{

    // lock the redundant server
    LockHelper lockAid( this->_aMutex );

    try
    {

        char buf[256];
        string sendData;

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();

        sprintf(buf, "%s %s %s %s\n", SYNC_REGISTER_ITEM_REQ, group.c_str(),
                name.c_str(), host.c_str());
        sendData += buf;
        cpLog(LOG_DEBUG, "Sending to remote server : (%s)", sendData.c_str());
        clientSocket->getConn().writeData(sendData);

    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send REGISTER to %s:%d.",
              _host.c_str(), _port);
        return ERR;

    }

    return OK;

}

int
RedundancyManager::sendRegisterGroup(const string &group,
                                   const string &host)
{

    // lock the redundant server
    LockHelper lockAid( this->_aMutex );

    try
    {

        char buf[256];
        string sendData;

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();

        sprintf(buf, "%s %s %s\n", SYNC_REGISTER_GROUP_REQ, group.c_str(),
                host.c_str());
        sendData += buf;
        cpLog(LOG_DEBUG, "Sending : (%s)", sendData.c_str());
        clientSocket->getConn().writeData(sendData);

    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send REGISTER to %s:%d.",
              _host.c_str(), _port);
        return ERR;

    }

    return OK;

}


int
RedundancyManager::sendUnregisterItem(const string &group, const string &name,
                                    const string &host)
{

    // lock the redundant server
    LockHelper lockAid( this->_aMutex );

    try
    {

        char buf[256];
        string sendData;

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();

        sprintf(buf, "%s %s %s %s\n", SYNC_UNREGISTER_ITEM_REQ, group.c_str(),
                name.c_str(), host.c_str());
        sendData += buf;
        cpLog(LOG_DEBUG, "Sending to remote server : (%s)", sendData.c_str());
        clientSocket->getConn().writeData(sendData);

    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send UNREGISTER to %s:%d.",
              _host.c_str(), _port);
        return ERR;

    }

    return OK;

}

int
RedundancyManager::sendUnregisterGroup(const string &group,
                                     const string &host)
{

    // lock the redundant server
    LockHelper lockAid( this->_aMutex );

    try
    {

        char buf[256];
        string sendData;

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();

        sprintf(buf, "%s %s %s\n", SYNC_UNREGISTER_ITEM_REQ, group.c_str(),
                host.c_str());
        sendData += buf;
        cpLog(LOG_DEBUG, "Sending : (%s)", sendData.c_str());
        clientSocket->getConn().writeData(sendData);

    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send UNREGISTER to %s:%d.",
              _host.c_str(), _port);
        return ERR;

    }

    return OK;

}


int
RedundancyManager::getRemoteRegistration(RegisterManager &man)
{

    char buf[256], msg[256];
    string sendData;
    int status, len, nread = 0;

    LockHelper lockAid( this->_aMutex );

    try
    {

        // send request for the informaion
        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        sprintf(buf, "%s %s\n", SYNC_GRLIST_REQ, VPP_VERSION);
        sendData += buf;
        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            buf[nread] = 0;
            cpLog(LOG_INFO, "Data read for remote registration : (%s)", buf);

            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {

                if (conn.readLine(buf, sizeof(buf), nread) <= 0)
                {
                    cpLog(LOG_ERR,
                          "Failed to read content len. -- Closing connection");
                    return ERR;
                }

                // get the content length
                sscanf(buf, "%s%d", msg, &len);

                if (len == 0)
                {
                    // no data on remote server, so just return OK
                    return OK;
                }

                // allocate buffer for contents
                char *bigbuf;

                LocalScopeAllocator lAllocator(&bigbuf, len+1);

                conn.readn((void*)bigbuf, len, nread);

                if (nread < len)
                {
                    cpLog(LOG_ERR, "Insufficent data on socket - closing");
                    return ERR;
                }

                bigbuf[len] = 0;

                char *token;
                char file[512], host[256];
                int mode = 0;

                cpLog(LOG_DEBUG, "Registration list is: (%s)", bigbuf);

                token = strtok(bigbuf, ",");
                if (token == NULL)
                {
                    cpLog(LOG_ERR, "Invalid data sent for list of reg.");
                    return ERR;
                }
                sscanf(token, "%s", file);

                if (string(file) != "++ITEM++")
                {
                    cpLog(LOG_ERR, "Parse error on registration data.");
                    return ERR;
                }


                while ((token = strtok(NULL, ",")) != NULL)
                {

                    sscanf(token, "%s", file);
                    if (string(file) == "++GROUP++")
                    {
                        mode = 1;
                        continue;
                    }
                    token = strtok(NULL, ",");
                    if (token == NULL)
                    {
                        cpLog(LOG_ERR, "Parse error on socket - closing");
                        return ERR;
                    }
                    sscanf(token, "%s", host);
                    if (mode == 0)
                    {
                        man.registerItemCat(file, host);
                    }
                    else
                    {
                        man.registerGroup(file, host);
                    }
                }

                
                return OK;

            }
            else
            {

                cpLog(LOG_DEBUG, "Failed to get remote reg list");
                return FAIL;

            }

        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to send get Remote Reg list",
              _host.c_str(), _port);
        return ERR;
    }

}


int
RedundancyManager::putLocalRegistration(RegisterManager &man)
{

    char buf[256], msg[256];
    string sendData;
    int status, nread = 0;

    LockHelper lockAid( this->_aMutex );
    cpLog(LOG_DEBUG, "Sending twin our configuration");
    try
    {

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        string contents;
        man.getList(contents);

        // format and send message
        sprintf(buf, "%s %s\n", SYNC_PRLIST_REQ, VPP_VERSION);
        sendData += buf;
        sprintf(buf, "%s %d\n", CONT_LENGTH, contents.size());
        sendData += buf;
        sendData += contents;
        sendData += "\n";
        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            // terminate string and read status
            buf[nread] = 0;
            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {
                // if ok, just return ok
                return OK;
            }
            else
            {
                cpLog(LOG_DEBUG, "SYNCPRLIST File returned FAILED");
                return FAIL;
            }

        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to write to redundant server %s:%d.",
              _host.c_str(), _port);
        return ERR;
    }

}


int
RedundancyManager::sendDeleteItem (const string &group, const string &name)
{

    char buf[256], msg[256];
    string sendData;
    int status, nread = 0;

    LockHelper lockAid( this->_aMutex );

    try
    {

        Sptr < TcpClientSocket > clientSocket = 0;
        clientSocket = new TcpClientSocket(*_networkAddress);
        clientSocket->connect();
        Connection &conn = clientSocket->getConn();

        // format and send message
        sprintf(buf, "%s %s %s %s\n", SYNC_DEL_REQ, group.c_str(),
                name.c_str(), VPP_VERSION);
        sendData += buf;

        conn.writeData(sendData);

        // read back the result and make sure we get a 200 OK
        if (conn.readLine(buf, sizeof(buf), nread) <= 0)
        {

            cpLog(LOG_ERR, "No response -- Closing connection");
            return ERR;

        }
        else
        {

            // terminate string and read status
            buf[nread] = 0;
            sscanf(buf, "%d %s", &status, msg);

            if (status == 200)
            {
                // if ok, just return ok
                return OK;
            }
            else
            {
                cpLog(LOG_DEBUG, "SYNCDEL File returned FAILED");
                return FAIL;
            }

        }
    }

    catch (VNetworkException &e)
    {
        cpLog(LOG_DEBUG, "Failed to write to redundant server %s:%d.",
              _host.c_str(), _port);
        return ERR;
    }


}
