/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */


static const char* const AAATransceiver_cxx_Version =
    "$Id: AAATransceiver.cxx,v 1.17.2.1 2002/05/24 23:50:10 sprajpat Exp $";

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "UdpStack.hxx"
#include "B2bConfig.hxx"
#include "AAATransceiver.hxx"
#include "AuthData.hxx"
#include "BillingData.hxx"


AAATransceiver::AAATransceiver(int recvPort, Sptr<Fifo<Sptr<SipProxyEvent> > > outputFifo)
    : myOutputFifo(outputFifo),
      myShutdownFlg(false),
      myRadiusIdCounter(0)
{
    myUdpStack = new UdpStack(0, recvPort);
    mySendThread.spawn(sendThreadWrapper, this);
    myRecvThread.spawn(rcvThreadWrapper, this);
}

void 
AAATransceiver::setResponse(Sptr<AAAEvent> event)
{
    myOutputFifo->add(event);
}

AAATransceiver::~AAATransceiver()
{
   myShutdownFlg = true;
   mySendFifo.add(0);

   mySendThread.join();
   myRecvThread.exit();
}

void* 
AAATransceiver::sendThreadWrapper(void *p)
{
    AAATransceiver* self = static_cast<AAATransceiver*>(p);
    return(self->sendMain());
}

void* 
AAATransceiver::rcvThreadWrapper(void *p)
{
    AAATransceiver* self = static_cast<AAATransceiver*>(p);
    return(self->receiveMain());
}

void* 
AAATransceiver::receiveMain()
{
    NetworkAddress sender;
    while (1)
    {
        int len = 0;
        static const string::size_type MTU = 3600;
        char buf[MTU+1];
        buf[0] = '\0';
        len = 0;
        len = myUdpStack->receiveFrom( buf,MTU , &sender );
        buf[len] = '\0';
 
        if (( len == 0) || ( len < 0 ))
        {
            return 0;
        }

        if(myShutdownFlg)
        {
            return 0;
        }

        RadiusData msgReceived(buf, len);
        const char* secret = 
              B2bConfig::instance().getStr(B2bConfig::RAD_BILLING_PASSWD).c_str();
        Sptr<RadiusMessage> responseMsg = 0;
        try
        {
            responseMsg = new RadiusMessage( msgReceived , secret); 
            u_int8_t rId = responseMsg->getIdentifier();
            if((responseMsg->type() == RP_ACCESS_ACCEPT) ||
               (responseMsg->type() == RP_ACCESS_REJECT))
            {
                Sptr<RadiusPayload> rPayload;
                myMutex.lock();
                if(myRetransMap[rId] == 0)
                {
                    cpLog(LOG_ERR,"Got response for id=%d, but no request found, ignoring", rId);
                    myMutex.unlock();
                    continue;
                }
                rPayload = myRetransMap[rId];
                cleanUp(rId);
                myMutex.unlock();
                Sptr<AAAEvent> aEvent = rPayload->aEvent;
                AuthAgent* authAgent = aEvent->getAuthAgent();
                if(authAgent == 0) continue;
                Sptr<BasicAAAData> basicData = authAgent->getAAAData();

                Sptr<AuthData> authData;
                authData.dynamicCast(basicData);

                if(authData == 0)
                {
                    //Got a delayed response, which does not have any
                    //pending transaction, ignore
                    cpLog(LOG_INFO, "Got a delayed response from radius server, which does not have any pending transaction, ignoring");
                    continue;
                }

                if(responseMsg->type() == RP_ACCESS_ACCEPT)
                {
                    RadiusData durationData = 
                           (responseMsg->get( RA_SESSION_TIMEOUT)).value();
                    u_int32_t duration=0;
                    memcpy(&duration, durationData.data(), durationData.length());
                    duration = ntohl(duration);
                    cpLog(LOG_DEBUG, "Got session timeout %ld", duration);

                    authData->setSessionTimeOut(duration);
                    authData->setStatus(A_ACCEPT);
                }
                else
                {
                    authData->setStatus(A_REJECT);
                }
                myOutputFifo->add(aEvent);
            }
            else
            {
                //Accounting response, do nothing
                myMutex.lock();
                cleanUp(rId);
                myMutex.unlock();
            }
        }
        catch(VException& e)
        {
           cpLog(LOG_WARNING, "Problem with Radius message ,reason:%s" ,e.getDescription().c_str());
        }

    }
}

void* 
AAATransceiver::sendMain()
{
    while(1)
    {
        Sptr<AAAEvent> aaaEvent =  mySendFifo.getNext();

        if(myShutdownFlg)
        {
            return 0;
        }

        udpSend(aaaEvent);

        //Retransmission 
        handleRetransmission(aaaEvent);

#if 0
        ///Simulate the response received
        vusleep(100000);
        Sptr<AuthData> authData;
        authData.dynamicCast(aaaEvent->getAuthAgent()->getAAAData());
        if(authData != 0)
        {
            authData->setStatus(A_ACCEPT);
            myOutputFifo->add(aaaEvent);
        }
        u_int8_t rId = aaaEvent->getReqId();
        Sptr<RadiusPayload> rPayload = myRetransMap[rId];
        rPayload->retransCount = 0; 
        //Cancel the retransmission event also
        mySendFifo.cancel(rPayload->myRetransId);
        cleanUp(rId);
        ///
#endif


    }
    return 0;
}

void
AAATransceiver::handleRetransmission(Sptr<AAAEvent> aaaEvent)
{
    u_int8_t rId = aaaEvent->getReqId();
    myMutex.lock();
    if(myRetransMap[rId] == 0)
    {
        //rId not found, ignore retransmission
        myMutex.unlock();
        return;
    }

    Sptr<RadiusPayload> rPayload = myRetransMap[rId];
    myMutex.unlock();
    if(rPayload->retransCount)
    {
        rPayload->retransCount--;
        cpLog(LOG_DEBUG,  "Scheduled for retransmission, count left:%d", rPayload->retransCount);
        //Schedule for retransmission after 3 seconds
        rPayload->myRetransId = mySendFifo.addDelayMs(aaaEvent, 3000);
    }
    else
    {
        cpLog(LOG_DEBUG,  "Retransmission exhausted");
        RadiusPacketType rType = rPayload->rMsg->type();
        if(rType == RP_ACCESS_REQUEST)
        {
            //Auth failed
            myMutex.lock();
            myRetransMap[rId] = 0;
            myMutex.unlock();
            myOutputFifo->add(aaaEvent);
        }
        else
        {
            //Failed to talk to billing server
            //if Billing is manadatory, no call should go through
            if(B2bConfig::instance().getStr(B2bConfig::PP_OPTIONS) == B2bConfig::PP_MANDATORY)
            {
                //TODO - Should log it locally for later consolidation
            }
        }
    }
}

void
AAATransceiver::cleanUp(u_int8_t rId)
{
    if(myRetransMap[rId] == 0)
    {
        return;
    }
    Sptr<RadiusPayload> rPayload = myRetransMap[rId];
    myRetransMap[rId] = 0;

    //Stop the retransmission
    rPayload->retransCount = 0; 
    //Cancel the retransmission event also
    mySendFifo.cancel(rPayload->myRetransId);
}

void
AAATransceiver::udpSend(Sptr<AAAEvent> aaaEvent)
{
    Sptr<RadiusPayload> rPayload =0;
    u_int8_t rId;

    myMutex.lock();
    rId = aaaEvent->getReqId();
    if(myRetransMap[rId] == 0)
    {
        cpLog(LOG_DEBUG, "Radius ID (%d) not in database, ignoring..", rId);
        myMutex.unlock();
        return;
    }
    rPayload = myRetransMap[rId];
    myMutex.unlock();

    if(rPayload->rMsg == 0) return;

    //Send it now
    //Get Radius Auth Host/Port
    B2bConfig& bc = B2bConfig::instance();
    RadiusPacketType rType = rPayload->rMsg->type();
    Sptr<NetworkAddress> dest = 0; 
    if(rType == RP_ACCESS_REQUEST)
    {
        dest = new NetworkAddress (bc.getStr(B2bConfig::RAD_BILLING_ADDR),
                        bc.getInt(B2bConfig::RAD_BILLING_AUTH_PORT));
    }
    else
    {
        dest = new NetworkAddress (bc.getStr(B2bConfig::RAD_BILLING_ADDR),
                        bc.getInt(B2bConfig::RAD_BILLING_ACCT_PORT));
    }
    cpLog(LOG_DEBUG, "Sending request for Radius id(%d), to %s:%d", rId, 
                      dest->getIpName().c_str(), dest->getPort());
    RawMessage udpMsg( rPayload->rMsg->data() );
    //Hack, I hate it
    const char* c = reinterpret_cast<const char*>(udpMsg.buffer);
    myUdpStack->transmitTo(c, ntohs(udpMsg.msgHdr.length), dest.getPtr());
}

void
AAATransceiver::buildThePayload(Sptr<AAAEvent> aaaEvent, 
                                u_int8_t type, u_int8_t subType)
{
    cpLog(LOG_DEBUG, "AAATransceiver::buildThePayload for type %d", type);
    Sptr<RadiusPayload> rPayload = new RadiusPayload();
    u_int8_t rId = getNextId();
    aaaEvent->setReqId(rId);
    switch(type)
    {
        case RP_ACCESS_REQUEST:
        {
            rPayload->rMsg = new RadiusMessage(RP_ACCESS_REQUEST);
            fillAAAData(aaaEvent, rPayload->rMsg);
        }
        break;
        case RP_ACCOUNTING_REQUEST:
        {
            rPayload->rMsg = new RadiusMessage(RP_ACCOUNTING_REQUEST);
            if(subType == RAS_START)
            {
                fillAccountStartData(aaaEvent, rPayload->rMsg);
            }
            else if(subType == RAS_STOP)
            {
                fillAccountStopData(aaaEvent, rPayload->rMsg);
            }
            else if(subType == RAS_INTERIM)
            {
                fillInterimAccountData(aaaEvent, rPayload->rMsg);
            }
        }
        break;
        default:
            cpLog(LOG_WARNING, "Can not handle the type %d, ignoring", type);
            return;
        break;
    }

    rPayload->aEvent = aaaEvent;
    rPayload->rMsg->setIdentifier(rId);
    const char* secret = B2bConfig::instance().getStr(B2bConfig::RAD_BILLING_PASSWD).c_str();        
    rPayload->rMsg->calcAuthenticator( secret );
    myMutex.lock();
    myRetransMap[rId] = rPayload;
    myMutex.unlock();
}

void 
AAATransceiver::sendAuthRequest(Sptr<AAAEvent> aaaEvent)
{
    cpLog(LOG_DEBUG, "AAATransceiver::sendAuthRequest");
    buildThePayload(aaaEvent, RP_ACCESS_REQUEST, 0);
    mySendFifo.add(aaaEvent);
}

void 
AAATransceiver::sendAcctStart(Sptr<AAAEvent> aaaEvent)
{
    cpLog(LOG_DEBUG, "AAATransceiver::sendAcctStart");
    buildThePayload(aaaEvent, RP_ACCOUNTING_REQUEST, RAS_START);
    mySendFifo.add(aaaEvent);
}

void 
AAATransceiver::sendAcctInterim(Sptr<AAAEvent> aaaEvent)
{
    cpLog(LOG_DEBUG, "AAATransceiver::sendAcctInterim");
    buildThePayload(aaaEvent, RP_ACCOUNTING_REQUEST, RAS_INTERIM);
    mySendFifo.add(aaaEvent);
}

void 
AAATransceiver::sendAcctStop(Sptr<AAAEvent> aaaEvent)
{
    cpLog(LOG_DEBUG, "AAATransceiver::sendAcctStop");
    buildThePayload(aaaEvent, RP_ACCOUNTING_REQUEST, RAS_STOP);
    mySendFifo.add(aaaEvent);
}

void
AAATransceiver::fillAAAData(Sptr<AAAEvent> aEvent, 
                            Sptr<RadiusMessage> rMsg) const
{
    Sptr<BasicAAAData> aData = aEvent->getAuthAgent()->getAAAData();
    //Set the user name, PUID
    const string& uId = aData->getUserName();
    RadiusData valUserName(uId.c_str(), uId.length());
    RadiusAttribute attrUserName( RA_USER_NAME, valUserName );
    rMsg->add( attrUserName );

    if(rMsg->type() == RP_ACCESS_REQUEST)
    {
        Sptr<AuthData> authData;
        authData.dynamicCast(aData);
        const string& userPasswd = authData->getUserPasswd();
        RadiusData valUserPassword( userPasswd.c_str() , userPasswd.length() );
        RadiusAttribute attrUserPassword( RA_USER_PASSWORD, valUserPassword );
        rMsg->add( attrUserPassword );
    }

    //NAS-IP
    u_int32_t ip = inet_addr(aData->getIp().c_str());
    RadiusData valNasIpAddress( &ip , sizeof(ip) );
    RadiusAttribute attrNasIpAddress( RA_NAS_IP_ADDRESS, valNasIpAddress );
    rMsg->add( attrNasIpAddress );

    ///Fixed port for a given server- Arbitrarily choosen
    u_int32_t port( htonl(1000));
    RadiusData valNasPort( &port , sizeof(port) );
    RadiusAttribute attrNasPort( RA_NAS_PORT, valNasPort );
    rMsg->add( attrNasPort );

    const string& calledId = aData->getCalledId();
    RadiusData valCallee( calledId.c_str(), calledId.length());
    RadiusAttribute attrCallee( RA_CALLED_STATION_ID, valCallee );
    rMsg->add( attrCallee );

    const string& callingId = aData->getCallingId(); 
    RadiusData valCaller( callingId.c_str(),callingId.length());
    RadiusAttribute attrCaller( RA_CALLING_STATION_ID, valCaller );
    rMsg->add( attrCaller );
}

void
AAATransceiver::fillAccountStartData(Sptr<AAAEvent> aEvent, 
                                     Sptr<RadiusMessage> rMsg) const
{
    fillAAAData(aEvent, rMsg);
    Sptr<BillingData> aData;
    aData.dynamicCast(aEvent->getAuthAgent()->getAAAData());
    assert(aData != 0);

    u_int32_t acctType( htonl(RAS_START) );
    RadiusData valAcctStatus( &acctType, sizeof(acctType) );
    RadiusAttribute attrAcctStatus( RA_ACCT_STATUS_TYPE, valAcctStatus );
    rMsg->add( attrAcctStatus );

    const string& sId = aData->getSessionId();
    RadiusData valAcctSessionId( sId.c_str(), sId.length() );
    RadiusAttribute attrAcctSessionId( RA_ACCT_SESSION_ID, valAcctSessionId );
    rMsg->add( attrAcctSessionId );
}

void
AAATransceiver::fillInterimAccountData(Sptr<AAAEvent> aEvent, 
                                  Sptr<RadiusMessage> rMsg) const
{
    fillAAAData(aEvent, rMsg);

    Sptr<BillingData> aData;
    aData.dynamicCast(aEvent->getAuthAgent()->getAAAData());
    assert(aData != 0);

    u_int32_t acctType = htonl(RAS_INTERIM);
    RadiusData valAcctStatus2( &acctType, sizeof(acctType) );
    RadiusAttribute attrAcctStatus2( RA_ACCT_STATUS_TYPE, valAcctStatus2 );
    rMsg->add( attrAcctStatus2 );

    const string& sId = aData->getSessionId();
    RadiusData valAcctSessionId( sId.c_str(), sId.length() );
    RadiusAttribute attrAcctSessionId( RA_ACCT_SESSION_ID, valAcctSessionId );
    rMsg->add( attrAcctSessionId );

    u_int32_t setupDelay( htonl(aData->getSetupDelay() ));
    RadiusData valSetupDelay( &setupDelay, sizeof(setupDelay) );
    RadiusAttribute attrSetupDelay( RA_ACCT_DELAY_TIME, valSetupDelay );
    rMsg->add( attrSetupDelay );
}

void
AAATransceiver::fillAccountStopData(Sptr<AAAEvent> aEvent, 
                                     Sptr<RadiusMessage> rMsg) const
{
    fillAAAData(aEvent, rMsg);
    Sptr<BillingData> aData;
    aData.dynamicCast(aEvent->getAuthAgent()->getAAAData());
    assert(aData != 0);

    u_int32_t acctType( htonl(RAS_STOP) );
    RadiusData valAcctStatus( &acctType, sizeof(acctType) );
    RadiusAttribute attrAcctStatus( RA_ACCT_STATUS_TYPE, valAcctStatus );
    rMsg->add( attrAcctStatus ) ;

    const string& sId = aData->getSessionId();
    RadiusData valAcctSessionId( sId.c_str(), sId.length() );
    RadiusAttribute attrAcctSessionId( RA_ACCT_SESSION_ID, valAcctSessionId );
    rMsg->add( attrAcctSessionId ) ;

    u_int32_t duration( htonl(aData->getDuration()));
    RadiusData valAcctSessionTime( &duration, sizeof(duration) );
    RadiusAttribute attrAcctSessionTime( RA_ACCT_SESSION_TIME, valAcctSessionTime );
    rMsg->add( attrAcctSessionTime );

    u_int32_t cause;
    if(aData->getTermCause() == BT_SESSION_TIMEOUT)
    { 
        cause = ( htonl(RATC_SESSION_TIMEOUT) );
    }
    else
    {
        cause = ( htonl(RATC_USER_REQUEST) );
    }
    
    RadiusData valAcctTerminateCause( &cause, sizeof(cause) );
    RadiusAttribute attrAcctTerminateCause( RA_ACCT_TERMINATE_CAUSE, valAcctTerminateCause );
    rMsg->add( attrAcctTerminateCause );
}
