/* ====================================================================
 * The Vovida Software License, Version 1.0 
 * 
 * Copyright (c) 2000 Vovida Networks, Inc.  All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 * 
 * 3. The names "VOCAL", "Vovida Open Communication Application Library",
 *    and "Vovida Open Communication Application Library (VOCAL)" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact vocal@vovida.org.
 *
 * 4. Products derived from this software may not be called "VOCAL", nor
 *    may "VOCAL" appear in their name, without prior written
 *    permission of Vovida Networks, Inc.
 * 
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE, TITLE AND
 * NON-INFRINGEMENT ARE DISCLAIMED.  IN NO EVENT SHALL VOVIDA
 * NETWORKS, INC. OR ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT DAMAGES
 * IN EXCESS OF $1,000, NOR FOR ANY INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
 * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
 * DAMAGE.
 * 
 * ====================================================================
 * 
 * This software consists of voluntary contributions made by Vovida
 * Networks, Inc. and many individuals on behalf of Vovida Networks,
 * Inc.  For more information on Vovida Networks, Inc., please see
 * <http://www.vovida.org/>.
 *
 */

static const char* const SipTransceiver_cxx_Version =
    "$Id: SipTransceiver.cxx,v 1.69.2.1 2002/05/15 22:21:05 chok Exp $";

#include "global.h"
#include <cstdlib>
#include <cassert>
#include <sys/time.h>
#include <string>
#include <Sptr.hxx>
#include "SipCommand.hxx"
#include "StatusMsg.hxx"
#include "InviteMsg.hxx"
#include "CancelMsg.hxx"
#include "AckMsg.hxx"
#include "ByeMsg.hxx"
#include "RegisterMsg.hxx"
#include "OptionsMsg.hxx"
#include "InfoMsg.hxx"
#include "ReferMsg.hxx"
#include "TransferMsg.hxx"
#include "SipVia.hxx"
#include "SipFrom.hxx"
#include "SipContact.hxx"
#include "SipCSeq.hxx"
#include "SipSnmpDetails.hxx"
#include "SipTransactionId.hxx"
#include "SipTransceiver.hxx"
#include "SystemInfo.hxx"

using namespace Vocal;


class TransactionDBStatus: public SipDebuggingStatus
{
    public:
        TransactionDBStatus(SipTransactionDB& x, Data label )
            :myItem(x),
             myLabel(label + ": ")
            {
            }

        Data status()
            {
                return myLabel + myItem.getDetails();
            }

    private:
        SipTransactionDB& myItem;
        Data myLabel;
};


class SipUdpStatus : public SipDebuggingStatus
{
    public:
        SipUdpStatus(SipUdpConnection& x, int port)
            :myItem(x),
             myLabel()
            {
                char buf[256];
                sprintf(buf, "UDP Stack (port %d): ", port);
                myLabel = buf;
            }
        Data status()
            {
                return myLabel + myItem.getDetails();
            }
    private:
        SipUdpConnection& myItem;
        Data myLabel;
};


SipAppContext SipTransceiver::myAppContext = APP_CONTEXT_GENERIC;

SipTransceiver::SipTransceiver( Data adata, 
				int siplistenPort, 
				bool nat  /*Default Agruments*/,
				SipAppContext aContext  /*Default Agruments*/)
    :
    recvdMsgsFifo(),
    udpConnection(0),
    tcpConnection(0),
    sentRequestDB(),
    sentResponseDB()
{
    myAppContext = aContext;
    cpLog( LOG_DEBUG_STACK, "SipStack listening on port %d ", siplistenPort );
    cpLog( LOG_DEBUG_STACK, "SipStack context %s ", (myAppContext == APP_CONTEXT_PROXY) ? "Proxy" : "Non-Proxy" );

    udpConnection = new SipUdpConnection( &recvdMsgsFifo, siplistenPort );

    if ( udpConnection == 0 )
    {
        cpLog(LOG_INFO, "SipTransceiver::unable to instantiate UDP connections");
    }

    tcpConnection = new SipTcpConnection( &recvdMsgsFifo, siplistenPort );

    if ( tcpConnection == 0 )
    {
        cpLog(LOG_INFO, "SipTransceiver::unable to instantiate TCP connections");
    }
    application = adata;
    Data nameport;
    nameport += adata;
    nameport += ":";
    nameport += siplistenPort;
#ifdef WIN32
    sipAgent = 0;
#else
    sipAgent = new SipAgent(nameport);
#endif
    natOn = nat;

    Vocal::theSystem.setSystemPort(siplistenPort);

    char* port;
    if((port = getenv("VOCAL_SIP_REPORTER")))
    {
        debugger = new SipDebuggingInterface(atoi(port)); 
        debugger->add(new TransactionDBStatus(sentRequestDB, "Request DB"));
        debugger->add(new TransactionDBStatus(sentResponseDB, "Response DB"));
        debugger->add(new SipUdpStatus(*udpConnection, siplistenPort));
    }

    myLocalNamePort = nameport;
}


SipTransceiver::~SipTransceiver()
{
    // xxx - this needs to close up shop
    SipTransactionGC::shutdown();
}


void
SipTransceiver::sendAsync(SipCommand& sipMessage)
{
    if ( !(sipMessage.checkMaxForwards()) )
    {
        cpLog(LOG_DEBUG_STACK, "SipTransceiver: sendAsync: %s",
              sipMessage.encode().logData());
        cpLog(LOG_DEBUG_STACK, "reached MaxForwards limit, not sending");

        return ;
    }

    SipMsgContainer* msgPtr 
	= sentRequestDB.processSend(
	    copyPtrtoSptr(&sipMessage));

    if(msgPtr)
    {
	//// should this only be for actually sent messages? /////////
	if (sipAgent != 0)
	{
	    updateSnmpData(copyPtrtoSptr(&sipMessage), OUTS);
	}
	//////////////////////////////////////////////////////////////
	
	send(msgPtr);
    }
}


void
SipTransceiver::sendAsync(Sptr<SipCommand> sipMessage, const Data& host,
                       const Data& port) 
{
    if ( !(sipMessage->checkMaxForwards()) )
    {
        cpLog(LOG_DEBUG_STACK, "SipTransceiver: sendAsync: %s",
              sipMessage->encode().logData());
        cpLog(LOG_DEBUG_STACK, "reached MaxForwards limit, not sending");

        return ;
    }

    SipMsgContainer* msgPtr = sentRequestDB.processSend(sipMessage);
    
    if(msgPtr)
    {
	//// should this only be for actually sent messages? /////////
	if (sipAgent != 0)
	{
	    updateSnmpData(sipMessage, OUTS);
	}
	//////////////////////////////////////////////////////////////

	send(msgPtr, host, port);
    }
}


void
SipTransceiver::sendReply(StatusMsg& sipMessage)
{
    //// should this not be symmetric to what we are doing in sendAsync????
    //// (its missing all the checks and the snmp update)

    SipMsgContainer *sipMsg 
	= sentResponseDB.processSend(
	    copyPtrtoSptr(&sipMessage));
    if(sipMsg)
    {
	send(sipMsg);
    }
}


void
SipTransceiver::sendReply(Sptr<StatusMsg> sipMessage)
{
    //// should this not be symmetric to what we are doing in sendAsync????
    //// (its missing all the checks and the snmp update)
    
    SipMsgContainer *sipMsg = sentResponseDB.processSend(sipMessage);
    if(sipMsg)
    {
	send(sipMsg);
    }
}


void
SipTransceiver::send(SipMsgContainer *sipMsg, const Data& host,
                       const Data& port)
{
    if ((sipMsg->msg.transport == TCP) ||
        (sipMsg->msg.transport == "tcp"))
    {
        if (tcpConnection != 0)
        {
            //send on tcp.
            tcpConnection->send(sipMsg, host, port);
        }
        else
        {
            cpLog(LOG_INFO, "TCP connection not instantiated");

	    /// hence delete the msg container
	    SipTransactionGC::instance()->
                collect(sipMsg, ORPHAN_CLEANUP_DELAY);
        }
    }
    else
    {
        if (udpConnection != 0)
        {
            //send on udp.
            udpConnection->send(sipMsg, host, port);
        }
        else
        {
            cpLog(LOG_INFO, "UDP connection not instantiated");

	    /// hence delete the msg container
	    SipTransactionGC::instance()->
                collect(sipMsg, ORPHAN_CLEANUP_DELAY);
        }
    }
}


///for debug printing
string 
Vocal::printMsgBrief(Sptr < SipMsg > msg)
{
    if (msg == 0)
    {
        return string("NO MSG");
    }
    string returnString;
    Method method = msg->getType();

    switch (method)
    {
        case SIP_UNKNOWN:
        returnString = "UNKNOWN";
        break;
        case SIP_INVITE:
        returnString = "INVITE";
        break;
        case SIP_ACK:
        returnString = "ACK";
        break;
        case SIP_OPTIONS:
        returnString = "OPTIONS";
        break;
        case SIP_BYE:
        returnString = "BYE";
        break;
        case SIP_CANCEL:
        returnString = "CANCEL";
        break;
        case SIP_REGISTER:
        returnString = "REGISTER";
        break;
        case SIP_INFO:
        returnString = "INFO";
        break;
        case SIP_TRANSFER:
        returnString = "TRANSFER";
        break;
        case SIP_REFER:
        returnString = "REFER";
        break;
        case SIP_NOTIFY:
        returnString = "NOTIFY";
        break;
        case SIP_SUBSCRIBE:
        returnString = "SUBSCRIBE";
        break;
        case SIP_STATUS:
        {
            Sptr < StatusMsg > p;
            p.dynamicCast(msg);
            if (p != 0)
            {
                char buf[20];
                buf[0] = '\0';
                sprintf(buf, "%d", p->getStatusLine().getStatusCode());
                returnString = "STATUS (";
                returnString += buf;
                returnString += ")";
            }
            else
            {
                returnString = "STATUS";
            }
            break;
        }
        default:
        returnString = "TYPE NOT PRINTABLE";
        break;
    }
    LocalScopeAllocator lo;
    returnString += " [" ;
    returnString += (msg->getCallId().getCallIdData().getData(lo));
    returnString += "]";

    return returnString;
}


Sptr < SipMsgQueue > SipTransceiver::receive(int timeOut)
{
    Sptr < SipMsgQueue > msgQPtr = 0;
  
    timeval start, now;
  
    if ( timeOut >= 0 )
    {
	gettimeofday(&start, 0);
    }
  
    while (msgQPtr == 0)
    {
	int timePassed = 0;
	if ( timeOut >= 0 )
        {
	    gettimeofday(&now, 0);
	  
	    timePassed = ( now.tv_sec - start.tv_sec ) * 1000
		+ ( now.tv_usec - start.tv_usec ) / 1000;
	  
	    if (timePassed >= timeOut)
            {
		cpLog(LOG_DEBUG, "timeout value exceeded");
		return 0;
            }
	}
	cpLog( LOG_DEBUG_STACK, "entering block");
	recvdMsgsFifo.block(timeOut);
	cpLog( LOG_DEBUG_STACK, "woke from block");
      
	if ( !recvdMsgsFifo.messageAvailable() )
	{
	    continue;
	}
      
	cpLog( LOG_DEBUG_STACK, "woke from messageAvailable");
      
	SipMsgContainer *msgPtr = recvdMsgsFifo.getNext();
	if ( msgPtr == 0)
	{
	    assert(0);
            cpLog(LOG_CRIT, "received NULL");
	    continue;
	}

	/********************** TO DO ****************************
	 * not doing 'coz need to bring some stuff from udp impl
	 * to here, so will do after 500 cps
         * msgPtr->msg.in = SipMsg::decode(msgPtr->msg.out);
	 *********************************************************/

	/*****************************************************************
	 * how does this affect the transactions, i.e. it is only being
	 * done on received messages, so should this be before or after
	 * going thru the data base (i.e. is it visible to transactionDB)?
	 *****************************************************************/

	/*********** decided to remove it from the stack ******************/
#if 1
      if ( natOn == true)
	{
	//changes for taking care of the NAT traversals
	  SipVia natVia = msgPtr->msg.in->getVia(0);
	  LocalScopeAllocator lo; 
	  string addr1 = natVia.getHost().getData(lo);
	  string addr2 = msgPtr->msg.in->getReceivedIPName().getData(lo);
	  NetworkAddress netaddr1(addr1);
	  NetworkAddress netaddr2(addr2);
	  
	  if ( netaddr1.getHostName() != netaddr2.getHostName())
	    {
	      natVia.setReceivedhost(msgPtr->msg.in->getReceivedIPName());
	      natVia.setReceivedport(msgPtr->msg.in->getReceivedIPPort());
	      //remove the first item from the via list
	      msgPtr->msg.in->removeVia(0);
	      //insert natvia in the vector via list
	      msgPtr->msg.in->setVia(natVia, 0);
	    }
	}
#endif
	//---NAT
/*	*********************************************************************/
	
      
	cpLog(LOG_DEBUG, "Received: %s",Vocal::printMsgBrief(msgPtr->msg.in).c_str());
	cpLog(LOG_DEBUG_STACK, "msg is; %s",msgPtr->msg.out.logData());
      
	SipMsgQueue *msgQ = 0;
	Sptr<SipMsg> sipPtr = msgPtr->msg.in;

	if(msgPtr->msg.in->getType() == SIP_STATUS)
	    msgQ = sentRequestDB.processRecv(msgPtr);
	else
	    msgQ = sentResponseDB.processRecv(msgPtr);
      
	if(msgQ)
	{
	    msgQPtr = msgQ;
	  
	    //need to have snmpDetails for this.
	    if (sipAgent != 0)
	    {
		updateSnmpData(sipPtr, INS);
	    }
	}
	else if(msgPtr->msg.in != 0)
	{
	    send(msgPtr);
	}
	else if(msgPtr->msg.out.length())
	{
	    send(msgPtr);
	}
	else
	{
	    delete msgPtr;
	}
    }
    return msgQPtr;
}


void
SipTransceiver::reTransOff()
{
    cpLog(LOG_DEBUG_STACK, "*** Retransmission is turned off ***");
    SipUdpConnection::reTransOff();
}


void
SipTransceiver::reTransOn()
{
    cpLog(LOG_DEBUG_STACK, "*** Retransmission is turned on ***");
    SipUdpConnection::reTransOn();
}


void
SipTransceiver::setRetransTime(int initial, int max)
{
    cpLog(LOG_DEBUG_STACK, "Initial time set to %d ms", initial);
    cpLog(LOG_DEBUG_STACK, "Max time set to %d ms", max);
    SipUdpConnection::setRetransTime(initial, max);
}


void
SipTransceiver::setRandomLosePercent(int percent)
{
    if (udpConnection != 0)
    {
        udpConnection->setRandomLosePercent(percent);
    }
}


SipTransactionDB::CallLegVector
SipTransceiver::getCallLegMsgs(Sptr < SipMsg > sipmsg)
{
    //// will it always be on SentRequest???
    return (sentRequestDB.getCallLegMsgs(sipmsg));
}


void
SipTransceiver::updateSnmpData(Sptr < SipMsg > sipMsg, SnmpType snmpType)
{
    //come here only if snmpAgent is valid.

    if (sipAgent == 0)
        return ;

    Method type = sipMsg->getType();
    if (type == SIP_STATUS)
    {
        if (snmpType == INS)
        {
            sipAgent->updateSipCounter(sipSummaryInResponses);
        }
        else if (snmpType == OUTS)
        {
            sipAgent->updateSipCounter(sipSummaryOutResponses);
        }

        //get the status code of this msg.
        Sptr < StatusMsg > statusMsg;
        statusMsg.dynamicCast(sipMsg);

        int statusCode = statusMsg->getStatusLine().getStatusCode();

        if (snmpType == INS)
        {
            sipAgent->updateStatusSnmpData(statusCode, INS);
        }
        else if (snmpType == OUTS)
        {
            sipAgent->updateStatusSnmpData(statusCode, OUTS);
	}

    }

    else
    {
        //this is a command.
        if (snmpType == INS)
        {
            sipAgent->updateSipCounter(sipSummaryInRequests);

            sipAgent->updateCommandSnmpData(type, INS);
        }
        else if (snmpType == OUTS)
        {
            sipAgent->updateSipCounter(sipSummaryOutRequests);

            sipAgent->updateCommandSnmpData(type, OUTS);
        }
    }
}


void
SipTransceiver::printSize()
{
    cpLog(LOG_ERR, "%s:::::\n%s\n%s\n\n%s\n%s\n\n%s\n%s",
	  "Sip Stack Size",
	  "SentResponseDB size...", sentResponseDB.getDetails().logData(),
	  "SentRequestDB size...", sentRequestDB.getDetails().logData(),
          "UDP stack details...", udpConnection->getDetails().logData());
}


Data
SipTransceiver::getLocalNamePort() const
{
    return myLocalNamePort;
}

/* Local Variables: */
/* c-file-style: "stroustrup" */
/* indent-tabs-mode: nil */
/* c-file-offsets: ((access-label . -) (inclass . ++)) */
/* c-basic-offset: 4 */
/* End: */
